SystemOrganization addCategory: #'Tweak-Basic-Collections'!
SystemOrganization addCategory: #'Tweak-Basic-Costumes'!
SystemOrganization addCategory: #'Tweak-Basic-Events'!
SystemOrganization addCategory: #'Tweak-Basic-Examples'!
SystemOrganization addCategory: #'Tweak-Basic-Support'!
SystemOrganization addCategory: #'Tweak-Basic-Tests'!
SystemOrganization addCategory: #'Tweak-Basic-Undo'!
SystemOrganization addCategory: #'Tweak-Basic-User-Kernel'!
SystemOrganization addCategory: #'Tweak-Basic-User-Objects'!
SystemOrganization addCategory: #'Tweak-Basic-Worlds'!

CObject subclass: #CCollectionBase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Collections'!

!CCollectionBase commentStamp: 'ar 3/16/2004 10:44' prior: 0!
CCollectionBase defines the generic interface for collections inside Tweak.!

CCollectionBase subclass: #CCollection
	instanceVariableNames: 'values cursor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Kernel'!

!CCollection commentStamp: 'ar 3/27/2004 18:04' prior: 0!
CCollection defines the generic collection implementation based on CCollectionBase (which defines the interface).

Instance variables:
	values		<Array>		The values of the collection.
	cursor		<Number>	The primary cursor of the collection.
!

CCollection defineFields: '<?xml  version="1.0" ?>
<fields>
	<field instanceVariable="true">values</field>
	<field instanceVariable="true" toGet="cursor" toSet="cursor:">cursor</field>
</fields>'!

!CCollection class methodsFor: 'instance creation' stamp: 'ar 2/12/2005 01:57'!
withAll: anArray
	^(self new) addAll: anArray; yourself! !

!CCollection methodsFor: 'adding' stamp: 'ar 3/24/2004 20:54'!
add: anObject at: index
	"Insert the given object before the specified index"
	| newData |
	newData := Array new: values size+1.
	newData replaceFrom: 1 to: index-1 with: values startingAt: 1.
	newData at: index put: anObject.
	newData replaceFrom: index+1 to: newData size with: values startingAt: index.
	values := newData.
	self contentsChanged: #insert with: index with: index.
	index <= cursor ifTrue:[self cursor: cursor + 1].
	^anObject! !

!CCollection methodsFor: 'adding' stamp: 'ar 12/10/2005 20:58'!
addAll: aCollection at: index
	"Insert the given object before the specified index"
	| sz newData |
	sz := aCollection size.
	newData := Array new: values size + sz.
	newData replaceFrom: 1 to: index-1 with: values startingAt: 1.
	newData replaceFrom: index to: index+sz-1 with: aCollection asArray startingAt: 1.
	newData replaceFrom: index+sz to: newData size with: values startingAt: index.
	values := newData.
	self contentsChanged: #insert with: index with: index+sz-1.
	index <= cursor ifTrue:[self cursor: cursor + sz].
	^aCollection! !

!CCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 10:59'!
at: aKey
	"Answer the element at the given key"
	^values at: aKey! !

!CCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 10:59'!
at: aKey ifAbsent: aBlock
	"Answer the element at the given key or evaluate the given block if there is no element with the given key"
	^values at: aKey ifAbsent: aBlock! !

!CCollection methodsFor: 'accessing' stamp: 'ar 3/24/2004 20:54'!
at: aKey put: anObject
	"Store anObject under the given key in the receiver"
	values at: aKey put: anObject.
	self contentsChanged: #replace with: aKey with: aKey.
	^anObject! !

!CCollection methodsFor: 'accessing' stamp: 'ar 3/7/2004 17:42'!
contentType
	"Answer the type of the elements stored in the receiver"
	^nil! !

!CCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 11:00'!
contents
	"Answer the contents of the receiver"
	^values shallowCopy! !

!CCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 11:01'!
contents: newContents
	"Set the contents of the receiver"
	| selected |
	newContents == self ifTrue:[^self].
	selected := self selection.
	self removeAll.
	self addAll: newContents.
	self selection: selected.
	^newContents! !

!CCollection methodsFor: 'private' stamp: 'ar 1/9/2008 17:57'!
contentsChanged: aSymbol with: firstIndex with: lastIndex
	self signalEvent: (ScriptCollectionEvent new 
		setSender: self 
		selector: #changed 
		arguments: (Array with: aSymbol with: firstIndex with: lastIndex)).
! !

!CCollection methodsFor: 'copying' stamp: 'ar 3/17/2004 02:26'!
copyFrom: firstIndex to: lastIndex
	"Answer a copy of a subset of the receiver, starting from element at 
	index start until element at index stop."
	| newCollection |
	newCollection := self species new.
	firstIndex to: lastIndex do:[:i| newCollection add: (self at: i)].
	^newCollection! !

!CCollection methodsFor: 'accessing' stamp: 'ar 3/11/2004 14:12'!
cursor
	^cursor! !

!CCollection methodsFor: 'accessing' stamp: 'ar 9/8/2004 16:21'!
cursor: aValue
	| oldCursor |
	cursor = aValue ifTrue:[^aValue].
	oldCursor := cursor.
	cursor := (aValue min: self size max: 0).
	self signalChanged: #cursorChanged from: oldCursor to: cursor.
	^aValue! !

!CCollection methodsFor: 'accessing' stamp: 'ar 6/28/2003 20:55'!
defaultElement
	^nil! !

!CCollection methodsFor: 'enumerating' stamp: 'ar 2/22/2004 01:55'!
do: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument."
	1 to: values size do:[:i| aBlock value: (values at: i)].! !

!CCollection methodsFor: 'searching' stamp: 'ar 3/13/2004 03:17'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement after start
	within the receiver. If the receiver does not contain anElement, 
	answer the 	result of evaluating the argument, exceptionBlock."
	start to: self size do:[:index |
		((self at: index) equals: anElement) ifTrue: [^ index]].
	^ exceptionBlock value! !

!CCollection methodsFor: 'initialize' stamp: 'ar 3/16/2004 12:38'!
initialize
	super initialize.
	values := #().
	cursor := 0.! !

!CCollection methodsFor: 'enumerating' stamp: 'ar 3/16/2004 11:07'!
keysAndValuesDo: aBlock
	"Evaluate aBlock with each of the receiver's keys and values"
	1 to: values size do:[:i| aBlock value: i value: (values at: i)].! !

!CCollection methodsFor: 'serializing' stamp: 'ar 3/22/2004 02:37'!
readFixedFieldsFrom: deSerializer
	values := deSerializer readField: #values.
	cursor := deSerializer readField: #cursor.
	^super readFixedFieldsFrom: deSerializer.! !

!CCollection methodsFor: 'removing' stamp: 'ar 3/24/2004 20:54'!
removeAll
	"Remove all elements from the receiver"
	| sz |
	sz := values size.
	values := #().
	sz > 0 ifTrue:[self contentsChanged: #remove with: 1 with: sz].
	self cursor: 0.! !

!CCollection methodsFor: 'removing' stamp: 'ar 3/13/2005 11:16'!
removeFrom: firstIndex to: lastIndex
	"Remove all the elements from firstIndex to lastIndex"
	| sz oldData newData |
	sz := lastIndex - firstIndex+1.
	sz <= 0 ifTrue:[^#()].
	oldData := values copyFrom: firstIndex to: lastIndex.
	newData := Array new: values size - sz.
	newData replaceFrom: 1 to: firstIndex-1 with: values startingAt: 1.
	newData replaceFrom: firstIndex to: newData size with: values startingAt: lastIndex+1.
	values := newData.
	self contentsChanged: #remove with: firstIndex with: lastIndex.
	cursor < firstIndex ifFalse:[
		cursor <= lastIndex
			ifTrue:[self cursor: 0]
			ifFalse:[self cursor: cursor - sz]].
	^oldData! !

!CCollection methodsFor: 'copying' stamp: 'ar 3/24/2004 20:54'!
replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart
	values replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart.
	self contentsChanged: #replace with: firstIndex with: lastIndex.! !

!CCollection methodsFor: 'enumerating' stamp: 'ar 3/16/2004 17:20'!
reverseDo: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument."
	values size to: 1 by: -1 do:[:i| aBlock value: (values at: i)].! !

!CCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 11:02'!
size
	"Answer the number of elements in the receiver"
	^values size! !

!CCollection methodsFor: 'sorting' stamp: 'bf 9/30/2004 13:42'!
sort: aSortBlock 
	"Sort this collection using aSortBlock. The block should take two arguments and return true if the first element should preceed the second one."

	values sort: aSortBlock! !

!CCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 12:39'!
wrapCursor
	^self propertyValueAt: #wrapCursor ifAbsent:[false]! !

!CCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 12:39'!
wrapCursor: aBool
	^self propertyValueAt: #wrapCursor put: aBool with: #wrapCursorChanged! !

!CCollection methodsFor: 'serializing' stamp: 'ar 3/22/2004 02:37'!
writeFixedFieldsOn: aSerializer
	aSerializer writeField: #values value: values.
	aSerializer writeField: #cursor value: cursor.
	^super writeFixedFieldsOn: aSerializer.! !

CCollectionBase defineFields: '<?xml  version="1.0" ?>
<fields>
	<field virtual="true" toGet="cursor" toSet="cursor:" changeEvent="cursorChanged">cursor</field>
	<field virtual="true" toGet="wrapCursor" toSet="wrapCursor:" changeEvent="wrapCursorChanged">wrapCursor</field>
	<field toGet="elements" toSet="elements:" changeEvent="elementsChanged">elements</field>
</fields>'!

!CCollectionBase methodsFor: 'adding' stamp: 'ar 3/16/2004 10:46'!
add: newObject 
	"Include newObject as one of the receiver's elements. Answer newObject."
	^self addLast: newObject! !

!CCollectionBase methodsFor: 'adding' stamp: 'ar 12/19/2004 11:54'!
add: newObject after: oldObject
	"Insert newObject after oldObject"
	^self add: newObject at: (self indexOf: oldObject)+1! !

!CCollectionBase methodsFor: 'adding' stamp: 'ar 3/16/2004 14:34'!
add: anObject at: anIndex
	"Insert the given object at the specified index"
	^elements add: anObject at: anIndex.! !

!CCollectionBase methodsFor: 'adding' stamp: 'ar 3/16/2004 14:38'!
add: newObject before: oldObject				 	 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just preceding oldObject. Answer newObject."
	^self add: newObject at: (self indexOf: oldObject ifAbsent:[1])! !

!CCollectionBase methodsFor: 'adding' stamp: 'ar 3/16/2004 14:32'!
addAll: aCollection 
	"Include all the elements of aCollection as the receiver's elements."
	^self addAll: aCollection at: self size+1! !

!CCollectionBase methodsFor: 'adding' stamp: 'ar 3/16/2004 15:12'!
addAll: aCollection after: oldObject
	"Insert the given object before the specified index"
	^self addAll: aCollection at: (self indexOf: oldObject ifAbsent:[self size]) + 1.! !

!CCollectionBase methodsFor: 'adding' stamp: 'ar 3/16/2004 14:43'!
addAll: aCollection at: anIndex
	"Insert the given object before the specified index"
	^elements addAll: aCollection at: anIndex! !

!CCollectionBase methodsFor: 'adding' stamp: 'ar 3/16/2004 15:12'!
addAll: aCollection before: oldObject
	"Insert the given object before the specified index"
	^self addAll: aCollection at: (self indexOf: oldObject ifAbsent:[1]).! !

!CCollectionBase methodsFor: 'adding' stamp: 'ar 3/16/2004 14:35'!
addFirst: newObject 
	"Include newObject as one of the receiver's elements. Answer newObject. 
	ArrayedCollections cannot respond to this message."
	^self add: newObject at: 1.! !

!CCollectionBase methodsFor: 'adding' stamp: 'ar 3/16/2004 14:35'!
addLast: newObject 
	"Include newObject as one of the receiver's elements. Answer newObject. 
	ArrayedCollections cannot respond to this message."
	self add: newObject at: self size+1.! !

!CCollectionBase methodsFor: 'accessing' stamp: 'tk 11/7/2004 12:48'!
after: anElement
	"Return the object after me or nil"

	^ self at: (self indexOf: anElement)+1 ifAbsent: [nil] ! !

!CCollectionBase methodsFor: 'matching' stamp: 'ar 3/16/2004 10:55'!
all
	"Answer a new all matcher"
	^CAllMatcher on: self! !

!CCollectionBase methodsFor: 'enumerating' stamp: 'ar 3/16/2004 10:54'!
allSatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns false for any element return false.
	Otherwise return true."
	self do: [:each | (aBlock value: each) ifFalse: [^false]].
	^true! !

!CCollectionBase methodsFor: 'matching' stamp: 'ar 3/16/2004 10:55'!
any
	"Answer a new any matcher"
	^CAnyMatcher on: self! !

!CCollectionBase methodsFor: 'enumerating' stamp: 'ar 3/16/2004 10:54'!
anySatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns true for any element return true.
	Otherwise return false."
	self do: [:each | (aBlock value: each) ifTrue: [^ true]].
	^ false! !

!CCollectionBase methodsFor: 'converting' stamp: 'ar 3/22/2004 21:14'!
asArray
	^Array withAll: self! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 15:36'!
at: aKey
	"Answer the element at the given key"
	^elements at: aKey! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 15:36'!
at: aKey ifAbsent: aBlock
	"Answer the element at the given key or evaluate the given block if there is no element with the given key"
	^elements at: aKey ifAbsent: aBlock! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 15:37'!
at: aKey put: anObject
	"Store anObject under the given key in the receiver"
	^elements at: aKey put: anObject! !

!CCollectionBase methodsFor: 'streaming' stamp: 'ar 3/16/2004 10:53'!
atEnd
	"answer whether the receiver is at its end"
	^cursor >= self size! !

!CCollectionBase methodsFor: 'streaming' stamp: 'ar 3/16/2004 10:53'!
atStart
	"answer whether the receiver is at its start"
	^cursor <= 1! !

!CCollectionBase methodsFor: 'enumerating' stamp: 'ar 3/16/2004 15:47'!
collect: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into a collection like the receiver. Answer  
	the new collection."
	^self collect: aBlock species: self species! !

!CCollectionBase methodsFor: 'enumerating' stamp: 'ar 3/16/2004 15:49'!
collect: aBlock species: aClass
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into a collection like aClass. Answer  
	the new collection."
	| newCollection |
	newCollection := aClass new.
	1 to: self size do:[:i | newCollection add: (aBlock value: (self at: i))].
	^ newCollection! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 10:45'!
contentType
	"Answer the type of the elements stored in the receiver"
	^elements contentType! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 11:01'!
contents
	"Answer the contents of the receiver"
	^self copyFrom: 1 to: self size! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 11:01'!
contents: newContents
	"Set the contents of the receiver"
	self removeAll.
	self addAll: newContents.! !

!CCollectionBase methodsFor: 'copying' stamp: 'ar 3/16/2004 10:55'!
copyFrom: firstIndex to: lastIndex
	"Answer a copy of a subset of the receiver, starting from element at 
	index start until element at index stop."
	^elements copyFrom: firstIndex to: lastIndex! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 15:38'!
cursor
	"Answer the receiver's cursor"
	<bewareOf: #cursorChanged>
	^elements cursor! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 15:38'!
cursor: aValue
	"Modifiy the receiver's cursor"
	^elements cursor: aValue! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 17:06'!
defaultElement
	"Answer the default element for the receiver"
	^nil! !

!CCollectionBase methodsFor: 'searching' stamp: 'ar 3/16/2004 10:48'!
detect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Answer the first element for which aBlock evaluates to true."
	^self detect: aBlock ifNone:[nil]! !

!CCollectionBase methodsFor: 'searching' stamp: 'ar 3/16/2004 10:48'!
detect: aBlock ifNone: exceptionBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Answer the first element for which aBlock evaluates to true. If none  
	evaluate to true, then evaluate the argument, exceptionBlock."
	self do: [:each | (aBlock value: each) ifTrue: [^ each]].
	^exceptionBlock value! !

!CCollectionBase methodsFor: 'enumerating' stamp: 'ar 3/16/2004 10:54'!
do: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument."
	^elements do: aBlock! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 10:57'!
elements
	"Answer the elements of the receiver"
	<bewareOf: #elementsChanged>
	^self propertyValueAt: #elements! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 10:57'!
elements: aCollectionClient
	"Answer the elements of the receiver"
	^self propertyValueAt: #elements put: aCollectionClient with: #elementsChanged! !

!CCollectionBase methodsFor: 'private' stamp: 'ar 5/22/2005 17:59'!
explorerContentsDo: aBlock
	1 to: self size do:[:i|
		aBlock value: i printString value: (self at: i).
	].
! !

!CCollectionBase methodsFor: 'searching' stamp: 'ar 3/16/2004 17:22'!
findFirstElement: aBlock
	"Answer my first element for which aBlock evaluates as true."
	self do:[:each| (aBlock value: each) ifTrue:[^each]].
	^self defaultElement! !

!CCollectionBase methodsFor: 'searching' stamp: 'ar 3/16/2004 17:23'!
findFirstIndex: aBlock
	"Answer the index of my first element for which aBlock evaluates as true."
	1 to: self size do:[:i| (aBlock value: (self at: i)) ifTrue:[^i]].
	^0! !

!CCollectionBase methodsFor: 'searching' stamp: 'ar 3/16/2004 17:22'!
findLastElement: aBlock
	"Answer my last element for which aBlock evaluates as true."
	self reverseDo:[:each| (aBlock value: each) ifTrue:[^each]].
	^self defaultElement! !

!CCollectionBase methodsFor: 'searching' stamp: 'ar 3/16/2004 17:24'!
findLastIndex: aBlock
	"Answer the index of my first element for which aBlock evaluates as true."
	self size to: 1 by: -1 do:[:i| (aBlock value: (self at: i)) ifTrue:[^i]].
	^0! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 10:46'!
first
	"Answer the first element of the receiver"
	^self at: 1! !

!CCollectionBase methodsFor: 'accessing' stamp: 'tk 11/2/2004 18:23'!
fourth
	^ self at: 4! !

!CCollectionBase methodsFor: 'testing' stamp: 'ar 1/14/2009 17:23'!
ifEmpty: aBlock
	"Evaluate the block if I'm empty"

	self isEmpty ifTrue: [^ aBlock value ]! !

!CCollectionBase methodsFor: 'testing' stamp: 'ar 1/14/2009 17:24'!
ifNotEmpty: aBlock
	"Evaluate the given block unless the receiver is empty.

      If the block has an argument, eval with the receiver as its argument,
      but it might be better to use ifNotEmptyDo: to make the code easier to
      understand"

	self isEmpty ifFalse: [^ aBlock valueWithPossibleArgument: self].
! !

!CCollectionBase methodsFor: 'testing' stamp: 'ar 4/13/2005 01:00'!
includes: anObject 
	"Answer whether anObject is one of the receiver's elements."
	^ self anySatisfy: [:each | each = anObject]! !

!CCollectionBase methodsFor: 'testing' stamp: 'ar 4/13/2005 01:00'!
includesAllOf: aCollection 
	"Answer whether all the elements of aCollection are in the receiver."
	aCollection do: [:elem | (self includes: elem) ifFalse: [^ false]].
	^ true! !

!CCollectionBase methodsFor: 'testing' stamp: 'ar 4/13/2005 01:00'!
includesAnyOf: aCollection 
	"Answer whether any element of aCollection is one of the receiver's elements."
	aCollection do: [:elem | (self includes: elem) ifTrue: [^ true]].
	^ false! !

!CCollectionBase methodsFor: 'searching' stamp: 'ar 3/16/2004 10:52'!
indexOf: anObject
	"Answer the index of the given object within the receiver"
	^self indexOf: anObject ifAbsent:[0]! !

!CCollectionBase methodsFor: 'searching' stamp: 'ar 3/16/2004 10:52'!
indexOf: anElement ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the 
	result of evaluating the argument, exceptionBlock."
	^self indexOf: anElement startingAt: 1 ifAbsent: exceptionBlock! !

!CCollectionBase methodsFor: 'searching' stamp: 'ar 3/16/2004 10:52'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement after start
	within the receiver. If the receiver does not contain anElement, 
	answer the 	result of evaluating the argument, exceptionBlock."
	^elements indexOf: anElement startingAt: start ifAbsent: exceptionBlock! !

!CCollectionBase methodsFor: 'testing' stamp: 'ar 4/13/2005 01:00'!
isCollection
	"Return true if the receiver is some sort of Collection and responds to basic collection messages such as #size and #do:"
	^true! !

!CCollectionBase methodsFor: 'testing' stamp: 'ar 3/16/2004 10:48'!
isEmpty
	"Answer whether the receiver contains any elements."
	^self size = 0! !

!CCollectionBase methodsFor: 'enumerating' stamp: 'ar 3/16/2004 10:55'!
keysAndValuesDo: aBlock
	"Evaluate aBlock with each of the receiver's keys and values"
	^elements keysAndValuesDo: aBlock! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 10:46'!
last
	"Answer the last element in the receiver"
	^self at: self size! !

!CCollectionBase methodsFor: 'streaming' stamp: 'bf 5/10/2004 17:44'!
next
	"go to the next element in the receiver"
	self isEmpty ifTrue:[^self defaultElement].
	self atEnd ifTrue:[
		wrapCursor ifFalse:[^self defaultElement].
		^self at: (cursor := 1).
	].
	^self at: (cursor := cursor  + 1)! !

!CCollectionBase methodsFor: 'testing' stamp: 'ar 4/13/2005 01:00'!
notEmpty
	"Answer whether the receiver contains any elements."
	^ self isEmpty not! !

!CCollectionBase methodsFor: 'streaming' stamp: 'ar 3/16/2004 10:53'!
peek
	"answer the element at the receiver's cursor"
	(cursor between: 1 and: self size) ifFalse:[^self defaultElement].
	^self at: cursor! !

!CCollectionBase methodsFor: 'streaming' stamp: 'bf 5/10/2004 17:45'!
prev
	"go to the previous element in the receiver"
	self isEmpty ifTrue:[^self defaultElement].
	self atStart ifTrue:[
		wrapCursor ifFalse:[^self defaultElement].
		^self at: (cursor := self size).
	].
	^self at: (cursor := cursor - 1)! !

!CCollectionBase methodsFor: 'printing' stamp: 'ar 3/16/2004 16:58'!
printElementsOn: aStream
	aStream nextPut: $(.
	self do: [:element | aStream print: element; space].
	self isEmpty ifFalse: [aStream skip: -1].
	aStream nextPut: $)! !

!CCollectionBase methodsFor: 'printing' stamp: 'ar 3/16/2004 16:58'!
printNameOn: aStream
	super printOn: aStream! !

!CCollectionBase methodsFor: 'printing' stamp: 'ar 3/16/2004 16:58'!
printOn: aStream 
	"Append a sequence of characters that identify the receiver to aStream."
	self printNameOn: aStream.
	self printElementsOn: aStream! !

!CCollectionBase methodsFor: 'removing' stamp: 'ar 3/16/2004 10:47'!
remove: oldObject 
	"Remove oldObject from the receiver's elements. Answer oldObject 
	unless no element is equal to oldObject, in which case, raise an error.
	ArrayedCollections cannot respond to this message."
	^ self remove: oldObject ifAbsent: [nil]! !

!CCollectionBase methodsFor: 'removing' stamp: 'ar 3/16/2004 10:48'!
remove: oldObject ifAbsent: aBlock 
	"Remove oldObject from the receiver's elements. If several of the 
	elements are equal to oldObject, only one is removed. If no element is 
	equal to oldObject, answer the result of evaluating anExceptionBlock. 
	Otherwise, answer the argument, oldObject. ArrayedCollections cannot 
	respond to this message."
	^self removeAt: (self indexOf: oldObject ifAbsent:[^aBlock value])! !

!CCollectionBase methodsFor: 'removing' stamp: 'ar 3/16/2004 10:48'!
removeAll
	"Remove all elements from the receiver"
	elements removeAll.! !

!CCollectionBase methodsFor: 'removing' stamp: 'ar 3/16/2004 10:48'!
removeAll: aCollection 
	"Remove each element of aCollection from the receiver. If successful for 
	each, answer aCollection. Otherwise create an error notification.
	ArrayedCollections cannot respond to this message."
	aCollection do: [:each | self remove: each].
! !

!CCollectionBase methodsFor: 'removing' stamp: 'ar 4/13/2005 00:58'!
removeAt: index
	"Remove an element from the receiver"
	^(self removeFrom: index to: index) at: 1! !

!CCollectionBase methodsFor: 'removing' stamp: 'ar 4/22/2005 10:38'!
removeFirst
	"Remove the first element of the receiver"
	^self removeAt: 1! !

!CCollectionBase methodsFor: 'removing' stamp: 'ar 3/16/2004 10:48'!
removeFrom: firstIndex to: lastIndex
	"Remove all the elements from firstIndex to lastIndex"
	^elements removeFrom: firstIndex to: lastIndex! !

!CCollectionBase methodsFor: 'removing' stamp: 'ar 3/16/2004 10:48'!
removeLast
	"Remove the last element from the receiver"
	^self removeAt: self size! !

!CCollectionBase methodsFor: 'replacing' stamp: 'ar 3/16/2004 12:49'!
replace: oldObject with: newObject
	"Replace oldObject with newObject"
	| index |
	oldObject == newObject ifTrue:[^self].
	newObject ifNil:[^self remove: oldObject].
	oldObject ifNil:[^self add: newObject].
	index := self indexOf: oldObject ifAbsent:[^nil].
	^self replaceFrom: index to: index with: (Array with: newObject)! !

!CCollectionBase methodsFor: 'replacing' stamp: 'ar 3/16/2004 10:53'!
replaceFrom: firstIndex to: lastIndex with: aCollection
	"Replace the elements in the receiver"
	^self replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: 1! !

!CCollectionBase methodsFor: 'replacing' stamp: 'ar 3/16/2004 10:53'!
replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart
	^elements replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart! !

!CCollectionBase methodsFor: 'enumerating' stamp: 'ar 3/16/2004 17:20'!
reverseDo: aBlock
	^elements do: aBlock! !

!CCollectionBase methodsFor: 'copying' stamp: 'ar 3/29/2005 23:53'!
reversed
	"Answer a copy of the receiver with element order reversed."
	| newCollection |
	newCollection := self species new.
	self reverseDo:[:each| newCollection add: each].
	^newCollection! !

!CCollectionBase methodsFor: 'accessing' stamp: 'tk 11/2/2004 18:23'!
second
	^ self at: 2! !

!CCollectionBase methodsFor: 'enumerating' stamp: 'ar 3/16/2004 10:55'!
select: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new collection like the receiver, only those elements for 
	which aBlock evaluates to true. Answer the new collection."
	| newCollection |
	newCollection := self species new.
	self do: [:each | (aBlock value: each) ifTrue: [newCollection add: each]].
	^newCollection! !

!CCollectionBase methodsFor: 'enumerating' stamp: 'ar 3/16/2004 15:48'!
select: aBlock species: aClass
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new collection like aClass, only those elements for 
	which aBlock evaluates to true. Answer the new collection."
	| newCollection |
	newCollection := aClass new.
	self do: [:each | (aBlock value: each) ifTrue: [newCollection add: each]].
	^newCollection! !

!CCollectionBase methodsFor: 'accessing' stamp: 'bf 5/4/2004 17:00'!
selection
	"Answer the currently selected object - that is the object at the cursor's position"
	<bewareOf: #cursorChanged>
	(self cursor between: 1 and: self size) ifFalse:[^self defaultElement].
	^self at: cursor! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 10:46'!
selection: anObject
	"Set the currently selected object - that is move the cursor's position to point to anObject"
	cursor := self indexOf: anObject.
	^anObject! !

!CCollectionBase methodsFor: 'accessing' stamp: 'bf 5/4/2004 17:00'!
selectionIndex
	"a convenient alias to the cursor"
	<bewareOf: #cursorChanged>
	^cursor! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 10:46'!
selectionIndex: newIndex
	"a convenient alias to the cursor"
	^cursor := newIndex! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 10:46'!
size
	"Delegate to our elements"
	^elements size! !

!CCollectionBase methodsFor: 'sorting' stamp: 'bf 9/30/2004 13:44'!
sort
	"Sort this collection into ascending order using the '<=' operator."

	self sort: [:a :b | a <= b]! !

!CCollectionBase methodsFor: 'sorting' stamp: 'bf 9/30/2004 13:45'!
sort: aSortBlock
	"Sort this collection using aSortBlock. The block should take two arguments and return true if the first element should preceed the second one."

	elements sort: aSortBlock! !

!CCollectionBase methodsFor: 'accessing' stamp: 'ar 3/16/2004 16:57'!
species
	^CCollection! !

!CCollectionBase methodsFor: 'accessing' stamp: 'tk 11/2/2004 18:23'!
third
	^ self at: 3! !

!CCollectionBase methodsFor: 'streaming' stamp: 'sw 5/17/2005 04:48'!
wrapCursor
	"Delegate wrapCursor to our elements.  Answer true (vacuously) if elements not yet initialized."
	<bewareOf: #wrapCursorChanged>
	^elements ifNil: [true] ifNotNil: [elements wrapCursor]! !

!CCollectionBase methodsFor: 'streaming' stamp: 'ar 3/16/2004 10:54'!
wrapCursor: aBool
	"Delegate wrapCursor to our elements"
	^elements wrapCursor: aBool! !

CCollectionBase subclass: #CPlayerCollection
	instanceVariableNames: 'costume'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Costumes'!

CPlayerCollection defineFields: '<?xml  version="1.0" ?>
<fields>
	<field instanceVariable="true" toGet="costume" toSet="costume:">costume</field>
</fields>'!

!CPlayerCollection class methodsFor: 'instance creation' stamp: 'ar 3/16/2004 12:35'!
costume: aCostume
	^(self new) costume: aCostume; yourself! !

!CPlayerCollection methodsFor: 'adding' stamp: 'tk 7/14/2004 11:25'!
add: anObject at: index
	^costume add: anObject asPrimCostume at: index! !

!CPlayerCollection methodsFor: 'adding' stamp: 'tk 7/14/2004 11:25'!
addAll: aCollection at: index
	^costume addAll: (aCollection collect:[:each| each asPrimCostume]) at: index! !

!CPlayerCollection methodsFor: 'accessing' stamp: 'ar 3/29/2004 22:37'!
at: aKey
	^(costume at: aKey) asPlayer! !

!CPlayerCollection methodsFor: 'accessing' stamp: 'tk 11/7/2004 12:57'!
at: aKey ifAbsent: aBlock
	^(costume contents at: aKey ifAbsent:[^aBlock value]) asPlayer! !

!CPlayerCollection methodsFor: 'accessing' stamp: 'tk 7/14/2004 11:26'!
at: index put: anObject
	^costume replaceFrom: index to: index with: {anObject asPrimCostume}! !

!CPlayerCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 11:46'!
contentType
	^CPlayer! !

!CPlayerCollection methodsFor: 'copying' stamp: 'bgf 1/10/2009 23:25'!
copyFrom: firstIndex to: lastIndex
	^(costume contents copyFrom: firstIndex to: lastIndex) collect:[:each| each player]! !

!CPlayerCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 12:36'!
costume
	^costume! !

!CPlayerCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 12:36'!
costume: aCostume
	costume := aCostume.! !

!CPlayerCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 17:06'!
cursor
	^self propertyValueAt: #cursor ifAbsent:[0]! !

!CPlayerCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 17:42'!
cursor: aNumber
	^self propertyValueAt: #cursor put: (aNumber min: self size max: 0) with: #cursorChanged! !

!CPlayerCollection methodsFor: 'enumerating' stamp: 'ar 3/29/2005 23:52'!
do: aBlock
	^costume do:[:each| aBlock value: each asPlayer]! !

!CPlayerCollection methodsFor: 'searching' stamp: 'tk 7/14/2004 11:26'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement after start
	within the receiver. If the receiver does not contain anElement, 
	answer the 	result of evaluating the argument, exceptionBlock.
	Note: There is no point in searching for anything but players
	since all others would have their costumes created on the fly."

	anElement isPlayer ifFalse:[^exceptionBlock value].		
	^costume indexOf: anElement asPrimCostume startingAt: start ifAbsent: exceptionBlock! !

!CPlayerCollection methodsFor: 'enumerating' stamp: 'ar 3/29/2004 22:37'!
keysAndValuesDo: aBlock
	^costume keysAndValuesDo:[:aKey :aValue| aBlock value: aKey value: aValue asPlayer]! !

!CPlayerCollection methodsFor: 'removing' stamp: 'ar 3/16/2004 12:17'!
removeAll
	^costume removeAll! !

!CPlayerCollection methodsFor: 'removing' stamp: 'ar 3/13/2005 11:17'!
removeFrom: firstIndex to: lastIndex
	^(costume removeFrom: firstIndex to: lastIndex) collect:[:each| each asPlayer]! !

!CPlayerCollection methodsFor: 'copying' stamp: 'tk 7/14/2004 11:26'!
replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart
	^costume 
		replaceFrom: firstIndex 
		to: lastIndex 
		with: (aCollection collect:[:each| each asPrimCostume])
		startingAt: replStart! !

!CPlayerCollection methodsFor: 'enumerating' stamp: 'ar 3/29/2004 22:37'!
reverseDo: aBlock
	^costume reverseDo:[:each| aBlock value: each asPlayer]! !

!CPlayerCollection methodsFor: 'accessing' stamp: 'ar 3/16/2004 12:16'!
size
	^costume size! !

CObject subclass: #CCollectionClient
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Collections'!

!CCollectionClient commentStamp: 'ar 3/15/2004 15:52' prior: 0!
CCollectionClient defines the implementation side of a full collection. It provides a much smaller interface which has to be implemented for 'types' of collections.!

CCollectionClient subclass: #CArrayedCollectionClient
	instanceVariableNames: 'values'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Collections'!

CArrayedCollectionClient defineFields: '<?xml  version="1.0" ?>
<fields>
	<field instanceVariable="true">values</field>
	<field toGet="cursor" toSet="cursor:" changeEvent="cursorChanged">cursor</field>
	<field toGet="wrapCursor" toSet="wrapCursor:" changeEvent="wrapCursorChanged">wrapCursor</field>
</fields>'!

!CArrayedCollectionClient methodsFor: 'adding' stamp: 'ar 1/9/2008 17:55'!
add: anObject at: index
	"Insert the given object before the specified index"
	| newData |
	newData := Array new: values size+1.
	newData replaceFrom: 1 to: index-1 with: values startingAt: 1.
	newData at: index put: anObject.
	newData replaceFrom: index+1 to: newData size with: values startingAt: index.
	values := newData.
	self signalEvent: (ScriptCollectionEvent new 
		setSender: self 
		selector: #changed 
		arguments: (Array with: #insert with: index with: index)).
	index <= cursor ifTrue:[self cursor: cursor + 1].
	^anObject! !

!CArrayedCollectionClient methodsFor: 'adding' stamp: 'ar 1/9/2008 17:55'!
addAll: aCollection at: index
	"Insert the given object before the specified index"
	| sz newData |
	sz := aCollection size.
	newData := Array new: values size + sz.
	newData replaceFrom: 1 to: index-1 with: values startingAt: 1.
	newData replaceFrom: index to: index+sz-1 with: aCollection startingAt: 1.
	newData replaceFrom: index+sz to: newData size with: values startingAt: index.
	values := newData.
	self signalEvent: (ScriptCollectionEvent new 
		setSender: self 
		selector: #changed 
		arguments: (Array with: #insert with: index with: index+sz-1)).
	index <= cursor ifTrue:[self cursor: cursor + sz].
	^aCollection! !

!CArrayedCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:54'!
at: aKey
	"Answer the element at the given key"
	^values at: aKey! !

!CArrayedCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 16:19'!
at: aKey ifAbsent: aBlock
	"Answer the element at the given key or evaluate the given block 
	if there is no element with the given key"
	(aKey < 1 or:[aKey > values size]) ifTrue:[^aBlock value].
	^self at: aKey! !

!CArrayedCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:55'!
at: aKey put: anObject
	"Store anObject under the given key in the receiver"
	values at: aKey put: anObject! !

!CArrayedCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 16:19'!
contentType
	"Actually, answering nil is incorrect since values might be homogenous... 
	but what the hey..."
	^nil! !

!CArrayedCollectionClient methodsFor: 'removing' stamp: 'ar 3/15/2004 16:00'!
copyFrom: firstIndex to: lastIndex
	"Answer a copy of a subset of the receiver, starting from element at 
	index start until element at index stop."
	| newCollection sz |
	sz := lastIndex - firstIndex + 1.
	newCollection := self species new: sz.
	newCollection replaceFrom: 1 to: sz with: values startingAt: firstIndex.
	^newCollection! !

!CArrayedCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 16:18'!
cursor
	"Answer the cursor of the receiver"
	<bewareOf: #cursorChanged>
	^self propertyValueAt: #cursor ifAbsent:[0]! !

!CArrayedCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 16:18'!
cursor: aValue
	"Modify the receiver's cursor"
	^self propertyValueAt: #cursor put: (aValue min: self size max: 0) with: #cursorChanged! !

!CArrayedCollectionClient methodsFor: 'enumerating' stamp: 'ar 3/15/2004 16:01'!
do: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument."
	1 to: self size do:[:i| aBlock value: (values at: i)]! !

!CArrayedCollectionClient methodsFor: 'initialize' stamp: 'ar 3/15/2004 16:17'!
initialize
	"Initialize the receiver"
	super initialize.
	values := #().! !

!CArrayedCollectionClient methodsFor: 'initialize' stamp: 'ar 3/15/2004 16:12'!
initialize: numElements
	"Initialize the receiver to a capacity of numElements"
	values := Array new: numElements.! !

!CArrayedCollectionClient methodsFor: 'enumerating' stamp: 'ar 3/15/2004 16:01'!
keysAndValuesDo: aBlock
	"Evaluate aBlock with all of the receiver's keys and values"
	1 to: self size do:[:i| aBlock value: i value: (values at: i)]! !

!CArrayedCollectionClient methodsFor: 'removing' stamp: 'ar 1/9/2008 17:56'!
removeAll
	| sz |
	sz := values size.
	values := #().
	sz > 0 ifTrue:[self signalEvent: (ScriptCollectionEvent new 
		setSender: self 
		selector: #changed 
		arguments: (Array with: #remove with: 1 with: sz))].
	self cursor: 0.! !

!CArrayedCollectionClient methodsFor: 'removing' stamp: 'ar 1/9/2008 17:56'!
removeFrom: firstIndex to: lastIndex
	| sz oldData newData |
	sz := lastIndex - firstIndex+1.
	sz <= 0 ifTrue:[^#()].
	oldData := values copyFrom: firstIndex to: lastIndex.
	newData := Array new: values size - sz.
	newData replaceFrom: 1 to: firstIndex-1 with: values startingAt: 1.
	newData replaceFrom: firstIndex to: newData size with: values startingAt: lastIndex+1.
	values := newData.
	self signalEvent: (ScriptCollectionEvent new 
		setSender: self 
		selector: #changed 
		arguments: (Array with: #remove with: firstIndex with: lastIndex)).
	cursor < firstIndex ifFalse:[
		cursor <= lastIndex
			ifTrue:[self cursor: 0]
			ifFalse:[self cursor: cursor - sz]].
	^oldData! !

!CArrayedCollectionClient methodsFor: 'removing' stamp: 'ar 1/9/2008 17:57'!
replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart
	values replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart.
	self signalEvent: (ScriptCollectionEvent new 
		setSender: self 
		selector: #changed 
		arguments: (Array with: #replace with: firstIndex with: lastIndex)).! !

!CArrayedCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:57'!
size
	"Answer the number of elements in the receiver"
	^values size! !

!CArrayedCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 16:19'!
wrapCursor
	"Answer the wrapCursor of the receiver"
	<bewareOf: #wrapCursorChanged>
	^self propertyValueAt: #wrapCursor ifAbsent:[false]! !

!CArrayedCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:58'!
wrapCursor: aValue
	"Modify the receiver's wrapCursor"
	^self propertyValueAt: #wrapCursor put: aValue with: #wrapCursorChanged! !

CCollectionClient defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!CCollectionClient methodsFor: 'adding' stamp: 'ar 3/15/2004 15:44'!
add: anObject at: anIndex
	"Insert the given object at the specified index"
	^self addAll: (Array with: anObject) at: anIndex! !

!CCollectionClient methodsFor: 'adding' stamp: 'ar 3/15/2004 15:44'!
addAll: aCollection at: anIndex
	"Insert the given collection before the specified index"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:41'!
at: aKey
	"Answer the element at the given key"
	^self subclassResponsibility
! !

!CCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:42'!
at: aKey ifAbsent: aBlock
	"Answer the element at the given key or evaluate the given block if there is no element with the given key"
	^self subclassResponsibility
! !

!CCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:42'!
at: aKey put: anObject
	"Store anObject under the given key in the receiver"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:52'!
contentType
	"Answer the type of the elements stored in the receiver"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'copying' stamp: 'ar 3/15/2004 15:45'!
copyFrom: firstIndex to: lastIndex
	"Answer a copy of a subset of the receiver, starting from element at 
	index start until element at index stop."
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:43'!
cursor
	"Answer the cursor"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:43'!
cursor: aValue
	"Set the cursor to the given value"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'enumerating' stamp: 'ar 3/15/2004 15:46'!
do: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument."
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'initialize' stamp: 'ar 3/15/2004 16:11'!
initialize: numElements
	"Initialize the receiver to a capacity of numElements"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'enumerating' stamp: 'ar 3/15/2004 15:46'!
keysAndValuesDo: aBlock
	"Evaluate aBlock with all of the receiver's keys and values"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'removing' stamp: 'ar 3/15/2004 15:45'!
removeAll
	"Remove all the elements from the receiver"
	^self removeFrom: 1 to: self size! !

!CCollectionClient methodsFor: 'removing' stamp: 'ar 3/15/2004 15:45'!
removeFrom: firstIndex to: lastIndex
	"Remove all the elements from firstIndex to lastIndex"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'removing' stamp: 'ar 3/15/2004 15:45'!
replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart
	"Replace the elements in the receiver with the given collection"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:43'!
size
	"Answer the number of elements in the receiver"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:43'!
wrapCursor
	"Answer whether the cursor should wrap when streaming"
	^self subclassResponsibility! !

!CCollectionClient methodsFor: 'accessing' stamp: 'ar 3/15/2004 15:44'!
wrapCursor: aBool
	"Indicate whether the cursor should wrap when streaming"
	^self subclassResponsibility! !

CObject subclass: #CDirectoryEntry
	instanceVariableNames: ''
	classVariableNames: 'DefaultIcon IconCache'
	poolDictionaries: ''
	category: 'Tweak-Basic-Support'!

!CDirectoryEntry commentStamp: 'ar 3/14/2004 02:28' prior: 0!
This ridiculous class exists for the mere purpose of the unavailability of the directory from a directory entry. This is so stupid I don't even know where to start...!

CDirectoryEntry defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="directory" toSet="directory:" changeEvent="directoryChanged">directory</field>
	<field toGet="entry" toSet="entry:" changeEvent="entryChanged">entry</field>
	<field toGet="icon" toSet="icon:" changeEvent="iconChanged">icon</field>
</fields>'!

!CDirectoryEntry class methodsFor: 'class initialization' stamp: 'ar 3/15/2004 01:46'!
initialize
	"CDirectoryEntry initialize"
	Smalltalk addToShutDownList: self.! !

!CDirectoryEntry class methodsFor: 'class initialization' stamp: 'ar 3/15/2004 01:46'!
shutDown
	"Flush the icon cache"
	IconCache := nil.! !

!CDirectoryEntry methodsFor: 'utilities' stamp: 'ar 3/14/2004 03:18'!
appIcon
	^self defaultIcon! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:30'!
creationTime
	^entry creationTime! !

!CDirectoryEntry methodsFor: 'utilities' stamp: 'ar 3/14/2004 03:18'!
defaultIcon
	^DefaultIcon ifNil:[DefaultIcon := Form extent: 64@64 depth: 8]! !

!CDirectoryEntry methodsFor: 'utilities' stamp: 'ar 3/14/2004 03:18'!
defaultImageIcon
	^self defaultIcon! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:28'!
directory
	"Answer the directory of the receiver"
	<bewareOf: #directoryChanged>
	^self propertyValueAt: #directory! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:28'!
directory: aValue
	"Modify the receiver's directory"
	^self propertyValueAt: #directory put: aValue with: #directoryChanged! !

!CDirectoryEntry methodsFor: 'utilities' stamp: 'ar 3/14/2004 03:19'!
directoryIcon
	^self defaultIcon! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:28'!
entry
	"Answer the entry of the receiver"
	<bewareOf: #entryChanged>
	^self propertyValueAt: #entry! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:28'!
entry: aValue
	"Modify the receiver's entry"
	^self propertyValueAt: #entry put: aValue with: #entryChanged! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:33'!
extension
	self isDirectory ifTrue:[^''].
	^entry name 
		copyFrom: (entry name lastIndexOf: directory class extensionDelimiter ifAbsent:[name size]) + 1 
		to: entry name size! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:30'!
fileName
	^entry name! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:31'!
fileSize
	^entry fileSize! !

!CDirectoryEntry methodsFor: 'utilities' stamp: 'ar 3/14/2004 03:17'!
getIcon
	"Retrieve the icon for this entry"
	| types mimeType |
	self isDirectory ifTrue:[^self directoryIcon].
	types := directory mimeTypesFor: name.
	types isEmptyOrNil ifTrue:[^self defaultIcon].
	mimeType := types first.
	('image/gif' = mimeType) ifTrue:[^self imageIcon].
	('image/bmp' = mimeType) ifTrue:[^self imageIcon].
	('image/png' = mimeType) ifTrue:[^self imageIcon].
	('image/jpeg' = mimeType) ifTrue:[^self imageIcon].
	('image/*' match: mimeType) ifTrue:[^self defaultImageIcon].
	('audio/*' match: mimeType) ifTrue:[^self soundIcon].
	('video/*' match: mimeType) ifTrue:[^self videoIcon].
	('application/*' match: mimeType) ifTrue:[^self appIcon].
	('text/*' match: mimeType) ifTrue:[^self textIcon].
	^self defaultIcon! !

!CDirectoryEntry methodsFor: 'utilities' stamp: 'ar 3/14/2004 04:04'!
getImageIcon
	| cacheDir hashName file form newIcon scale extent hash |
	hash := self md5Hash.
	IconCache ifNil:[IconCache := WeakValueDictionary new].
	newIcon := IconCache at: hash ifAbsent:[nil].
	newIcon ifNotNil:[^newIcon].

	cacheDir := FileDirectory default directoryNamed: '.thumbs'.
	cacheDir assureExistence.

	"Stage 1: Try to find image in cache"
	hashName := hash hex allButFirst: 3.
	file := [cacheDir readOnlyFileNamed: hashName] 
		on: FileDoesNotExistException do:[:ex| ex return: nil].
	file ifNotNil:[
		[
			newIcon := [Form fromBinaryStream: file] on: Error do:[:ex| ex return: nil].
		] ensure:[file close].
		newIcon ifNotNil:[IconCache at: hash put: newIcon].
		^newIcon ifNil:[self defaultImageIcon]
	].

	"Stage 2: Create a new icon"
	file := [directory readOnlyFileNamed: self fileName] on: Error do:[:ex| ex return: nil].
	file ifNil:[^self defaultImageIcon].
	[
		[form := Form fromBinaryStream: file] ensure:[file close].
		newIcon := Form extent: 64@64 depth: form depth.
		scale := 64.0 / (form width max: form height) min: 8.
		extent := (form extent * scale) truncated.
		form drawOn: (CTransformCanvas on: newIcon) in: (newIcon extent - extent // 2 extent: extent).
	] on: Error do:[:ex| ex return].

	"If reading the image failed create a new (empty) cache entry 
	to indicate broken file and avoid future read attempts"
	file := cacheDir forceNewFileNamed: hashName.
	[
		newIcon ifNotNil:[PNGReadWriter putForm: newIcon onStream: file].
	] ensure:[file close].
	IconCache at: hash put: newIcon.
	^newIcon ifNil:[self defaultImageIcon]! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 03:21'!
icon
	"Answer the icon of the receiver"
	<bewareOf: #iconChanged>
	^(self propertyValueAt: #icon) ifNil:[
		icon := self getIcon.
		icon]! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 03:17'!
icon: aValue
	"Modify the receiver's icon"
	^self propertyValueAt: #icon put: aValue with: #iconChanged! !

!CDirectoryEntry methodsFor: 'utilities' stamp: 'ar 11/13/2006 18:10'!
imageIcon
	^([self getImageIcon] on: Error do:[:ex| ex return: nil]) ifNil:[DefaultIcon]! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:31'!
isDirectory
	^entry isDirectory! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/25/2004 19:27'!
md5Hash
	^CMD5Hasher hashMessage: entry name, entry fileSize, entry modificationTime.! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 2/16/2006 17:39'!
mimeTypes
	self isDirectory ifTrue:[^#()].
	^(directory mimeTypesFor: name) ifNil:[#()]
! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:31'!
modificationTime
	^entry modificationTime! !

!CDirectoryEntry methodsFor: 'accessing' stamp: 'ar 3/14/2004 02:31'!
name
	^entry name! !

!CDirectoryEntry methodsFor: 'utilities' stamp: 'ar 3/14/2004 03:19'!
soundIcon
	^self defaultIcon! !

!CDirectoryEntry methodsFor: 'utilities' stamp: 'ar 3/14/2004 03:19'!
textIcon
	^self defaultIcon! !

!CDirectoryEntry methodsFor: 'utilities' stamp: 'ar 3/14/2004 03:19'!
videoIcon
	^self defaultIcon! !

CObject subclass: #CEventBehavior
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Kernel'!

!CEventBehavior commentStamp: 'ar 5/3/2004 22:44' prior: 0!
An event behavior describes a set of "standalone scripts" which are triggered by events. Event behaviors can be attached to arbitrary players and are an easy way to generalize (event-triggered) behaviors to generic players.

See CExampleBehavior for a simple example of how to use a behavior.

Instance variables:
	target	<CObject>		The event source of the object.
!

CEventBehavior defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="target" toSet="target:" changeEvent="targetChanged">target</field>
</fields>'!

!CEventBehavior class methodsFor: 'instance creation' stamp: 'ar 4/30/2004 17:53'!
attachTo: anObject
	^self new attachTo: anObject! !

!CEventBehavior methodsFor: 'initialize' stamp: 'ar 4/30/2004 17:51'!
attachTo: anObject
	"Attach the receiver to anObject"
	target := anObject.
	anObject attachBehavior: self.! !

!CEventBehavior methodsFor: 'initialize' stamp: 'ar 4/30/2004 17:52'!
detach
	"Detach the receiver from its target"
	| oldTarget |
	target ifNil:[^self]. "I am not attached"
	oldTarget := target.
	target := nil.
	oldTarget detachBehavior: self.! !

!CEventBehavior methodsFor: 'accessing' stamp: 'ar 4/30/2004 17:42'!
target
	"Answer the target of the receiver"
	<bewareOf: #targetChanged>
	^self propertyValueAt: #target! !

!CEventBehavior methodsFor: 'accessing' stamp: 'ar 4/30/2004 17:42'!
target: aValue
	"Modify the receiver's target"
	^self propertyValueAt: #target put: aValue with: #targetChanged! !

CEventBehavior subclass: #CExampleBehavior
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Examples'!

!CExampleBehavior commentStamp: 'ar 5/3/2004 22:44' prior: 0!
CExampleBehavior is an example for what an event behavior can do. It will change the alpha value of its target to be translucent when the mouse cursor enters or leaves the target.

To run evaluate the following:

	CExampleBehavior attachTo: CWorkspace open.

!

CExampleBehavior defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!CExampleBehavior methodsFor: 'events' stamp: 'ar 4/30/2004 17:57'!
onMouseEnter
	"The mouse goes over the target - make it opaque"
	<on: mouseEnter in: target>
	target fillAlpha := 1.
! !

!CExampleBehavior methodsFor: 'events' stamp: 'ar 4/30/2004 17:57'!
onMouseLeave
	"The mouse leaves the target - make it translucent"
	<on: mouseLeave in: target>
	target fillAlpha := 0.5.
! !

!CObject methodsFor: '*Tweak-Basic' stamp: 'ar 1/22/2005 19:06'!
fieldNamed: fieldName
	"answer a reference to the field named fieldName"
	^(CFieldReference new)
		target: self;
		fieldName: fieldName;
		yourself! !

CObject subclass: #CPlayer
	instanceVariableNames: 'myCostume myPlayer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Kernel'!

!CPlayer commentStamp: '<historical>' prior: 0!
A Player is a fundamental kind of graphical object.  
It is an assembly of various "aspects" including:  elements are the objects inside the player.  geometry describes where the player is on the screen.  fill tells how is it filled with color or texture.  border around the object.  text describes how it holds text.  layout is how it arranges its elements.  grid tells what kind of grid lines are inside this object.  shadow describes a drop shadow, if any.  highlights tells if it has any extra graphical highlights.[general]
This Player is wearing a graphic image.  It may be a picture that the author painted.  To modify it, get the halo, and choose 'repaint' from the red handle menu.
	To start a new painting, choose 'Painter' from the Tools menu.[image]
This Player is being used as a text holder.[text]

### Allow Users to see the above explanation ###!

CPlayer subclass: #CCostume
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Kernel'!

!CCostume commentStamp: 'ar 3/25/2004 01:15' prior: 0!
CCostume is a convenience class which can be used as superclass for objects which are intended to be used as costumes for other players. Since "being a costume" is a role rather than a type there is no other significance to CCostume than marking an object so that users know they should think about this object as a costume which has a player. Because of this, CCostume makes #defaultPlayer a responsibility of its subclasses so that these subclasses properly document their requirements.!

CCostume defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!CCostume methodsFor: 'initialize' stamp: 'ar 3/13/2004 00:35'!
defaultColor
	^Color white! !

!CCostume methodsFor: 'initialize' stamp: 'ar 3/7/2004 22:02'!
defaultPlayer
	^self subclassResponsibility! !

!CCostume methodsFor: 'events' stamp: 'ar 3/25/2004 19:51'!
onPlayerChanged
	<on: playerChanged>
! !

!CCostume methodsFor: 'serializing' stamp: 'ar 9/8/2004 16:21'!
postBuildWith: aMaker
	"Now that the player is fully loaded, make sure we're initialized correctly"
	self autoStartScripts.
	self setupCostume.
	"For now, call #makeDefaultCostume so that we keep compatibility"
		self makeDefaultCostume.
	self signalChanged: #playerChanged from: nil to: myPlayer.! !

!CCostume methodsFor: 'serializing' stamp: 'ar 1/22/2005 22:31'!
readFieldsFrom: deSerializer
	"By default, costumes only remember their player since they can be perfectly recreated after the player came back in. Note that the process is a little tricky since the player may not have been fully loaded yet and therefore we cannot initialize the receiver quite yet. This only happens in #postBuildWith: which is invoked once everything is loaded"
	myPlayer := deSerializer readField: #player ifAbsent:[myPlayer].
	myPlayer privateChangeCostumeTo: self.
	self initialize.
	transform := deSerializer readField: #transform ifAbsent:[transform].
	extent := deSerializer readField: #extent ifAbsent:[extent].! !

!CCostume methodsFor: 'initialize' stamp: 'ar 4/30/2004 15:58'!
setupCostume
	self color: self defaultColor.
	self clipping: true.
! !

!CCostume methodsFor: 'serializing' stamp: 'ar 1/22/2005 22:31'!
writeFieldsOn: aSerializer
	"By default, costumes only remember their player since they can be perfectly recreated after the player came back in"
	aSerializer writeField: #player value: player.
	aSerializer writeField: #transform value: transform.
	aSerializer writeField: #extent value: extent.! !

CPlayer subclass: #CEllipsePlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Objects'!

!CEllipsePlayer commentStamp: 'tk 3/1/2005 10:56' prior: 0!
A Ellipse is an oval.  If the height and width are equal, it is a circle.  Like other players, it can be embedded inside another object, and other objects can be embedded in it.[general]


### Allow Users to see the above explanation ###!

CEllipsePlayer defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!CEllipsePlayer methodsFor: 'testing' stamp: 'ar 2/6/2005 19:52'!
containsPoint: aPoint
	| radius other delta xOverY box |
	box := 0@0 extent: extent.
	(box containsPoint: aPoint) ifFalse: [^ false].  "quick elimination"
	(box width = 1 or: [box height = 1])
		ifTrue: [^ true].  "Degenerate case -- code below fails by a bit"

	radius := box height asFloat / 2.
	other := box width asFloat / 2.
	delta := aPoint - (other@radius).
	xOverY := box width asFloat / box height asFloat.
	^ (delta x asFloat / xOverY) squared + delta y squared <= radius squared! !

!CEllipsePlayer methodsFor: 'drawing' stamp: 'ar 2/6/2005 19:50'!
drawOn: aCanvas in: drawingBounds
	aCanvas fillOval: drawingBounds fill: self fill border: self border! !

!CEllipsePlayer methodsFor: 'initialize' stamp: 'ar 2/6/2005 19:11'!
initialize
	super initialize.
	self userDraw := true.
	self color := Color yellow.
	self borderColor := Color black.
	self borderStyle := #simple.
	self borderWidth := 2.! !

CPlayer subclass: #CFieldReference
	instanceVariableNames: 'value target fieldDefinition'
	classVariableNames: 'CachedFields FailToken'
	poolDictionaries: ''
	category: 'Tweak-Basic-Support'!

!CFieldReference commentStamp: 'ar 3/27/2004 17:25' prior: 0!
CFieldReference represents a link to a particular field in some object. Its value is kept in sync with the field it observes via the change event associated with the field. Field references are created using #fieldNamed:, e.g.,

	xField := Sketch fieldNamed: #x.
	xField value. "will answer the value of Sketch's x"
	xField value: 100. "will set Sketch's x field to 100"

Field references can be used as players for many of the primitive value containing widgets, for example:

	xSpinner := CSpinner new.
	xSpinner player: xField.
	xSpinner open.
!

CFieldReference defineFields: '<?xml  version="1.0" ?>
<fields>
	<field instanceVariable="true" toGet="value" toSet="value:">value</field>
	<field instanceVariable="true" toGet="target" toSet="target:" changeEvent="targetChanged">target</field>
	<field instanceVariable="true" toGet="fieldDefinition" toSet="fieldDefinition:" changeEvent="fieldDefinitionChanged">fieldDefinition</field>
	<field virtual="true" toGet="fieldName" toSet="fieldName:">fieldName</field>
</fields>'!

!CFieldReference class methodsFor: 'instance creation' stamp: ''!
compileReference: aField
	"Compile a uni-class for the given field reference"
	| scripted |
	scripted := self new.
	scripted becomeUniClass.
	scripted class compileUnlogged:
				('readValue\	^target ', aField toGet) withCRs
			classified: nil
			notifying: nil.
	aField toSet ifNotNil:[
		scripted class compileUnlogged:
				('writeValue\	^target ', aField toSet,' value') withCRs
			classified: nil
			notifying: nil.
	].
	^scripted class! !

!CFieldReference class methodsFor: 'instance creation' stamp: 'ar 3/12/2003 21:32'!
forField: aField
	"Answer a field reference for the given field"
	| key cached |
	aField accessKey ifNotNil:[^self new]. "indexed!!"
	key := aField toGet.
	aField toSet ifNotNil:[
		key := key, ' - ', aField toSet.
		key := key asSymbol.
	].
	cached := CachedFields at: key ifAbsent:[nil].
	cached ifNil:[
		cached := self compileReference: aField.
		CachedFields at: key put: cached].
	^cached new.! !

!CFieldReference class methodsFor: 'class initialization' stamp: 'ar 3/11/2003 19:34'!
initialize
	"CFieldReference initialize"
	CachedFields := WeakValueDictionary new.
	FailToken := Object new.! !

!CFieldReference methodsFor: 'private' stamp: 'bf 3/30/2004 12:31'!
fieldDefinition
	<bewareOf: #fieldDefinitionChanged>
	^fieldDefinition! !

!CFieldReference methodsFor: 'private' stamp: 'bf 3/30/2004 12:31'!
fieldDefinition: aFieldDefinition
	self stopScript: #onUpdateFromTarget.
	fieldDefinition := aFieldDefinition.
	self signal: #fieldDefinitionChanged.
	self onTargetChanged.
	^fieldDefinition! !

!CFieldReference methodsFor: 'accessing' stamp: 'ar 3/11/2004 14:12'!
fieldName
	^self fieldDefinition name! !

!CFieldReference methodsFor: 'accessing' stamp: 'bf 3/30/2004 14:02'!
fieldName: aString
	self fieldDefinition: (self target class fieldNamed: aString).
	^aString! !

!CFieldReference methodsFor: 'testing' stamp: 'ar 3/12/2003 15:29'!
isFieldReference
	^true! !

!CFieldReference methodsFor: 'testing' stamp: 'ar 6/22/2003 21:06'!
isReadOnly
	^fieldDefinition isReadOnly! !

!CFieldReference methodsFor: 'accessing' stamp: 'ar 9/24/2003 13:53'!
name
	^(self propertyValueAt: #name) ifNil:[target name,'''s ', fieldDefinition name].! !

!CFieldReference methodsFor: 'events' stamp: 'ar 7/31/2004 17:50'!
onTargetChanged
	| script eventName eventVariable varDef |
	target ifNil:[^self].
	fieldDefinition ifNil:[^self].
	eventName := fieldDefinition changeEvent.
	eventVariable := nil.
	self stopScript: #onUpdateFromTarget.
	eventVariable ifNil:[
		self startScript: #onUpdateFromTarget 
			when:{target. eventName}.
	] ifNotNil:[
		varDef := target class bindingOf: eventVariable.
		varDef ifNil:[^self].
		varDef class == CFieldDefinition ifTrue:[
			script := self startScript: #onUpdateFromTarget 
							when: {target getField: varDef. eventName}.
			script startScript: #onUpdateTrigger:from: 
				when: {target. varDef changeEvent}.
		] ifFalse:[
			script := self startScript: #onUpdateFromTarget 
							when: {varDef value. eventName}.
		].
	].
	self onUpdateFromTarget.
! !

!CFieldReference methodsFor: 'events' stamp: 'ar 9/8/2004 16:22'!
onUpdateFromTarget
	|  aValue |
	aValue := [self readValue] on: Error do:[:ex| ex return: FailToken].
	aValue == FailToken ifTrue:[^self].
	aValue == value ifTrue:[^self].
	value := aValue.
	self signalChanged: #valueChanged from: value to: aValue.! !

!CFieldReference methodsFor: 'private' stamp: 'ar 9/8/2004 16:22'!
privateSetValueTo: aValue
	"Yes, I *do* know what I'm doing here"
	| oldValue |
	value == aValue ifTrue:[^self].
	oldValue := value.		"must come before any other action"
	myPlayer ifNotNil:[myPlayer value: aValue].
	self signalChanged: #valueChanged from: value to: aValue.
	value := aValue.
	[
		self writeValue.
		"Notify any interested parties that we pushed a new value to our target"
		self signal: #valueSaved with: self.
		"Note: the error handler will prevent any clients 
		to hear #valueSaved unless it was successful."
		target tweaked: fieldDefinition toSet from: oldValue to: value.	"record the tweak"

	] on: Error do:[:ex| ex return].
	^aValue! !

!CFieldReference methodsFor: 'accessing' stamp: 'ar 3/11/2003 19:14'!
readValue
	^target getField: fieldDefinition! !

!CFieldReference methodsFor: 'accessing' stamp: 'ar 3/28/2004 23:36'!
target
	<bewareOf: #targetChanged>
	^target! !

!CFieldReference methodsFor: 'accessing' stamp: 'ar 9/8/2004 16:22'!
target: aTarget
	target == aTarget ifTrue:[^aTarget].
	self stopScript: #onUpdateFromTarget.
	self signalChanged: #targetChanged from: target to: aTarget.
	target := aTarget.
	self onTargetChanged.
	^aTarget! !

!CFieldReference methodsFor: 'accessing' stamp: 'ar 3/17/2004 17:17'!
value
	^value! !

!CFieldReference methodsFor: 'accessing' stamp: 'ar 3/17/2004 19:22'!
value: newValue
	"Modify the receiver's value"
	^self privateSetValueTo: newValue! !

!CFieldReference methodsFor: 'accessing' stamp: 'ar 3/19/2004 15:57'!
valueType
	^fieldDefinition valueType ifNil:[super valueType]! !

!CFieldReference methodsFor: 'copying' stamp: 'tk 2/13/2005 23:57'!
veryDeepCheck: deepCopier
	"Restart events and scripts."

	super veryDeepCheck: deepCopier.

	self onTargetChanged.! !

!CFieldReference methodsFor: 'accessing' stamp: 'ar 6/22/2003 21:12'!
writeValue
	self isReadOnly ifTrue:[^value := self readValue].
	target setField: fieldDefinition to: value.! !

CPlayer subclass: #CHolder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Kernel'!

!CHolder commentStamp: 'ar 3/27/2004 18:05' prior: 0!
CHolder provides the full collection interface for use with its elements. Subclasses are typically thought about as being containers of some sort - for any kind of either graphical or non-graphical objects.!

CHolder defineFields: '<?xml  version="1.0" ?>
<fields>
	<field virtual="true" toGet="cursor" toSet="cursor:" changeEvent="cursorChanged">cursor</field>
	<field virtual="true" toGet="wrapCursor" toSet="wrapCursor:" changeEvent="wrapCursorChanged">wrapCursor</field>
</fields>'!

!CHolder methodsFor: 'matching' stamp: 'ar 3/16/2004 15:46'!
all
	"Answer a new all matcher"
	^CAllMatcher on: self! !

!CHolder methodsFor: 'enumerating' stamp: 'ar 3/14/2005 20:13'!
allSatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns false for any element return false.
	Otherwise return true."
	^self collection allSatisfy: aBlock! !

!CHolder methodsFor: 'matching' stamp: 'ar 3/15/2004 15:36'!
any
	"Answer a new any matcher"
	^CAnyMatcher on: self! !

!CHolder methodsFor: 'enumerating' stamp: 'ar 3/14/2005 20:14'!
anySatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns true for any element return true.
	Otherwise return false."
	^self collection anySatisfy: aBlock! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:14'!
at: aKey
	"Answer the element at the given key"
	^self collection at: aKey! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:14'!
at: aKey ifAbsent: aBlock
	"Answer the element at the given key or evaluate the given block if there is no element with the given key"
	^self collection at: aKey ifAbsent: aBlock! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:13'!
at: aKey put: anObject
	"Store anObject under the given key in the receiver"
	^self collection at: aKey put: anObject! !

!CHolder methodsFor: 'streaming' stamp: 'ar 3/14/2005 20:13'!
atEnd
	"answer whether the receiver is at its end"
	^self collection atEnd! !

!CHolder methodsFor: 'streaming' stamp: 'ar 3/14/2005 20:14'!
atStart
	"answer whether the receiver is at its start"
	^self collection atStart! !

!CHolder methodsFor: 'enumerating' stamp: 'ar 3/16/2004 15:50'!
collect: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into a collection like the receiver. Answer  
	the new collection."
	^self collect: aBlock species: self species! !

!CHolder methodsFor: 'enumerating' stamp: 'ar 3/14/2005 20:13'!
collect: aBlock species: aClass
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into a collection like aClass. Answer  
	the new collection."
	^self collection collect: aBlock species: aClass! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:14'!
contentType
	"Answer the type of the elements stored in the receiver"
	^self collection contentType! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:14'!
contents
	"Answer the contents of the receiver"
	^self collection copyFrom: 1 to: self size! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:13'!
contents: newContents
	"Set the contents of the receiver"
	^self collection contents: newContents! !

!CHolder methodsFor: 'copying' stamp: 'ar 3/14/2005 20:13'!
copyFrom: firstIndex to: lastIndex
	"Answer a copy of a subset of the receiver, starting from element at 
	index start until element at index stop."
	^self collection copyFrom: firstIndex to: lastIndex! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:13'!
cursor
	"Answer the receiver's cursor"
	<bewareOf: #cursorChanged>
	^self collection cursor! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:14'!
cursor: aValue
	"Modifiy the receiver's cursor"
	^self collection cursor: aValue! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:14'!
defaultElement
	"Answer the default element for the receiver"
	^self collection defaultElement! !

!CHolder methodsFor: 'searching' stamp: 'ar 3/14/2005 20:14'!
detect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Answer the first element for which aBlock evaluates to true."
	^self collection detect: aBlock! !

!CHolder methodsFor: 'searching' stamp: 'ar 3/14/2005 20:13'!
detect: aBlock ifNone: exceptionBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Answer the first element for which aBlock evaluates to true. If none  
	evaluate to true, then evaluate the argument, exceptionBlock."
	^self collection detect: aBlock ifNone: exceptionBlock! !

!CHolder methodsFor: 'enumerating' stamp: 'ar 3/14/2005 20:13'!
do: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument."
	^self collection do: aBlock! !

!CHolder methodsFor: 'searching' stamp: 'ar 3/14/2005 20:14'!
findFirstElement: aBlock
	"Answer my first element for which aBlock evaluates as true."
	^self collection findFirstElement: aBlock! !

!CHolder methodsFor: 'searching' stamp: 'ar 3/14/2005 20:13'!
findFirstIndex: aBlock
	"Answer the index of my first element for which aBlock evaluates as true."
	^self collection findFirstIndex: aBlock! !

!CHolder methodsFor: 'searching' stamp: 'ar 3/14/2005 20:13'!
findLast: aBlock
	"Answer the index of my last element for which aBlock evaluates as true."
	^self collection findLast: aBlock! !

!CHolder methodsFor: 'searching' stamp: 'ar 3/14/2005 20:14'!
findLastElement: aBlock
	"Answer my last element for which aBlock evaluates as true."
	^self collection findLastElement: aBlock! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:13'!
first
	"Answer the first element of the receiver"
	^self collection first! !

!CHolder methodsFor: 'searching' stamp: 'ar 3/14/2005 20:13'!
includes: anObject 
	"Answer whether anObject is one of the receiver's elements."
	^self collection includes: anObject! !

!CHolder methodsFor: 'searching' stamp: 'ar 3/14/2005 20:14'!
indexOf: anObject
	"Answer the index of the given object within the receiver"
	^self collection indexOf: anObject! !

!CHolder methodsFor: 'searching' stamp: 'ar 3/14/2005 20:13'!
indexOf: anElement ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the 
	result of evaluating the argument, exceptionBlock."
	^self collection indexOf: anElement ifAbsent: exceptionBlock! !

!CHolder methodsFor: 'searching' stamp: 'ar 3/14/2005 20:13'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement after start
	within the receiver. If the receiver does not contain anElement, 
	answer the 	result of evaluating the argument, exceptionBlock."
	^self collection indexOf: anElement startingAt: start ifAbsent: exceptionBlock! !

!CHolder methodsFor: 'testing' stamp: 'ar 3/14/2005 20:14'!
isEmpty
	"Answer whether the receiver contains any elements."
	^self collection isEmpty! !

!CHolder methodsFor: 'enumerating' stamp: 'ar 3/14/2005 20:13'!
keysAndValuesDo: aBlock
	"Evaluate aBlock with each of the receiver's keys and values"
	^self collection keysAndValuesDo: aBlock! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:14'!
last
	"Answer the last element in the receiver"
	^self collection last! !

!CHolder methodsFor: 'streaming' stamp: 'ar 3/14/2005 20:14'!
next
	"go to the next element in the receiver"
	^self collection next! !

!CHolder methodsFor: 'streaming' stamp: 'ar 3/14/2005 20:13'!
peek
	"answer the element at the receiver's cursor"
	^self collection peek! !

!CHolder methodsFor: 'streaming' stamp: 'ar 3/14/2005 20:13'!
prev
	"go to the previous element in the receiver"
	^self collection prev! !

!CHolder methodsFor: 'replacing' stamp: 'ar 3/14/2005 20:12'!
replace: oldObject with: newObject
	"Replace oldObject with newObject"
	^self collection replace: oldObject with: newObject! !

!CHolder methodsFor: 'replacing' stamp: 'ar 3/14/2005 20:13'!
replaceFrom: firstIndex to: lastIndex with: aCollection
	"Replace the elements in the receiver"
	^self collection replaceFrom: firstIndex to: lastIndex with: aCollection! !

!CHolder methodsFor: 'replacing' stamp: 'ar 3/14/2005 20:12'!
replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart
	^self collection replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart! !

!CHolder methodsFor: 'enumerating' stamp: 'ar 3/16/2004 15:50'!
select: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new collection like the receiver, only those elements for 
	which aBlock evaluates to true. Answer the new collection."
	^self select: aBlock species: self species! !

!CHolder methodsFor: 'enumerating' stamp: 'ar 3/14/2005 20:14'!
select: aBlock species: aClass
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new collection like the receiver, only those elements for 
	which aBlock evaluates to true. Answer the new collection."
	^self collection select: aBlock species: aClass! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:13'!
selection
	"Answer the currently selected object - that is the object at the cursor's position"
	^self collection selection! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:14'!
selection: anObject
	"Set the currently selected object - that is move the cursor's position to point to anObject"
	^self collection selection: anObject! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:14'!
selectionIndex
	"A convenient alias to the cursor"
	^self collection selectionIndex! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:13'!
selectionIndex: newIndex
	"A convenient alias to the cursor"
	^self collection selectionIndex: newIndex! !

!CHolder methodsFor: 'accessing' stamp: 'ar 3/14/2005 20:14'!
size
	"Answer the number of elements of the receiver"
	^self collection size! !

!CHolder methodsFor: 'streaming' stamp: 'ar 3/14/2005 20:14'!
wrapCursor
	"Answer whether the cursor can wrap around"
	<bewareOf: #wrapCursorChanged>
	^self collection wrapCursor! !

!CHolder methodsFor: 'streaming' stamp: 'ar 3/14/2005 20:13'!
wrapCursor: aBool
	"Indicate whether the cursor can wrap around"
	^self collection wrapCursor: aBool! !

CPlayer subclass: #CImagePlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Objects'!

!CImagePlayer commentStamp: 'tk 3/1/2005 11:36' prior: 0!
An Image holds a picture.  When you import a photo or GIF or JPEG file from a web page, it becomes an Image in Tweak.[general]


### Allow Users to see the above explanation ###

This class is a stand-in for objects representing pictures. It is intended to be used wherever we create or import an image. At some point we WILL want to have different costumes for 'images' and this will be the base type to be used for it.!

CImagePlayer defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

CPlayer subclass: #CMoviePlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Objects'!

!CMoviePlayer commentStamp: 'ar 3/14/2004 16:10' prior: 0!
This class is a stand-in for objects representing movies. It is intended to be used wherever we create or import a movie. At some point we WILL want to have different costumes for 'movies' and this will be the base type to be used for it.!

CMoviePlayer defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!CMoviePlayer methodsFor: 'initialize' stamp: 'ar 3/14/2004 17:47'!
initialize
	super initialize.
	self color: Color white.! !

CPlayer defineFields: '<?xml  version="1.0" ?>
<fields>
	<field instanceVariable="true">myCostume</field>
	<field instanceVariable="true">myPlayer</field>
	<field virtual="true" toGet="player" toSet="player:" changeEvent="playerChanged" color="FF00FF">player</field>
	<field virtual="true" toGet="costume" toSet="costume:" changeEvent="costumeChanged" color="FF00FF">costume</field>
	<field virtual="true" toGet="container" toSet="container:" changeEvent="containerChanged" color="FF00FF">container</field>
	<field virtual="true" toGet="elements" toSet="elements:" changeEvent="elementsChanged" color="FF00FF">elements</field>
	<field virtual="true" toGet="hand" color="FF00FF">hand</field>
	<field virtual="true" toGet="world" changeEvent="containerChanged" color="FF00FF">world</field>
	<field virtual="true" toGet="balloonHelp" toSet="balloonHelp:" changeEvent="balloonHelpChanged" color="FF00FF">balloonHelp</field>
	<field virtual="true" toGet="mouseCursor" toSet="mouseCursor:" changeEvent="mouseCursorChanged" color="FF00FF">mouseCursor</field>
	<field virtual="true" toGet="border" toSet="border:" changeEvent="borderChanged" color="00FF00">border</field>
	<field virtual="true" toGet="borderColor" toSet="borderColor:" changeEvent="borderChanged" color="00FF00">borderColor</field>
	<field virtual="true" toGet="borderStyle" toSet="borderStyle:" changeEvent="borderChanged" color="00FF00">borderStyle</field>
	<field virtual="true" toGet="borderWidth" toSet="borderWidth:" changeEvent="borderChanged" color="00FF00">borderWidth</field>
	<field virtual="true" serializable="false" toGet="geometry" toSet="geometry:" changeEvent="geometryChanged" color="00FFFF">geometry</field>
	<field virtual="true" toGet="position" toSet="position:" changeEvent="geometryChanged" color="00FFFF">position</field>
	<field virtual="true" toGet="width" toSet="width:" changeEvent="geometryChanged" color="00FFFF">width</field>
	<field virtual="true" toGet="height" toSet="height:" changeEvent="geometryChanged" color="00FFFF">height</field>
	<field virtual="true" toGet="extent" toSet="extent:" changeEvent="geometryChanged" color="00FFFF">extent</field>
	<field virtual="true" toGet="angle" toSet="angle:" changeEvent="geometryChanged" color="00FFFF">angle</field>
	<field virtual="true" toGet="scale" toSet="scale:" changeEvent="geometryChanged" color="00FFFF">scale</field>
	<field virtual="true" toGet="heading" toSet="heading:" changeEvent="geometryChanged" color="00FFFF">heading</field>
	<field virtual="true" serializable="false" toGet="direction" toSet="direction:" changeEvent="geometryChanged" color="00FFFF">direction</field>
	<field virtual="true" serializable="false" toGet="normal" toSet="normal:" changeEvent="geometryChanged" color="00FFFF">normal</field>
	<field virtual="true" toGet="left" toSet="left:" changeEvent="geometryChanged" color="00FFFF">left</field>
	<field virtual="true" toGet="right" toSet="right:" changeEvent="geometryChanged" color="00FFFF">right</field>
	<field virtual="true" toGet="top" toSet="top:" changeEvent="geometryChanged" color="00FFFF">top</field>
	<field virtual="true" toGet="bottom" toSet="bottom:" changeEvent="geometryChanged" color="00FFFF">bottom</field>
	<field virtual="true" toGet="origin" toSet="origin:" changeEvent="geometryChanged" color="00FFFF">origin</field>
	<field virtual="true" toGet="corner" toSet="corner:" changeEvent="geometryChanged" color="00FFFF">corner</field>
	<field virtual="true" toGet="center" toSet="center:" changeEvent="geometryChanged" color="00FFFF">center</field>
	<field virtual="true" toGet="topLeft" toSet="topLeft:" changeEvent="geometryChanged" color="00FFFF">topLeft</field>
	<field virtual="true" toGet="topRight" toSet="topRight:" changeEvent="geometryChanged" color="00FFFF">topRight</field>
	<field virtual="true" toGet="bottomLeft" toSet="bottomLeft:" changeEvent="geometryChanged" color="00FFFF">bottomLeft</field>
	<field virtual="true" toGet="bottomRight" toSet="bottomRight:" changeEvent="geometryChanged" color="00FFFF">bottomRight</field>
	<field virtual="true" toGet="bounds" toSet="bounds:" changeEvent="geometryChanged" color="00FFFF">bounds</field>
	<field virtual="true" serializable="false" toGet="rotation" toSet="rotation:" changeEvent="geometryChanged" color="00FFFF">rotation</field>
	<field virtual="true" toGet="transform" toSet="transform:" changeEvent="geometryChanged" color="00FFFF">transform</field>
	<field virtual="true" toGet="color" toSet="color:" changeEvent="fillChanged" color="FFFF00">color</field>
	<field virtual="true" toGet="fill" toSet="fill:" changeEvent="fillChanged" color="FFFF00">fill</field>
	<field virtual="true" toGet="fillAlpha" toSet="fillAlpha:" changeEvent="fillChanged" color="FFFF00">fillAlpha</field>
	<field virtual="true" toGet="fillAngle" toSet="fillAngle:" changeEvent="fillChanged" color="FFFF00">fillAngle</field>
	<field virtual="true" toGet="fillColor" toSet="fillColor:" changeEvent="fillChanged" color="FFFF00">fillColor</field>
	<field virtual="true" toGet="fillColors" toSet="fillColors:" changeEvent="fillChanged" color="FFFF00">fillColors</field>
	<field virtual="true" toGet="fillOffset" toSet="fillOffset:" changeEvent="fillChanged" color="FFFF00">fillOffset</field>
	<field virtual="true" serializable="false" toGet="fillStyle" toSet="fillStyle:" changeEvent="fillChanged" color="FFFF00">fillStyle</field>
	<field virtual="true" serializable="false" toGet="penDown" toSet="penDown:" changeEvent="penChanged" color="0000FF">penDown</field>
	<field virtual="true" serializable="false" toGet="penColor" toSet="penColor:" changeEvent="penChanged" color="0000FF">penColor</field>
	<field virtual="true" serializable="false" toGet="penWidth" toSet="penWidth:" changeEvent="penChanged" color="0000FF">penWidth</field>
	<field virtual="true" serializable="false" toGet="penTrailStyle" toSet="penTrailStyle:" changeEvent="penChanged" color="0000FF">penTrailStyle</field>
	<field virtual="true" serializable="false" toGet="rotationStyle" toSet="rotationStyle:" changeEvent="rotationStyleChanged" color="FF0000">rotationStyle</field>
	<field virtual="true" toGet="aspectRatio" toSet="aspectRatio:" changeEvent="geometryChanged" color="FF0000">aspectRatio</field>
	<field virtual="true" toGet="graphic" toSet="graphic:" changeEvent="graphicChanged" color="FF0000">graphic</field>
	<field virtual="true" toGet="graphicAnchorPoint" toSet="graphicAnchorPoint:" changeEvent="graphicChanged" color="FF0000">graphicAnchorPoint</field>
	<field virtual="true" toGet="graphicFit" toSet="graphicFit:" changeEvent="graphicChanged" color="FF0000">graphicFit</field>
	<field virtual="true" toGet="graphicForm" toSet="graphicForm:" changeEvent="graphicChanged" color="FF0000">graphicForm</field>
	<field virtual="true" toGet="graphicOffset" toSet="graphicOffset:" changeEvent="graphicChanged" color="FF0000">graphicOffset</field>
	<field virtual="true" toGet="hResizing" toSet="hResizing:" color="00FF00">hResizing</field>
	<field virtual="true" toGet="vResizing" toSet="vResizing:" color="00FF00">vResizing</field>
	<field virtual="true" toGet="layout" toSet="layout:" color="00FF00">layout</field>
	<field virtual="true" toGet="clipping" toSet="clipping:" changeEvent="flagsChanged" color="FF00FF">clipping</field>
	<field virtual="true" toGet="floating" toSet="floating:" changeEvent="flagsChanged" color="FF00FF">floating</field>
	<field virtual="true" toGet="infinite" toSet="infinite:" changeEvent="flagsChanged" color="FF00FF">infinite</field>
	<field virtual="true" toGet="locked" toSet="locked:" changeEvent="flagsChanged" color="FF00FF">locked</field>
	<field virtual="true" toGet="sticky" toSet="sticky:" changeEvent="flagsChanged" color="FF00FF">sticky</field>
	<field virtual="true" toGet="visible" toSet="visible:" changeEvent="flagsChanged" color="FF00FF">visible</field>
	<field virtual="true" toGet="collisionsEnabled" toSet="collisionsEnabled:" changeEvent="collisionsEnabledChanged" color="00FF00">collisionsEnabled</field>
	<field virtual="true" toGet="preciseCollisions" toSet="preciseCollisions:" changeEvent="preciseCollisionsChanged" color="00FF00">preciseCollisions</field>
	<field virtual="true" toGet="roundCorners" toSet="roundCorners:" changeEvent="roundCornersChanged" color="0000FF">roundCorners</field>
	<field virtual="true" toGet="roundCornerSpec" toSet="roundCornerSpec:" changeEvent="roundCornersChanged" color="0000FF">roundCornerSpec</field>
	<field toGet="hScrollable" toSet="hScrollable:" changeEvent="hScrollableChanged" color="00FF00">hScrollable</field>
	<field toGet="vScrollable" toSet="vScrollable:" changeEvent="vScrollableChanged" color="00FF00">vScrollable</field>
	<field toGet="scrollDelta" toSet="scrollDelta:" changeEvent="scrollDeltaChanged" color="00FF00">scrollDelta</field>
	<field virtual="true" toGet="scrollingDisabled" toSet="scrollingDisabled:" changeEvent="scrollingDisabledChanged" color="00FF00">scrollingDisabled</field>
	<field virtual="true" toGet="scrollOffset" toSet="scrollOffset:" changeEvent="scrollOffsetChanged" color="00FF00">scrollOffset</field>
	<field virtual="true" toGet="dropShadow" toSet="dropShadow:" changeEvent="dropShadowChanged" color="FF0000">dropShadow</field>
	<field virtual="true" toGet="dropShadowColor" toSet="dropShadowColor:" changeEvent="shadowChanged" color="FF0000">dropShadowColor</field>
	<field virtual="true" toGet="dropShadowOffset" toSet="dropShadowOffset:" changeEvent="shadowChanged" color="FF0000">dropShadowOffset</field>
	<field virtual="true" toGet="dropShadowEnabled" toSet="dropShadowEnabled:" changeEvent="dropShadowChanged" color="FF0000">dropShadowEnabled</field>
	<field virtual="true" toGet="textHolder" toSet="textHolder:" changeEvent="textChanged" color="0000FF">textHolder</field>
	<field virtual="true" toGet="text" toSet="text:" changeEvent="textChanged" color="0000FF">text</field>
	<field virtual="true" toGet="textAnchorPoint" toSet="textAnchorPoint:" changeEvent="textChanged" color="0000FF">textAnchorPoint</field>
	<field virtual="true" toGet="textAttributes" toSet="textAttributes:" changeEvent="textChanged" color="0000FF">textAttributes</field>
	<field virtual="true" toGet="textBorderColor" toSet="textBorderColor:" changeEvent="textChanged" color="0000FF">textBorderColor</field>
	<field virtual="true" toGet="textBorderWidth" toSet="textBorderWidth:" changeEvent="textChanged" color="0000FF">textBorderWidth</field>
	<field virtual="true" toGet="textColor" toSet="textColor:" changeEvent="textChanged" color="0000FF">textColor</field>
	<field virtual="true" toGet="textEmphasis" toSet="textEmphasis:" changeEvent="textChanged" color="0000FF">textEmphasis</field>
	<field virtual="true" toGet="textFont" toSet="textFont:" changeEvent="textChanged" color="0000FF">textFont</field>
	<field virtual="true" toGet="textInset" toSet="textInset:" changeEvent="textChanged" color="0000FF">textInset</field>
	<field virtual="true" toGet="textOffset" toSet="textOffset:" changeEvent="textChanged" color="0000FF">textOffset</field>
	<field virtual="true" toGet="textWrap" toSet="textWrap:" changeEvent="textChanged" color="0000FF">textWrap</field>
	<field toGet="eventBehaviors" toSet="eventBehaviors:" changeEvent="eventBehaviorsChanged" color="FF00FF">eventBehaviors</field>
	<field toGet="selectionColor" toSet="selectionColor:" changeEvent="selectionColorChanged">selectionColor</field>
	<field virtual="true" serializable="false" toGet="cartesianXY" toSet="cartesianXY:" changeEvent="geometryChanged" color="00FFFF">cartesianXY</field>
	<field virtual="true" serializable="false" toGet="cartesianX" toSet="cartesianX:" changeEvent="geometryChanged" color="00FFFF">cartesianX</field>
	<field virtual="true" serializable="false" toGet="cartesianY" toSet="cartesianY:" changeEvent="geometryChanged" color="00FFFF">cartesianY</field>
	<field virtual="true" serializable="false" toGet="cartesianLeft" toSet="cartesianLeft:" changeEvent="geometryChanged" color="00FFFF">cartesianLeft</field>
	<field virtual="true" serializable="false" toGet="cartesianRight" toSet="cartesianRight:" changeEvent="geometryChanged" color="00FFFF">cartesianRight</field>
	<field virtual="true" serializable="false" toGet="cartesianTop" toSet="cartesianTop:" changeEvent="geometryChanged" color="00FFFF">cartesianTop</field>
	<field virtual="true" serializable="false" toGet="cartesianBottom" toSet="cartesianBottom:" changeEvent="geometryChanged" color="00FFFF">cartesianBottom</field>
</fields>'!

!CPlayer class methodsFor: 'instance creation' stamp: 'ar 5/14/2005 15:57'!
asCostumeFor: aPlayer
	"Create a new costume for the given player"
	^(self basicNew) 
		autoStartScripts;
		privateChangePlayerTo: aPlayer; 
		initialize;
		autoStartTriggers;
	yourself
! !

!CPlayer class methodsFor: 'instance creation' stamp: 'ar 5/14/2005 16:11'!
asPlayerFor: aCostume
	^(self basicNew) 
		privateChangeCostumeTo: aCostume; 
		autoStartScripts;
		initialize;
		autoStartTriggers;
	yourself! !

!CPlayer class methodsFor: 'instance creation' stamp: 'ar 2/6/2005 19:21'!
newSample
	^self new! !

!CPlayer class methodsFor: 'instance creation' stamp: 'sw 2/8/2005 05:13'!
newbornPartToDropIn: aContainer
	"Answer a new instance of the receiver suitable for dropping into the given container.  If the result of simply sending #newSample to the receiver (which is the default implementation here) is not adequate, subclasses can reimplement this."

	^ self newSample
! !

!CPlayer class methodsFor: 'opening' stamp: 'ar 2/6/2005 19:21'!
open
	^self new open! !

!CPlayer class methodsFor: 'opening' stamp: ''!
openInHand
	^self new openInHand! !

!CPlayer class methodsFor: 'opening' stamp: 'ar 6/5/2005 12:15'!
openInHand: aHand
	^self new openInHand: aHand! !

!CPlayer class methodsFor: 'opening' stamp: ''!
openOn: aValue
	^(self on: aValue) open! !

!CPlayer class methodsFor: 'opening' stamp: 'tk 2/2/2005 10:15'!
wearEllipse
	"Create a Player wearing a Ellipse costume"

	^ (self new) shape: (CEllipseShape new); yourself
! !

!CPlayer class methodsFor: 'opening' stamp: 'tk 2/1/2005 22:07'!
wearPolygon
	"Create a Player wearing a Polygon costume"

	^ (self new) shape: (CPolygonShape new); yourself
! !

!CPlayer methodsFor: 'adding' stamp: 'ar 4/2/2004 14:27'!
add: newObject 
	"Include newObject as one of the receiver's elements. Answer newObject."
	^self addLast: newObject! !

!CPlayer methodsFor: 'adding' stamp: 'ar 3/14/2005 20:12'!
add: newObject after: oldObject
	"Insert newObject after oldObject"
	^self collection add: newObject after: oldObject! !

!CPlayer methodsFor: 'adding' stamp: 'ar 3/14/2005 20:12'!
add: anObject at: anIndex
	"Insert the given object at the specified index"
	^self collection add: anObject at: anIndex.! !

!CPlayer methodsFor: 'adding' stamp: 'ar 3/14/2005 20:11'!
add: newObject before: oldObject				 	 
	"Add the argument, newObject, as an element of the receiver."
	^self collection add: newObject before: oldObject! !

!CPlayer methodsFor: 'adding' stamp: 'ar 3/14/2005 20:11'!
addAll: aCollection 
	"Include all the elements of aCollection as the receiver's elements."
	^self collection addAll: aCollection! !

!CPlayer methodsFor: 'adding' stamp: 'ar 3/14/2005 20:11'!
addAll: aCollection after: oldObject
	"Insert the given object before the specified index"
	^self collection addAll: aCollection after: oldObject! !

!CPlayer methodsFor: 'adding' stamp: 'ar 3/14/2005 20:11'!
addAll: aCollection at: anIndex
	"Insert the given object before the specified index"
	^self collection addAll: aCollection at: anIndex! !

!CPlayer methodsFor: 'adding' stamp: 'ar 3/14/2005 20:11'!
addAll: aCollection before: oldObject
	"Insert the given object before the specified index"
	^self collection addAll: aCollection before: oldObject! !

!CPlayer methodsFor: 'adding' stamp: 'ar 3/14/2005 20:11'!
addFirst: newObject 
	"Include newObject as the first of the receiver's elements. Answer newObject."
	^self collection addFirst: newObject! !

!CPlayer methodsFor: 'highlights' stamp: 'ar 3/24/2004 01:58'!
addHighlight: aHighlight
	"Add the given highlight to the receiver"
	^costume addHighlight: aHighlight! !

!CPlayer methodsFor: 'adding' stamp: 'ar 3/14/2005 20:11'!
addLast: newObject
	"Include newObject as the last of the receiver's elements."
	^self collection addLast: newObject! !

!CPlayer methodsFor: 'queries' stamp: 'ar 3/25/2004 22:04'!
allContainersDo: aBlock
	"Evaluate aBlock with all containers of the receiver"
	| parent |
	parent := container.
	[parent == nil] whileFalse:[
		aBlock value: parent.
		parent := parent container].! !

!CPlayer methodsFor: 'queries' stamp: 'tk 10/28/2004 18:59'!
allElementsDo: aBlock
	"Evaluate aBlock with all elements and sub-elements of the receiver. Depth first."

	elements do: [:ele |
		aBlock value: ele.
		ele allElementsDo: aBlock]! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/27/2005 14:55'!
angle
	"Answer the receiver's angle, measured counter-clockwise from east in degrees"
	<bewareOf: #geometryChanged>
	^geometry angle! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/27/2005 14:55'!
angle: aValue
	"Indicate the receiver's angle, measured counter-clockwise from east in degrees"
	^geometry angle: aValue.
! !

!CPlayer methodsFor: 'converting' stamp: 'ar 2/12/2004 16:41'!
asCostume
	"Answer myself as a costume"
	^self costume asCostume! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 5/13/2005 13:28'!
asCostumeFor: aPlayer
	"Make myself the costume for aPlayer"
	player := aPlayer.
	^self! !

!CPlayer methodsFor: 'converting' stamp: 'ar 3/16/2004 15:31'!
asForm
	"Answer a form representing the receiver"
	^costume asForm! !

!CPlayer methodsFor: 'converting' stamp: 'ar 3/29/2004 22:36'!
asPlayer
	"Answer this object's player representation."
	myPlayer == nil ifTrue:[^self].
	myPlayer getCostume == self ifFalse:[^self].
	^myPlayer asPlayer! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 5/14/2005 16:12'!
asPlayerFor: aCostume
	"Make myself the player for aCostume"
	costume := aCostume.
	^self! !

!CPlayer methodsFor: 'converting' stamp: 'tk 7/14/2004 11:25'!
asPrimCostume
	"Answer my lowest level costume"

	^self costume asPrimCostume! !

!CPlayer methodsFor: 'graphic' stamp: 'ar 3/23/2004 23:57'!
aspectRatio
	<bewareOf: #geometryChanged>
	^self costume aspectRatio! !

!CPlayer methodsFor: 'graphic' stamp: 'tk 9/15/2004 23:19'!
aspectRatio: aValue
	"When an object such as a photograph has a width and height that need to stay in proportion, aspectRatio holds (width/height).  When the height is changed, knowing the aspectRatio allows Tweak to find the new width."
	^costume aspectRatio: aValue.
! !

!CPlayer methodsFor: 'events' stamp: 'ar 4/30/2004 17:48'!
attachBehavior: anEventBehavior
	anEventBehavior target == self ifFalse:[^anEventBehavior attachTo: self].
	self eventBehaviors: (self eventBehaviors copyWith: anEventBehavior).! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/23/2004 23:57'!
balloonHelp
	"Answer the balloon help message"
	<bewareOf: #balloonHelpChanged>
	^self costume balloonHelp! !

!CPlayer methodsFor: 'basic' stamp: 'tk 2/22/2005 16:49'!
balloonHelp: aText
	"balloon help is the text that appears when you hold the mouse still over an object.  The help message can be changed."
	^self costume balloonHelp: aText! !

!CPlayer methodsFor: 'opening' stamp: 'ar 2/6/2005 23:59'!
beOnScreen
	"Make sure the receiver shows on the screen (and not outside)"
	costume beOnScreen.! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 5/29/2005 19:33'!
becomeUniClass
	"Make the receiver an instance of a unique class"
	| mgr |	
	super becomeUniClass.
	mgr := Processor activeProject scriptManager.
	mgr ifNotNil: [mgr uniClassCreatedFor: self]! !

!CPlayer methodsFor: 'border' stamp: 'ar 3/23/2004 17:27'!
border
	<bewareOf: #borderChanged>
	^costume border! !

!CPlayer methodsFor: 'border' stamp: 'tk 1/4/2005 11:58'!
border: aValue
	"The object that is in charge of this Player's border.  It can be a kind of CostumeBorder, or it can be nil."

	^costume border: aValue.
! !

!CPlayer methodsFor: 'border' stamp: 'ar 3/23/2004 17:27'!
borderColor
	<bewareOf: #borderChanged>
	^border color! !

!CPlayer methodsFor: 'border' stamp: 'das 11/20/2009 17:01'!
borderColor: aValue
	"borderColor is the main color to use for the border of this object. broderWidth must be greater than zero, and borderStyle must not be #none, for the border to be seen.[general]
	In the Border section, drag in the color space and move the mouse to a color.  The red box shows what color is selected.[colorTool]"

	^border color: (self ensureColor: aValue).
! !

!CPlayer methodsFor: 'border' stamp: 'ar 3/23/2004 17:28'!
borderStyle
	<bewareOf: #borderChanged>
	^border style! !

!CPlayer methodsFor: 'border' stamp: 'tk 1/4/2005 12:00'!
borderStyle: aValue
	"borderStyle is the 'look' of the border.  The ColorTool shows the various styles graphically.  They are called complexInset, complexAltFramed, complexAltRaised, complexAltInset, complexRaised, complexFramed, inset, raised, simple, and none.  broderWidth must be greater than zero for the border to be seen.  If you can't see a border, the border color may be transparent.[general]
	In the Border section, at the bottom, click on a border style button.[colorTool]"

	^border style: aValue.
! !

!CPlayer methodsFor: 'border' stamp: 'ar 3/25/2004 17:10'!
borderWidth
	<bewareOf: #borderChanged>
	^border width! !

!CPlayer methodsFor: 'border' stamp: 'tk 9/14/2004 10:56'!
borderWidth: aValue
	"borderWidth tells a Player's border how wide to be.  It is in screen dots, also called pixels.  The border appears on all four sides of the Player.  When borderWidth is zero, the Player has no border.  When borderWidth is not zero, and you still can't see a border, the borderColor may be transparent, or the borderStyle may be #none.[general]
	In the Border section, change the Width by typing or clicking the arrows.[colorTool]"

	^border width: aValue.! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:19'!
bottom
	<bewareOf: #geometryChanged>
	^geometry bottom! !

!CPlayer methodsFor: 'geometry' stamp: 'tk 8/11/2004 17:50'!
bottom: aValue
	"bottom is the position of lowest part of this object on the screen.  It is the number of screen dots from the top edge of the object's container to the bottom edge of this object.  Increasing bottom moves the object downward." 

	^geometry bottom: aValue.
! !

!CPlayer methodsFor: 'geometry' stamp: 'bgf 2/23/2008 12:43'!
bottomCenter
	<bewareOf: #geometryChanged>
	^geometry bottomCenter! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:57'!
bottomCenter: aValue

	^geometry bottomCenter: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:21'!
bottomLeft
	<bewareOf: #geometryChanged>
	^geometry bottomLeft! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:57'!
bottomLeft: aValue
	"bottomLeft is the location of the lower lefthand corner of the object.  It is a Point whose value is (left@bottom)."

	^geometry bottomLeft: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:21'!
bottomRight
	<bewareOf: #geometryChanged>
	^geometry bottomRight! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:57'!
bottomRight: aValue
	"bottomRight is the location of lower right corner of the object.  It is a Point and is measured from the container's origin.  bottomRight is the same as corner."

	^geometry bottomRight: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'animating' stamp: 'ar 6/25/2005 20:04'!
bounce
	"Bounce myself in my container."
	| oldDirection newDirection |
	oldDirection := self direction.
	newDirection := self bounce: oldDirection.
	"Set my direction to the new direction.
	This is optimized since changing the direction is somewhat expensive."
	newDirection = oldDirection 
		ifFalse:[self direction: newDirection].
! !

!CPlayer methodsFor: 'animating' stamp: 'ar 3/27/2004 21:11'!
bounce: aHeading
	"Bounce off my container's wall using the given heading.
	Answer the new heading after the bounce, or the original
	one if no change in heading occurs."
	| innerBox outerBox newHeading |
	innerBox := self bounds.
	outerBox := 0@0 extent: self container extent.
	(outerBox containsRect: innerBox)
		ifTrue:[^aHeading].
	newHeading := aHeading.

	(innerBox left < 0 and:[aHeading x < 0]) "left"
		ifTrue:[newHeading := aHeading x negated @ aHeading y].

	(innerBox right > outerBox width and:[aHeading x > 0]) "right" 
		ifTrue:[newHeading := aHeading x negated @ aHeading y].

	(innerBox top < 0 and:[aHeading y < 0])  "top"
		ifTrue:[newHeading := aHeading x @ aHeading y negated].

	(innerBox bottom > outerBox height and:[aHeading y > 0]) "bottom"
		ifTrue:[newHeading := aHeading x @ aHeading y negated].

	^newHeading! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:21'!
bounds
	<bewareOf: #geometryChanged>
	^geometry bounds! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:58'!
bounds: newBounds
	"bounds is a box that tells where the object is on the screen.  bounds is a Rectangle that goes from the origin to the corner of the object.  It is measured from the container's origin."

	^geometry bounds: (self ensureRect: newBounds).
! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:38'!
cartesianBottom
	"Answer the cartesian left value"
	<bewareOf: #geometryChanged>
	^self cartesianBounds bottom! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:50'!
cartesianBottom: yValue
	"Set the cartesian bottom value"
	self moveBy: 0 @ (self cartesianBottom - yValue).
	^yValue! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:37'!
cartesianBounds
	"Answer the cartesian bounds"
	| base cc box |
	<bewareOf: #geometryChanged>
	box := self bounds.
	base := (cc := container) ifNil:[self bottomLeft] ifNotNil:[cc cartesianOrigin].
	^Rectangle 
		origin: (box left - base x) @ (base y - box top)
		corner: (box right - base x) @ (base y - box bottom)! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:37'!
cartesianLeft
	"Answer the cartesian left value"
	<bewareOf: #geometryChanged>
	^self cartesianBounds left! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:48'!
cartesianLeft: xValue
	"Set the cartesian left value"
	self moveBy: (xValue - self cartesianLeft) @ 0.
	^xValue! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 23:04'!
cartesianOrigin
	"Answer the origin of the cartesian coordinate system for my elements"
	^0@self height "bottom-left corner"! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:41'!
cartesianRight
	"Answer the cartesian right value"
	<bewareOf: #geometryChanged>
	^self cartesianBounds right! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:48'!
cartesianRight: xValue
	"Set the cartesian right value"
	self moveBy: (xValue - self cartesianRight) @ 0.
	^xValue! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:41'!
cartesianTop
	"Answer the cartesian top value"
	<bewareOf: #geometryChanged>
	^self cartesianBounds top! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:50'!
cartesianTop: yValue
	"Set the cartesian top value"
	self moveBy: 0 @ (self cartesianTop - yValue).
	^yValue! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:36'!
cartesianX
	"Answer the cartesian x value"
	<bewareOf: #geometryChanged>
	^self cartesianXY x! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:49'!
cartesianX: xValue
	"Set the cartesian x value"
	self moveBy: (xValue - self cartesianX) @ 0.
	^xValue! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:35'!
cartesianXY
	"Answer the cartesian position"
	| base pos cc |
	<bewareOf: #geometryChanged>
	pos := self position.
	base := (cc := container) ifNil:[self bottomLeft] ifNotNil:[cc cartesianOrigin].
	^(pos x - base x) @ (base y - pos y)! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 23:50'!
cartesianXY: aPoint
	"Answer the cartesian y value"
	| delta |
	delta := aPoint - self cartesianXY.
	self moveBy: delta x @ delta y negated.
	^aPoint! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:36'!
cartesianY
	"Answer the cartesian y value"
	<bewareOf: #geometryChanged>
	^self cartesianXY y! !

!CPlayer methodsFor: 'cartesian' stamp: 'ar 6/27/2005 22:50'!
cartesianY: yValue
	"Answer the cartesian y value"
	self moveBy: 0 @ (self cartesianY - yValue).
	^yValue! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:21'!
center
	<bewareOf: #geometryChanged>
	^geometry center! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:58'!
center: aValue
	"center is the location of the center of an object.  It is a Point and is measured from the container's origin.  When you change the angle of an object, it rotates about its center."

	^geometry center: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'properties' stamp: 'ar 3/23/2004 23:01'!
clipping
	"Answer the receiver's clipping property"
	<bewareOf: #flagsChanged>
	^costume clipping! !

!CPlayer methodsFor: 'properties' stamp: 'tk 2/24/2005 16:26'!
clipping: aBool
	"clipping tells whether elements of this object can be seen when they are outside of the border.  True means this object cuts off the view of anything that sticks out.[general]
	In the upper section, find 'clipping'. A checkmark means do cut off the view of elements that sticks out.[colorTool]"

	^costume clipping: aBool.! !

!CPlayer methodsFor: 'opening' stamp: 'bgf 11/14/2009 22:16'!
close
	"Close this player, e.g. remove it from the screen"
	myCostume ifNotNil:[
		myCostume close.
		self signal: #playerClosed
	].
! !

!CPlayer methodsFor: 'shape' stamp: 'tk 2/1/2005 18:08'!
closed: aBoolean
	"closed tells if the object's polygon shape will be a filled area, or just a segmented line (or a curve).  You must give this object a shape for this command to have any effect (self shape: CPolygonShape new)."

	^costume closed: aBoolean.
! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/15/2005 11:28'!
collection
	"Answer the delegate responsible for receiving collection messages.
	For generic players, this will be the elements."
	^elements! !

!CPlayer methodsFor: 'collisions' stamp: 'ar 3/23/2004 23:49'!
collisionsEnabled
	"Answer whether the receiver's will participate in collision detection"
	<bewareOf: #collisionsEnabledChanged>
	^costume collisionsEnabled! !

!CPlayer methodsFor: 'collisions' stamp: 'ar 3/23/2004 23:49'!
collisionsEnabled: aBool
	"Indicate whether the receiver's will participate in collision detection"
	^costume collisionsEnabled: aBool! !

!CPlayer methodsFor: 'color and fill' stamp: 'ar 3/23/2004 17:25'!
color
	<bewareOf: #fillChanged>
	^fill color! !

!CPlayer methodsFor: 'color and fill' stamp: 'das 11/20/2009 17:03'!
color: aValue
	"color is the Color of the interior of the object.  When a gradient (shading from one color to another) is used, this color is the color at the left or top edge.  If Fill Alpha is zero, the color can't be seen, since the object will be transparent.[general]
	In the Fill section, drag in the color space and move the mouse to a color.  The red box shows what color is selected.[colorTool]"


"When aValue = Color transparent, (fill color:) will set the fillStyle to be #none.  This is to force there to be a single way to be transparent."

	^fill color: (self ensureColor: aValue).
! !

!CPlayer methodsFor: 'queries' stamp: 'ar 3/23/2004 23:55'!
colorAt: aPoint
	^self colorAt: aPoint behind: nil! !

!CPlayer methodsFor: 'queries' stamp: 'ar 3/23/2004 23:56'!
colorAt: aPoint behind: aPlayer
	^self costume colorAt: aPoint behind: aPlayer! !

!CPlayer methodsFor: 'animating' stamp: 'ar 5/28/2005 21:43'!
comeToFront
	"Make me the front-most object"
	^costume comeToFront! !

!CPlayer methodsFor: 'basic' stamp: 'ar 1/12/2005 22:54'!
container
	"Answer the container of the receiver"
	<bewareOf: #containerChanged>
	"@@@ FIXME: Costumes should really answer proper players by default but the primitive ones still have too many assumptions about it. So we probably send #asPlayer WAY to often, we need to check for nil containers and all that crap. What this SHOULD say (and bless you if you fix the primitive costumes to make that work!!) is:
		^costume container
	but we ain't quite there yet..."
	myCostume ifNil:[^ nil].	"If no costume, it has no container."
	^(costume container ifNil:[^nil]) asPlayer! !

!CPlayer methodsFor: 'basic' stamp: 'tk 8/11/2004 11:04'!
container: aPlayer
	"A Player's container is the larger graphical object that holds the Player as a element.  When the container is moved on the screen, all of its elements, including this object, move with it."

	(aPlayer isIn: self) 
		ifTrue:[aPlayer container: self container].
	aPlayer elements add: self.
	^aPlayer! !

!CPlayer methodsFor: 'drawing' stamp: 'bf 5/10/2004 13:49'!
containsPoint: aPoint
	"Answer whether aPoint is in the receiver.
	Used for hit-detection in userDraw components."
	^self costume containsPoint: aPoint! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:21'!
corner
	<bewareOf: #geometryChanged>
	^geometry corner! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:58'!
corner: aValue
	"corner is the location of lower right corner of an object.  It is a Point and is measured from the container's origin."
 
	^geometry corner: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/16/2004 15:21'!
costume
	"Answer the receiver's costume"
	<bewareOf: #costumeChanged>
	^myCostume ifNil:[self wearDefaultCostume].! !

!CPlayer methodsFor: 'basic' stamp: 'ar 5/13/2005 13:24'!
costume: newCostume
	"A Player can have a Costume to determine how it looks on the screen.  A CheckBox has a CheckBoxCostume that keeps the checkmark object.  Many Players can themselves be costumes for other Players.  Some costumes are primitive costumes."

	^self privateChangeCostumeTo: (newCostume asCostumeFor: self)! !

!CPlayer methodsFor: 'private' stamp: 'ar 6/22/2005 23:24'!
costumeChanged
	"My costume has changed. Signal various events that originate directly or indirectly from the costume."
	self signal: #flagsChanged.
	self signal: #fillChanged.
	self signal: #geometryChanged.
	self signal: #borderChanged.
	self signal: #textChanged.
	self signal: #graphicChanged.
	self signal: #gridChanged.
	self signal: #extentChanged.! !

!CPlayer methodsFor: 'styling' stamp: 'ar 7/7/2005 12:12'!
cssTag
	"Set the css tag of the receiver"
	^costume cssTag! !

!CPlayer methodsFor: 'styling' stamp: 'ar 7/7/2005 12:12'!
cssTag: aStyleName
	"Set the css tag of the receiver"
	costume cssTag: aStyleName! !

!CPlayer methodsFor: 'queries' stamp: 'ar 3/24/2004 00:30'!
cursorPoint
	^self globalToLocal: hand position! !

!CPlayer methodsFor: 'opening' stamp: 'ar 2/23/2004 22:51'!
defaultContainer
	"Answer the default container to use when open the receiver"
	^self! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 3/24/2004 16:34'!
defaultCostume
	"Answer the default costume maker for this kind of player"
	^CPrimitiveCostume! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 3/16/2004 15:17'!
defaultPlayer
	"Answer the default player maker for the receiver."
	^nil! !

!CPlayer methodsFor: 'selecting' stamp: 'ar 3/23/2004 23:43'!
deselect
	"Deselect the receiver"
	^self costume deselect! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 1/12/2005 22:56'!
destroy
	"Destroy this object"
	| oldObject |
	self close.
	oldObject := myCostume.
	myCostume := nil.
	oldObject ifNotNil:[oldObject destroy].
	oldObject := myPlayer.
	myPlayer := nil.
	oldObject ifNotNil:[oldObject destroy].
	super destroy.! !

!CPlayer methodsFor: 'events' stamp: 'ar 4/30/2004 17:50'!
detachBehavior: anEventBehavior
	anEventBehavior target == nil ifFalse:[
		anEventBehavior target == self 
			ifFalse:[^self error: anEventBehavior name, ' is not owned by me'].
		^anEventBehavior detach].
	self eventBehaviors: (self eventBehaviors copyWithout: anEventBehavior).! !

!CPlayer methodsFor: 'geometry' stamp: 'bgf 1/3/2009 10:19'!
dialogCenter
	"The default center at which to place dialogs, progress popups, etc.
	 This can be overridden in any player serving as a world, to change the policy for that world."
	^ self center! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/25/2005 20:57'!
direction
	"Answer the direction of the receiver"
	<bewareOf: #geometryChanged>
	^geometry direction! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/25/2005 20:58'!
direction: aPoint
	"Indicate the direction of the receiver"
	^geometry direction: aPoint! !

!CPlayer methodsFor: 'selecting' stamp: 'ar 3/23/2004 23:43'!
disable
	"Disable the receiver"
	self costume disable.! !

!CPlayer methodsFor: 'private' stamp: 'ar 2/7/2004 21:28'!
dispatchEvent: event
	^self costume dispatchEvent: event! !

!CPlayer methodsFor: 'queries' stamp: 'ar 3/23/2004 23:56'!
distanceTo: aPlayer
	^self position distanceTo: aPlayer position! !

!CPlayer methodsFor: 'private' stamp: 'bgf 2/24/2009 16:41'!
dragMoveShadow
	"Make a drop-shadow for hilighting during an interactive move.
	 This is weaker than a regular drop shadow, it just needs to provide a bit of reinforcement."
	| shadow |
	shadow := CSoftShadowHighlight new.
	shadow name: #dragMoveShadow.
	shadow color: (shadow color copy alpha: 0.25).
	shadow offset: 8.
	shadow direction: 1@4.
	^ shadow.! !

!CPlayer methodsFor: 'drawing' stamp: 'bf 5/10/2004 13:49'!
drawOn: aCanvas in: drawingBounds
	"Draw the receiver in the provided bounds on aCanvas. 
	NOTE: This will only work if userDraw is enabled."
	^self costume drawOn: aCanvas in: drawingBounds! !

!CPlayer methodsFor: 'shadow' stamp: 'ar 3/23/2004 21:02'!
dropShadow
	<bewareOf: #dropShadowChanged>
	^costume dropShadow! !

!CPlayer methodsFor: 'shadow' stamp: 'bf 3/30/2004 11:41'!
dropShadow: aDropShadow
	^costume dropShadow: aDropShadow! !

!CPlayer methodsFor: 'shadow' stamp: 'ar 3/23/2004 21:04'!
dropShadowColor
	<bewareOf: #shadowChanged>
	^dropShadow color! !

!CPlayer methodsFor: 'shadow' stamp: 'das 11/20/2009 17:11'!
dropShadowColor: aValue
	^dropShadow color: (self ensureColor:aValue).! !

!CPlayer methodsFor: 'shadow' stamp: 'ar 3/23/2004 21:01'!
dropShadowEnabled
	<bewareOf: #dropShadowChanged>
	^dropShadow notNil! !

!CPlayer methodsFor: 'shadow' stamp: 'bgf 2/24/2009 00:45'!
dropShadowEnabled: aBool
	"dropShadowEnabled tells if the object has a shadow outside of its right and bottom edges.  The color and size of the shadow can be changed with dropShadowColor: and dropShadowOffset:.[general]
	In the upper section, find 'shadow'.  A checkmark means do have a drop shadow.[colorTool]"

	self dropShadowEnabled == aBool ifFalse:[
		dropShadow := aBool ifTrue:[CSoftShadowHighlight new] ifFalse:[nil].
	].
	^aBool
! !

!CPlayer methodsFor: 'shadow' stamp: 'ar 3/23/2004 21:05'!
dropShadowOffset
	<bewareOf: #shadowChanged>
	^dropShadow offset! !

!CPlayer methodsFor: 'shadow' stamp: 'bf 3/30/2004 11:34'!
dropShadowOffset: aValue
	^dropShadow offset: aValue.
! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/16/2004 15:21'!
elements
	"Answer the receiver's elements"
	<bewareOf: #elementsChanged>
	^costume elements! !

!CPlayer methodsFor: 'basic' stamp: 'tk 9/2/2004 18:09'!
elements: aCollection
	"elements is the list of objects that a Player owns as graphical parts.  Objects in this list are contained 'inside' the Player on the screen.  When the Player moves, the elements move with it.  Each object that is an element can look in its 'container' field to find the Player that owns it."

	^elements contents: aCollection! !

!CPlayer methodsFor: 'selecting' stamp: 'ar 3/23/2004 23:43'!
enable
	"(Re-)enable the receiver"
	self costume enable! !

!CPlayer methodsFor: 'selecting' stamp: 'bgf 3/1/2008 17:51'!
enabled: bEnabled
	"(Re-)enable or disable the receiver"
	bEnabled ifTrue: [ self enable ] ifFalse: [ self disable ].! !

!CPlayer methodsFor: 'Python' stamp: 'das 11/20/2009 16:54'!
ensureColor: rgba
	"Private conversion method to preserved 0-valued black"
	| alpha clr |
	rgba isColor ifTrue:[^ rgba].
	alpha := rgba bitShift: -24.
	alpha = 0 ifTrue:[^Color transparent].
	clr := Color 
				r: (rgba bitAnd: 255) 
				g: ((rgba bitShift: -8) bitAnd: 255) 
				b: ((rgba bitShift: -16) bitAnd: 255) 
				range: 255.
	alpha = 255 
		ifTrue:[^clr]
		ifFalse:[^clr alpha: alpha / 255.0]! !

!CPlayer methodsFor: 'Python' stamp: 'das 11/20/2009 16:53'!
ensurePoint: pt

	^ pt isPoint ifTrue:[pt] ifFalse:[(pt at:1)@(pt at:2)].
! !

!CPlayer methodsFor: 'Python' stamp: 'das 11/23/2009 16:44'!
ensureRect: rect

	^ rect isRectangle ifTrue:[rect] ifFalse:[(rect at:1)@(rect at:2) extent: (rect at:3)@(rect at:4)].! !

!CPlayer methodsFor: 'events' stamp: 'ar 4/30/2004 17:50'!
eventBehaviors
	"Answer the eventBehaviors of the receiver"
	<bewareOf: #eventBehaviorsChanged>
	^self propertyValueAt: #eventBehaviors ifAbsent:[#()]! !

!CPlayer methodsFor: 'events' stamp: 'ar 4/30/2004 17:50'!
eventBehaviors: aValue
	"Modify the receiver's eventBehaviors"
	^self propertyValueAt: #eventBehaviors put: aValue with: #eventBehaviorsChanged! !

!CPlayer methodsFor: 'private' stamp: 'ar 5/22/2005 17:59'!
explorerContentsDo: aBlock
	aBlock value: #costume value: myCostume.
	aBlock value: #elements value: (myCostume ifNotNil:[elements]).
	aBlock value: #container value: (myCostume ifNotNil:[container]).
	aBlock value: #player value: player.
	CPlayer instSize+1 to: self class instSize do:[:i|
		aBlock value: (self class allInstVarNames at: i) value: (self instVarAt: i).
	].
	myProperties ifNil:[^self].
	myProperties keysAndValuesDo: aBlock.! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:21'!
extent
	<bewareOf: #geometryChanged>
	^geometry extent! !

!CPlayer methodsFor: 'geometry' stamp: 'tk 8/15/2004 17:56'!
extent: aValue
	"extent is the size of the object on the screen.  The width and height are expressed as a Point.[general]
	bring up its halo, and drag the yellow handle to change the size.[halo]"

	^geometry extent: aValue.
! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/23/2004 22:09'!
fill
	<bewareOf: #fillChanged>
	^costume fill! !

!CPlayer methodsFor: 'basic' stamp: 'tk 9/15/2004 23:08'!
fill: aValue
	"the fill is the object that is in charge of filling the background inside this object.  It is usually a CCostumeFill, and may produce a single color or a shaded gradient."
	^costume fill: aValue.! !

!CPlayer methodsFor: 'color and fill' stamp: 'ar 3/23/2004 17:25'!
fillAlpha
	<bewareOf: #fillChanged>
	^fill alpha! !

!CPlayer methodsFor: 'color and fill' stamp: 'tk 1/4/2005 11:44'!
fillAlpha: aValue
	"fillAlpha tells how transparent the object will be.  Zero is completely transparent, and 1.0 is non-transparent, which is also called opaque.[general]
	In the Fill section, click the up and down arrows next to Fill Alpha, or type a number between 0.0 and 1.0 into the input box and press return. [colorTool]"
	fill alpha: aValue.
	(aValue = 0.0 and: [fillStyle ~~ #none]) ifTrue: [self fillStyle: #none].
	^ aValue.
! !

!CPlayer methodsFor: 'color and fill' stamp: 'ar 3/23/2004 17:25'!
fillAngle
	<bewareOf: #fillChanged>
	^fill angle! !

!CPlayer methodsFor: 'color and fill' stamp: 'tk 1/4/2005 11:44'!
fillAngle: aValue
	"fillAngle tells which direction a gradient fill will go.  A horizontal gradient fill has 0 degrees fill angle.  A vertical gradient has 90 degrees.  A slanted gradient has 45 degrees.  Try other values too.[general]
	In the Fill section, click the up and down arrows next to Fill Angle, or type a number of degrees into the input box and press return. [colorTool]"

	^fill angle: aValue.
! !

!CPlayer methodsFor: 'color and fill' stamp: 'ar 3/24/2004 00:26'!
fillColor
	<bewareOf: #fillChanged>
	^fill color! !

!CPlayer methodsFor: 'color and fill' stamp: 'tk 8/20/2004 16:00'!
fillColor: aValue
	"fillColor is the Color of the interior of the object.  It is identical to the color field.  When a gradient (shading from one color to another) is used, this color is the color at the left or top edge.  If Fill Alpha is zero, the color can't be seen, since the object will be transparent.[general]
	In the Fill section, drag in the color space and move the mouse to a color.  The red box shows what color is selected.[colorTool]"

	^fill color: aValue.
! !

!CPlayer methodsFor: 'color and fill' stamp: 'ar 3/23/2004 17:25'!
fillColors
	<bewareOf: #fillChanged>
	^fill colors! !

!CPlayer methodsFor: 'color and fill' stamp: 'tk 9/16/2004 14:06'!
fillColors: aValue
	"fillColors is an array of Colors.  The first entry is the same as the color field, which is the main color of the object.  When a gradient (shading from one color to another) is used, the second entry in fillColors is the color at the right or bottom edge.  The gradient goes from the first color to the second color.  If Fill Alpha is zero, the color can't be seen, since the object will be transparent.[general]
	In the Fill section, there are two tiny boxes in the color space.  One is for each entry in fillColors.  Move the mouse over one of the boxes.  When the cursor turns to a +, you can move that box.  If you just drag, you are setting the first entry (fillColor).  The red box is the most recent color you selected.[colorTool]"

	^fill colors: aValue.
! !

!CPlayer methodsFor: 'color and fill' stamp: 'ar 3/23/2004 17:26'!
fillOffset
	<bewareOf: #fillChanged>
	^fill offset! !

!CPlayer methodsFor: 'color and fill' stamp: 'tk 9/15/2004 23:13'!
fillOffset: aValue
	"When an object has a fillStyle of radial gradient, fillOffset tells the location of the center of the circles of color.  It is measured in screen dots from the upperLeft of the object."
	^fill offset: aValue.
! !

!CPlayer methodsFor: 'color and fill' stamp: 'ar 3/23/2004 17:26'!
fillStyle
	<bewareOf: #fillChanged>
	^fill style! !

!CPlayer methodsFor: 'color and fill' stamp: 'tk 3/7/2005 22:03'!
fillStyle: aValue
	"fillStyle tells how the object will be filled with color.  #simple and #none use fillColor.  gradients blend gradually from one color to another.  The colors are stored in fillColors. The gradients are .[general]
	In the Fill section, click a button in the row of square buttons below the color space. [colorTool]"

	fill style: aValue.
	(aValue == #none and: [fillAlpha > 0.0]) ifTrue: [self fillAlpha: 0.0].
	^ aValue.
! !

!CPlayer methodsFor: 'queries' stamp: 'ar 5/22/2005 18:05'!
findDraggablePlayerAt: aPoint
	"Search the receiver to find the first draggable player located at the given position."
	^costume findDraggablePlayerAt: aPoint
! !

!CPlayer methodsFor: 'queries' stamp: 'ar 3/23/2004 23:58'!
findPlayerAt: aPoint
	"Search the receiver to find the first player located at the given position."
	^costume findPlayerAt: aPoint! !

!CPlayer methodsFor: 'selecting' stamp: 'ar 3/23/2004 23:44'!
fire
	"Any player can be asked to act as a button and fire"
	self signal: #fire.! !

!CPlayer methodsFor: 'selecting' stamp: 'ar 3/6/2004 22:59'!
flash
	^self costume flash! !

!CPlayer methodsFor: 'properties' stamp: 'ar 3/23/2004 23:02'!
floating
	"Answer the receiver's floating property"
	<bewareOf: #flagsChanged>
	^costume floating! !

!CPlayer methodsFor: 'properties' stamp: 'tk 2/24/2005 16:23'!
floating: aBool
	"clipping tells if this object appears in front of all normal objects in this project.  The ColorTool and other control panels often have floating set to true.[general]
	In the upper section, find 'floating'.  A checkmark means the object appears in front of all normal objects.[colorTool]"

	^costume floating: aBool! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/24/2004 15:02'!
font
	^self textFont! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/24/2004 15:02'!
font: aFont
	^self textFont: aFont! !

!CPlayer methodsFor: 'animating' stamp: 'ar 6/25/2005 20:04'!
forwardBy: amount
	"Move me forward by the specified amount"
	self moveBy: self direction * amount.! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:21'!
fullBounds
	^geometry fullBounds! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/23/2004 22:08'!
geometry
	"Answer the receiver's geometry"
	^costume geometry! !

!CPlayer methodsFor: 'basic' stamp: 'tk 1/4/2005 13:14'!
geometry: aGeometry
	"Geometry is the aspect of an object that determines it's size, location on the screen, and angle of rotation.  Usually the geometry object is the costume, which delegates the work to its primitive costume.  Because geometry messages, such as 'center', are used so often, many of them can be sent to a Player instead of being sent the Player's geometry.  You can say (self center) instead of (self geometry center)."

	^costume geometry: aGeometry! !

!CPlayer methodsFor: 'private' stamp: 'ar 3/22/2004 02:33'!
getCostume
	^myCostume! !

!CPlayer methodsFor: 'private' stamp: 'ar 3/22/2004 02:33'!
getPlayer
	^myPlayer! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:23'!
globalBounds
	<bewareOf: #frameChanged>
	^geometry globalBounds! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:58'!
globalBounds: aRectangle
	^geometry globalBounds: (self ensureRect: aRectangle)! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:23'!
globalPosition
	<bewareOf: #frameChanged>
	^geometry globalPosition! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:59'!
globalPosition: aPoint
	^geometry globalPosition: (self ensurePoint: aPoint)! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 12:46'!
globalToLocal
	^costume globalToLocal! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 12:47'!
globalToLocal: anObject
	^costume globalToLocal: anObject! !

!CPlayer methodsFor: 'graphic' stamp: 'ar 3/23/2004 19:19'!
graphic
	<bewareOf: #graphicChanged>
	^costume graphic! !

!CPlayer methodsFor: 'graphic' stamp: 'tk 9/15/2004 23:22'!
graphic: aValue
	"graphic is a picture (also called an image, a bitmap, or a Form) that this object will display."
	^costume graphic: aValue.
! !

!CPlayer methodsFor: 'graphic' stamp: 'ar 3/23/2004 19:20'!
graphicAnchorPoint
	<bewareOf: #graphicChanged>
	^graphic anchorPoint! !

!CPlayer methodsFor: 'graphic' stamp: 'bf 3/30/2004 11:31'!
graphicAnchorPoint: aValue
	^graphic anchorPoint: aValue.
! !

!CPlayer methodsFor: 'graphic' stamp: 'ar 3/23/2004 20:26'!
graphicEnabled
	^graphic notNil! !

!CPlayer methodsFor: 'graphic' stamp: 'bf 3/30/2004 11:31'!
graphicEnabled: aBool
	self graphicEnabled = aBool ifFalse:[
		graphic := aBool ifTrue:[CCostumeGraphic new] ifFalse:[nil].
	].
	^aBool! !

!CPlayer methodsFor: 'graphic' stamp: 'ar 3/23/2004 19:20'!
graphicFit
	<bewareOf: #graphicChanged>
	^graphic fit! !

!CPlayer methodsFor: 'graphic' stamp: 'tk 1/4/2005 12:10'!
graphicFit: aValue
	"Determine how the graphic should be drawn within this object's bounds:
		#rigid - use original size
		#scale - scale to size (preserving aspect ratio)
		#stretch - stretch to bounds.
		#tile - draw the graphic repeatedly
		#autoScale - be rigid when drawing into an area bigger than my natural size
				- scale when drawing into an smaller area.
	"
	^graphic fit: aValue.
! !

!CPlayer methodsFor: 'graphic' stamp: 'ar 3/23/2004 19:33'!
graphicForm
	<bewareOf: #graphicChanged>
	^(graphic ifNil:[^nil]) form! !

!CPlayer methodsFor: 'graphic' stamp: 'tk 1/4/2005 12:12'!
graphicForm: aForm
	"The Form (picture, bitmap) that will be shown on the screen."

	^ graphic 
		ifNil:[graphic := aForm]
		ifNotNil:[graphic form: aForm].
! !

!CPlayer methodsFor: 'graphic' stamp: 'ar 3/23/2004 19:20'!
graphicOffset
	<bewareOf: #graphicChanged>
	^graphic offset! !

!CPlayer methodsFor: 'graphic' stamp: 'tk 1/4/2005 12:16'!
graphicOffset: aPoint
	"graphicOffset is a Point that tells where the 'position' of this graphic is.  0@0 means the upper left.  To make a graphic claim that its position is in its center, set graphicOffset to be (width/2)@(height/2)."
	^graphic offset: aPoint.
! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/23/2004 15:18'!
grid
	"Answer the receiver's grid"
	^costume grid! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/23/2004 15:27'!
grid: aGrid
	"Modify the receiver's grid"
	^costume grid: aGrid! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/23/2004 15:25'!
gridEnabled
	"Answer true if the receiver's grid is present and enabled"
	| myGrid |
	myGrid := self grid.
	myGrid ifNil:[^false].
	^myGrid enabled! !

!CPlayer methodsFor: 'basic' stamp: 'bf 3/30/2004 11:26'!
gridEnabled: aBool
	"Enable gridding. If no grid is present and aBool is true, associate a default grid."
	| myGrid |
	self gridEnabled = aBool ifTrue:[^self].
	myGrid := self grid.
	myGrid ifNil:[
		myGrid := CCostumeGrid new.
		self grid: myGrid.
	].
	^myGrid enabled: aBool.! !

!CPlayer methodsFor: 'layout' stamp: 'ar 3/11/2004 12:44'!
hResizing
	^self costume hResizing! !

!CPlayer methodsFor: 'layout' stamp: 'tk 1/4/2005 12:23'!
hResizing: aValue
	"hResizing tells how I will change width when my elements or my container change.  My container is the object I am inside of.  Meaning depends on what layout I have, such as TableLayout.  
	#rigid		 - do not change my size
	#spaceFill	 - resize to fill my container's available space
	#shrinkWrap - resize to fit my elements
	"
	^costume hResizing: aValue.
! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/19/2004 13:21'!
hScrollable
	"Answer the hScrollable of the receiver"
	<bewareOf: #hScrollableChanged>
	^self propertyValueAt: #hScrollable! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/19/2004 13:21'!
hScrollable: aValue
	"Modify the receiver's hScrollable"
	^self propertyValueAt: #hScrollable put: aValue with: #hScrollableChanged! !

!CPlayer methodsFor: 'basic' stamp: 'pe 7/13/2005 16:50'!
haloTarget
	"Answer the halo's default target when clicking on me"
	^self! !

!CPlayer methodsFor: 'basic' stamp: 'tk 1/4/2005 11:54'!
hand
	"The object that represents the user's cursor.  It looks like an arrow and moves when you move the mouse.  When you pick up an object and move it, it is carried in the hand."

	^self costume hand! !

!CPlayer methodsFor: 'private' stamp: 'ar 5/10/2005 14:26'!
handleError: anError ifNone: aBlock
	"Handle an exception by allowing our costume to handle #error events"
	(self handlesEvent: #error) ifTrue:[
		"Why synchronous? The event handler may need to invoke the exception response methods (abort, restart, resume) and right now those aren't safe across processes. Invoking them from the wrong place can have dire consequences so we use a synchronous signal.
		In addition, using the synchronous signal means there will only be one handler response ever be executed - multiple handlers might be registered but they won't execute."
		^self synchronousSignal: #error with: anError.
	].
	^aBlock value! !

!CPlayer methodsFor: 'tabbing' stamp: 'ar 9/24/2007 11:55'!
handleTab: evt from: aPlayer after: previous
	"Try to find the next tab handler in the receiver"
	| handler |
	handler := self tabStopAfter: previous.
	handler ifNil:[
		container ifNotNil:[container handleTab: evt from: aPlayer after: self].
	].
	handler == previous ifFalse:[
		aPlayer ifNotNil:[aPlayer asPrimCostume signal: #tabLeave with: evt with: handler].
		handler ifNotNil:[handler asPrimCostume signal: #tabEnter with: evt with: aPlayer].
	].
! !

!CPlayer methodsFor: 'private' stamp: 'ar 5/6/2004 16:19'!
handlesAnyOf: eventNames
	(super handlesAnyOf: eventNames) ifTrue:[^true].
	myPlayer ifNil:[^false].
	myPlayer getCostume == self ifFalse:[^false].
	^myPlayer handlesAnyOf: eventNames! !

!CPlayer methodsFor: 'private' stamp: 'ar 5/6/2004 16:18'!
handlesEvent: eventName
	(super handlesEvent: eventName) ifTrue:[^true].
	myPlayer ifNil:[^false].
	myPlayer getCostume == self ifFalse:[^false].
	^myPlayer handlesEvent: eventName! !

!CPlayer methodsFor: 'queries' stamp: 'bf 5/13/2005 18:16'!
hasKeyboardFocus
	^costume hasKeyboardFocus
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/27/2005 14:55'!
heading
	"Answer the receiver's angle, measured clockwise from north in degrees"
	<bewareOf: #geometryChanged>
	^geometry heading! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/27/2005 14:55'!
heading: newHeading
	"Indicate the receiver's angle, measured clockwise from north in degrees"
	^geometry heading: newHeading! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:19'!
height
	<bewareOf: #geometryChanged>
	^geometry height! !

!CPlayer methodsFor: 'geometry' stamp: 'tk 8/11/2004 17:14'!
height: aValue
	"height tells how tall the object is.  It is measured in screen dots."

	^geometry height: aValue.
! !

!CPlayer methodsFor: 'highlights' stamp: 'ar 6/26/2005 18:53'!
highlightNamed: aSymbol
	"Answer the highlight with the given name"
	^costume highlightNamed: aSymbol! !

!CPlayer methodsFor: 'drawing' stamp: 'ar 3/11/2004 12:44'!
icon
	<bewareOf: #iconChanged>
	^(self propertyValueAt: #icon) ifNil:[self makeIcon]! !

!CPlayer methodsFor: 'drawing' stamp: 'ar 3/13/2003 22:31'!
icon: newIcon
	^self propertyValueAt: #icon put: newIcon with: #iconChanged! !

!CPlayer methodsFor: 'properties' stamp: 'ar 3/23/2004 23:03'!
infinite
	"Answer the receiver's infinite property"
	<bewareOf: #flagsChanged>
	^costume infinite! !

!CPlayer methodsFor: 'properties' stamp: 'tk 2/24/2005 16:23'!
infinite: aBool
	"When an object has infinite set to true, it is treated as an infinite plane.  It's extent is the same as its container at all times.[general]
	In the upper section, find 'infinite'.  A checkmark means for object always fills its container.[colorTool]"

	^costume infinite: aBool! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 3/15/2005 11:52'!
initialize
	"Initialize the receiver"
	super initialize.
	myPlayer ifNil:[
		myPlayer := self makeDefaultPlayer.
		myPlayer ifNotNil:[myPlayer costume: self]].! !

!CPlayer methodsFor: 'drawing' stamp: 'ar 3/15/2005 11:53'!
invalidate
	"Invalidate the receiver"
	myCostume ifNotNil:[myCostume invalidate]! !

!CPlayer methodsFor: 'drawing' stamp: 'ar 3/15/2005 11:57'!
invalidate: aRectangle
	"Invalidate the receiver"
	myCostume ifNotNil:[myCostume invalidate: aRectangle]! !

!CPlayer methodsFor: 'layout' stamp: 'ar 12/19/2004 19:17'!
invalidateLayout
	"Invalidate my layout"
	myCostume ifNotNil:[myCostume invalidateLayout].! !

!CPlayer methodsFor: 'testing' stamp: 'ar 3/16/2004 15:23'!
isIn: aContainer
	"Answer whether the receiver inside the given container"
	| cc |
	"Answer true for a query against nil since ultimately all player hierarchies end up there"
	aContainer ifNil:[^true].
	"Open-coded for speed."
	cc := self container.
	[cc == nil] whileFalse:[
		aContainer == cc ifTrue:[^true].
		cc := cc container].
	^false! !

!CPlayer methodsFor: 'testing' stamp: 'ar 8/3/2004 17:20'!
isKeyPressed: aKey
	"Answer whether the given key is currently pressed"
	"FIXME: This is most definitely not correct ...."
	| w |
	(w := self world) ifNotNil:[^w costume pressedKeys includes: aKey].
	^false
! !

!CPlayer methodsFor: 'opening' stamp: 'tk 7/14/2004 11:25'!
isOpen
	"Answer whether I exist in some world"
	^self asPrimCostume isInWorld! !

!CPlayer methodsFor: 'testing' stamp: 'ar 3/16/2004 15:24'!
isPlayer
	"Answer whether the receiver represents a player"
	^true! !

!CPlayer methodsFor: 'tabbing' stamp: 'ar 1/25/2007 14:55'!
isTabStop
	"Answer whether tabbing should stop at the receiver"
	^costume isTabStop! !

!CPlayer methodsFor: 'testing' stamp: 'tk 11/20/2004 20:03'!
isTileLike
	"Answer whether the receiver represents a tile-scripting element (player or costume)"
	^false! !

!CPlayer methodsFor: 'testing' stamp: 'ar 3/16/2004 15:24'!
isTilePlayer
	"Answer whether the receiver represents a tile-scripting element"
	^false! !

!CPlayer methodsFor: 'testing' stamp: 'ar 3/16/2004 15:25'!
isUnderMouse
	"Answer whether the receiver is currently under the hand"
	<bewareOf: #mouseOverChanged>
	^self hand isOver: self! !

!CPlayer methodsFor: 'testing' stamp: 'ar 7/15/2003 16:36'!
isWorld
	"Do I act as a world?
	Note: Later we will modify this based on what exactly a 'world' means"
	^self costume isWorld! !

!CPlayer methodsFor: 'layout' stamp: 'bf 3/30/2004 11:42'!
layout
	^costume layout! !

!CPlayer methodsFor: 'layout' stamp: 'bf 3/30/2004 11:42'!
layout: aLayout
	^costume layout: aLayout! !

!CPlayer methodsFor: 'layout' stamp: 'ar 3/24/2004 02:15'!
layoutBounds
	^costume layoutBounds! !

!CPlayer methodsFor: 'layout' stamp: 'ar 3/24/2004 02:15'!
layoutBounds: newBounds
	^costume layoutBounds! !

!CPlayer methodsFor: 'layout' stamp: 'ar 3/11/2004 12:43'!
layoutFrame
	^self costume layoutFrame! !

!CPlayer methodsFor: 'layout' stamp: 'bf 4/27/2004 01:14'!
layoutFrame: aFrame
	^ self costume layoutFrame: aFrame! !

!CPlayer methodsFor: 'layout' stamp: 'ar 3/24/2004 03:40'!
layoutInset
	^costume layoutInset! !

!CPlayer methodsFor: 'layout' stamp: 'ar 3/24/2004 03:40'!
layoutInset: aPoint
	costume layoutInset: aPoint! !

!CPlayer methodsFor: 'layout' stamp: 'ar 3/24/2004 02:08'!
layoutNeeded
	^costume layoutNeeded! !

!CPlayer methodsFor: 'layout' stamp: 'ar 3/24/2004 02:08'!
layoutNeeded: aBool
	costume layoutNeeded: aBool! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:19'!
left
	<bewareOf: #geometryChanged>
	^geometry left! !

!CPlayer methodsFor: 'geometry' stamp: 'tk 9/14/2004 14:32'!
left: aValue
	"left is the position of this object in the horizontal direction.  left is another name for x.  It is the number of screen dots from the left edge of the object's container to the left edge of this object.  Increasing left moves the object to the right.  left is the ""x part"" of the object's position." 

	^geometry left: aValue.
! !

!CPlayer methodsFor: 'geometry' stamp: 'bgf 2/23/2008 12:43'!
leftCenter
	<bewareOf: #geometryChanged>
	^geometry leftCenter! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:59'!
leftCenter: aValue

	^geometry leftCenter: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'private' stamp: 'ar 3/25/2004 12:05'!
listItemValue: aValue
	"Request the receiver to represent a particular value in a list. In most situations, we will simply carry the value as our own (since many, if not most, elements in lists will be bare-bone objects rather than players) but in some situations we may want to interpret it differently. Most importantly we may want to have the lists element represent another costume for the object in the list and this situation cannot be handled well via a #valueChanged handler since setting one's player via #valueChanged will implicitly modify the receiver's value (which causes all sorts of trouble since it is never clear if the value is now the player or the player's value). For this situations, one can simply implement #listItemValue: to provide the appropriate conversion. Here, we simply assume that most list items will show their value."
	self value: aValue! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:21'!
localBounds
	^geometry localBounds! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 12:46'!
localToGlobal
	^costume localToGlobal! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 12:47'!
localToGlobal: anObject
	^costume localToGlobal: anObject! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 5/22/2005 18:06'!
localToWorld
	^costume localToWorld
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 5/22/2005 18:06'!
localToWorld: anObject
	^costume localToWorld: anObject
! !

!CPlayer methodsFor: 'properties' stamp: 'ar 3/23/2004 23:04'!
locked
	"Answer the receiver's locked property"
	<bewareOf: #flagsChanged>
	^costume locked! !

!CPlayer methodsFor: 'properties' stamp: 'tk 2/24/2005 16:22'!
locked: aBool
	"When an object is locked, it does not respond to mouse clicks.  It lets its container handle mouse events.  It is not possible to bring up a halo on a locked object.  If you lock an object in a Project, and move the halo to some other object, it is very difficult to get the object unlocked again!![general]
	In the upper section, find 'locked'.  A checkmark means the object is locked.[colorTool]"

	^costume locked: aBool.! !

!CPlayer methodsFor: 'private' stamp: 'ar 4/30/2004 16:09'!
makeDefaultCostume
	"This method is deprecated, but for now keep it around for compatibility"
! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 5/14/2005 16:11'!
makeDefaultPlayer
	"Create a default player for the receiver if one is needed"
	| playerMaker |
	playerMaker := self defaultPlayer ifNil:[^nil].
	^playerMaker asPlayerFor: self! !

!CPlayer methodsFor: 'drawing' stamp: 'ar 3/9/2003 21:38'!
makeIcon
	"Create an icon for the receiver"
	^self makeIcon: 32@32! !

!CPlayer methodsFor: 'drawing' stamp: 'tk 7/14/2004 11:25'!
makeIcon: iconExtent
	"Create an icon for the receiver"
	| form |
	form := self asPrimCostume makeThumbnail: iconExtent-2.
	form := form shapeBorder: Color black width: 1.
	^form! !

!CPlayer methodsFor: 'animating' stamp: 'ar 3/21/2004 14:03'!
makeSound: soundName
	| snd |
	snd := SampledSound soundNamed: soundName.
	(snd == nil or:[snd isPlaying]) ifTrue:[^nil].
	snd play.! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/16/2004 15:22'!
menu
	"Answer the receiver's menu"
	<bewareOf: #menuChanged>
	^self propertyValueAt: #menu! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/16/2004 15:22'!
menu: aValue
	"Modify the receiver's menu"
	^self propertyValueAt: #menu put: aValue with: #menuChanged! !

!CPlayer methodsFor: 'layout' stamp: 'ar 10/8/2005 11:13'!
minHeight
	"Answer the minimal height the receiver can be layed out in."
	^costume minHeight! !

!CPlayer methodsFor: 'layout' stamp: 'ar 10/8/2005 11:13'!
minHeight: aNumber
	"Answer the minimal height the receiver can be layed out in."
	costume minHeight: aNumber.! !

!CPlayer methodsFor: 'layout' stamp: 'ar 10/8/2005 11:12'!
minWidth
	"Answer the minimal width the receiver can be layed out in."
	^costume minWidth! !

!CPlayer methodsFor: 'layout' stamp: 'ar 10/8/2005 11:13'!
minWidth: aNumber
	"Set the minimal width the receiver can be layed out in."
	costume minWidth: aNumber.! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/24/2004 00:30'!
mouseCursor
	<bewareOf: #mouseCursorChanged>
	^self costume mouseCursor! !

!CPlayer methodsFor: 'basic' stamp: 'tk 1/4/2005 11:57'!
mouseCursor: aCursor
	"A special cursor that shows when the user's mouse is inside this object.  The cursor is changed to this when the pointer enters, and changed back when it leaves."

	^ self costume mouseCursor: aCursor.
! !

!CPlayer methodsFor: 'animating' stamp: 'ar 3/29/2003 22:45'!
moveBy: delta
	^self costume moveBy: delta asPoint! !

!CPlayer methodsFor: 'animating' stamp: 'ar 6/27/2005 13:43'!
moveBy: delta duration: seconds
	| dt lastTime nextTime amount remain |
	seconds <= 0 ifTrue:[^self moveBy: delta].
	dt := 1.0 / (seconds * 1000.0).
	remain := 1.0.
	lastTime := Time millisecondClockValue.
	[remain > 0.0] whileTrue:[
		self waitTick.
		nextTime := Time millisecondClockValue.
		amount := ((nextTime - lastTime max: 0) * dt) min: remain.
		self moveBy: delta * amount.
		remain := remain - amount.
		lastTime := nextTime.
	].
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/25/2005 20:58'!
normal
	"Answer the normal of the receiver"
	<bewareOf: #geometryChanged>
	^geometry normal! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:59'!
normal: aPoint
	"Indicate the normal of the receiver"
	^geometry normal: (self ensurePoint: aPoint)! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/25/2005 19:51'!
nose
	<bewareOf: #geometryChanged>
	^geometry nose! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/25/2005 19:52'!
nose: aPoint
	^geometry nose: aPoint! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 2/9/2003 01:55'!
on: aValue
	"Initialize the receivers 'value'"
	self value: aValue.! !

!CPlayer methodsFor: 'private' stamp: 'ar 6/27/2005 13:44'!
onDrag
	"Drag the receiver"
	| myContainer myTransform myIndex myPosition evt |
	container ifNil:[^self].
	self floating ifTrue:[^self onDragMove].
	"Remember my container and my transform for drop rejection"
	myContainer := container.
	myIndex := container elements indexOf: self.
	myTransform := self transform.
	myPosition := container localToGlobal: self position.

	self hand dragContents: self asPlayer.
	self onDragMove.

	"Catch any attempts to reject the drop operation"
	evt := self catchAnyOf:{self. #dropReject} during:[
		self hand dragContents: nil.
		self hand drop: self.
		"Wait for a single tick after the drop to see if it should be rejected"
		self waitTick.
	].
	evt ifNotNil:["Yup, it got rejected"
		self moveBy: (myPosition - self position) duration: 0.2.
		myContainer add: self at: myIndex.
		self transform: myTransform.
	].
! !

!CPlayer methodsFor: 'private' stamp: 'bgf 9/30/2009 09:45'!
onDragMove
	"Drag the receiver but do not drop it when the move completes"
	| nextPos moved cursor delta wasLocked offset |
	container ifNil:[^self].
	wasLocked := self locked.

	self transformBy: container localToWorld.
	self world add: self.
	self locked: true.

	cursor := self mouseCursor.
	self mouseCursor: (cursor ifNil:[Cursor normal]) asForm.
	self hand cursorFocus: self.

	offset := self cursorPoint.
	" shadow := self dragMoveShadow.   *** retracted, causes jump/jumpback when first added,
									 when running on slow-graphics machines under openGL."
	moved := false.
	self hand forEach: #mouseMove do:[
		nextPos := self cursorPoint.
		delta := nextPos - offset.
		moved ifFalse:[
			" self addHighlight: shadow. 		*** see above "
			moved := true.
		].
		self transform: ((COffsetTransform withOffset: delta) transformedBy: self transform).
	] until: #mouseUp.
	self mouseCursor: cursor.
	self hand releaseMouseFocus: self.
	" self removeHighlight: #dragMoveShadow.	*** see above "
	self locked: wasLocked.
	self hand releaseCursorFocus: self.! !

!CPlayer methodsFor: 'events' stamp: 'ar 3/24/2004 01:36'!
onKeyDown: keySpec signal: eventName
	"Signal the given event when the key specified by keySpec is pressed.
	Examples:
		editor onKeyDown: $1 signal: #onePressed.
		editor onKeyDown: #(cmd $1) signal: #cmdOnePressed.
		editor onKeyDown: #(ctrl shift a) signal: #ctrlShiftAPressed.
		editor onKeyDown: #(13) signal: #crPressed.
		editor onKeyDown: #(27) signal: #cancel.
	To remove a key specification use nil as event name, e.g.,
		editor onKeyDown: $1 signal: nil.
		editor onKeyDown: #(13) signal: nil.
	This method can be used by clients who don't want to handle *all* keyDown events."
	^costume onKeyDown: keySpec signal: eventName! !

!CPlayer methodsFor: 'events' stamp: 'ar 3/24/2004 01:19'!
onKeyStroke: keySpec signal: eventName
	"Signal the given event when the key specified by keySpec is input.
	To remove a key specification use nil as event name.
	This method can be used by clients who don't want to handle *all* keyStroke events."
	^costume onKeyStroke: keySpec signal: eventName! !

!CPlayer methodsFor: 'events' stamp: 'ar 3/24/2004 01:07'!
onKeyUp: keySpec signal: eventName
	"Signal the given event when the key specified by keySpec is released.
	To remove a key specification use nil as event name.
	This method can be used by clients who don't want to handle *all* keyUp events."
	^costume onKeyUp: keySpec signal: eventName! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 3/7/2004 18:19'!
onPlayerChanged
	"backstop"! !

!CPlayer methodsFor: 'opening' stamp: 'ar 2/7/2005 01:00'!
open
	^self open: self defaultContainer! !

!CPlayer methodsFor: 'opening' stamp: 'ar 12/20/2004 19:30'!
open: aPlayer
	aPlayer == self ifFalse:[^aPlayer open].
	^self open: aPlayer at: hand position! !

!CPlayer methodsFor: 'opening' stamp: 'ar 5/26/2005 18:19'!
open: aPlayer at: aPoint
	"FIXME: Where to open the player should probably be a policy decision but where would we put it???"
	^self open: aPlayer in: Processor activeProject world at: aPoint! !

!CPlayer methodsFor: 'opening' stamp: 'ar 12/20/2004 19:31'!
open: aPlayer in: aContainer
	aPlayer == self ifFalse:[
		aPlayer openIn: aContainer.
		^self].
	^self open: self defaultContainer in: aContainer at: self hand position - 100! !

!CPlayer methodsFor: 'opening' stamp: 'bgf 5/6/2007 12:30'!
open: aPlayer in: aContainer at: aPoint
	| parent |
	aPlayer == self ifFalse:[
		aPlayer openAt: aPoint in: aContainer.
		^self].
	parent := aContainer ifNil:[self error: 'Fix this - nil container does not work'].

	aPlayer fullBounds.	"Workaround, so the geometry is up to date - without this,
							 reopening a player can place it differently than when new."

	aPlayer topLeft: (parent globalToLocal: aPoint).
	(aContainer ifNil:[self world]) add: aPlayer.
	aPlayer beOnScreen.! !

!CPlayer methodsFor: 'opening' stamp: 'ar 2/23/2004 22:51'!
openAt: aPoint
	self open: self defaultContainer at: aPoint! !

!CPlayer methodsFor: 'opening' stamp: 'ar 2/23/2004 22:51'!
openAt: aPoint in: aContainer
	self open: self defaultContainer in: aContainer at: aPoint! !

!CPlayer methodsFor: 'opening' stamp: 'ar 2/23/2004 22:51'!
openCenteredAt: aPoint
	^self open: self defaultContainer at: aPoint - (self costume fullBounds extent // 2)! !

!CPlayer methodsFor: 'opening' stamp: 'bgf 1/3/2009 10:24'!
openCenteredIn: aWorld
	"Appear, centered on the dialog centering point of the world."
	^self open: self defaultContainer
		   in: aWorld 
		   at: aWorld dialogCenter - (self costume fullBounds extent // 2)! !

!CPlayer methodsFor: 'opening' stamp: 'ar 2/23/2004 22:52'!
openIn: aContainer
	^self open: self defaultContainer in: aContainer! !

!CPlayer methodsFor: 'opening' stamp: 'ar 2/23/2004 22:52'!
openIn: aContainer at: aPoint
	^self open: self defaultContainer in: aContainer at: aPoint! !

!CPlayer methodsFor: 'opening' stamp: 'ar 6/5/2005 12:14'!
openInHand
	^self openInHand: self hand! !

!CPlayer methodsFor: 'opening' stamp: 'ar 6/5/2005 12:14'!
openInHand: aHand

	aHand attach: self! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:21'!
origin
	<bewareOf: #geometryChanged>
	^geometry origin! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 16:59'!
origin: aValue
	"origin is the top left of an object as if it were not rotated by an angle.  When angle is zero, origin is the same as position and the same as topLeft."

	^geometry origin: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/26/2005 12:07'!
pivot
	"Answer the pivot point, e.g., the distance by which the position 
	is offset from the origin of the local coordinate system"
	<bewareOf: #geometryChanged>
	^geometry pivot
! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 17:00'!
pivot: aPoint
	"Set the pivot point, e.g., the distance by which the position 
	is offset from the origin of the local coordinate system"
	^geometry pivot: (self ensurePoint: aPoint)
! !

!CPlayer methodsFor: 'basic' stamp: 'tk 8/11/2004 11:00'!
player
	"When an object is being used as a costume, this field holds the Player it is a costume for.  The Player itself has nil in its player field."

	<bewareOf: #playerChanged>
	^myPlayer! !

!CPlayer methodsFor: 'basic' stamp: 'tk 8/11/2004 11:01'!
player: newPlayer
	"When an object is being used as a costume, this field holds the Player it is a costume for.  The Player itself has nil in its player field."

	^ self privateChangePlayerTo: newPlayer.! !

!CPlayer methodsFor: 'opening' stamp: 'sw 5/3/2005 11:03'!
playersForTrash
	"If the user makes a 'throwing away' gesture, via halo or via drag and drop, on this player, answer a list of players that should be individually added to any 'trash respository' that the containing project might be maintaining.  Usually this will be just the player itself, as per this default implementation, but it could be more or fewer players -- see other implementors."

	^ OrderedCollection with: self! !

!CPlayer methodsFor: 'geometry' stamp: 'tk 9/14/2004 14:28'!
position
	"position is the location of the upper left corner of an object on the screen.  It is a Point, and is measured in screen dots, also called pixels.  When an object is an element of a container, the position is measured from the upper left of the container.  See the comments for x and y.[general]
	bring up its halo, and drag the brown handle to move the object to the proper location.[halo]"

	<bewareOf: #geometryChanged>
	^geometry position! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 17:00'!
position: aValue
	^geometry position: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'collisions' stamp: 'ar 3/23/2004 23:50'!
preciseCollisions
	"Answer whether the receiver will do pixel-precise collisions"
	<bewareOf: #preciseCollisionsChanged>
	^self costume preciseCollisions! !

!CPlayer methodsFor: 'collisions' stamp: 'bf 3/30/2004 11:42'!
preciseCollisions: aBool
	"Indicate whether the receiver will do pixel-precise collisions"
	^costume preciseCollisions: aBool.! !

!CPlayer methodsFor: 'layout' stamp: 'ar 8/18/2005 21:27'!
preferredExtent
	^costume preferredExtent! !

!CPlayer methodsFor: 'layout' stamp: 'ar 8/18/2005 21:27'!
preferredHeight
	^self preferredExtent y! !

!CPlayer methodsFor: 'layout' stamp: 'ar 8/18/2005 21:27'!
preferredWidth
	^self preferredExtent x! !

!CPlayer methodsFor: 'private' stamp: 'tk 7/1/2004 16:16'!
printOn: aStream
	super printOn: aStream.
	self componentBinding ifNotNil:[
		aStream nextPut:$[; 
			nextPutAll: self componentBinding name; 
		nextPut: $].
	].! !

!CPlayer methodsFor: 'private' stamp: 'ar 6/20/2005 20:52'!
privateChangeCostumeTo: newCostume
	"Private!! Change my costume to be newCostume.

	WARNING: This method must maintain the fundamental invariants of the framework and changing it is likely to get you entirely unpredictable results. Don't do it okay? At least not until you know precisely what you're doing."
	| oldCostume |

	"get out quickly if we can"
	(oldCostume := myCostume) == newCostume ifTrue:[^self].

	"Set my costume"
	myCostume := newCostume.

	"Update my internal constraints"
	self privateReplacedCostume: oldCostume by: newCostume.

	"Notify anyone interested in the change"
	self signalChanged: #costumeChanged from: oldCostume to: newCostume.

	^newCostume! !

!CPlayer methodsFor: 'private' stamp: 'ar 9/8/2004 16:24'!
privateChangePlayerTo: newPlayer
	"Private!! Change my player to be newPlayer.

	WARNING: This method must maintain the fundamental invariants of the framework and changing it is likely to get you entirely unpredictable results. Don't do it okay? At least not until you know precisely what you're doing."
	| oldPlayer |

	"get out quickly if we can"
	(oldPlayer := myPlayer) == newPlayer ifTrue:[^newPlayer].

	"Release myPlayer here to prevent any recursion problems"
	myPlayer := nil.

	oldPlayer ifNotNil:[
		"If I was the costume (primary view) for the old player release it"
		oldPlayer privateReleaseCostume: self.
		"Stop my value updating script."
		self stopScript: #privateChangeValueTo:.
	].

	"Prevent another set of recursion issues by setting newPlayer here"
	myPlayer := newPlayer.

	"Note: Do NOT set the new costume here - inside this method 
	we don't know if a sender really wants us to be the primary 
	costume for the new player."

	"But signal #playerChanged here and now before we update the value."
	self signalChanged: #playerChanged from: oldPlayer to: newPlayer.

	"Update my value from the new player"
	newPlayer ifNotNil:[
		value := newPlayer value.
		"And start the value updating script"
		self startScript: #privateChangeValueTo:
			when:{newPlayer. #valueChanged}.
	].
	^newPlayer! !

!CPlayer methodsFor: 'private' stamp: 'ar 5/29/2005 16:38'!
privateChangeValueTo: newValue
	"Private. Really. Change my value to be newValue after our player's value has changed. 

	WARNING: This method must maintain the fundamental invariants of the framework and changing it is likely to get you entirely unpredictable results. Don't do it okay? At least not until you know precisely what you're doing.

	This method must avoid pushing the value synchronously back to the player since it takes 'one tick' to get the information from the player and at this point our player might already have new information (which we will see in the next tick)."
	| oldPlayer |
	oldPlayer := myPlayer.
	myPlayer := nil.
	value := newValue.
	myPlayer := oldPlayer.
! !

!CPlayer methodsFor: 'private' stamp: 'ar 3/26/2004 00:58'!
privateGetValue
	"Private!! Answer my value.

	WARNING: This method must maintain the fundamental invariants of the framework and changing it is likely to get you entirely unpredictable results. Don't do it okay? At least not until you know precisely what you're doing."
	^super value! !

!CPlayer methodsFor: 'private' stamp: 'ar 3/17/2004 18:30'!
privateReleaseCostume: aCostume
	"Private!! Release aCostume as my primary costume if it had acted in this role.

	WARNING: This method must maintain the fundamental invariants of the framework and changing it is likely to get you entirely unpredictable results. Don't do it okay? At least not until you know precisely what you're doing."
	myCostume == aCostume ifTrue:[costume := nil].! !

!CPlayer methodsFor: 'private' stamp: 'ar 6/26/2005 19:00'!
privateReplacedCostume: oldCostume by: newCostume
	"My oldCostume has just been replaced by newCostume. Do whatever necessary."
	| cc shadow |
	oldCostume ifNil:[^nil].
	newCostume ifNil:[^nil].

	"Argh. This should not be necessary. But it seems that if costumes have disappeared they probably will be invalidated and showing up in the world will not broadcast the pending invalidation."
	costume signal: #invalidate.

	"Preserve transform properties"
	newCostume transform: oldCostume transform.
	newCostume position: oldCostume position.
	newCostume nose: oldCostume nose.

	"Preserve locking"
	newCostume locked: oldCostume locked.

	shadow := oldCostume highlightNamed: #dragMoveShadow.
	shadow ifNotNil:[
		oldCostume removeHighlight: #dragMoveShadow.
		newCostume addHighlight: shadow.
	].

	"Preserve container"
	(cc := oldCostume container) ifNotNil:[
		| pausedScripts |
		"And script ticking status"
		pausedScripts := self myScripts select:[:each| each isTickingScript and:[each isPaused]].
		cc asPrimCostume replace: oldCostume with: newCostume.
		pausedScripts do:[:each| each pauseScript].
	].

	self costumeChanged.! !

!CPlayer methodsFor: 'private' stamp: 'ar 5/12/2005 16:59'!
privateSetValueTo: newValue
	"Private!! Change my value to be newValue. 

	WARNING: This method must maintain the fundamental invariants of the framework and changing it is likely to get you entirely unpredictable results. Don't do it okay? At least not until you know precisely what you're doing.

	OPTIMIZATIONS: Actually, there is absolutely no reason to push the #valueChanged message over to the player since setting its value anyway. We could just nil out myPlayer while we write our own value. I haven't done this (yet) simply because having the extra notification pointed out an important bug in the architecture so I'll leave it in until it hurts too much."
	| oldPlayer |
	"First set my value - this allows synchronously acting players to
	signal a valueChanged *after* our valueChanged which can be critical."
	"This code now also implements the above optimization. It turned out that it was not entirely optional since pushing back valueChanged events to the player can have dire consequences if there is a listener for the valueChanged event in the player and the costume notification takes a tick or two."
	oldPlayer := myPlayer. myPlayer := nil.
		super value: newValue.
	myPlayer := oldPlayer.

	"Now notify my player"
	myPlayer ifNotNil:[myPlayer value: newValue].

	"And answer the new value"
	^newValue! !

!CPlayer methodsFor: 'basic' stamp: 'tk 6/9/2005 17:05'!
project
	^ Processor activeProject! !

!CPlayer methodsFor: 'serializing' stamp: 'ar 3/14/2004 15:03'!
readFixedFieldsFrom: deSerializer
	| varNames vv |
	super readFixedFieldsFrom: deSerializer.
	myPlayer := deSerializer readField: #player ifAbsent:[myPlayer].
	myCostume := deSerializer readField: #costume ifAbsent:[myCostume].
	CPlayer instSize+1 to: self class instSize do:[:i|
		varNames ifNil:[varNames := self class allInstVarNames].
		vv := deSerializer readField: (varNames at: i) ifAbsent:[self instVarAt: i].
		self instVarAt: i put: vv.
	].! !

!CPlayer methodsFor: 'removing' stamp: 'ar 3/14/2005 20:12'!
remove: oldObject 
	"Remove oldObject from the receiver's elements."
	^self collection remove: oldObject! !

!CPlayer methodsFor: 'removing' stamp: 'ar 3/14/2005 20:12'!
remove: oldObject ifAbsent: aBlock 
	"Remove oldObject from the receiver's elements. If several of the 
	elements are equal to oldObject, only one is removed. If no element is 
	equal to oldObject, answer the result of evaluating anExceptionBlock."
	^self collection remove: oldObject ifAbsent: aBlock! !

!CPlayer methodsFor: 'removing' stamp: 'ar 3/14/2005 20:12'!
removeAll
	"Remove all elements from the receiver"
	self collection removeAll.! !

!CPlayer methodsFor: 'removing' stamp: 'ar 3/14/2005 20:12'!
removeAll: aCollection 
	"Remove each element of aCollection from the receiver."
	self collection removeAll: aCollection! !

!CPlayer methodsFor: 'removing' stamp: 'ar 3/14/2005 20:11'!
removeAt: index
	"Remove an element from the receiver"
	^self collection removeAt: index! !

!CPlayer methodsFor: 'removing' stamp: 'ar 3/14/2005 20:12'!
removeFirst
	"Remove the first element of the receiver"
	^self collection removeFirst! !

!CPlayer methodsFor: 'removing' stamp: 'ar 3/14/2005 20:12'!
removeFrom: firstIndex to: lastIndex
	"Remove all the elements from firstIndex to lastIndex"
	^self collection removeFrom: firstIndex to: lastIndex! !

!CPlayer methodsFor: 'highlights' stamp: 'ar 3/24/2004 01:59'!
removeHighlight: highlightName
	"Remove the highlight from the receiver"
	^costume removeHighlight: highlightName! !

!CPlayer methodsFor: 'removing' stamp: 'ar 3/14/2005 20:12'!
removeLast
	"Remove the last element from the receiver"
	^self collection removeLast! !

!CPlayer methodsFor: 'basic' stamp: 'bgf 9/14/2008 21:00'!
resizeMoveBy: xy width: w height: h
	"Window resizer protocol.
	 Represents an update during CWindowResizeBehavior tracking,
	 when the window resizer is set for #resizeWithContents."
	self moveBy: xy.
	self width: w.
	self height: h.! !

!CPlayer methodsFor: 'opening' stamp: 'ar 10/18/2004 20:30'!
resumeTickingScripts
	"Resume all ticking scripts for the receiver and its costume(s)."
	| aPlayer |
	aPlayer := self.
	[aPlayer == nil] whileFalse:[
		aPlayer myScripts do:[:each| each isTickingScript ifTrue:[each resumeScript]].
		aPlayer := aPlayer getCostume.
	].
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:20'!
right
	<bewareOf: #geometryChanged>
	^geometry right! !

!CPlayer methodsFor: 'geometry' stamp: 'tk 8/11/2004 17:49'!
right: aValue
	"right is the position of rightmost edge of this object on the screen.  It is the number of screen dots from the left edge of the object's container to the right edge of this object.  Increasing right moves the object to the right." 

	^geometry right: aValue.
! !

!CPlayer methodsFor: 'geometry' stamp: 'bgf 2/23/2008 12:43'!
rightCenter
	<bewareOf: #geometryChanged>
	^geometry rightCenter! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 17:00'!
rightCenter: aValue

	^geometry rightCenter: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/24/2004 00:31'!
root
	^self costume root! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 6/27/2005 15:09'!
rotation
	"Answer the angle (in degrees) by which the object's costume appears rotated on the screen"
	<bewareOf: #geometryChanged>
	^geometry rotation! !

!CPlayer methodsFor: 'geometry' stamp: 'abc 4/16/2007 00:00'!
rotation: aNumber
	"Indicate the angle (in degrees) by which the object's costume appears rotated on the screen"
	^geometry rotation: aNumber! !

!CPlayer methodsFor: 'basic' stamp: 'ar 6/26/2005 13:21'!
rotationStyle
	"Answer the style by which the receiver should be rotated"
	<bewareOf: #rotationStyleChanged>
	^costume rotationStyle! !

!CPlayer methodsFor: 'basic' stamp: 'ar 6/26/2005 13:21'!
rotationStyle: aSymbol
	"Indicate the style by which the receiver should be rotated"
	^costume rotationStyle: aSymbol! !

!CPlayer methodsFor: 'corners' stamp: 'ar 3/24/2004 14:27'!
roundCornerSpec
	<bewareOf: #roundCornersChanged>
	^costume roundCornerSpec! !

!CPlayer methodsFor: 'corners' stamp: 'ar 5/19/2010 13:57'!
roundCornerSpec: anArray
	"Specification for round corners. Now accepts either the bit constants
	or the symbolic names:

		aPlayer roundCornerSpec: #(1 2 3 4).
		aPlayer roundCornerSpec: #(topLeft topRight).

	For compatibility, the getter always returns the numeric array."

	^costume roundCornerSpec: anArray.! !

!CPlayer methodsFor: 'corners' stamp: 'ar 3/24/2004 14:27'!
roundCorners
	"Answer true if the receiver should round its corners"
	<bewareOf: #roundCornersChanged>
	^costume roundCorners! !

!CPlayer methodsFor: 'corners' stamp: 'tk 2/24/2005 16:22'!
roundCorners: aBool
	"When roundedCorners is true, the object will have small curves at its corners.  This is different from the object being a RoundRectPlayer.[general]
	In the upper section, find 'rounded'.  A checkmark means the corners will be rounded.[colorTool]"

	^costume roundCorners: aBool.! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:20'!
scale
	<bewareOf: #geometryChanged>
	^geometry scale! !

!CPlayer methodsFor: 'geometry' stamp: 'tk 8/11/2004 17:22'!
scale: aValue
	"scale is the magnification of an object on the screen.  Normally it is 1.0."

	^geometry scale: aValue.
! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/24/2004 01:50'!
scrollArea
	"Answer the area that should be scrolled in the receiver"
	^(self propertyValueAt: #scrollArea) ifNil:[self costume scrollArea].
! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/24/2004 01:51'!
scrollArea: newBounds
	"Indicate the area that should be scrolled in the receiver"
	^self propertyValueAt: #scrollArea put: newBounds! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/24/2004 01:46'!
scrollDelta
	"Answer the amount by which the receiver should be scrolled"
	<bewareOf: #scrollDeltaChanged>
	^(self propertyValueAt: #scrollDelta) ifNil:[1@self textFont height].! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/24/2004 01:46'!
scrollDelta: aPoint
	"Indicate the amount by which the receiver should be scrolled"
	^self propertyValueAt: #scrollDelta put: aPoint with: #scrollDeltaChanged
! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/24/2004 01:46'!
scrollOffset
	"Answer the amount by which the receiver is scrolled"
	<bewareOf: #scrollOffsetChanged>
	^self costume scrollOffset! !

!CPlayer methodsFor: 'scrolling' stamp: 'bf 3/30/2004 14:02'!
scrollOffset: aPoint
	"Indicate the amount by which the receiver is scrolled"
	^self costume scrollOffset: aPoint! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 12/19/2004 19:16'!
scrollToShow: aRectangleOrPlayer
	"Scroll the receiver to show the given player"
	| box trial xOffset yOffset |

	aRectangleOrPlayer ifNil:[^self].
	aRectangleOrPlayer isPlayer 
		ifTrue:[box := aRectangleOrPlayer bounds]
		ifFalse:[box := aRectangleOrPlayer].

	trial := self visibleScrollArea.
	(trial containsRect: box) ifTrue:[^self].

	xOffset := yOffset := 0.

	"check right/bottom first"
	box right > trial right 
		ifTrue:[xOffset := box right - trial right].
	box bottom > trial bottom 
		ifTrue:[yOffset := box bottom - trial bottom].

	"if the box doesn't fit, prefer left/top over bottom/right"
	box := box origin - (xOffset @ yOffset) extent: box extent.

	"check left/top"
	box left < trial left 
		ifTrue:[xOffset := xOffset + box left - trial left].
	box top < trial top
		ifTrue:[yOffset := yOffset + box top - trial top].

	xOffset := xOffset + self scrollOffset x.
	yOffset := yOffset + self scrollOffset y.

	"make sure that a player which will grow to accomodate contents does not scroll"
	(self hResizing == #rigid or:[self hResizing == #spaceFill and:[self clipping]])
		ifFalse:[self invalidateLayout. xOffset := 0].
	(self vResizing == #rigid or:[self vResizing == #spaceFill and:[self clipping]])
		ifFalse:[self invalidateLayout. yOffset := 0].

	self scrollOffset: xOffset @ yOffset.! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/24/2004 01:47'!
scrollingDisabled
	"Answer whether the receiver should NOT be scrolled in its owner"
	<bewareOf: #scrollingDisabledChanged>
	^self costume scrollingDisabled
! !

!CPlayer methodsFor: 'scrolling' stamp: 'bf 3/30/2004 14:01'!
scrollingDisabled: aBool
	"Indicate whether the receiver should NOT be scrolled in its owner"
	^self costume scrollingDisabled: aBool! !

!CPlayer methodsFor: 'selecting' stamp: 'ar 3/23/2004 23:42'!
select
	"Select the receiver"
	^self costume select! !

!CPlayer methodsFor: 'selecting' stamp: 'ar 11/12/2004 21:04'!
selectionColor
	"Answer the selectionColor of the receiver"
	<bewareOf: #selectionColorChanged>
	^(self propertyValueAt: #selectionColor) ifNil:[CWidgetLibrary default selectionColor]! !

!CPlayer methodsFor: 'selecting' stamp: 'ar 11/12/2004 22:12'!
selectionColor: aValue
	"Modify the receiver's selectionColor"
	^self propertyValueAt: #selectionColor put: aValue with: #selectionColorChanged! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 4/30/2004 16:00'!
setupCostume
	"Set up the receiver's costume when it is lazily instantiated. "! !

!CPlayer methodsFor: 'shape' stamp: 'tk 2/2/2005 00:04'!
shape

	^costume shape! !

!CPlayer methodsFor: 'shape' stamp: 'tk 2/1/2005 18:16'!
shape: aValue
	"The object that is in charge of this Player's shape.  It can be a CPolygonShape, or it can be nil.  Balloon2D (Flash) objects will be shapes also."

	^costume shape: aValue.
! !

!CPlayer methodsFor: 'shape' stamp: 'tk 2/1/2005 18:14'!
showHandles: aBoolean
	"showHandles tells if the object's polygon shape will have knobs on each corner.  You can drag a knob to move the corner.  Drag a knob in the middle of a line to get a new corner (also called a vertex).  
	You must give this object a shape for this command to have any effect (self shape: CPolygonShape new)."

	^costume showHandles: aBoolean.
! !

!CPlayer methodsFor: 'private' stamp: 'ar 5/6/2004 16:19'!
signalEvent: anEvent
	super signalEvent: anEvent.
	myPlayer ifNil:[^nil].
	myPlayer getCostume == self ifFalse:[^nil].
	anEvent setSender: myPlayer.
	myPlayer signalEvent: anEvent.! !

!CPlayer methodsFor: 'shape' stamp: 'tk 2/1/2005 18:09'!
smoothCurve: aBoolean
	"smoothCurve tells if the object's polygon shape will a smooth curve as its edge, or straight line segments.  You must give this object a shape for this command to have any effect (self shape: CPolygonShape new)."

	^costume smoothCurve: aBoolean.
! !

!CPlayer methodsFor: 'properties' stamp: 'ar 3/23/2004 23:04'!
sticky
	"Answer the receiver's sticky property"
	<bewareOf: #flagsChanged>
	^costume sticky! !

!CPlayer methodsFor: 'properties' stamp: 'tk 2/24/2005 16:22'!
sticky: aBool
	"sticky tells whether an object can be dragged around with the mouse.  Many objects that are parts of other objects have been made sticky.  To move a sticky object, get a halo on it, and drag the brown or black handle.[general]
	In the upper section, find 'sticky'.  A checkmark means the object cannot be dragged.[colorTool]"

	^costume sticky: aBool! !

!CPlayer methodsFor: 'styling' stamp: 'ar 7/7/2005 09:25'!
styleSheet
	"Answer the receiver's style sheet"
	^costume styleSheet! !

!CPlayer methodsFor: 'styling' stamp: 'ar 7/7/2005 09:25'!
styleSheet: aStyleSheet
	"Answer the receiver's style sheet"
	costume styleSheet: aStyleSheet! !

!CPlayer methodsFor: 'opening' stamp: 'ar 10/18/2004 20:31'!
suspendTickingScripts
	"Suspend all ticking scripts for the receiver and its costume(s)."
	| aPlayer |
	aPlayer := self.
	[aPlayer == nil] whileFalse:[
		aPlayer myScripts do:[:each| each isTickingScript ifTrue:[each pauseScript]].
		aPlayer := aPlayer getCostume.
	].! !

!CPlayer methodsFor: 'tabbing' stamp: 'ar 1/25/2007 14:46'!
tabStopAfter: aPlayer
	"Find the next tab stop after aPlayer which is a child of the receiver"
	| index nextChild target stopIndex |
	elements size = 0 ifTrue:[^nil].
	stopIndex := index := elements indexOf: aPlayer.
	index = 0 ifTrue:[
		"cover first element explicitly"
		nextChild := elements at: 1.
		nextChild isTabStop ifTrue:[^nextChild].
		target := nextChild tabStopAfter: nil.
		target ifNotNil:[^target].
		index := stopIndex := 1.
	].
	[(index := index \\ elements size + 1) = stopIndex] whileFalse:[
		nextChild := elements at: index.
		nextChild isTabStop ifTrue:[^nextChild].
		target := nextChild tabStopAfter: nil.
		target ifNotNil:[^target].
	].
	^nil! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:15'!
text
	<bewareOf: #textChanged>
	^(textHolder ifNil:[^'']) text! !

!CPlayer methodsFor: 'text' stamp: 'bf 3/30/2004 11:33'!
text: aText
	^textHolder
		ifNil:[textHolder := aText]
		ifNotNil:[textHolder text: aText].
! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:09'!
textAnchorPoint
	<bewareOf: #textChanged>
	^textHolder anchorPoint! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:09'!
textAnchorPoint: aValue
	^textHolder anchorPoint: aValue.! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:09'!
textAttributes
	<bewareOf: #textChanged>
	^textHolder attributes! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:10'!
textAttributes: aValue
	^textHolder attributes: aValue.! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:10'!
textBorderColor
	<bewareOf: #textChanged>
	^textHolder borderColor! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:10'!
textBorderColor: aValue
	^textHolder borderColor: aValue.! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:10'!
textBorderWidth
	<bewareOf: #textChanged>
	^textHolder borderWidth! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/28/2004 23:37'!
textBorderWidth: aValue
	^textHolder borderWidth: aValue.! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:11'!
textColor
	<bewareOf: #textChanged>
	^textHolder color! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:11'!
textColor: aValue
	^textHolder color: aValue.! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:30'!
textEditor
	^textHolder editor! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:11'!
textEmphasis
	<bewareOf: #textChanged>
	^textHolder emphasis! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:11'!
textEmphasis: aValue
	^textHolder emphasis: aValue.! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:23'!
textEnabled
	^textHolder notNil! !

!CPlayer methodsFor: 'text' stamp: 'bf 3/30/2004 11:40'!
textEnabled: aBool
	self textEnabled = aBool ifFalse:[
		textHolder := aBool ifTrue:[Text new] ifFalse:[nil].
	].
	^ aBool! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/24/2004 01:45'!
textFont
	<bewareOf: #textChanged>
	^(textHolder ifNil:[^TextStyle defaultFont]) font! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:11'!
textFont: aValue
	^textHolder font: aValue.! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/23/2004 22:10'!
textHolder
	<bewareOf: #textChanged>
	^costume textHolder! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/23/2004 22:10'!
textHolder: aTextHolder
	^costume textHolder: aTextHolder.! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:12'!
textInset
	<bewareOf: #textChanged>
	^textHolder inset! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:12'!
textInset: aValue
	^textHolder inset: aValue.! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:12'!
textOffset
	<bewareOf: #textChanged>
	^textHolder offset! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:12'!
textOffset: aValue
	^textHolder offset: aValue.! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:12'!
textWrap
	<bewareOf: #textChanged>
	^textHolder wrapContents! !

!CPlayer methodsFor: 'text' stamp: 'ar 3/23/2004 20:12'!
textWrap: aValue
	^textHolder wrapContents: aValue.! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:20'!
top
	<bewareOf: #geometryChanged>
	^geometry top! !

!CPlayer methodsFor: 'geometry' stamp: 'tk 9/14/2004 14:33'!
top: aValue
	"top is the position of this object in the vertical direction.  top is another name for y.  It is the number of screen dots from the top of the object's container to the top of this object.  Increasing top moves the object downward.  (Yes, this is the opposite direction from Y in a math class.)  top is the ""y part"" of the object's position." 

	^geometry top: aValue.
! !

!CPlayer methodsFor: 'geometry' stamp: 'bgf 2/23/2008 12:43'!
topCenter
	<bewareOf: #geometryChanged>
	^geometry topCenter! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 17:00'!
topCenter: aValue

	^geometry topCenter: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:22'!
topLeft
	<bewareOf: #geometryChanged>
	^geometry topLeft! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 17:00'!
topLeft: aValue
	"topLeft is the location of the upper left corner of an object on the screen.  It is a Point.  When the angle is zero, topLeft is the same as origin."

	^geometry topLeft: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:22'!
topRight
	<bewareOf: #geometryChanged>
	^geometry topRight! !

!CPlayer methodsFor: 'geometry' stamp: 'das 11/20/2009 17:00'!
topRight: aValue
	"topRight is the location of the upper righthand corner of the object.  It is a Point whose value is (right@top)."


	^geometry topRight: (self ensurePoint: aValue).
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:20'!
transform
	<bewareOf: #geometryChanged>
	^geometry transform! !

!CPlayer methodsFor: 'geometry' stamp: 'tk 9/15/2004 23:05'!
transform: aTransform
	"the transform holds a special object that can convert from the coordinate system of this object to the absolute screen coordinates of Tweak.  It uses the messages localPointToGlobal: and globalPointToLocal: to convert a Point you give it.  The simplest kind of transform is a COffsetTransform."
	^geometry transform: aTransform! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 12:47'!
transformBy: aTransform
	^costume transformBy: aTransform! !

!CPlayer methodsFor: 'animating' stamp: 'ar 3/28/2004 16:10'!
turnBy: degrees
	self costume turnBy: degrees! !

!CPlayer methodsFor: 'animating' stamp: 'bf 7/22/2005 15:22'!
turnTo: target
	"Make me point towards the target"
	| delta |
	target ifNil:[^self].
	target isNumber
		ifTrue:[delta := target - heading]
		ifFalse:[delta := ((self globalToLocal: target globalPosition) - self pivot) degrees - self nose degrees].
	self turnBy: delta! !

!CPlayer methodsFor: 'private' stamp: 'ar 3/24/2004 00:34'!
tweakedCostume: aCostume
	"Broadcast tweaks happened to aCostume.
	This feels wrong ... see the senders"
	myCostume == aCostume 
		ifTrue:[self tweaked: #costume from: aCostume to: aCostume].! !

!CPlayer methodsFor: 'properties' stamp: 'bf 3/30/2004 11:28'!
userDraw
	"Answer the receiver's userDraw property"
	<bewareOf: #flagsChanged>
	^costume userDraw! !

!CPlayer methodsFor: 'properties' stamp: 'bf 3/30/2004 11:28'!
userDraw: aBool
	"Modify the receiver's userDraw property"
	^costume userDraw: aBool! !

!CPlayer methodsFor: 'layout' stamp: 'ar 3/11/2004 12:44'!
vResizing
	^self costume vResizing! !

!CPlayer methodsFor: 'layout' stamp: 'tk 1/4/2005 12:43'!
vResizing: aValue
	"vResizing tells how I will change height when my elements or my container change.  My container is the object I am inside of.  Meaning depends on what layout I have, such as TableLayout.  
	#rigid		 - do not change my size
	#spaceFill	 - resize to fill my container's available space
	#shrinkWrap - resize to fit my elements
	"
	^self costume vResizing: aValue.
	! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/19/2004 13:22'!
vScrollable
	"Answer the vScrollable of the receiver"
	<bewareOf: #vScrollableChanged>
	^self propertyValueAt: #vScrollable! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/19/2004 13:22'!
vScrollable: aValue
	"Modify the receiver's vScrollable"
	^self propertyValueAt: #vScrollable put: aValue with: #vScrollableChanged! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/26/2004 00:51'!
value
	"Answer the receiver's value"
	^self privateGetValue! !

!CPlayer methodsFor: 'basic' stamp: 'tk 1/4/2005 12:55'!
value: newValue
	"The value field of an object is used to store the part of the object that matters the most.  The value represents different things about different kinds of objects.  For a CheckBox, the value tells whether the box is currently checked (true if it is checked, and false otherwise)."

	^self privateSetValueTo: newValue! !

!CPlayer methodsFor: 'copying' stamp: 'tk 2/16/2005 12:18'!
veryDeepCheck: deepCopier
	"Give CObject a chance to look for unmapped elements"

	| |
	(deepCopier newGuys includes: self) ifFalse: [self error: 'not a new guy'].
	myCostume ifNotNil: [
		self elements do: [:ele |
			ele isPlayer ifTrue: [
				(deepCopier newGuys includes: ele) ifFalse: [self error: 'not a new guy']]]].

	"Set 'value' in costumes, and activate privateChangeValueTo:"
	myCostume ifNotNil: [
		(deepCopier newGuys includes: myCostume) ifFalse: [self error: 'not a new guy'].
		"oldCost := myCostume.
		myCostume := nil.
		self costume: oldCost"].

	myPlayer ifNotNil: [
		"(deepCopier newGuys includes: myPlayer) ifFalse: [self inform: 'a second costume']."
		"Sometimes Player is not being duplicated.  A costume that points at an outside player"
		"oldPlayer := myPlayer.
		myPlayer := nil.
		self player: oldPlayer"].

! !

!CPlayer methodsFor: 'copying' stamp: 'tk 1/24/2005 22:34'!
veryDeepCopyWith: deepCopier
	"All Players are weakly copied, except when in the elements of another Player being dulicated.  myPlayer and myCostume are always duplicated."

	"Duplication never makes a new uniClass in Tweak.  See CObject."

	deepCopier references at: self ifPresent: [:newer | ^ newer]. 		"already mapped me"
	self container ifNil: [^ super veryDeepCopyWith: deepCopier].		"do duplicate"
	deepCopier references at: self container ifPresent: [:newer | 
			^ super veryDeepCopyWith: deepCopier].		"I am an element of an object being copied"
	^ self 	"will be mapped in veryDeepFixup:"

"Unusual case:  If a Player is being used as a shared system resource, but it's container is nil, and therefore it is being duplicated, just enclose it in another player.  Its container will not be nil, and it won't be duplicated."! !

!CPlayer methodsFor: 'copying' stamp: 'tk 1/20/2005 12:10'!
veryDeepForSure: copier
	"Force a veryDeepCopy, even if the default for this class is to be weakly copied.  I am an element."

	"Duplication never makes a new uniClass in Tweak.  See CObject."
	^ super veryDeepCopyWith: copier! !

!CPlayer methodsFor: 'copying' stamp: 'tk 2/16/2005 12:21'!
veryDeepInner: deepCopier 
	"Yes, I am being duplicated.  Make sure my costume and player are also."

	(deepCopier newGuys includes: self) ifFalse: [self error: 'not a new guy'].
	myCostume := myCostume veryDeepForSure: deepCopier.	"force it"
	myPlayer ifNotNil: [
		(myPlayer getCostume == self or: [
			(deepCopier mapIfCan: myPlayer getCostume) == self]) 
				ifTrue: ["I am the primary costume"
					myPlayer := myPlayer veryDeepForSure: deepCopier]
				ifFalse: ["I am a secondary costume"
					myPlayer := myPlayer veryDeepCopyWith: deepCopier]].	"may be outside the tree"

	super veryDeepInner: deepCopier.	"copy others and nil scripts"
! !

!CPlayer methodsFor: 'properties' stamp: 'ar 3/23/2004 23:05'!
visible
	"Answer the receiver's visible property"
	<bewareOf: #flagsChanged>
	^costume visible! !

!CPlayer methodsFor: 'properties' stamp: 'tk 2/24/2005 16:22'!
visible: aValue
	"visible tells whether the object shows on the screen.  For elements of an object that need to appear and disappear, changing visible is a good way to do it.  Invisible objects do not take any time to display.[general]
	In the upper section, find 'visible'.  A checkmark means the object is visible.[colorTool]"

	^costume visible: aValue.! !

!CPlayer methodsFor: 'scrolling' stamp: 'ar 3/24/2004 01:54'!
visibleScrollArea
	"Answer the visible portion of the scroll area"
	^costume visibleScrollArea! !

!CPlayer methodsFor: 'initialize' stamp: 'ar 5/13/2005 14:12'!
wearDefaultCostume
	"Make me wear my default costume"
	costume := self defaultCostume.
	myCostume onPlayerChanged.
	self setupCostume.
	"For now, call #makeDefaultCostume so that we keep compatibility"
		self makeDefaultCostume.
	^myCostume! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 3/24/2004 00:20'!
width
	<bewareOf: #geometryChanged>
	^geometry width! !

!CPlayer methodsFor: 'geometry' stamp: 'tk 8/11/2004 17:14'!
width: aValue
	"width tells how wide the object is.  It is measured in screen dots."

	^geometry width: aValue.
! !

!CPlayer methodsFor: 'basic' stamp: 'ar 3/24/2004 00:17'!
windowColor
	"Answer the receiver's window color"
	container ifNil:[^color] ifNotNil:[^container windowColor].! !

!CPlayer methodsFor: 'basic' stamp: 'bf 3/30/2004 11:27'!
windowColor: aColor
	"Make the receiver use the given window color"
	^costume windowColor: aColor! !

!CPlayer methodsFor: 'basic' stamp: 'tk 6/9/2005 17:08'!
world
	"The outermost container of this object.  It is an EToysWorld or a CPlayer, and its extent is what you see on the screen for this project."
	<bewareOf: #containerChanged>
	myCostume ifNil:[^nil].
	^self costume world! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 5/22/2005 18:07'!
worldBounds
	<bewareOf: #frameChanged>
	^geometry worldBounds
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 5/22/2005 18:08'!
worldToLocal
	^costume worldToLocal
! !

!CPlayer methodsFor: 'geometry' stamp: 'ar 5/22/2005 18:08'!
worldToLocal: anObject
	^costume worldToLocal: anObject
! !

!CPlayer methodsFor: 'animating' stamp: 'ar 8/17/2003 15:34'!
wrap
	"Wrap me in my container"
	| innerBox outerBox |
	innerBox := self bounds.
	outerBox := 0@0 extent: self container extent.

	innerBox right < 0 "left"
		ifTrue:[self left: outerBox width].

	innerBox left > outerBox width "right"
		ifTrue:[self right: 0].

	innerBox bottom < 0 "top"
		ifTrue:[self top: outerBox height].

	innerBox top > outerBox height "bottom"
		ifTrue:[self bottom: 0].
! !

!CPlayer methodsFor: 'serializing' stamp: 'ar 3/14/2004 15:03'!
writeFixedFieldsOn: aSerializer
	| varNames |
	super writeFixedFieldsOn: aSerializer.
	aSerializer writeField: #player value: myPlayer.
	aSerializer writeField: #costume value: myCostume.
	CPlayer instSize+1 to: self class instSize do:[:i|
		varNames ifNil:[varNames := self class allInstVarNames].
		aSerializer writeField: (varNames at: i) value: (self instVarAt: i).
	].
! !

CPlayer subclass: #CPolygonPlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Objects'!

!CPolygonPlayer commentStamp: 'tk 3/1/2005 11:40' prior: 0!
A Polygon is very versitle.  It can be a complex shape with many straight lines as its edges.  It can be a smooth curved shape.  If it is 'open', it looks like a jagged line or like a smooth wiggly line.[general]


### Allow Users to see the above explanation ###!

CPolygonPlayer defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="vertices" toSet="vertices:" changeEvent="verticesChanged">vertices</field>
	<field toGet="closed" toSet="closed:" changeEvent="closedChanged">closed</field>
	<field toGet="smoothCurve" toSet="smoothCurve:" changeEvent="smoothCurveChanged">smoothCurve</field>
	<field toGet="arrowStyle" toSet="arrowStyle:" changeEvent="arrowStyleChanged">arrowStyle</field>
	<field toGet="arrowSpec" toSet="arrowSpec:" changeEvent="arrowSpecChanged">arrowSpec</field>
	<field toGet="borderDashSpec" toSet="borderDashSpec:" changeEvent="borderDashSpecChanged">borderDashSpec</field>
	<field toGet="curveState" toSet="curveState:" changeEvent="curveStateChanged">curveState</field>
	<field toGet="arrowForms" toSet="arrowForms:" changeEvent="arrowFormsChanged">arrowForms</field>
	<field toGet="borderForm" toSet="borderForm:" changeEvent="borderFormChanged">borderForm</field>
	<field toGet="filledForm" toSet="filledForm:" changeEvent="filledFormChanged">filledForm</field>
</fields>'!

!CPolygonPlayer class methodsFor: 'instance creation' stamp: 'ar 6/25/2005 21:45'!
arrowPointsFrom: priorPoint to: endPoint
	^self arrowPointsFrom: priorPoint to: endPoint using: 5@4.! !

!CPolygonPlayer class methodsFor: 'instance creation' stamp: 'ar 6/25/2005 21:36'!
arrowPointsFrom: priorPoint to: endPoint using: arrowSpec
	"Answer a triangle oriented along the line from priorPoint to endPoint."
	| d v deg wingBase length w |
	v := endPoint - priorPoint.
	deg := v degrees.
	d := 1.
	length := arrowSpec x abs.  w := arrowSpec y abs.
	wingBase := endPoint + (Point r: d * length degrees: deg + 180.0).
	arrowSpec x >= 0
		ifTrue: [^ {	endPoint.
					wingBase + (Point r: d * w degrees: deg + 125.0).
					wingBase + (Point r: d * w degrees: deg - 125.0) }]
		ifFalse: ["Negative length means concave base."
				^ {	endPoint.
					wingBase + (Point r: d * w degrees: deg + 125.0).
					wingBase.
					wingBase + (Point r: d * w degrees: deg - 125.0) }]! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/4/2005 23:43'!
arrowBoundsAt: endPoint from: priorPoint 
	"Answer a triangle oriented along the line from priorPoint to endPoint."
	| d v deg wingBase length w |
	v := endPoint - priorPoint.
	deg := v degrees.
	d := borderWidth max: 1.
	length := arrowSpec x abs.  w := arrowSpec y abs.
	wingBase := endPoint + (Point r: d * length degrees: deg + 180.0).
	arrowSpec x >= 0
		ifTrue: [^ {	endPoint.
					wingBase + (Point r: d * w degrees: deg + 125.0).
					wingBase + (Point r: d * w degrees: deg - 125.0) }]
		ifFalse: ["Negative length means concave base."
				^ {	endPoint.
					wingBase + (Point r: d * w degrees: deg + 125.0).
					wingBase.
					wingBase + (Point r: d * w degrees: deg - 125.0) }]! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/4/2005 23:51'!
arrowForms
	"Answer the arrowForms of the receiver"
	<bewareOf: #arrowFormsChanged>
	^(self propertyValueAt: #arrowForms) ifNil:[self computeArrowForms]! !

!CPolygonPlayer methodsFor: 'private' stamp: 'tk 2/7/2005 13:06'!
arrowForms: aValue
	"Modify the receiver's arrowForms"
	^self propertyValueAt: #arrowForms put: aValue with: #arrowFormsChanged! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/4/2005 23:42'!
arrowSpec
	"Answer the arrowSpec of the receiver"
	<bewareOf: #arrowSpecChanged>
	^self propertyValueAt: #arrowSpec ifAbsent:[5@4]! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
arrowSpec: aValue
	"Modify the receiver's arrowSpec"
	^self propertyValueAt: #arrowSpec put: aValue with: #arrowSpecChanged! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
arrowStyle
	"Answer the arrowStyle of the receiver"
	<bewareOf: #arrowStyleChanged>
	^self propertyValueAt: #arrowStyle! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
arrowStyle: aValue
	"Modify the receiver's arrowStyle"
	^self propertyValueAt: #arrowStyle put: aValue with: #arrowStyleChanged! !

!CPolygonPlayer methodsFor: 'dashes' stamp: 'ar 2/4/2005 23:21'!
borderDashOffset
	borderDashSpec size < 4 ifTrue: [^0.0].
	^(borderDashSpec fourth) asFloat! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
borderDashSpec
	"Answer the borderDashSpec of the receiver"
	<bewareOf: #borderDashSpecChanged>
	^self propertyValueAt: #borderDashSpec! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
borderDashSpec: aValue
	"Modify the receiver's borderDashSpec"
	^self propertyValueAt: #borderDashSpec put: aValue with: #borderDashSpecChanged! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/4/2005 23:51'!
borderForm
	"Answer the borderForm of the receiver"
	<bewareOf: #borderFormChanged>
	^(self propertyValueAt: #borderForm) ifNil:[self computeBorderForm]! !

!CPolygonPlayer methodsFor: 'private' stamp: 'tk 2/7/2005 13:06'!
borderForm: aValue
	"Modify the receiver's borderForm"
	^self propertyValueAt: #borderForm put: aValue with: #borderFormChanged! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
closed
	"Answer the closed of the receiver"
	<bewareOf: #closedChanged>
	^self propertyValueAt: #closed! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
closed: aValue
	"Modify the receiver's closed"
	^self propertyValueAt: #closed put: aValue with: #closedChanged! !

!CPolygonPlayer methodsFor: 'geometry' stamp: 'ar 2/4/2005 23:30'!
closestPointTo: aPoint 
	| curvePoint closestPoint dist minDist |
	closestPoint := minDist := nil.
	self lineSegmentsDo:[:p1 :p2 | 
		curvePoint := aPoint nearestPointOnLineFrom: p1 to: p2.
		dist := curvePoint dist: aPoint.
		(closestPoint isNil or: [dist < minDist]) ifTrue:[
			closestPoint := curvePoint.
			minDist := dist]].
	^closestPoint! !

!CPolygonPlayer methodsFor: 'spline' stamp: 'ar 2/5/2005 21:36'!
coefficients
	"Compute an array for the coefficients.  This is copied from Flegal's old
	code in the Spline class."
	| length extras verts coefficients cc |
	curveState ifNotNil: [^curveState first].
	verts := (closed and:[vertices first ~= vertices last])
				ifTrue: [vertices copyWith: vertices first]
				ifFalse: [vertices].
	length := verts size.
	extras := 0.
	coefficients := Array new: 8.
	1 to: 8 do: [:i | coefficients at: i put: (Array new: length + extras)].
	1 to: 5 by: 4 do:[:k | 
		1 to: length do:[:i | 
			k = 1 
				ifTrue: [cc := (verts at: i) x asFloat]
				ifFalse: [cc := (verts at: i) y asFloat].
			(coefficients at: k) at: i put: cc].
		1 to: extras do:[:i | 
			(coefficients at: k) at: length + i put: ((coefficients at: k) at: i + 1)].
			self
				derivs: (coefficients at: k)
				first: (coefficients at: k + 1)
				second: (coefficients at: k + 2)
				third: (coefficients at: k + 3)].
	extras > 0  ifTrue:[
		1 to: 8 do:[:i | 
			coefficients at: i put: ((coefficients at: i) copyFrom: 2 to: length + 1)]].
	curveState := { 
		coefficients.
		nil.
		nil}.
	self computeNextToEndPoints.
	^coefficients! !

!CPolygonPlayer methodsFor: 'private' stamp: 'tk 4/28/2005 13:52'!
computeArrowFormAt: endPoint from: priorPoint 
	"Compute a triangle oriented along the line from priorPoint to  
	endPoint. Then draw those lines in a form and return that  
	form, with appropriate offset"

	| p1 pts box arrowForm bb pos |
	pts := self arrowBoundsAt: endPoint from: priorPoint.
	box := ((pts first rect: pts last) encompass: (pts second)) expandBy: 2.
	arrowForm := Form extent: box extent asIntegerPoint.
	bb := (BitBlt current toForm: arrowForm)
				sourceForm: nil;
				fillColor: Color black;
				combinationRule: Form over;
				width: 1;
				height: 1.
	pos := box topLeft.
	p1 := pts last - pos.
	pts do: 
			[:p | 
			bb drawFrom: p1 to: p - pos.
			p1 := p - pos].
	arrowForm convexShapeFill: Color black.
	^arrowForm offset: box topLeft! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/4/2005 23:51'!
computeArrowForms
	"ArrowForms are computed only upon demand"
	arrowForms := Array new.
	(closed or: [arrowStyle == #none or: [vertices size < 2]]) ifTrue:
		[^ arrowForms].
	(arrowStyle == #forward or: [arrowStyle == #both]) ifTrue:
		[arrowForms := arrowForms copyWith:
			(self computeArrowFormAt: vertices last from: self nextToLastPoint)].
	(arrowStyle == #back or: [arrowStyle == #both]) ifTrue:
		[arrowForms := arrowForms copyWith:
			(self computeArrowFormAt: vertices first from: self nextToFirstPoint)].
	^ arrowForms! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/4/2005 23:51'!
computeBorderForm
	"A form must be created for drawing the border whenever the borderColor is translucent."
	| borderCanvas |
	borderCanvas := (CTransformCanvas extent: bounds extent depth: 1)
		shadowColor: Color black.
	borderCanvas translateBy: bounds topLeft negated
		during:[:tempCanvas| self drawBorderOn: tempCanvas].
	borderForm := borderCanvas form.
	self arrowForms do:
		[:f |  "Eliminate overlap between line and arrowheads if transparent."
		borderForm copy: f boundingBox from: f to: f offset - self position rule: Form erase].
	^ borderForm! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/6/2005 15:23'!
computeBounds
	| oldBounds newBounds |
	vertices ifNil: [^ self].
	oldBounds := bounds.
	self releaseCachedState.
	newBounds := self curveBounds.
	self arrowForms do:[:f | newBounds := newBounds merge: (f offset extent: f extent)].
	newBounds origin isZero ifFalse:[
		vertices := vertices collect:[:pt| pt - newBounds origin].
		self releaseCachedState.
	].
	newBounds := self asPrimCostume childToParent: newBounds.
	oldBounds = newBounds ifFalse:[bounds := newBounds].
	self invalidate: (oldBounds merge: bounds).! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/6/2005 15:51'!
computeFilledForm
	"Note: The filled form is actually 2 pixels bigger than bounds, and the point corresponding to this morphs' position is at 1@1 in the form.  This is due to the details of the fillig routines, at least one of which requires an extra 1-pixel margin around the outside.  Computation of the filled form is done only on demand."
	| bb pos |
	closed ifFalse: [^nil].
	filledForm := Form extent: extent+2.

	"Draw the border..."
	bb := (BitBlt current toForm: filledForm) sourceForm: nil; fillColor: Color black;
			combinationRule: Form over; width: 1; height: 1.
	pos := -1@-1.
	self lineSegmentsDo: [:p1 :p2 | bb drawFrom: p1 asIntegerPoint-pos
										to: p2 asIntegerPoint-pos].

	"Fill it in..."
	filledForm convexShapeFill: Color black.

	(borderColor isColor and: [borderColor isTranslucentColor]) ifTrue:
		["If border is stored as a form, then erase any overlap now."
		filledForm copy: self borderForm boundingBox from: self borderForm
			to: 1@1 rule: Form erase].

	^ filledForm! !

!CPolygonPlayer methodsFor: 'spline' stamp: 'ar 2/4/2005 23:39'!
computeNextToEndPoints
	| pointAfterFirst pointBeforeLast |
	pointAfterFirst := nil.
	self lineSegmentsDo: 
			[:p1 :p2 | 
			pointAfterFirst isNil ifTrue: [pointAfterFirst := p2 asIntegerPoint].
			pointBeforeLast := p1 asIntegerPoint].
	curveState at: 2 put: pointAfterFirst.
	curveState at: 3 put: pointBeforeLast! !

!CPolygonPlayer methodsFor: 'testing' stamp: 'ar 2/6/2005 15:52'!
containsPoint: aPoint

	(costume containsPoint: aPoint) ifFalse: [^ false].

	closed & color isTransparent not ifTrue:[
		^(self filledForm pixelValueAt: aPoint + 1) > 0].

	self lineSegmentsDo:[:p1 :p2 |
		(aPoint onLineFrom: p1 to: p2 within: (3 max: borderWidth+1//2) asFloat)
				ifTrue: [^ true]].

	self arrowForms do:[:f | 
		(f pixelValueAt: aPoint - f offset) > 0 ifTrue: [^ true]].

	^ false! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/5/2005 00:44'!
curveBounds
	| curveBounds pointAfterFirst pointBeforeLast |
	smoothCurve ifFalse:[^(Rectangle encompassing: vertices) expandBy: (borderWidth + 1) // 2].

	"Compute the bounds from actual curve traversal, with leeway for borderWidth.
	Also note the next-to-first and next-to-last points for arrow directions."
	curveState := nil.	"Force recomputation"
	curveBounds := vertices first corner: vertices last.
	pointAfterFirst := nil.
	self lineSegmentsDo:[:p1 :p2 | 
		pointAfterFirst isNil ifTrue:[
				pointAfterFirst := p2 asIntegerPoint.
				curveBounds := curveBounds encompass: p1 asIntegerPoint].
			curveBounds := curveBounds encompass: p2 asIntegerPoint.
			pointBeforeLast := p1 asIntegerPoint].
	curveState at: 2 put: pointAfterFirst.
	curveState at: 3 put: pointBeforeLast.
	^curveBounds expandBy: (borderWidth + 1) // 2! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
curveState
	"Answer the curveState of the receiver"
	<bewareOf: #curveStateChanged>
	^self propertyValueAt: #curveState! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
curveState: aValue
	"Modify the receiver's curveState"
	^self propertyValueAt: #curveState put: aValue with: #curveStateChanged! !

!CPolygonPlayer methodsFor: 'dashes' stamp: 'ar 2/4/2005 23:21'!
dashedBorder: dashSpec
	"A dash spec is a 3- or 5-element array with
		{ length of normal border color.
		length of alternate border color.
		alternate border color.
		starting offset.
		amount to add to offset at each step }
	Starting offset is usually = 0, but changing it moves the dashes along the curve."
	borderDashSpec := dashSpec.
	self invalidate.! !

!CPolygonPlayer methodsFor: 'spline' stamp: 'ar 2/4/2005 23:39'!
derivs: a first: point1 second: point2 third: point3 
	"Compute the first, second and third derivitives (in coeffs) from
	the Points in this Path (coeffs at: 1 and coeffs at: 5)."

	| len v anArray |
	len := a size.
	len < 2 ifTrue: [^self].
	len > 2 
		ifTrue: 
			[v := Array new: len.
			v at: 1 put: 4.0.
			anArray := Array new: len.
			anArray at: 1 put: 6.0 * (a first - (a second * 2.0) + (a third)).
			2 to: len - 2
				do: 
					[:i | 
					v at: i put: 4.0 - (1.0 / (v at: i - 1)).
					anArray at: i
						put: 6.0 * ((a at: i) - ((a at: i + 1) * 2.0) + (a at: i + 2)) 
								- ((anArray at: i - 1) / (v at: i - 1))].
			point2 at: len - 1 put: (anArray at: len - 2) / (v at: len - 2).
			len - 2 to: 2
				by: 0 - 1
				do: 
					[:i | 
					point2 at: i
						put: ((anArray at: i - 1) - (point2 at: i + 1)) / (v at: i - 1)]].
	point2 at: 1 put: (point2 at: len put: 0.0).
	1 to: len - 1
		do: 
			[:i | 
			point1 at: i
				put: (a at: i + 1) - (a at: i) 
						- (((point2 at: i) * 2.0 + (point2 at: i + 1)) / 6.0).
			point3 at: i put: (point2 at: i + 1) - (point2 at: i)]! !

!CPolygonPlayer methodsFor: 'drawing' stamp: 'ar 2/6/2005 15:44'!
drawArrowOn: aCanvas at: endPoint from: priorPoint 
	"Draw a triangle oriented along the line from priorPoint to  
	endPoint. Answer the wingBase."

	| pts spec wingBase |
	pts := self arrowBoundsAt: endPoint from: priorPoint.
	wingBase := pts size = 4 
				ifTrue: [pts third]
				ifFalse: [(pts copyFrom: 2 to: 3) average].
	spec := self arrowSpec.
	spec x sign = spec y sign 
		ifTrue: [aCanvas drawPolygon: pts fillStyle: borderColor]
		ifFalse: 
			[aCanvas 
				drawPolygon: pts
				fillStyle: Color transparent
				borderWidth: (borderWidth + 1) // 2
				borderColor: borderColor].
	^wingBase! !

!CPolygonPlayer methodsFor: 'drawing' stamp: 'ar 2/4/2005 23:23'!
drawArrowsOn: aCanvas 
	"Answer (possibly modified) endpoints for border drawing"
	"ArrowForms are computed only upon demand"
	| array |
	(closed or: [arrowStyle == #none or: [vertices size < 2]]) ifTrue: [^ self].
	"Nothing to do"
	borderColor isColor ifFalse: [^ self].
	array := Array new: 2.
	"Prevent crashes for #raised or #inset borders"
	array at: 2 put: ((arrowStyle == #forward or: [arrowStyle == #both])
		ifTrue: [ self
					drawArrowOn: aCanvas
					at: vertices last
					from: self nextToLastPoint]
		ifFalse: [ vertices last ]).
	array at: 1 put: ((arrowStyle == #back or: [arrowStyle == #both])
		ifTrue: [self
					drawArrowOn: aCanvas
					at: vertices first
					from: self nextToFirstPoint]
		ifFalse: [ vertices first ]).
	^array! !

!CPolygonPlayer methodsFor: 'drawing' stamp: 'ar 2/4/2005 23:23'!
drawBorderOn: aCanvas 
	self
		drawClippedBorderOn: aCanvas
		usingEnds: (Array with: vertices first with: vertices last)! !

!CPolygonPlayer methodsFor: 'drawing' stamp: 'ar 2/4/2005 23:52'!
drawBorderOn: aCanvas usingEnds: anArray 
	"Display my border on the canvas."
	"NOTE: Much of this code is also copied in drawDashedBorderOn:  
	(should be factored)"
	| bigClipRect p1i p2i |
	borderDashSpec
		ifNotNil: [^ self drawDashedBorderOn: aCanvas usingEnds: anArray].
	bigClipRect := aCanvas clipRect expandBy: self borderWidth + 1 // 2.
	self lineSegmentsDo: [:p1 :p2 | 
		p1i := p1 asIntegerPoint.
		p2i := p2 asIntegerPoint.
		(arrowStyle ~= #none and: [closed not])
			ifTrue: ["Shorten line ends so as not to interfere with tip of arrow."
					((arrowStyle == #back
								or: [arrowStyle == #both])
							and: [p1 = vertices first])
						ifTrue: [p1i := anArray first asIntegerPoint].
					((arrowStyle == #forward
								or: [arrowStyle == #both])
							and: [p2 = vertices last])
						ifTrue: [p2i := anArray last asIntegerPoint]].
		(closed or: ["bigClipRect intersects: (p1i rect: p2i) optimized:"
			((p1i min: p2i) max: bigClipRect origin)
				<= ((p1i max: p2i) min: bigClipRect corner)])
				ifTrue: [border drawLineFrom: p1i to: p2i on: aCanvas]]! !

!CPolygonPlayer methodsFor: 'drawing' stamp: 'ar 2/4/2005 23:24'!
drawClippedBorderOn: aCanvas usingEnds: anArray 
	self drawBorderOn: aCanvas usingEnds: anArray.! !

!CPolygonPlayer methodsFor: 'drawing' stamp: 'ar 2/4/2005 23:24'!
drawDashedBorderOn: aCanvas 
	self
		drawDashedBorderOn: aCanvas
		usingEnds: (Array with: vertices first with: vertices last)! !

!CPolygonPlayer methodsFor: 'drawing' stamp: 'ar 2/4/2005 23:28'!
drawDashedBorderOn: aCanvas usingEnds: anArray 
	"Display my border on the canvas. NOTE: mostly copied from  
	drawBorderOn:"

	| lineColor bevel topLeftColor bottomRightColor bigClipRect p1i p2i segmentOffset |
	(borderColor isNil  or: [borderColor isColor and: [borderColor isTransparent]]) ifTrue: [^self].
	lineColor := borderColor.
	bevel := false.
	"Border colors for bevelled effects depend on CW ordering of vertices"
	topLeftColor := border topLeftColor.
	bottomRightColor := border bottomRightColor.
	bevel := topLeftColor ~= bottomRightColor.
	bigClipRect := aCanvas clipRect expandBy: (self borderWidth + 1) // 2.
	segmentOffset := self borderDashOffset.
	self lineSegmentsDo:[:p1 :p2 | 
		p1i := p1 asIntegerPoint.
		p2i := p2 asIntegerPoint.
		(arrowStyle ~= #none and: [closed not]) ifTrue:[
			"Shorten line ends so as not to interfere with tip of arrow."
			((arrowStyle == #back or: [arrowStyle == #both]) and: [p1 = vertices first]) 
				ifTrue: [p1i := anArray first asIntegerPoint].
			((arrowStyle == #forward or: [arrowStyle == #both]) and: [p2 = vertices last]) 
				ifTrue: [p2i := anArray last asIntegerPoint]].
			(closed or: ["bigClipRect intersects: (p1i rect: p2i) optimized:"
					((p1i min: p2i) max: bigClipRect origin) 
						<= ((p1i max: p2i) min: bigClipRect corner)]) 
				ifTrue:[
					bevel ifTrue:[
						lineColor := (p1i quadrantOf: p2i) > 2 
								ifTrue: [topLeftColor]
								ifFalse: [bottomRightColor]].
					segmentOffset := aCanvas line: p1i to: p2i width: borderWidth 
										color: lineColor dashLength: borderDashSpec first
										secondColor: borderDashSpec third
										secondDashLength: borderDashSpec second
										startingOffset: segmentOffset]]! !

!CPolygonPlayer methodsFor: 'drawing' stamp: 'ar 2/5/2005 20:52'!
drawOn: aCanvas in: drawingBounds
	"Display the receiver, a spline curve, approximated by straight 
	line segments."
	| array |
	vertices size <= 1 ifTrue:[^super drawOn: aCanvas in: drawingBounds].
	closed ifTrue:[
		aCanvas drawPolygon: self getVertices fillStyle: self fill.
		aCanvas isShadowDrawing ifTrue: [^ self]].
	array := self drawArrowsOn: aCanvas.
	self drawClippedBorderOn: aCanvas usingEnds: array.
! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/4/2005 23:50'!
filledForm
	"Answer the filledForm of the receiver"
	<bewareOf: #filledFormChanged>
	^(self propertyValueAt: #filledForm) ifNil:[self computeFilledForm]! !

!CPolygonPlayer methodsFor: 'private' stamp: 'tk 2/7/2005 13:06'!
filledForm: aValue
	"Modify the receiver's filledForm"
	^self propertyValueAt: #filledForm put: aValue with: #filledFormChanged! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/4/2005 23:47'!
getVertices

	smoothCurve ifFalse: [^ vertices].

	"For curves, enumerate the full set of interpolated points"
	^ Array streamContents:
		[:s | self lineSegmentsDo: [:pt1 :pt2 | s nextPut: pt1]]! !

!CPolygonPlayer methodsFor: 'initialize' stamp: 'ar 2/6/2005 19:15'!
initialize
	super initialize.
	self userDraw := true.
	vertices := Array with: 12@0 with: 50@20 with: 0@40.
	borderWidth := 2.
	borderColor := Color r: 0.0 g: 0.419 b: 0.935.
	borderStyle := #simple.
	closed := true.
	smoothCurve := false.
	arrowStyle := #none.
	color := Color orange.
	self computeBounds.
! !

!CPolygonPlayer methodsFor: 'testing' stamp: 'ar 2/4/2005 23:18'!
isClosed
	^ closed! !

!CPolygonPlayer methodsFor: 'testing' stamp: 'ar 2/4/2005 23:19'!
isCurve
	^ smoothCurve! !

!CPolygonPlayer methodsFor: 'private' stamp: 'ar 2/4/2005 23:47'!
lineSegments
	| lineSegments |
	lineSegments := OrderedCollection new.
	self lineSegmentsDo: [:p1 :p2 | lineSegments addLast: (Array with: p1 with: p2)].
	^ lineSegments! !

!CPolygonPlayer methodsFor: 'spline' stamp: 'ar 2/5/2005 21:28'!
lineSegmentsDo: endPointsBlock 
	"Emit a sequence of segment endpoints into endPointsBlock."

	| n t xx yy x1 x2 x3 y1 y2 y3 beginPoint endPoint cs |
	smoothCurve ifFalse:[
		beginPoint := nil.
		vertices do:[:vert | 
			beginPoint ifNotNil: [endPointsBlock value: beginPoint value: vert].
			beginPoint := vert].
		(closed or: [vertices size = 1]) 
			ifTrue: [endPointsBlock value: beginPoint value: vertices first].
		^self].

	"For curves we include all the interpolated sub segments."
	vertices size < 1 ifTrue: [^self].
	cs := self coefficients.
	beginPoint := (xx := cs first first) @ (yy := cs fifth first).
	1 to: cs first size - 1 do:[:i | 
		"taylor series coefficients"
		x1 := cs second at: i.
		y1 := cs sixth at: i.
		x2 := (cs third at: i) / 2.0.
		y2 := (cs seventh at: i) / 2.0.
		x3 := (cs fourth at: i) / 6.0.
		y3 := ((cs eighth) at: i) / 6.0.
		"guess n"
		n := 5 max: (((x2 abs + y2 abs) * 2.0 + (cs third at: i + 1) abs 
						+ (cs seventh at: i + 1) abs) / 100.0) rounded.
		1 to: n - 1 do:[:j | 
			t := j asFloat / n.
			endPoint := (((x3 * t + x2) * t + x1) * t + xx) 
						@ (((y3 * t + y2) * t + y1) * t + yy).
			endPointsBlock value: beginPoint value: endPoint.
			beginPoint := endPoint].
		endPoint := (xx := cs first at: i + 1) @ (yy := cs fifth at: i + 1).
		endPointsBlock value: beginPoint value: endPoint.
		beginPoint := endPoint]
! !

!CPolygonPlayer methodsFor: 'spline' stamp: 'ar 2/4/2005 23:40'!
nextToFirstPoint
	"For arrow direction"

	smoothCurve 
		ifTrue: 
			[curveState ifNil: [self coefficients].
			^curveState second]
		ifFalse: [^vertices second]! !

!CPolygonPlayer methodsFor: 'spline' stamp: 'ar 2/4/2005 23:40'!
nextToLastPoint
	"For arrow direction"

	smoothCurve 
		ifTrue: 
			[curveState ifNil: [self coefficients].
			^curveState third]
		ifFalse: [^vertices at: vertices size - 1]! !

!CPolygonPlayer methodsFor: 'events' stamp: 'ar 2/6/2005 15:24'!
onArrowSpecChanged
	<on: arrowSpecChanged>
	self computeBounds.
	self invalidate.! !

!CPolygonPlayer methodsFor: 'events' stamp: 'ar 2/6/2005 15:45'!
onArrowStyleChanged
	<on: arrowStyleChanged>
	self computeBounds.
	self invalidate.! !

!CPolygonPlayer methodsFor: 'events' stamp: 'ar 4/28/2005 13:16'!
onBorderChanged
	| radius oldBox newBox tfm |
	<on: borderChanged>
	radius := (borderWidth + 1) // 2.
	oldBox := Rectangle encompassing: self getVertices.
	newBox := (0@0 extent: width@height) insetBy: radius.
	(oldBox width < 1 or:[oldBox height < 1]) ifTrue:[
		extent := (oldBox expandBy: radius) extent.
		^self].
	(newBox width < 1 or:[newBox height < 1]) ifTrue:[
		extent := ((oldBox origin extent: 1@1) expandBy: radius) extent.
		^self].
	tfm := CMatrixTransform localRect: oldBox toGlobal: newBox.
	vertices := vertices collect:[:p | tfm localPointToGlobal: p].! !

!CPolygonPlayer methodsFor: 'events' stamp: 'ar 2/6/2005 15:24'!
onClosedChanged
	<on: closedChanged>
	self computeBounds.
	self invalidate.! !

!CPolygonPlayer methodsFor: 'events' stamp: 'ar 2/6/2005 18:52'!
onExtentChanged
	| newBox vtxScale outerBox innerBox inset |
	<on: extentChanged>
	"Compute inner and outer bounds"
	outerBox := self curveBounds.
	self arrowForms do:[:f | outerBox := outerBox merge: (f offset extent: f extent)].
	innerBox := Rectangle encompassing: self getVertices.
	"Compute inset of vertices (since we do not reize border or arrows)"
	inset := innerBox origin -  outerBox origin corner: outerBox corner - innerBox corner.
	newBox := (0@0 extent: width@height) insetBy: inset.
	(newBox width < 1.0 or:[newBox height < 1.0]) ifTrue:[
		newBox := newBox origin extent: (newBox extent max: 1@1).
	].
	vtxScale := [newBox extent asFloatPoint / innerBox extent] 
		on: ZeroDivide do:[:ex| ex resume: 1].
	self pauseScript: #onVerticesChanged.
		vertices := vertices collect:[:p | p * vtxScale].
		self computeBounds. "recompute from scratch"
	self resumeScript: #onVerticesChanged.
! !

!CPolygonPlayer methodsFor: 'events' stamp: 'ar 4/25/2005 13:27'!
onFillChanged
	<on: fillChanged>
	(self closed not and:[border width > 0]) ifTrue:[border color: self color].! !

!CPolygonPlayer methodsFor: 'events' stamp: 'ar 6/25/2005 19:55'!
onGeometryChanged
	"A polygon's transform is always identity - so absorb changes here"
	| tfm |
	<on: geometryChanged>
	tfm := transform.
	tfm isPureTranslation ifTrue:[^self].
	transform := COffsetTransform withOffset: tfm offset.
	tfm offset: 0@0.
	self pauseScript: #onVerticesChanged.
	self nose: (tfm transformDirection: self nose).
	vertices := vertices collect:[:v| tfm transformPoint: v].
	self computeBounds.
	self resumeScript: #onVerticesChanged.
! !

!CPolygonPlayer methodsFor: 'events' stamp: 'ar 2/6/2005 15:24'!
onSmoothCurveChanged
	<on: smoothCurveChanged>
	self computeBounds.
	self invalidate.! !

!CPolygonPlayer methodsFor: 'events' stamp: 'ar 2/6/2005 15:24'!
onVerticesChanged
	<on: verticesChanged>
	self computeBounds.
	self invalidate.! !

!CPolygonPlayer methodsFor: 'initialize' stamp: 'ar 2/4/2005 23:48'!
releaseCachedState
	filledForm := nil.
	arrowForms := nil.
	borderForm := nil.
	curveState := nil.
! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
smoothCurve
	"Answer the smoothCurve of the receiver"
	<bewareOf: #smoothCurveChanged>
	^self propertyValueAt: #smoothCurve! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
smoothCurve: aValue
	"Modify the receiver's smoothCurve"
	^self propertyValueAt: #smoothCurve put: aValue with: #smoothCurveChanged! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
vertices
	"Answer the vertices of the receiver"
	<bewareOf: #verticesChanged>
	^self propertyValueAt: #vertices! !

!CPolygonPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:55'!
vertices: aValue
	"Modify the receiver's vertices"
	^self propertyValueAt: #vertices put: aValue with: #verticesChanged! !

CPlayer subclass: #CRectanglePlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Objects'!

!CRectanglePlayer commentStamp: 'tk 3/1/2005 11:37' prior: 0!
A Rectangle is a basic player.  It can be embedded inside another object, and other objects can be embedded in it.  Use the Color Tool to change its color and border.[general]


### Allow Users to see the above explanation ###!

CRectanglePlayer defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!CRectanglePlayer methodsFor: 'initialize' stamp: 'ar 2/6/2005 19:12'!
initialize
	super initialize.
	self color := Color blue.
	self borderWidth := 2.
	self borderColor := Color black.
	self borderStyle := #simple.
! !

CPlayer subclass: #CRoundRectPlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Objects'!

!CRoundRectPlayer commentStamp: 'tk 3/1/2005 11:33' prior: 0!
A RoundRect is a rectangle with rounded corners.  Like other players, it can be embedded inside another object, and other objects can be embedded in it.[general]


### Allow Users to see the above explanation ###!

CRoundRectPlayer defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="radius" toSet="radius:" changeEvent="radiusChanged">radius</field>
</fields>'!

!CRoundRectPlayer methodsFor: 'drawing' stamp: 'ar 2/6/2005 19:06'!
drawOn: aCanvas in: drawingBounds
	aCanvas fillRoundRect: drawingBounds radius: radius fill: self fill border: self border! !

!CRoundRectPlayer methodsFor: 'initialize' stamp: 'ar 2/6/2005 19:13'!
initialize
	super initialize.
	self radius := 16.
	self userDraw := true.
	self color := Color green.
	self borderWidth := 2.
	self borderColor := Color black.
	self borderStyle := #simple.
! !

!CRoundRectPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:03'!
radius
	"Answer the radius of the receiver"
	<bewareOf: #radiusChanged>
	^self propertyValueAt: #radius! !

!CRoundRectPlayer methodsFor: 'accessing' stamp: 'ar 2/6/2005 19:03'!
radius: aValue
	"Modify the receiver's radius"
	^self propertyValueAt: #radius put: aValue with: #radiusChanged! !

CPlayer subclass: #CSoundPlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Objects'!

!CSoundPlayer commentStamp: 'ar 3/14/2004 16:08' prior: 0!
This class is a stand-in for objects representing sounds. It is intended to be used wherever we create or import a sound. At some point we WILL want to have different costumes for 'sounds' and this will be the base type to be used for it.!

CSoundPlayer defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!CSoundPlayer methodsFor: 'initialize' stamp: 'ar 3/14/2004 17:47'!
initialize
	super initialize.
	self color: Color white.! !

CPlayer subclass: #CSprite
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Objects'!

!CSprite commentStamp: '<historical>' prior: 0!
A sprite is an animated object. It contains a list of spriteCostumes upon which it chooses when its cursor changes.!

CSprite defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="spriteCostumes" toSet="spriteCostumes:" changeEvent="spriteCostumesChanged">spriteCostumes</field>
	<field toGet="cursor" toSet="cursor:" changeEvent="cursorChanged">cursor</field>
</fields>'!

!CSprite methodsFor: 'accessing' stamp: 'ar 6/21/2005 19:40'!
cursor
	"Answer the cursor of the receiver"
	<bewareOf: #cursorChanged>
	^self propertyValueAt: #cursor! !

!CSprite methodsFor: 'accessing' stamp: 'ar 6/21/2005 19:40'!
cursor: aValue
	"Modify the receiver's cursor"
	^self propertyValueAt: #cursor put: aValue with: #cursorChanged! !

!CSprite methodsFor: 'initialize' stamp: 'ar 6/21/2005 22:40'!
defaultCostume
	^CSpriteCostume! !

!CSprite methodsFor: 'initialize' stamp: 'ar 6/21/2005 20:07'!
initialize
	super initialize.
	self define: #spriteCostumes as: CCollection new.
	spriteCostumes wrapCursor: true.
	cursor := 0.! !

!CSprite methodsFor: 'events' stamp: 'ar 6/21/2005 19:41'!
onCostumeChanged
	<on: costumeChanged>
	spriteCostumes selection: costume.! !

!CSprite methodsFor: 'events' stamp: 'ar 6/21/2005 21:08'!
onCostumesCursorChanged
	<on: cursorChanged in: spriteCostumes>
	cursor := spriteCostumes cursor.! !

!CSprite methodsFor: 'events' stamp: 'ar 6/21/2005 23:09'!
onCursorChanged
	| max index |
	<on: cursorChanged>
	max := spriteCostumes size.
	max = 0 
		ifTrue:[cursor := 0]
		ifFalse:[cursor := cursor - 1 \\ max + 1].
	index := cursor truncated.

	"pause these scripts while changing the cursor"
	self pauseScript: #onCostumesCursorChanged.
	self pauseScript: #onCostumeChanged.
		spriteCostumes cursor: index.
		costume := spriteCostumes selection ifNil:[self defaultCostume].
	self resumeScript: #onCostumesCursorChanged.
	self resumeScript: #onCostumeChanged.! !

!CSprite methodsFor: 'events' stamp: 'ar 6/21/2005 22:47'!
onSpriteCostumesChanged
	"This is a little watcher to see if we removed the first or last element from the list"
	<on: changed in: spriteCostumes>
	spriteCostumes size = 0 ifTrue:[^cursor := 0].
	cursor = 0 ifTrue:[^cursor := 1].
! !

!CSprite methodsFor: 'accessing' stamp: 'ar 6/21/2005 19:39'!
spriteCostumes
	"Answer the spriteCostumes of the receiver"
	<bewareOf: #spriteCostumesChanged>
	^self propertyValueAt: #spriteCostumes! !

!CSprite methodsFor: 'accessing' stamp: 'ar 6/21/2005 19:39'!
spriteCostumes: aValue
	"Modify the receiver's spriteCostumes"
	^self propertyValueAt: #spriteCostumes put: aValue with: #spriteCostumesChanged! !

CPlayer subclass: #CSpriteCostume
	instanceVariableNames: ''
	classVariableNames: 'DefaultForm'
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Objects'!

!CSpriteCostume commentStamp: '<historical>' prior: 0!
I represent the default costume for a sprite just so that there is something to see when a new sprite is created or all costumes have been removed from the sprite.!

CSpriteCostume defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!CSpriteCostume class methodsFor: 'class initialization' stamp: 'ar 6/21/2005 22:39'!
defaultFormDataA
	^'R0lGODlhRwBQAOcAAAAAABgYGLW1/+cACPf39wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAkAAAAALAAAAABHAFAAAAj+AAEIHEiwoMGD
AwcoXMgQocOHECMeZEix4kKJGDNOtMiRo8aPETuK7AiyZMGRKD2aBJmyZcWVGikKEEBAgMuR
MCUynEmTQE2UPHmqHBggJ8GdM336nCkyaFCfFIkGmFrU6MKnSn9ydMozK0WqYKdqrJpQodOs
Wi1yTep1YdiwEuECuMq17da1aG0OeCvXId+pZtf2BIpXqd6/YB8iDiy4peClNhEnRig5wIDG
N+k6dVuZbMHKl7lmrhiUYWfPUkFvHp3y9MHTjFm3dG0QtuzMtD93vt2RgMXcBG3zpqi0IvDU
kodXLP618+vdyhkyN+28NvToCqdzrvwcNPbsUJv+c7fuPbNvjtoVHhco3LFh9OG3S+6evGnd
pTWFgj9PHTHl66Q9lldp6e01H32LySQYVoadFV9/b/kFYGgLdpXVffxB2Bd59VFYoYUXMqhX
VBv+l6CHH7IVIohMRcUeRt6liGGDIo6oUEkJyjgjjSrqdyNIi+m4I2QstvgjjmBpJiSIPPaU
n405FYVUhR3RlBSGPg5g1FxTrnYTli/lFGCWtzlpkVXf9fbgRWKmCV+GbMLkpppwatnmnMut
aaeceOZZJ5p9SqfnniYFStygdxo6wHsuvvhRl246aeSPJT7U5aS8FWmjfyExJppsWE7KKURK
LvhlqCOO95pApcpo0ZPFheWnYYQcWtbqkk42KemsfAX31606oqWrdsCdiOKSwsbKK60AlOep
kMmWtuiDuTkL7ILR2khsde0pKGOIh8JJW7dVuuqneAeSm2aB407oJruuqYtdgeo5J69ydqHL
6b235avvX+y5O5y///barMD90rusXPy6JCxK8TZMp7B1cjSexPtRjNZNFyMcrsaD4nSwwSN3
OJLGtyHHsMeG1uqrtYoSWhuCb8Xc6FgH2gzTqFyeOdyWPB/lkEtbulz00TAirfRBAQEAOw=='! !

!CSpriteCostume class methodsFor: 'class initialization' stamp: 'ar 6/21/2005 22:39'!
defaultFormDataB
	^'R0lGODlhSwBIAOcAAAAAABAQECn/AFKU//f39///awAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAkAAAAALAAAAABLAEgAAAj+AAEIHEiwoMGD
CAUoXIiwocOHECMKXEixokKJGDNKtMixosaPIAF0HEmxYYCQEk8eJGmxgMsCBC4WDECTJkqT
NVUOpPiSQMyWPX3KBJCz6E2CRXNOXBhUKM+XMJ0OTFrz6FSqKpm69Mn1p0KoXQtcxFr1KFma
Wrd29SoAbNeFZ3WijJtWLdenTcUKiGtzLt+vUKOyBRwYLt+bfAMQLsyRscLEiP8u1svSImS/
dNNW7ng5ZOLNoDt7Pgs69OHIZEubJmuVKFbVq7G2Tg17M+vWAovWLn0bd+7SA4ILD65wsGWl
vguSHM58eNcBFQfLTb60YvPrwtcyv1uSOkGL2MP+a3fOfah3kRTDi3++3SnD8+gXqsc+nrx7
AfDjK5x/fS2B5uWZl5x1/LXHnn1snUdggQj61N99CqbHYIP/AXgfftQtOOEA/j0IYYYSbshh
fQZ2N2CIG3booUcn7idicP5VWKKJrWmYoooWJuibjRPGSN+HuPHIoI8r0niTkD2SaB+LViE5
pJLZXVgjii+OeOCM7x3l5JNXLslkSOBVGWWXY1qk0XQciTmmg0WaOJ1BRum3ZYFkQhfgWMjN
9FplVZJZ3IW95bYnny9yBZ1Fdwaa2W6VGbdQgHFdRRujlEJ6lqCLUlpbeZB9pumm93UqmaaO
jmRppJ7CZt+hLCV6WKqjwK3Kakd37vXqqKDlKBxJtYqa6Wa6EmfqhbaiiiuwBu5KK7HFJvYr
ss4NN6yjzh4b63XTUlstaYxiuyyzzW6b1Keg1WqYuFSR2yi4j6Gbrrrfshuuu2jBS1GMLNE7
LrkxvmXbpbAy2q+5HLnGWsAC+1caUrJhOqi9sOlJlaQPQ2xmfA/FyfDE1XXUcW0gvemwxnIa
+fFI+RncV0Ip+yZyy7gFBAA7'! !

!CSpriteCostume class methodsFor: 'class initialization' stamp: 'ar 6/21/2005 22:39'!
initialize
	"CSpriteCostume initialize"
	| data |
	data := Base64MimeConverter mimeDecode: self defaultFormDataB as: ByteString.
	DefaultForm := GIFReadWriter formFromStream: (ReadStream on: data asByteArray).! !

!CSpriteCostume methodsFor: 'initialize' stamp: 'ar 6/29/2005 09:54'!
setupCostume
	| form |
	super setupCostume.
	form := Form extent: 8@8 depth: 8.
	form fillWhite.
	form fill: (0@0 corner: 4@4) fillColor: Color gray.
	form fill: (4@4 corner: 8@8) fillColor: Color gray.
	graphic := form.
	graphicFit := #tile.
	self extent: 40@40.! !

CPlayer subclass: #CValuePlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-User-Kernel'!

!CValuePlayer commentStamp: 'ar 3/25/2004 12:09' prior: 0!
CValuePlayer is provided as a convenient superclass for documenting that some player is intended to represent a particular value and now more. There is no other significance to it - all players represent a value and therefore many players all over the places will commonly refer to their values.!

CValuePlayer defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

CPlayer subclass: #CWorldPlayer
	instanceVariableNames: 'sensor hands scheduler maxLayoutRounds damageRecorder bltRects eventQueue display ownerMorph frameNumber activeProject killFrame'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Worlds'!

CWorldPlayer defineFields: '<?xml  version="1.0" ?>
<fields>
	<field instanceVariable="true">sensor</field>
	<field instanceVariable="true" toGet="hands">hands</field>
	<field instanceVariable="true" toGet="scheduler">scheduler</field>
	<field instanceVariable="true" toGet="maxLayoutRounds" toSet="maxLayoutRounds:">maxLayoutRounds</field>
	<field instanceVariable="true">damageRecorder</field>
	<field instanceVariable="true">bltRects</field>
	<field instanceVariable="true" toGet="eventQueue" toSet="eventQueue:">eventQueue</field>
	<field instanceVariable="true" toGet="display" toSet="display:">display</field>
	<field instanceVariable="true" toGet="ownerMorph" toSet="ownerMorph:">ownerMorph</field>
	<field instanceVariable="true">frameNumber</field>
	<field instanceVariable="true" toGet="activeProject" toSet="activeProject:" changeEvent="activeProjectChanged">activeProject</field>
	<field instanceVariable="true" toGet="killFrame" toSet="killFrame:">killFrame</field>
	<field toGet="process" toSet="process:" changeEvent="processChanged">process</field>
	<field toGet="doneSema" toSet="doneSema:" changeEvent="doneSemaChanged">doneSema</field>
</fields>'!

!CWorldPlayer class methodsFor: 'running' stamp: 'ar 11/24/2004 09:25'!
primitiveError: aString 
	"This method is called when the error handling results in a recursion in 
	calling on error: or halt or halt:."

	| context |
	(String
		streamContents: 
			[:s |
			s nextPutAll: '***System error handling failed***'.
			s cr; nextPutAll: aString.
			context := thisContext sender sender.
			20 timesRepeat: [context == nil ifFalse: [s cr; print: (context := context sender)]].
			s cr; nextPutAll: '-------------------------------'.
			s cr; nextPutAll: 'Type CR to enter an emergency evaluator.'.
			s cr; nextPutAll: 'Type any other character to restart.'])
		displayAt: 0 @ 0.
	[Sensor keyboardPressed] whileFalse.
	Sensor keyboard = Character cr ifTrue: [Transcripter emergencyEvaluator].! !

!CWorldPlayer class methodsFor: 'running' stamp: 'bf 12/5/2004 14:48'!
run
	"CWorldPlayer run"
	| world |
	world := self new.
	world bootstrap:[
		world onBootstrap. 	"install ProjectBuilder"
		world startScript: [
			[
				world scheduler shutDown.	"terminate all scripts"
				world process terminate.		"terminate Tweak process"
				World install. 				"reinstall Morphic"
				Project spawnNewProcess.	"fork Morphic process"
			] forkAt: Processor userInterruptPriority. "run at higher priority"
		] when: #close.
	].
	world forkRun.
	Processor terminateActive.! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 3/28/2004 23:37'!
activeProject
	"Answer the currently active project"
	<bewareOf: #activeProjectChanged>
	^activeProject! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'bf 3/30/2004 11:32'!
activeProject: aProject
	"Set the currently active project"
	activeProject ifNotNil:[activeProject signal: #passivate with: self].
	activeProject := aProject.
	self signal: #activeProjectChanged.
	activeProject ifNotNil:[activeProject signal: #activate with: self].
	^activeProject! !

!CWorldPlayer methodsFor: 'hands' stamp: 'ar 7/14/2003 11:01'!
addHand: aHand
	"@@@@: Fix this. Needs to handle #outOfWorld etc. :@@@@"
	aHand container ifNotNil:[aHand container removeHand: aHand].
	aHand costume setContainer: self costume.
	hands := (hands copyWithout: aHand) copyWith: aHand.
! !

!CWorldPlayer methodsFor: 'initialize' stamp: 'ar 5/22/2005 18:11'!
bootstrap: aBlock
	"Bootstrap message for otherworldly senders"
	scheduler bootstrap:[
		Processor activeProcess hand: self primaryHand.
		aBlock value].
! !

!CWorldPlayer methodsFor: 'Morphic-support' stamp: 'ar 12/18/2004 22:26'!
close
	| w |
	super close.
	self signal: #close.		"#close is not signalled for worlds because they have no container"
	self syncScript.
	"In case of running stand-alone we will not come here. See CWorldPlayer class>>run and CProjectBuilder>>quit"
	ownerMorph ifNil:[^self].
	self fullScreenMode: false.
	w := ownerMorph.
	[(w isSystemWindow or: [w owner isPlayfieldLike])
		ifTrue: [^w delete].
	w isNil] whileFalse: [w := w owner].! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 7/14/2003 11:01'!
display
	"Answer the display this world is shown on"
	^display! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 9/8/2004 16:25'!
display: aDisplayScreen
	"Set my display this world is shown on"
	display == aDisplayScreen ifTrue:[^self].
	self signalChanged: #displayChanged from: display to: aDisplayScreen.
	display := aDisplayScreen.
	display ifNotNil:[self costume extent: display extent].
	self costume localCache form: display.
! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 2/7/2004 21:59'!
displayDepth: bpp
	display setExtent: display extent depth: bpp.
	self invalidate.! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 7/15/2003 20:25'!
displayExtent
	^display extent! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 7/15/2003 20:25'!
displayExtent: aPoint
	display setExtent: aPoint depth: display depth.
	self invalidate.
! !

!CWorldPlayer methodsFor: 'drawing' stamp: 'ar 11/12/2004 21:09'!
displayWorld
	"@@@@: This needs to be moved into a special cache - it doesn't belong here :@@@@"
	| myCanvas rectList validList rectToFill flashList recorder |
	killFrame == true ifTrue:[killFrame := false.^#()].
	frameNumber := frameNumber + 1 bitAnd: SmallInteger maxVal.
	display deferUpdates: true.
	myCanvas := CTransformCanvas on: display.
	recorder := self costume localCache damageRecorder.
	rectList := recorder invalidRectsFullBounds: self costume bounds.
	"sort by areas to draw largest portions first"
	rectList := rectList asArray sort:[:r1 :r2| r1 area > r2 area].
	recorder reset.
	validList := OrderedCollection new: rectList size.
	rectToFill := nil.
	rectList do: [:dirtyRect |
		dirtyRect allAreasOutsideList: validList do:[:r| 
			rectToFill 
				ifNil:[rectToFill := r copy]
				ifNotNil:[rectToFill := rectToFill quickMerge: r].
			validList add: r]].
	rectToFill ifNotNil:[
		myCanvas clipBy: rectToFill during:[:cc| 
			self costume topDownFullDrawOn: cc damageList: validList.
			hands do:[:h| h costume fullDrawOn: myCanvas].
			"hands do:[:h| h costume topDownFullDrawOn: myCanvas damageList: validList]."
		].
		"hands do:[:h| h fullDrawOn: myCanvas damageList: validList]."
	].
	myCanvas finish.
	flashList := validList.
	bltRects ifNotNil:[
		"Merge the blt rects"
		validList := OrderedCollection new: flashList size + bltRects size.
		validList addAll: bltRects.
		flashList do:[:rect| rect allAreasOutsideList: bltRects do:[:r| validList add: r]].
		bltRects := nil].
	display forceDamageToScreen: validList.
	false ifTrue:[self flashRects: flashList color: Color cyan].
	display deferUpdates: false.
	"Sender might need the damage rectangles"
	^validList! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 12/21/2004 01:00'!
doneSema
	"Answer the doneSema of the receiver"
	<bewareOf: #doneSemaChanged>
	^self propertyValueAt: #doneSema! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 12/21/2004 01:00'!
doneSema: aValue
	"Modify the receiver's doneSema"
	^self propertyValueAt: #doneSema put: aValue with: #doneSemaChanged! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 7/14/2003 11:01'!
eventQueue
	"Answer the event queue being used for retrieving input events"
	^eventQueue! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 7/14/2003 11:01'!
eventQueue: aSharedQueue
	"Install the event queue being used for retrieving input events"
	eventQueue := aSharedQueue! !

!CWorldPlayer methodsFor: 'drawing' stamp: 'ar 7/14/2003 11:01'!
flashRects: rectangleList color: aColor
	"For testing. Flashes the given list of rectangles on the Display so you can watch incremental redisplay at work."
	"Details: Uses two reverses so that the display is restored to its original state. This is necessary when in deferred update mode."
	| blt |
	blt := (BitBlt toForm: display)
		fillColor: aColor;
		sourceOrigin: 0@0;
		combinationRule: Form reverse.
	rectangleList do: [:r | blt destRect: r; copyBits].
	rectangleList do:[:r| display forceToScreen: r].
	display forceDisplayUpdate.
	(Delay forMilliseconds: 100) wait.
	rectangleList do: [:r | blt destRect: r; copyBits].
	rectangleList do:[:r| display forceToScreen: r].
	display forceDisplayUpdate.! !

!CWorldPlayer methodsFor: 'initialize' stamp: 'ar 11/24/2004 14:21'!
forkRun
	process := [self run] forkAt: Processor userSchedulingPriority.! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 1/23/2005 10:33'!
fullScreenMode: aBool
	"delegate to morph if present"
	(ownerMorph respondsTo: #fullScreenMode:)
		ifTrue: [ownerMorph fullScreenMode: aBool].! !

!CWorldPlayer methodsFor: 'hands' stamp: 'ar 7/14/2003 11:01'!
hands
	^hands! !

!CWorldPlayer methodsFor: 'initialize' stamp: 'ar 5/28/2005 14:42'!
initialize
	"Bootstrap a new world"
	super initialize.
	scheduler ifNil:[
		scheduler := ScriptScheduler new.
		"set the context for any newly spawned scripts"
		scheduler run.
	].
	"temporary hand until we have a real one"
	hands := Array with: CHandPlayer basicNew.
	self bootstrap:[
		frameNumber := 0.
		self name: 'World'.
		maxLayoutRounds := 3.

		self costume: CWorldCostume new.
		self costume localCache: CFormCache new.
		self clipping: true.
		self userDraw: true.
		self color: (Color gray: 0.5).

		hands := #().
		self addHand: CHandPlayer new.

		"Default to Display"
		self display: Display.
		"Default to primary input queue"
		self eventQueue: Sensor eventQueue.
	].
! !

!CWorldPlayer methodsFor: 'private' stamp: 'ar 7/14/2003 11:01'!
isSpecialSystemScript: scriptDescriptor
	"Some scripts need to be run for debugging and cannot stopped for debugging"
	^(super isSpecialSystemScript: scriptDescriptor) or:[scriptDescriptor == #displayWorld]! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 10/17/2004 17:47'!
killFrame
	"Answer the killFrame of the receiver"
	^killFrame! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 10/17/2004 17:47'!
killFrame: aValue
	"Modify the receiver's killFrame"
	killFrame := aValue! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 12/7/2003 18:14'!
label
	ownerMorph ifNil:[^''].
	^ownerMorph label! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'bf 7/23/2004 12:19'!
label: aString
	ownerMorph ifNil:[^nil].
	(ownerMorph respondsTo: #label:) ifFalse: [^nil].
	^ownerMorph label: aString! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 7/14/2003 11:01'!
maxLayoutRounds
	"Answer the maximum number of retries for computing layouts"
	^maxLayoutRounds! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 7/14/2003 11:01'!
maxLayoutRounds: aNumber
	"Set the maximum number of retries for computing layouts"
	maxLayoutRounds := aNumber max: 1.! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 3/2/2005 10:26'!
minCycleTime
	"Answer the minimal time one world cycle should take. If the current cycle is faster then we may give the host OS a few cycles back"
	^Preferences higherPerformance
		ifTrue: [0]
		ifFalse: [20]
! !

!CWorldPlayer methodsFor: 'initialize' stamp: 'ar 6/24/2005 21:05'!
onBootstrap
	"Load auto load targets"
	| project |
	self startScript: #onDrop: when: {self. #dragDrop}.
	project := Smalltalk at: #CProjectBuilder ifPresent:[:cls| cls new].
	project window: self.
	project topLeft: 0@0.
	project costume isWorld: true.
	self add: project.
	self activeProject: project.! !

!CWorldPlayer methodsFor: 'running' stamp: 'ar 7/14/2003 11:40'!
onDrop: aPlayer
	aPlayer signal: #dropReject.! !

!CWorldPlayer methodsFor: 'Morphic-support' stamp: 'ar 7/14/2003 11:01'!
ownerMorph
	^ownerMorph! !

!CWorldPlayer methodsFor: 'Morphic-support' stamp: 'ar 7/14/2003 11:01'!
ownerMorph: aMorph
	ownerMorph := aMorph! !

!CWorldPlayer methodsFor: 'hands' stamp: 'ar 7/14/2003 11:01'!
primaryHand
	^hands first! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 11/24/2004 14:21'!
process
	"Answer the process of the receiver"
	<bewareOf: #processChanged>
	^self propertyValueAt: #process! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 11/24/2004 14:21'!
process: aValue
	"Modify the receiver's process"
	^self propertyValueAt: #process put: aValue with: #processChanged! !

!CWorldPlayer methodsFor: 'hands' stamp: 'ar 7/14/2003 11:01'!
removeHand: aHand
	"@@@@: Fix this. Needs to handle #outOfWorld etc. :@@@@"
	(hands includes: aHand) ifFalse: [^self].
	aHand invalidate.
	hands := hands copyWithout: aHand.! !

!CWorldPlayer methodsFor: 'running' stamp: 'ar 3/2/2005 10:27'!
run
	"Run a world cycle"
	| time |
	doneSema := nil.
	self display: Display.
	self extent: Display extent.
	[doneSema == nil] whileTrue:[
		"TODO: Wait for activity. How it should work: Figure out when a) an input event comes in or b) some script gets activated (e.g., put onto the schedulers 'active' list).
		@@@: This might be tricky since we are trying to prevent unneeded call chains (for process switching) in places like #waitProcessOn: Duh... how can we tell if some process just went onto that particular suspending list in order to wake us up??? Oh, well, I'll figure something out. @@@@"
		time := Time millisecondsToRun:[self runWorldCycle].
		self minCycleTime > 0
			ifTrue: [(Delay forMilliseconds: (self minCycleTime - time max: 1)) wait]
			ifFalse: [Processor yield].
	].
	doneSema signal.! !

!CWorldPlayer methodsFor: 'running' stamp: 'ar 7/14/2003 11:01'!
run: seconds
	| timeOut |
	timeOut := Time millisecondClockValue + (seconds*1000).
	[Time millisecondClockValue <= timeOut] 
		whileTrue:[self runWorldCycle].! !

!CWorldPlayer methodsFor: 'Morphic-support' stamp: 'ar 1/8/2008 17:48'!
runCroquetCycle
	"Run a world cycle in the presence of Morphic.
	The event queue is being hijacked for this..."
	| script activeHand prior |
	^[
		"This of course is all broken and only for compatibility"
		Processor activeIsland project: self.
		prior := Processor activeProcess scheduler.
		Processor activeIsland scheduler: scheduler.

		self signalTick: #tick.

		"Process pending events - only the primary hand here,
		and we dispatch directly since the event has already been constructed."
		activeHand := hands first.
		script := activeHand startScript: #processEventsFrom: withArguments:{eventQueue}.
		script hand: activeHand.

		"Run activated scripts"
		scheduler run.

		"Update layout"
		self runLayoutCycle.

		Processor activeIsland scheduler: prior.

	] on: Error do:[:ex|
		activeHand interrupted.
		ex pass.
	].! !

!CWorldPlayer methodsFor: 'running' stamp: 'ar 5/28/2005 14:44'!
runLayoutCycle
	"Run a layout cycle"
	"NOTE: Scripts triggered by #layoutChanged may not work correctly if they don't complete before the world is displayed. In order to solve this problem, we recompute the layout changes here and do so repeatedly (!!) up to maxLayoutRounds in order to prevent infinite recursion. In addition, we randomize the number of layout rounds in order to supply visual feedback for conflicting constraints."
	| layoutRepeat layoutMax script |
	layoutMax := maxLayoutRounds + (frameNumber bitAnd: 1).
	layoutRepeat := 0.
	[self costume layoutNeeded and:[
		(layoutRepeat := layoutRepeat + 1) <= layoutMax]] whileTrue:[
		script := self startScript: #fullBounds.
		script hand: self primaryHand.
		scheduler run.
	].
! !

!CWorldPlayer methodsFor: 'Morphic-support' stamp: 'ar 1/8/2008 17:48'!
runMorphicCycle
	"Run a world cycle in the presence of Morphic.
	The event queue is being hijacked for this..."
	| displaySize script activeHand prior dirtyList |
	"(eventQueue isEmpty not or:[scheduler hasActiveScripts]) ifFalse:[^#()]."
	^[
		"This of course is all broken and only for compatibility"
		Processor activeIsland project: self.
		prior := Processor activeProcess scheduler.
		Processor activeIsland scheduler: scheduler.

		"Check if the display size has changed"
		displaySize := display actualScreenSize.
		displaySize = display extent ifFalse:[
			display setExtent: displaySize depth: display depth.
		].

		self extent = display extent ifFalse:[self extent: displaySize].

		self signalTick: #tick.

		"Process pending events - only the primary hand here,
		and we dispatch directly since the event has already been constructed."
		activeHand := hands first.
		script := activeHand startScript: #processEventsFrom: withArguments:{eventQueue}.
		script hand: activeHand.

		"Run activated scripts"
		scheduler run.

		"Update layout"
		self runLayoutCycle.

		"Restore display - this will return the damage rectangles for the host"
		dirtyList := self displayWorld.
		Processor activeIsland scheduler: prior.
		dirtyList
	] on: Error do:[:ex|
		activeHand interrupted.
		ex pass.
	].! !

!CWorldPlayer methodsFor: 'running' stamp: 'ar 1/8/2008 17:49'!
runWorldCycle
	"Run a world cycle"
	| displaySize script prior result |

	"This of course is all broken and only for compatibility"
	Processor activeIsland project: self.
	prior := Processor activeProcess scheduler.
	Processor activeIsland scheduler: scheduler.

	"Check if the display size has changed"
	displaySize := display actualScreenSize.
	displaySize = display extent ifFalse:[
		display setExtent: displaySize depth: display depth.
	].

	self extent = display extent ifFalse:[self extent: displaySize].

	self signalTick: #tick.

	"TODO: Wait for activity. How it should work: Figure out when a) an input event comes in or b) some script gets activated (e.g., put onto the schedulers 'active' list).
	@@@: This might be tricky since we are trying to prevent unneeded call chains (for process switching) in places like #waitProcessOn: Duh... how can we tell if some process just went onto that particular suspending list in order to wake us up??? Oh, well, I'll figure something out. For now let's just go for 100% CPU usage ;-) @@@@"

	"Process pending events"
	hands do:[:h| 
		script := h startScript: #processEvents.
		script hand: h.
	].

	"Note: Using #startScript: instead of just #signaling the occurance of an event serves the purpose of being able to debug any problems in event handling. If we would just #signal the event here any error in the hand would result in suspending that hand - meaning that there wouldn't be anyone who could potentially handle the events for the debugger (oops...).
	@@@@: This needs to be changed - at least for the non-primary hands. I need to see if we can just 'restart' the primary hand's process event script each cycle. @@@@"

	"Run activated scripts"
	scheduler run.

	"Update layout"
	self runLayoutCycle.

	"Restore display"
	result := self displayWorld.

	Processor activeIsland scheduler: prior.
	^result! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 7/14/2003 11:01'!
scheduler
	"Answer this world's script scheduler"
	^scheduler! !

!CWorldPlayer methodsFor: 'private' stamp: 'ar 7/14/2003 11:01'!
setScheduler: aScriptScheduler
	scheduler := aScriptScheduler.! !

!CWorldPlayer methodsFor: 'world' stamp: 'ar 1/22/2005 22:37'!
topWindow
	^nil! !

!CWorldPlayer methodsFor: 'world' stamp: 'ar 1/22/2005 22:37'!
topWindow: aWindow
	^nil! !

!CWorldPlayer methodsFor: 'accessing' stamp: 'ar 5/26/2005 19:49'!
windowManager
	^nil! !

!FileDirectory methodsFor: '*Tweak-Basic' stamp: 'ar 3/13/2004 02:16'!
directories
	^self directoryNames collect:[:each| self directoryNamed: each].! !

!FileDirectory methodsFor: '*Tweak-Basic' stamp: 'ar 9/3/2010 16:20'!
directoryContentsFor: fullPath do: aBlock
	"Return a collection of directory entries for the files and directories in the directory with the given path. See primLookupEntryIn:index: for further details."
	"FileDirectory default directoryContentsFor: ''"

	| index done entryArray path |
	index := 1.
	done := false.
	path := 	fullPath asVmPathName.
	[done] whileFalse: [
		entryArray := self primLookupEntryIn: path index: index.
		#badDirectoryPath = entryArray ifTrue: [
			^(InvalidDirectoryError pathName: pathName) signal].
		entryArray == nil
			ifTrue: [done := true]
			ifFalse: [aBlock value: (DirectoryEntry fromArray: entryArray) convertFromSystemName].
		index := index + 1].! !

!FileDirectory methodsFor: '*Tweak-Basic' stamp: 'ar 1/31/2005 11:18'!
displayText
	pathName isEmptyOrNil ifTrue:[^'[root]'].
	^pathName asString asText! !

!FileDirectory methodsFor: '*Tweak-Basic' stamp: 'ar 3/12/2004 19:28'!
hasSubDirectories
	"Answer whether this directory has any subdirectories"
	self directoryContentsFor: pathName do:[:entry|
		entry isDirectory ifTrue:[^true].
	].
	^false! !

!FileDirectory methodsFor: '*Tweak-Basic' stamp: 'ar 2/28/2005 16:56'!
makeIconsInforming: bar
	| ic |
	bar value: pathName asString.
	self entries do:[:each|
		ic := (CDirectoryEntry new)
			directory: self;
			entry: each;
		icon.
		ic displayOn: Display at: 0@0 rule: Form over.
	].
	self directoryNames do:[:each|
		(self directoryNamed: each) makeIconsInforming: bar.
	].! !

!SmalltalkImage methodsFor: '*Tweak-Basic' stamp: 'ar 6/21/2010 11:17'!
tweakLowSpaceWatcher
	"Wait until the low space semaphore is signalled, then take appropriate actions."

	| free |
	self garbageCollectMost <= self lowSpaceThreshold ifTrue: [
		self garbageCollect <= self lowSpaceThreshold ifTrue: [
			"free space must be above threshold before starting low space watcher"
			^ self beep]].

	LowSpaceSemaphore := Semaphore new.
	self primLowSpaceSemaphore: LowSpaceSemaphore.
	self primSignalAtBytesLeft: self lowSpaceThreshold.  "enable low space interrupts"

	LowSpaceSemaphore wait.  "wait for a low space condition..."

	self primSignalAtBytesLeft: 0.  "disable low space interrupts"
	self primLowSpaceSemaphore: nil.
	LowSpaceProcess := nil.
	"Note: user now unprotected until the low space watcher is re-installed"

	self memoryHogs isEmpty ifFalse: [
		free := self bytesLeft.
		self memoryHogs do: [ :hog | hog freeSomeSpace ].
		self bytesLeft > free ifTrue: [ ^ self installLowSpaceWatcher ]].

	CWorldPlayer interruptName: 'Space is low'.! !

ProtoObject subclass: #CCollectionMatcher
	instanceVariableNames: 'target'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Collections'!

!CCollectionMatcher commentStamp: 'ar 3/27/2004 17:21' prior: 0!
I am an abstract base class for collection matchers such as the Any or the All matcher.!

CCollectionMatcher subclass: #CAllMatcher
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Collections'!

!CAllMatcher commentStamp: 'ar 3/27/2004 17:20' prior: 0!
I am a simple helper which allows to use the "all" message with collections, e.g.,
	#('foo.zip' 'bar.zip') all endsWith: '.zip'
!

!CAllMatcher methodsFor: 'private' stamp: 'ar 3/14/2004 16:24'!
pvtMatchSelector: aSelector arguments: arguments
	^target allSatisfy:[:each| each perform: aSelector withArguments: arguments]! !

CCollectionMatcher subclass: #CAnyMatcher
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Collections'!

!CAnyMatcher commentStamp: 'ar 3/27/2004 17:20' prior: 0!
I am a simple helper which allows to use the "all" message with collections, e.g.,
	#('foo.zip' 'bar.zip') any endsWith: '.zip'
!

!CAnyMatcher methodsFor: 'private' stamp: 'ar 3/14/2004 16:24'!
pvtMatchSelector: aSelector arguments: arguments
	^target anySatisfy:[:each| each perform: aSelector withArguments: arguments]! !

!CCollectionMatcher class methodsFor: 'instance creation' stamp: 'ar 3/14/2004 16:31'!
on: aCollection
	^(self new) privateSetCollectionTarget: aCollection! !

!CCollectionMatcher methodsFor: 'private' stamp: 'ar 3/14/2004 16:23'!
doesNotUnderstand: aMessage
	^self pvtMatchSelector: aMessage selector arguments: aMessage arguments.! !

!CCollectionMatcher methodsFor: 'private' stamp: 'ar 3/14/2004 16:30'!
privateSetCollectionTarget: aCollection
	target := aCollection.! !

!CCollectionMatcher methodsFor: 'private' stamp: 'ar 3/14/2004 16:23'!
pvtMatchSelector: aSelector arguments: arguments
	^false! !

!InputSensor methodsFor: '*Tweak-Basic' stamp: 'bf 4/4/2005 17:31'!
tweakInterruptWatcher
	"Wait for user interrupts and open a notifier on the active process when one occurs."

	[true] whileTrue: [
		InterruptSemaphore wait.
		Preferences cmdDotEnabled ifTrue: [
			Display deferUpdates: false.
			Smalltalk at: #SoundPlayer ifPresent: [:theClass | theClass shutDown].
			CWorldPlayer interruptName: 'User Interrupt']]
! !

!MethodReference methodsFor: '*Tweak-Basic' stamp: 'ar 3/14/2004 15:38'!
displayText
	^stringVersion! !

!String methodsFor: '*Tweak-Basic' stamp: 'ar 12/6/2005 15:14'!
displayText
	^Text string: self attributes: #()! !

TestCase subclass: #TweakTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Tests'!

!TweakTests methodsFor: 'utilities' stamp: 'bf 3/30/2004 12:22'!
allClassesWithFieldsDo: aBlock

	CProtoObject withAllSubclassesDo: [:class |
		class classFields do: [:field | aBlock value: class value: field]]! !

!TweakTests methodsFor: 'utilities' stamp: 'bf 3/30/2004 12:22'!
allFieldGetterMethodsDo: aBlock

	self allClassesWithFieldsDo: [:class :field |
		field toGet ifNotNilDo: [:selector |
			class withAllSubclassesDo: [:subclass |
				(subclass compiledMethodAt: selector ifAbsent: [nil])
					ifNotNilDo: [:method |
						aBlock
							value: subclass
							value: selector
							value: method]]]]! !

!TweakTests methodsFor: 'utilities' stamp: 'bf 3/30/2004 12:22'!
allFieldSetterMethodsDo: aBlock

	self allClassesWithFieldsDo: [:class :field |
		field toSet ifNotNilDo: [:selector |
			class withAllSubclassesDo: [:subclass |
				(subclass compiledMethodAt: selector ifAbsent: [nil])
					ifNotNilDo: [:method |
						aBlock
							value: subclass
							value: selector
							value: method]]]]! !

!TweakTests methodsFor: 'field tests' stamp: 'abc 4/16/2007 00:37'!
testFieldGettersBewareOfAnnotation
	"Field getters should have a bewareOf annotation"

	| missingAnnotations |
true ifTrue:[^nil].
	missingAnnotations := Array streamContents: [:s |
		self allFieldGetterMethodsDo: [:cls :sel :meth |
			(meth propertyValueAt: #bewareOf) ifNil: [
				s nextPut: (MethodReference new
					setStandardClass: cls 
					methodSymbol: sel)]]].

	self assert: missingAnnotations isEmpty.
	
	"SystemNavigation default
		browseMessageList: missingAnnotations
		name: 'Field getters without bewareOf Annotation'
		autoSelect: nil"! !

!TweakTests methodsFor: 'field tests' stamp: 'ar 4/15/2007 21:11'!
testFieldSettersBewareOfAnnotation
	"Field setters should not have a bewareOf annotation"

	| spuriousAnnotations |
	spuriousAnnotations := Array streamContents: [:s |
		self allFieldSetterMethodsDo: [:cls :sel :meth |
			(meth propertyValueAt: #bewareOf) ifNotNil: [
				s nextPut: (MethodReference new
					setStandardClass: cls 
					methodSymbol: sel)]]].

	self assert: spuriousAnnotations isEmpty.
	
	"SystemNavigation default
		browseMessageList: spuriousAnnotations
		name: 'Field setters with bewareOf Annotation'
		autoSelect: nil"! !

!TweakTests methodsFor: 'field tests' stamp: 'abc 4/16/2007 00:37'!
testFieldSettersReturn
	"Field setters must return their argument. This is because they can be used in an assignment like
		var := field := aValue
	which is compiled as  
		var := self field: aValue
	so #field: must return its argument. 
	
	We try to detect some obvious mistakes like [^self] here."
	
	| analyzer dubiousReturns |
true ifTrue:[^nil].
	analyzer := TweakTestMethodAnalyzer new.
	dubiousReturns := Array streamContents: [:s |
		self allFieldSetterMethodsDo: [:cls :sel :meth |
			(analyzer dubiousReturnInFieldSetter: meth) ifTrue: [
				s nextPut: (MethodReference new
					setStandardClass: cls 
					methodSymbol: sel)]]].

	self assert: dubiousReturns isEmpty.
	
	"SystemNavigation default
		browseMessageList: dubiousReturns
		name: 'Field setters with dubious return'
		autoSelect: nil"! !

!UndefinedObject methodsFor: '*Tweak-Basic' stamp: 'ar 5/13/2005 14:00'!
asCostumeFor: aPlayer
	"backstop"
	^nil! !

!UndefinedObject methodsFor: '*Tweak-Basic' stamp: 'ar 5/14/2005 16:12'!
asPlayerFor: aCostume
	"backstop"
	^nil! !

!UndefinedObject methodsFor: '*Tweak-Basic' stamp: 'ar 5/13/2005 11:21'!
displayText
	^''! !

!CTableLayout methodsFor: '*Tweak-Basic' stamp: 'tk 7/14/2004 18:14'!
fieldNamed: fieldName
	"pretend I have a field of this name.  I do have the getter and setter methods!!"

	| fldDfn |
	fldDfn := CFieldDefinition new.	"for my properties' instance vars"
	fieldName last == $: 
		ifTrue: ["allow a setter like borderWidth:, to retrieve field borderWidth"
			fldDfn toGet: (fieldName copyFrom: 1 to: (fieldName indexOf: $:)-1) asSymbol.
			fldDfn toSet: fieldName]
		ifFalse: [
			fldDfn toGet: fieldName.
			fldDfn toSet: (fieldName, ':') asSymbol].
	fldDfn name: fldDfn toGet.

	^(CFieldReference new)
		target: self;
		fieldDefinition: fldDfn;
		yourself! !

!Cursor class methodsFor: '*Tweak-Basic' stamp: 'ar 3/13/2004 21:30'!
drag
	^CCursorLibrary dragCursor! !

!Cursor class methodsFor: '*Tweak-Basic' stamp: 'ar 3/13/2004 21:30'!
text
	"Cursor text showWhile: [Sensor waitButton]"
 	^CCursorLibrary textCursor! !

!Collection methodsFor: '*Tweak-Basic' stamp: 'ar 3/14/2004 16:25'!
all
	"Answer a new all matcher"
	^CAllMatcher on: self! !

!Collection methodsFor: '*Tweak-Basic' stamp: 'ar 3/14/2004 16:25'!
any
	"Answer a new any matcher"
	^CAnyMatcher on: self! !

InstructionClient subclass: #TweakTestMethodAnalyzer
	instanceVariableNames: 'dubiousReturn'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Tests'!

!TweakTestMethodAnalyzer methodsFor: 'initialize-release' stamp: 'bf 3/30/2004 14:09'!
dubiousReturnInFieldSetter: method
	"A field setter must return its argument. Try to find cases where this might be violated."

	| scanner end |
	method isQuick ifTrue: [^true].
	scanner := InstructionStream on: method.
	end := method endPC.
	dubiousReturn := false.
	[scanner pc <= end]
		whileTrue: [
			scanner interpretNextInstructionFor: self.
			dubiousReturn ifTrue: [^true]].
	^false! !

!TweakTestMethodAnalyzer methodsFor: 'instruction decoding' stamp: 'bf 3/30/2004 13:53'!
methodReturnConstant: value 
	"Return Constant bytecode."

	dubiousReturn := true! !

!TweakTestMethodAnalyzer methodsFor: 'instruction decoding' stamp: 'bf 3/30/2004 13:53'!
methodReturnReceiver
	"Return Self bytecode."

	dubiousReturn := true! !

!TweakTestMethodAnalyzer methodsFor: 'instruction decoding' stamp: 'bf 3/30/2004 14:12'!
methodReturnTop
	"Return Top Of Stack bytecode."

	"Returning the argument (temp 0) is always okay. All other cases would require further analysis. We don't know, so do not raise a flag here ..."
! !

CPlayer subclass: #CHandPlayer
	instanceVariableNames: 'mouseFocus keyboardFocus cursorFocus mouseOverHandler lastEvent lastEventBuffer dragContents pressedKeys balloon'
	classVariableNames: 'EventStats ShowEvents'
	poolDictionaries: 'EventSensorConstants'
	category: 'Tweak-Basic-Worlds'!

CHandPlayer defineFields: '<?xml  version="1.0" ?>
<fields>
	<field instanceVariable="true" toGet="mouseFocus" toSet="mouseFocus:">mouseFocus</field>
	<field instanceVariable="true" toGet="keyboardFocus" toSet="keyboardFocus:">keyboardFocus</field>
	<field instanceVariable="true" toGet="cursorFocus" toSet="cursorFocus:">cursorFocus</field>
	<field instanceVariable="true" toGet="mouseOverHandler">mouseOverHandler</field>
	<field instanceVariable="true" toGet="lastEvent">lastEvent</field>
	<field instanceVariable="true">lastEventBuffer</field>
	<field instanceVariable="true" toGet="dragContents" toSet="dragContents:">dragContents</field>
	<field instanceVariable="true" toGet="pressedKeys">pressedKeys</field>
	<field instanceVariable="true" toGet="balloon">balloon</field>
	<field toGet="balloonTarget" toSet="balloonTarget:" changeEvent="balloonTargetChanged">balloonTarget</field>
	<field toGet="showHardwareCursor" toSet="showHardwareCursor:" changeEvent="showHardwareCursorChanged">showHardwareCursor</field>
</fields>'!

!CHandPlayer class methodsFor: 'initialize' stamp: 'ar 11/12/2004 11:03'!
initialize
	"CHandPlayer intialize"
	self initializePreferences.! !

!CHandPlayer class methodsFor: 'initialize' stamp: 'ar 11/12/2004 11:04'!
initializePreferences
	Preferences addPreference: #autoKeyboardFocus categories: #('Tweak') default: false balloonHelp: 'When true, the keyboard focus will move along with the mouse.'.! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:32'!
anyButtonPressed
	^lastEvent anyButtonPressed! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:32'!
anyModifierKeyPressed
	^lastEvent anyModifierKeyPressed! !

!CHandPlayer methodsFor: 'grabbing/dropping' stamp: 'ar 6/26/2005 18:49'!
attach: aPlayer
	"Position the center of the given costume under this hand, then grab it. This method is used to grab far away or newly created costumes."
	| m pos |
	m := aPlayer asPlayer.
	m openIn: self world.
	pos := self position.
	m container ifNotNil:[pos := m container worldToLocal: pos].
	m position: pos.
	self releaseMouseFocus.
	m startScript: #onDrag.
! !

!CHandPlayer methodsFor: 'accessing' stamp: 'ar 11/12/2004 11:05'!
autoKeyboardFocus
	^Preferences valueOfFlag: #autoKeyboardFocus! !

!CHandPlayer methodsFor: 'accessing' stamp: 'ar 2/7/2005 20:45'!
balloon
	^balloon! !

!CHandPlayer methodsFor: 'balloon help' stamp: 'ar 2/6/2005 21:40'!
balloonTarget
	"Answer the balloonTarget of the receiver"
	<bewareOf: #balloonTargetChanged>
	^self propertyValueAt: #balloonTarget! !

!CHandPlayer methodsFor: 'balloon help' stamp: 'ar 2/6/2005 21:40'!
balloonTarget: aValue
	"Modify the receiver's balloonTarget"
	^self propertyValueAt: #balloonTarget put: aValue with: #balloonTargetChanged! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:33'!
blueButtonChanged
	^lastEvent blueButtonChanged! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:33'!
blueButtonPressed
	^lastEvent blueButtonPressed! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:33'!
commandKeyPressed
	^lastEvent commandKeyPressed! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:33'!
controlKeyPressed
	^lastEvent controlKeyPressed! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'ar 7/8/2003 19:00'!
cursorFocus
	"answer the current cursor focus"
	^cursorFocus! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'ar 7/8/2003 19:00'!
cursorFocus: aPlayer
	"indicate the current cursor focus"
	cursorFocus := aPlayer.! !

!CHandPlayer methodsFor: 'private' stamp: 'ar 3/26/2004 02:03'!
disableGenieFocus: aMorph! !

!CHandPlayer methodsFor: 'private events' stamp: 'bgf 12/14/2009 22:02'!
dispatchEvent: anEvent
	"@@@@: Fix me. Need to notify listeners (based on event type etc) since they have gone away. Event needs to be signaled at least two times (for generic event listeners and specific event listeners) - alternatively just signal event and leave it to client to figure out... :@@@@"
	| evt ofs |
	self container ifNil:[^self].
	evt := anEvent.
	evt setHand: self.
	EventStats ifNil:[EventStats := IdentityDictionary new].
	EventStats at: #count put: (EventStats at: #count ifAbsent:[0]) + 1.
	EventStats at: evt type put: (EventStats at: evt type ifAbsent:[0]) + 1.
	evt isMouseOver ifTrue:[^self sendMouseEvent: evt].
	ShowEvents == true ifTrue:[
		ofs := (self container hands indexOf: self) - 1 * 60.
		evt printString displayAt: (0@ofs) + (evt isKeyboard ifTrue:[0@30] ifFalse:[0@0]).
		self keyboardFocus printString displayAt: (0@ofs)+(0@45).
	].
	self signal: evt type with: evt.
	evt isKeyboard ifTrue:[
		lastEvent := evt.
		^self sendKeyboardEvent: evt].
	evt isDropEvent ifTrue:[^self sendEvent: evt focus: nil].
	evt isMouse ifTrue:[
		lastEvent := evt.
		self signal: #updateMouseOver.
	].

	evt isMove ifTrue:[
		self position: evt position.
		self sendMouseEvent: evt.
	] ifFalse:[
		"Issue a synthetic move event (if appropriate), 
		 or at least shift the hand position,  if not at the position of the event"
		(evt position = self position) ifFalse:[
			evt type = #mouseUp ifTrue:[
				self moveToEvent: evt			"Synthetic move, to complete the drag-cycle"
			] ifFalse: [
				self position: (evt position)		"Don't mouseMove, since it would be out of phase."
			]].
		self sendMouseEvent: evt.
	].
	ShowEvents == true ifTrue:[
		self mouseFocus printString displayAt: (0@ofs) + (0@15).
		self playersUnder printString displayAt: (0@60) + (0@15).
	].! !

!CHandPlayer methodsFor: 'drag drop' stamp: 'ar 3/18/2003 12:38'!
dragContents
	"Answer the dragContents of this hand"
	^dragContents! !

!CHandPlayer methodsFor: 'drag drop' stamp: 'ar 3/23/2004 23:14'!
dragContents: anObject
	"Set the dragContents of this hand."
	"Note: The event dispatch below essentially models that we move out into the sky, pick up the object and move back to where we were. This is needed in cases where objects may react to dragEnter/dragLeave and should be notified about the change in state."
	| evt |
	dragContents == anObject ifTrue:[^self].
	evt := lastEvent asMouseLeave.
	self playersUnder do:[:each| 
		each locked ifFalse:[
			each costume handleEvent: (evt transformedBy: each costume globalToLocal)]].
	dragContents := anObject.
	evt := lastEvent asMouseEnter.
	self playersUnder reverseDo:[:each| 
		each locked ifFalse:[
			each costume handleEvent: (evt transformedBy: each costume globalToLocal)]].
	^dragContents! !

!CHandPlayer methodsFor: 'grabbing/dropping' stamp: 'ar 8/28/2005 20:43'!
drop: aPlayer
	"Drop this costume"
	| wasLocked event |
	wasLocked := aPlayer locked.
	aPlayer locked: true.
	event := CDropEvent new setPosition: self position contents: aPlayer hand: self.
	self sendEvent: event focus: nil.
	aPlayer locked: wasLocked.! !

!CHandPlayer methodsFor: 'mouse' stamp: 'ar 3/26/2004 01:50'!
forEachAndEveryMouseMoveDo: aBlock
	"A simple way of querying for *all* mouse move events where it matters"
	| moveEvent |
	self forEach: #mouseMove do:[:scriptEvent|
		"capture events in the order they were received"
		scriptEvent withDroppedEvents reverseDo:[:possiblyDropped|
			"get the move event out of the argument of the script event"
			moveEvent := possiblyDropped arguments first.
			moveEvent trail do:[:aPoint| aBlock value: aPoint].
			"and pass the last one along explicitly"
			aBlock value: moveEvent position.
		].
	] until: #mouseUp.! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 8/28/2005 20:43'!
generateDropFilesEvent: evtBuf
	"Generate the appropriate mouse event for the given raw event buffer"
	"Note: This is still in an experimental phase and will need more work"
	| pos buttons modifiers stamp numFiles dragType |
	stamp := (evtBuf at: 2).
	stamp = 0 ifTrue:[stamp := Time millisecondClockValue].
	dragType := evtBuf at: 3.
	pos := (evtBuf at: 4) @ (evtBuf at: 5).
	buttons := 0.
	modifiers := (evtBuf at: 6).
	buttons := buttons bitOr: (modifiers bitShift: 3).
	numFiles := (evtBuf at: 7).
	dragType = 4 ifTrue:[
		"e.g., drop"
		self container borderWidth: 0.
		^CDropFilesEvent new 
			setPosition: pos 
			contents: numFiles 
			hand: self.
	].
	"the others are currently not handled by costumes themselves"
	dragType = 1 ifTrue:[
		"experimental drag enter"
		self container borderWidth: 4; borderColor: self container color negated.
	].
	dragType = 2 ifTrue:[
		"experimental drag move"
	].
	dragType = 3 ifTrue:[
		"experimental drag leave"
		self container borderWidth: 0.
	].
	^nil! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 7/12/2006 17:27'!
generateKeyboardEvent: evtBuf
	"Generate the appropriate mouse event for the given raw event buffer"
	| buttons modifiers type keyValue pressType stamp |
	stamp := (evtBuf at: 2).
	stamp = 0 ifTrue:[stamp := Time millisecondClockValue].
	keyValue := evtBuf at: 3.
	pressType := evtBuf at: 4.
	pressType = EventKeyDown ifTrue:[type := #keyDown].
	pressType = EventKeyUp ifTrue:[type := #keyUp].
	pressType = EventKeyChar ifTrue:[type := #keystroke].
	modifiers := evtBuf at: 5.
	"Keyboard events do not report mouse buttons"
	buttons := (modifiers bitShift: 3) bitOr: (lastEvent buttons bitAnd: 7).
	^lastEvent := CKeyboardEvent new
		setType: type
		buttons: buttons
		position: self position
		keyValue: keyValue
		hand: self
		stamp: stamp! !

!CHandPlayer methodsFor: 'private events' stamp: 'jdm 4/1/2008 15:47'!
generateMouseEvent: evtBuf
	"Generate the appropriate mouse event for the given raw event buffer"
	| pos buttons modifiers type trail stamp oldButtons evtChanged nrClicks |
Transcript show: 'CHandPlayer generateMouseEvent';cr.
	(evtBuf at: 1) = (lastEventBuffer at: 1) ifTrue:[
		"Workaround for Mac VM bug, *always* generating 3 events on clicks"
		evtChanged := false.
		3 to: evtBuf size do:[:i| (lastEventBuffer at: i) = (evtBuf at: i) ifFalse:[evtChanged := true]].
		evtChanged ifFalse:[^nil]].
	stamp := (evtBuf at: 2).
	stamp = 0 ifTrue:[stamp := Time millisecondClockValue].
	pos := (evtBuf at: 3) @ (evtBuf at: 4).
	buttons := (evtBuf at: 5).
	modifiers := (evtBuf at: 6).
	oldButtons := lastEventBuffer at: 5.
	(buttons = oldButtons) ifTrue:[
		type := #mouseMove.
	] ifFalse:[
		(buttons bitClear: oldButtons) > 0
			ifTrue:[type := #mouseDown]
			ifFalse:[type := #mouseUp].
	].
	buttons := buttons bitOr: (modifiers bitShift: 3).
	oldButtons := (lastEventBuffer at: 5) bitOr: ((lastEventBuffer at: 6) bitShift: 3).
	lastEventBuffer := evtBuf.
	type == #mouseMove ifTrue:[
		trail := self mouseTrailFrom: evtBuf.
		^lastEvent := CMouseMoveEvent new 
			setType: type 
			startPoint: trail first
			endPoint: trail last 
			trail: trail
			buttons: buttons
			hand: self
			stamp: stamp.
	].

	nrClicks := evtBuf at: 7.
	nrClicks ifNil: [ nrClicks := 0 ].
	(nrClicks > 1)
		ifTrue: [ Transcript show: 'CHandPlayer>>generateMouseEvent: nrClicks=', nrClicks; cr. ].

	^lastEvent := CMouseButtonEvent new 
		setType: type 
		position: pos
		which: (oldButtons bitXor: buttons)
		buttons: buttons 
		nrClicks: nrClicks
		hand: self
		stamp: stamp! !

!CHandPlayer methodsFor: 'grabbing/dropping' stamp: 'ar 3/18/2003 12:26'!
grab: aPlayer
	"Grab the given costume (i.e., add it to this hand and remove it from its current container) without changing its position. This is used to pick up a costume under the hand's current position, versus attach: which is used to pick up a costume that may not be near this hand."
	| grabbed |
	grabbed := aPlayer asPlayer.
	grabbed ifNil:[^self].
	^self grab: grabbed from: grabbed container! !

!CHandPlayer methodsFor: 'grabbing/dropping' stamp: 'ar 6/26/2005 18:49'!
grab: aPlayer from: formerOwner
	"Grab the given costume (i.e., add it to this hand and remove it from its current owner) without changing its position. This is used to pick up a costume under the hand's current position, versus attach: which is used to pick up a costume that may not be near this hand."
	| grabbed |
	formerOwner ifNil:[^self attach: aPlayer].
	grabbed := aPlayer asPlayer.
	self container add: grabbed.
	grabbed transformBy: formerOwner costume localToGlobal.
	self mouseFocus: grabbed.
	grabbed startScript: #onDrag.! !

!CHandPlayer methodsFor: 'drag drop' stamp: 'ar 3/18/2003 12:39'!
hasDragContents
	"Answer whether this hand is in a drag and drop transaction"
	^dragContents notNil! !

!CHandPlayer methodsFor: 'initialize' stamp: 'ar 8/28/2005 20:43'!
initialize
	| script |
	super initialize.
	self name: 'Hand'.
	keyboardFocus := nil.
	pressedKeys := Set new.
	mouseFocus := nil.
	mouseOverHandler := nil.
	balloon := CBalloonHelp new.
	lastEvent := CMouseEvent new setType: #mouseMove position: 0@0 buttons: 0 hand: self.
	lastEventBuffer := {1. 0. 0. 0. 0. 0. nil. nil}.
	"Bootstrap the cursor"
	script := self startScript: #mouseCursor: withArguments:{Cursor normal}.
	script hand: self.! !

!CHandPlayer methodsFor: 'private' stamp: 'ar 3/18/2003 12:33'!
interrupted
	"Something went wrong - we're about to bring up a debugger. 
	Release some stuff that could be problematic."
	self releaseAllFoci. "or else debugger might not handle clicks"
	dragContents := nil.! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 8/3/2004 17:20'!
isKeyPressed: aKey
	^self pressedKeys includes: aKey! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'ar 8/17/2003 13:56'!
isMouseOver: aPlayer
	"Return true if the receiver is over aPlayer"
	^self mouseOverHandler mouseOverList anySatisfy:[:over| aPlayer == over].! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'ar 4/22/2005 12:16'!
isOver: aPlayer
	"Return true if the receiver is over aPlayer"
	^self mouseOverHandler mouseOverList anySatisfy:[:over| aPlayer asPrimCostume == over asPrimCostume].! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'ar 3/18/2003 12:28'!
isOverPlayer: aPlayer
	"Return true if the receiver is over aCostume"
	^self mouseOverHandler mouseOverList anySatisfy:[:over| aPlayer == over].! !

!CHandPlayer methodsFor: 'initialize' stamp: 'ar 3/18/2003 12:20'!
isSpecialSystemScript: scriptDescriptor
	"Some scripts need to be run for debugging and cannot stopped for debugging"
	^(super isSpecialSystemScript: scriptDescriptor) or:[scriptDescriptor == #processEvents]! !

!CHandPlayer methodsFor: 'keyboard' stamp: 'ar 3/18/2003 12:37'!
keyboardFocus				 
	^ keyboardFocus! !

!CHandPlayer methodsFor: 'keyboard' stamp: 'ar 9/8/2004 16:23'!
keyboardFocus: focusHolder
	"Make the given costume the new keyboard focus, canceling the previous keyboard focus if any. If the argument is nil, the current keyboard focus is cancelled."
	| oldFocus newFocus |
	oldFocus := self keyboardFocus.
	newFocus := focusHolder ifNotNil:[focusHolder asPrimCostume].
	oldFocus == newFocus ifTrue:[^self].
	self setKeyboardFocus: newFocus.
	oldFocus ifNotNil:[oldFocus keyboardFocusLeave].
	newFocus ifNotNil:[newFocus keyboardFocusEnter].
	self signalChanged: #keyboardFocusChanged from: oldFocus to: newFocus.! !

!CHandPlayer methodsFor: 'accessing' stamp: 'ar 3/18/2003 12:22'!
lastEvent
	^ lastEvent! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:33'!
macOptionKeyPressed
	^lastEvent macOptionKeyPressed! !

!CHandPlayer methodsFor: 'accessing' stamp: 'ar 6/28/2005 18:17'!
mouseCursor: aCursor
	| oldPosition newCursor oldCursor |
	newCursor := aCursor ifNil:[Cursor normal].
	oldCursor := self mouseCursor ifNil:[Cursor normal].
	oldCursor == newCursor ifTrue:[^aCursor].
	self invalidate.
	oldPosition := self position.
	super mouseCursor: newCursor.
	self extent: newCursor extent.
	self pivot: newCursor offset negated.
	self graphic: mouseCursor asForm.
	self position: oldPosition.
	self asPrimCostume invalidationSuppressed: (newCursor isCursor and:[showHardwareCursor]).
	self invalidate.
	^newCursor! !

!CHandPlayer methodsFor: 'mouse' stamp: 'ar 3/18/2003 12:38'!
mouseFocus
	^mouseFocus! !

!CHandPlayer methodsFor: 'mouse' stamp: 'ar 9/8/2004 16:23'!
mouseFocus: focusHolder
	"Make the given costume the new mouse focus, canceling the previous mouse focus if any. If the argument is nil, the current mouse focus is cancelled."
	| newFocus oldFocus |
	newFocus := focusHolder ifNotNil:[focusHolder asPrimCostume].
	oldFocus := self mouseFocus.
	self setMouseFocus: newFocus.
	self signalChanged: #mouseFocusChanged from: oldFocus to: newFocus.
	lastEvent anyButtonPressed ifTrue:[oldFocus signal: #buttonFocusLost].
	^focusHolder! !

!CHandPlayer methodsFor: 'accessing' stamp: 'ar 3/18/2003 12:23'!
mouseOverHandler
	^mouseOverHandler ifNil:[mouseOverHandler := CMouseOverHandler new].! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 3/18/2003 12:30'!
mouseTrailFrom: currentBuf
	"Current event, a mouse event buffer, is about to be processed.  If there are other similar mouse events queued up, then drop them from the queue, and report the positions inbetween."
	| nextEvent trail |
	trail := WriteStream on: (Array new: 1).
	trail nextPut: ((currentBuf at: 3) @ (currentBuf at: 4)).
	[(nextEvent := Sensor peekEvent) == nil] whileFalse:[
		(nextEvent at: 1) = (currentBuf at: 1) ifFalse:
			[^ trail contents  "different event type"].
		(nextEvent at: 5) = (currentBuf at: 5) ifFalse:
			[^ trail contents  "buttons changed"].
		(nextEvent at: 6) = (currentBuf at: 6) ifFalse:
			[^ trail contents  "modifiers changed"].
		"nextEvent is similar.  Remove it from the queue, and check the next."
		nextEvent := Sensor nextEvent.
		trail nextPut: ((nextEvent at: 3) @ (nextEvent at: 4))
	].
	^trail contents! !

!CHandPlayer methodsFor: 'private events' stamp: 'das 12/7/2009 16:41'!
moveToEvent: anEvent
	"Issue a mouse move event to make the receiver appear at the given position"
	self dispatchEvent: (lastEvent := CMouseMoveEvent new
		setType: #mouseMove 
		startPoint: self position 
		endPoint: anEvent position 
		trail: (Array with: self position with: anEvent position)
		buttons: (lastEvent ifNotNil:[lastEvent buttons] ifNil:[anEvent buttons])
		hand: self
		stamp: anEvent timeStamp)! !

!CHandPlayer methodsFor: 'testing' stamp: 'ar 8/23/2007 11:43'!
needsSoftwareCursor
	"Answer true if this hand needs to be rendered using the software cursor.
	This is a quick utility function for clients who need to know."
	(self showHardwareCursor and:[self mouseCursor isCursor]) ifTrue:[^false]. "hardware"
	"Possibly more conditions to come here"
	^true! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 3/18/2003 12:30'!
noticeMouseOver: aPlayer event: anEvent
	mouseOverHandler ifNil:[^self].
	mouseOverHandler noticeMouseOver: aPlayer event: anEvent.! !

!CHandPlayer methodsFor: 'balloon help' stamp: 'bf 7/25/2005 16:31'!
onBalloonTargetChanged
	<on: balloonTargetChanged>
	balloonTarget ifNil:[
		balloon ifNotNil:[balloon close].
		balloon := nil.
	] ifNotNil:[
		Preferences balloonHelpEnabled
			ifFalse: [^self].
		balloon ifNil:[
			[	balloonTarget ifNil: [^self].
				(self waitUntilAnyOf: {balloonTarget. #mouseMove. self. #balloonTargetChanged}
					timeout: Preferences balloonHelpDelayTime) == nil
			] whileFalse.
			balloon := CBalloonHelp new].
		balloon label: ((balloonTarget balloonHelp asString copy
				withNoLineLongerThan: Preferences maxBalloonHelpLineLength)
					 asText addAttribute: TextAlignment centered).
		balloon target: balloonTarget.
		balloon openIn: self world.
	].
! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'ar 6/28/2005 17:51'!
onUpdateMouseCursor
	"update the current mouse cursor"
	| aCursor |
	cursorFocus ifNotNil:[^self mouseCursor: cursorFocus mouseCursor].
	mouseFocus ifNil:[
		self playersUnder do:[:aPlayer|
			aCursor := aPlayer asPlayer mouseCursor.
			aCursor ifNotNil:[^self mouseCursor: aCursor].
		].
		^self mouseCursor: nil
	].
	"if we have a mouse focus then only children of the current mouse focus can hold the cursor or one of the current containers of the mouse cursor"
	self playersUnder do:[:aPlayer|
		(aPlayer == mouseFocus or:[aPlayer isIn: mouseFocus]) ifTrue:[
			aCursor := aPlayer asPlayer mouseCursor.
			aCursor ifNotNil:[^self mouseCursor: aCursor].
		].
	].
	mouseFocus withAllContainersDo:[:each|
		aCursor := each asPlayer mouseCursor.
		aCursor ifNotNil:[^self mouseCursor: aCursor].
	].
	^self mouseCursor: nil! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'ar 9/23/2003 01:05'!
onUpdateMouseOver
	"update the mouse over state"
	<on: updateMouseOver>
	self mouseOverHandler processMouseOver: lastEvent.
	self onUpdateMouseCursor.! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'ar 3/18/2003 12:28'!
playersUnder
	"Return the costumes currently under the hand"
	^self mouseOverHandler mouseOverList! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'tk 7/14/2004 11:24'!
playersUnder: topPlayer
	"Return the players currently under the hand starting with topPlayer"
	| aList |
	aList := self mouseOverHandler mouseOverList.
	^aList copyFrom: 1 to: (aList indexOf: topPlayer asPrimCostume)! !

!CHandPlayer methodsFor: 'accessing' stamp: 'ar 8/3/2004 17:19'!
pressedKeys
	^pressedKeys ifNil:[pressedKeys := Set new].! !

!CHandPlayer methodsFor: 'private events' stamp: 'jdm 4/1/2008 15:40'!
processEvents
	"Process user input events from the local input devices."
	| evt evtBuf type script |
	[(evtBuf := Sensor nextEvent) == nil] whileFalse:[
		evt := nil. "for unknown event types"
		type := evtBuf at: 1.
		(type = EventTypeMouse) 
			ifTrue:[evt := self generateMouseEvent: evtBuf].
		(type = EventTypeKeyboard) 
			ifTrue:[evt := self generateKeyboardEvent: evtBuf].
		(type = EventTypeDragDropFiles)
			ifTrue:[evt := self generateDropFilesEvent: evtBuf].
		"All other events are ignored"
		evt == nil ifTrue: [^ self].
		"Finally, handle it"
		script := self startScript: #dispatchEvent: withArguments: (Array with: evt).
		script hand: self.
		script waitUntil: #done.
		"For better user feedback, return immediately after a mouse event has been processed."
		evt isMouse ifTrue: [^self].
	].
	"note: if we come here we didn't have any mouse events"
	self signal: #updateMouseOver.! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 1/29/2005 11:07'!
processEventsFrom: aQueue
	"Process events from the given queue"
	| script event |
	[event := aQueue nextOrNil.
	event == nil] whileFalse:[
		script := self startScript: #dispatchEvent: withArguments:{event clone}.
		script hand: self.
		script waitUntil: #done.
		"For better user feedback, return immediately after a mouse event has been processed."
	].
	self signal: #updateMouseOver! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 2/18/2006 19:25'!
processEventsLoop
	"Process user input events from the local input devices."
	[true] whileTrue:[
		self startScript: #processEvents.
		self waitTick.
	].! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:33'!
redButtonChanged
	^lastEvent redButtonChanged! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:33'!
redButtonPressed
	^lastEvent redButtonPressed! !

!CHandPlayer methodsFor: 'private' stamp: 'ar 3/18/2003 12:34'!
releaseAllFoci
	mouseFocus := nil.
	keyboardFocus := nil.! !

!CHandPlayer methodsFor: 'balloon help' stamp: 'ar 2/6/2005 21:56'!
releaseBalloonTarget: aPlayer
	balloonTarget == aPlayer ifTrue:[balloonTarget := nil].! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'ar 7/8/2003 19:00'!
releaseCursorFocus
	"release the current cursor focus unconditionally"
	cursorFocus := nil.! !

!CHandPlayer methodsFor: 'mouse over' stamp: 'ar 7/8/2003 19:00'!
releaseCursorFocus: aPlayer
	"release the current cursor focus"
	aPlayer == cursorFocus ifTrue:[self releaseCursorFocus].! !

!CHandPlayer methodsFor: 'keyboard' stamp: 'ar 3/18/2003 12:37'!
releaseKeyboardFocus
	"Release the current keyboard focus unconditionally"
	self keyboardFocus: nil.! !

!CHandPlayer methodsFor: 'keyboard' stamp: 'ar 1/25/2007 15:00'!
releaseKeyboardFocus: aPlayer
	"If the given costume had the keyboard focus before, release it"
	self keyboardFocus asPlayer == aPlayer asPlayer ifTrue:[self releaseKeyboardFocus].! !

!CHandPlayer methodsFor: 'mouse' stamp: 'ar 3/18/2003 12:38'!
releaseMouseFocus
	"Release the current mouse focus unconditionally."
	self mouseFocus: nil.! !

!CHandPlayer methodsFor: 'mouse' stamp: 'tk 7/14/2004 11:24'!
releaseMouseFocus: aCostume
	"If the given costume had the mouse focus before, release it"
	self mouseFocus == aCostume asPrimCostume ifTrue:[self releaseMouseFocus].! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 3/18/2003 12:30'!
sendEvent: anEvent focus: focusHolder
	"Send the event to the costume currently holding the focus, or if none to the owner of the hand."
	^self sendEvent: anEvent focus: focusHolder clear:[nil]! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 5/22/2005 18:04'!
sendEvent: anEvent focus: focusHolder clear: aBlock
	"Send the event to the costume currently holding the focus, or if none to the container of the hand."
	focusHolder ifNotNil:[^self sendFocusEvent: anEvent to: focusHolder clear: aBlock].
	^self container dispatchEvent: anEvent.! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 5/22/2005 18:04'!
sendFocusEvent: anEvent to: focusHolder clear: aBlock
	"Send the event to the costume currently holding the focus"
	| cc |
	cc := self container asPrimCostume.
	(focusHolder rejectsEvent: anEvent) ifTrue:[^aBlock value].
	(focusHolder == cc or:[focusHolder isIn: cc]) ifFalse:[^aBlock value].
	^focusHolder 
		handleFocusEvent: (anEvent transformedBy: focusHolder globalToLocal)! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 8/3/2004 17:21'!
sendKeyboardEvent: anEvent
	"Send the event to the costume currently holding the focus, or if none to the owner of the hand."
	| keyValue focus context |
	keyValue := anEvent keyValue.
	focus := self keyboardFocus.
	anEvent isKeyDown ifTrue:[
		self pressedKeys add: keyValue.
		context := focus ifNotNil:[focus world].
		context ifNotNil:[context costume pressedKeys add: keyValue].
	].
	anEvent isKeyUp ifTrue:[
		self pressedKeys remove: keyValue ifAbsent:[].
		context := focus ifNotNil:[focus world].
		context ifNotNil:[context costume pressedKeys remove: keyValue ifAbsent:[]].
	].
	^self sendEvent: anEvent focus: self keyboardFocus clear:[self keyboardFocus: nil]! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 3/18/2003 12:31'!
sendListenEvent: anEvent to: listenerGroup
	"Send the event to the given group of listeners"
	listenerGroup ifNil:[^self].
	listenerGroup do:[:listener| 
		listener ifNotNil:[listener handleListenEvent: anEvent copy]].! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 2/17/2004 19:16'!
sendMouseEvent: anEvent
	"Send the event to the costume currently holding the focus, or if none to the owner of the hand."
	^self sendEvent: anEvent focus: self mouseFocus clear:[self mouseFocus: nil].! !

!CHandPlayer methodsFor: 'private events' stamp: 'ar 5/22/2005 18:04'!
sendMouseOverEvent: anEvent
	"Note: Sending mouse overs is slightly different since we want to determine over state for all costumes within the focus but not for those on top (which will get leave/enter events only *after* the current mouse transition is completed)"
	| focus |
	focus := self mouseFocus.
	focus ifNil:[^self container dispatchEvent: anEvent].
	focus container ifNil:[^focus dispatchEvent: anEvent].
	"Note: dispatchEvent: expects parent frame not child so transform into parent local frame"
	^focus costume dispatchEvent: (anEvent transformedBy: focus container asPrimCostume globalToLocal)! !

!CHandPlayer methodsFor: 'private' stamp: 'ar 3/18/2003 12:34'!
setKeyboardFocus: newFocus
	keyboardFocus := newFocus! !

!CHandPlayer methodsFor: 'private' stamp: 'ar 3/18/2003 12:34'!
setMouseFocus: newFocus
	mouseFocus := newFocus.! !

!CHandPlayer methodsFor: 'initialize' stamp: 'ar 6/28/2005 17:59'!
setupCostume
	super setupCostume.
	costume mouseCursor := Cursor blank.
	self color: Color transparent.
	self graphicEnabled: true.! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:33'!
shiftPressed
	^lastEvent shiftPressed! !

!CHandPlayer methodsFor: 'accessing' stamp: 'ar 6/28/2005 18:17'!
showHardwareCursor
	"Answer the showHardwareCursor of the receiver"
	<bewareOf: #showHardwareCursorChanged>
	^self propertyValueAt: #showHardwareCursor ifAbsent:[false]! !

!CHandPlayer methodsFor: 'accessing' stamp: 'ar 6/28/2005 18:16'!
showHardwareCursor: aValue
	"Modify the receiver's showHardwareCursor"
	^self propertyValueAt: #showHardwareCursor put: aValue with: #showHardwareCursorChanged! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:33'!
yellowButtonChanged
	^lastEvent yellowButtonChanged! !

!CHandPlayer methodsFor: 'input state' stamp: 'ar 3/18/2003 12:33'!
yellowButtonPressed
	^lastEvent yellowButtonPressed! !

Object subclass: #CChangeRecord
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Undo'!

CChangeRecord subclass: #CPropertyChange
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Undo'!

!CPropertyChange methodsFor: 'as yet unclassified' stamp: 'tk 6/23/2004 10:39'!
printOn: strm

	super printOn: strm.  strm nextPut: $(.
	value printOn: strm.  strm nextPut: $); space.! !

!CPropertyChange methodsFor: 'accessing' stamp: 'ar 2/10/2004 13:57'!
value
	^value! !

!CPropertyChange methodsFor: 'accessing' stamp: 'ar 2/10/2004 13:57'!
value: aValue
	value := aValue! !

CChangeRecord subclass: #CScriptChange
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Undo'!

!CScriptChange methodsFor: 'accessing' stamp: 'ar 2/14/2004 22:53'!
value
	^value! !

!CScriptChange methodsFor: 'accessing' stamp: 'ar 2/14/2004 22:54'!
value: aString
	value := aString! !

Object subclass: #CCursorLibrary
	instanceVariableNames: ''
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Tweak-Basic-Support'!

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:27'!
bottomLeftCursor
	"CCursorLibrary bottomLeftCursor show"
	^self topRightCursor! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:26'!
bottomRightCursor
	"CCursorLibrary bottomRightCursor show"
	^self topLeftCursor! !

!CCursorLibrary class methodsFor: 'accessing' stamp: 'ar 3/13/2004 21:26'!
default
	^Default ifNil:[Default := IdentityDictionary new]! !

!CCursorLibrary class methodsFor: 'accessing' stamp: 'ar 3/13/2004 21:15'!
default: aDictionary
	Default := aDictionary.! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:27'!
diagonalCursor
	"CCursorLibrary diagonalCursor show"
	^Default at: #DiagonalCursor ifAbsentPut:[
		(Cursor extent: 16@16 fromArray: #(
			2r0000000000000000
			2r0000000000000000
			2r0000000000000000
			2r0000111000111000
			2r0000110000011000
			2r0000101000101000
			2r0000000101000000
			2r0000000010000000
			2r0000000101000000
			2r0000101000101000
			2r0000110000011000
			2r0000111000111000
			2r0000000000000000
			2r0000000000000000
			2r0000000000000000
			2r0000000000000000)
		offset: -7@-7) withMask
	].
! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'bgf 2/11/2007 11:31'!
downLeftCursor
	"CCursorLibrary downLeftCursor show"
	^Default at: #DownLeftCursor ifAbsentPut:[
		(Cursor
			extent: 16@16
			fromArray: #(
		2r0000001100000000
		2r0000001100000000
		2r0000011000000000
		2r0000011000000000
		2r0000110000000000
		2r0000110000000000
		2r1001100000000000
		2r1111100000000000
		2r1111100000000000
		2r1111111000000000
		2r1111110000000000
		2r1111100000000000
		2r1111000000000000
		2r1110000000000000
		2r1100000000000000
		2r1000000000000000)
			offset: 0@-15).
	].

	
	
! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:27'!
dragCursor
	"CCursorLibrary dragCursor show"
	^Default at: #DragCursor ifAbsentPut:[
		(Cursor extent: 16@16 fromArray: #(
			2r00000000000000000
			2r00000000100000000
			2r00000001110000000
			2r00000011111000000
			2r00000000100000000
			2r00001000100010000
			2r00011000100011000
			2r00111111111111100
			2r00011000100011000
			2r00001000100010000
			2r00000000100000000
			2r00000011111000000
			2r00000001110000000
			2r00000000100000000
			2r00000000000000000
			2r00000000000000000 )
		offset: -7@-7 ) withMask
	].
! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:36'!
fistCursor
	"CCursorLibrary fistCursor show"
 	^Default at:	#FistCursor ifAbsentPut:[
		((CursorWithMask
		extent: 16@16
		depth: 1
		fromArray: #( 0 0 0 0 0 229638144 306970624 269090816 134348800 402784256 537001984 537133056 268697600 134742016 67633152 67633152)
		offset: -5@-9) setMaskForm: (Form
		extent: 16@16
		depth: 1
		fromArray: #( 0 0 0 0 0 229638144 536608768 536739840 268304384 536739840 1073610752 1073479680 536608768 267911168 133693440 133693440)
		offset: 0@0))
	].! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:36'!
handCursor
	"CCursorLibrary handCursor show"
 	^Default at:	#HandCursor ifAbsentPut:[
		((CursorWithMask
		extent: 16@16
		depth: 1
		fromArray: #( 25165824 443547648 642252800 642383872 307036160 306774016 1745420288 2550202368 2281832448 1073872896 537001984 537133056 268697600 134742016 67633152 67633152)
		offset: -5@-9) setMaskForm: (Form
		extent: 16@16
		depth: 1
		fromArray: #( 25165824 468713472 1073217536 1073348608 536805376 536805376 1878982656 4294901760 4294836224 2147352576 1073610752 1073479680 536608768 267911168 133693440 133693440)
		offset: 0@0))
	].! !

!CCursorLibrary class methodsFor: 'initialize' stamp: 'ar 3/13/2004 21:26'!
initialize
	"CCursorLibrary initialize"
	Default := IdentityDictionary new.! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 2/12/2007 15:43'!
leftArrowCursor
	"CCursorLibrary leftArrowCursor show"
	^Default at: #LeftArrowCursor ifAbsentPut:[
		(Cursor
			extent: 16@16
			fromArray: #(
		2r0000000000000000
		2r0000000000000000
		2r0000000000000000
		2r0000000000001000
		2r0000000001111000
		2r0000001111110000
		2r0001111111110000
		2r0111111111111111
		2r0001111111110000
		2r0000001111110000
		2r0000000001111000
		2r0000000000001000
		2r0000000000000000
		2r0000000000000000
		2r0000000000000000
		2r0000000000000000)
			offset: -1@-7).
	].

	
	
! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:27'!
leftRightCursor
	"CCursorLibrary leftRightCursor show"
	^Default at: #LeftRightCursor ifAbsentPut:[
		(Cursor extent: 16@16 fromArray: #(
			2r0000000000000000
			2r0000000000000000
			2r0000000000000000
			2r0000000000000000
			2r0000010000010000
			2r0000110000011000
			2r0001110000011100
			2r0011111111111110
			2r0001110000011100
			2r0000110000011000
			2r0000010000010000
			2r0000000000000000
			2r0000000000000000
			2r0000000000000000
			2r0000000000000000
			2r0000000000000000 )
		offset: -7@-7 ) withMask
	].! !

!CCursorLibrary class methodsFor: 'accessing' stamp: 'ar 4/27/2004 15:10'!
resizeForEdge: aSymbol
	aSymbol == #top ifTrue:[^self topBottomCursor].
	aSymbol == #bottom ifTrue:[^self topBottomCursor].
	aSymbol == #topCenter ifTrue:[^self topBottomCursor].
	aSymbol == #bottomCenter ifTrue:[^self topBottomCursor].

	aSymbol == #left ifTrue:[^self leftRightCursor].
	aSymbol == #right ifTrue:[^self leftRightCursor].
	aSymbol == #leftCenter ifTrue:[^self leftRightCursor].
	aSymbol == #rightCenter ifTrue:[^self leftRightCursor].

	aSymbol == #topLeft ifTrue:[^self topLeftCursor].
	aSymbol == #bottomRight ifTrue:[^self topLeftCursor].

	aSymbol == #topRight ifTrue:[^self topRightCursor].
	aSymbol == #bottomLeft ifTrue:[^self topRightCursor].

	^self dragCursor! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:28'!
textCursor
	"CCursorLibrary textCursor show"
 	^Default at:	#TextCursor ifAbsentPut:[
		(Cursor extent: 16@16 fromArray: #(
				2r0000000000000000
				2r0000011011000000
				2r0000000100000000
				2r0000000100000000
				2r0000000100000000
				2r0000000100000000
				2r0000000100000000
				2r0000000100000000
				2r0000000100000000
				2r0000000100000000
				2r0000000100000000
				2r0000000100000000
				2r0000000100000000
				2r0000000100000000
				2r0000011011000000
				2r0000000000000000)
		offset: -7@-7) withMask
	].! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:28'!
topBottomCursor
	"CCursorLibrary topBottomCursor show"
	^Default at: #TopBottomCursor ifAbsentPut:[
		(Cursor extent: 16@16 fromArray: #(
			2r0000000000000000
			2r0000000010000000
			2r0000000111000000
			2r0000001111100000
			2r0000011111110000
			2r0000000010000000
			2r0000000010000000
			2r0000000010000000
			2r0000000010000000
			2r0000000010000000
			2r0000011111110000
			2r0000001111100000
			2r0000000111000000
			2r0000000010000000
			2r0000000000000000
			2r0000000000000000 )
		offset: -7@-7 ) withMask
	].

! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:28'!
topLeftCursor
	"CCursorLibrary topLeftCursor show"
	^Default at: #TopLeftCursor ifAbsentPut:[
		(Cursor extent: 16@16 fromArray: #(
			2r0000000000000000
			2r0000000000000000
			2r0001111100000000
			2r0001111000000000
			2r0001110000000000
			2r0001101000000000
			2r0001000100000000
			2r0000000010000000
			2r0000000001000100
			2r0000000000101100
			2r0000000000011100
			2r0000000000111100
			2r0000000001111100
			2r0000000000000000
			2r0000000000000000
			2r0000000000000000)
		offset: -7@-7) withMask
	].
! !

!CCursorLibrary class methodsFor: 'cursors' stamp: 'ar 3/13/2004 21:28'!
topRightCursor
	"CCursorLibrary topRightCursor show"
	^Default at: #TopRightCursor ifAbsentPut:[
		(Cursor extent: 16@16 fromArray: #(
			2r0000000000000000
			2r0000000000000000
			2r0000000001111100
			2r0000000000111100
			2r0000000000011100
			2r0000000000101100
			2r0000000001000100
			2r0000000010000000
			2r0001000100000000
			2r0001101000000000
			2r0001110000000000
			2r0001111000000000
			2r0001111100000000
			2r0000000000000000
			2r0000000000000000
			2r0000000000000000)
		offset: -7@-7) withMask.
	].

! !

Object subclass: #CDeepCopier
	instanceVariableNames: 'references newGuys tweakScripts uniClasses newUniClasses forbiddenIsland'
	classVariableNames: 'NextVariableCheckTime'
	poolDictionaries: ''
	category: 'Tweak-Basic-Support'!

!CDeepCopier commentStamp: '<historical>' prior: 0!
DeepCopier does a veryDeepCopy.  

It is a complete tree copy using a dictionary.  Any object that is in the tree twice is only copied once.  All references to the object in the copy of the tree will point to the new copy.  See Object|veryDeepCopy which calls (self veryDeepCopyWith: aDeepCopier).

references = he dictionary of objects that have been seen, holds the correspondance (uniClass -> new uniClass).
newGuys = Set of values in references (the new objects)
forbiddenIsland = resolve all references to this Island.  Return a tree of objects with no references to this island.

When a tree of CPlayers points at a CPlayer outside of itself, that CPlayer should not be copied.  Use our own kind of weak pointers for the 'potentially outside' CPlayers.   Default is that any new class will have all of its fields deeply copied.  If a field needs to be weakly copied, define veryDeepInner: and veryDeepFixupWith:.
     veryDeepInner: has the loop that actually copies the fields.  If a class defines its own copy of veryDeepInner: (to leave some fields out), then veryDeepFixupWith: will be called on that object at the end.  veryDeepInner: can compute an alternate object to put in a field.  (Object veryDeepCopyWith: discovers which superclasses did not define veryDeepInner:, and very deeply copies the variables defined in those classes).
	To decide if a class needs veryDeepInner: and veryDeepFixupWith:, ask this about an instance:  If I duplicate this object, does that mean that I also want to make duplicates of the things it holds onto?  If yes, (i.e. a Paragraph does want a new copy of its Text) then do nothing.  If no, (i.e. an undo command does not want to copy the objects it acts upon), then define veryDeepInner: and veryDeepFixupWith:.
	
Here is an analysis for the specific case of a CPlayer being held by another morph.  
Does field X contain a morph (or a Player whose costume is a morph)?  If not, no action needed.
Is the morph in field X already a submorph of the object?  Is it down lower in the submorph tree?
	If so, no action needed.
Could the morph in field X every appear on the screen (be a submorph of some other morph)?
	If not, no action needed.
	If it could, you must write the methods veryDeepFixupWith:   and   veryDeepInner:, and in them, refrain from sending veryDeepCopyWith: to the contents of field X.

newUniClasses =   true in the normal case.  Every duplicated Player gets a new class.  When false, all duplicates will be siblings (sister instances) of existing players.

newGuys =  IdentitySet of objects that we created.  All objects in the value side of references.  Need to check here to make sure that a new CPlayer is not being duplicated for a second time!!
tweakScripts = IdentityDictionary from a new object to an array of scripts the original object had.  Used when a Tweak script is 'non triggered'.  These have to be copied separately. 


----- Things Ted is still considering -----
Rule:  If a morph stores a uniClass class (Player 57) as an object in a field, the new uniClass will not be stored there.   Each uniClass instance does have a new class created for it.  (fix this by putting the old class in references and allow lookup?  Wrong if encounter it before seeing an instance?)

Rule: If object A has object C in a field, and A says (^ C) for the copy, but object B has A in a normal field and it gets deepCopied, and A in encountered first, then there will be two copies of C.  (just be aware of it)

Dependents are now fixed up.  Suppose a model has a dependent view.  In the DependentFields dictionary, model -> (view ...).  
	If only the model is copied, no dependents are created (no one knows about the new model).  
	If only the view is copied, it is inserted into DependentFields on the right side.  model -> (view  copiedView ...).  
	If both are copied, the new model has the new view as its dependent.
	If additional things depend on a model that is copied, the caller must add them to its dependents.
!

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'tk 10/4/2001 13:54'!
checkBasicClasses
	"Check that no indexes of instance vars have changed in certain classes.  If you get an error in this method, an implementation of veryDeepCopyWith: needs to be updated.  The idea is to catch a change while it is still in the system of the programmer who made it.  
	DeepCopier new checkVariables	"

	| str str2 objCls morphCls playerCls |
	str := '|veryDeepCopyWith: or veryDeepInner: is out of date.'.
	(objCls := self objInMemory: #Object) ifNotNil: [
		objCls instSize = 0 ifFalse: [self error: 
			'Many implementers of veryDeepCopyWith: are out of date']].
	(morphCls := self objInMemory: #Morph) ifNotNil: [
		morphCls superclass == Object ifFalse: [self error: 'Morph', str].
		(morphCls instVarNames copyFrom: 1 to: 6) = #('bounds' 'owner' 'submorphs' 
				'fullBounds' 'color' 'extension') 
			ifFalse: [self error: 'Morph', str]].	"added ones are OK"

	str2 := 'Player|copyUniClassWith: and DeepCopier|mapUniClasses are out of date'.
	(playerCls := self objInMemory: #Player) ifNotNil: [
		playerCls class instVarNames = #('scripts' 'slotInfo')
			ifFalse: [self error: str2]].
! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'tk 3/7/2001 15:42'!
checkClass: aClass
	| meth |
	"Check that no indexes of instance vars have changed in certain classes.  If you get an error in this method, an implementation of veryDeepCopyWith: needs to be updated.  The idea is to catch a change while it is still in the system of the programmer who made it."

	self checkBasicClasses.	"Unlikely, but important to catch when it does happen."

	"Every class that implements veryDeepInner: must copy all its inst vars.  Danger is that a user will add a new instance variable and forget to copy it.  So check that the last one is mentioned in the copy method."
	(aClass includesSelector: #veryDeepInner:) ifTrue: [ 
		((aClass compiledMethodAt: #veryDeepInner:) writesField: aClass instSize) ifFalse: [
			aClass instSize > 0 ifTrue: [
				self warnIverNotCopiedIn: aClass sel: #veryDeepInner:]]].
	(aClass includesSelector: #veryDeepCopyWith:) ifTrue: [
		meth := aClass compiledMethodAt: #veryDeepCopyWith:.
		(meth size > 20) & (meth literals includes: #veryDeepCopyWith:) not ifTrue: [
			(meth writesField: aClass instSize) ifFalse: [
				self warnIverNotCopiedIn: aClass sel: #veryDeepCopyWith:]]].
! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'dvf 8/23/2003 11:52'!
checkDeep
	"Write exceptions in the Transcript.  Every class that implements veryDeepInner: must copy all its inst vars.  Danger is that a user will add a new instance variable and forget to copy it.  This check is only run by hand once in a while to make sure nothing was forgotten.  
(Please do not remove this method.)
	DeepCopier new checkDeep 	"

	| mm |
	Transcript
		cr;
		show: 'Instance variables shared with the original object when it is copied'.
	(self systemNavigation allClassesImplementing: #veryDeepInner:) do: 
			[:aClass | 
			(mm := aClass instVarNames size) > 0 
				ifTrue: 
					[aClass instSize - mm + 1 to: aClass instSize
						do: 
							[:index | 
							((aClass compiledMethodAt: #veryDeepInner:) writesField: index) 
								ifFalse: 
									[Transcript
										cr;
										show: aClass name;
										space;
										show: (aClass allInstVarNames at: index)]]]]! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'dvf 8/23/2003 11:53'!
checkVariables
	"Check that no indexes of instance vars have changed in certain classes.  If you get an error in this method, an implementation of veryDeepCopyWith: needs to be updated.  The idea is to catch a change while it is still in the system of the programmer who made it.  
	DeepCopier new checkVariables	"

	| meth |
	self checkBasicClasses.

	"Every class that implements veryDeepInner: must copy all its inst vars.  Danger is that a user will add a new instance variable and forget to copy it.  So check that the last one is mentioned in the copy method."
	(self systemNavigation allClassesImplementing: #veryDeepInner:) do: 
			[:aClass | 
			((aClass compiledMethodAt: #veryDeepInner:) writesField: aClass instSize) 
				ifFalse: 
					[aClass instSize > 0 
						ifTrue: [self warnIverNotCopiedIn: aClass sel: #veryDeepInner:]]].
	(self systemNavigation allClassesImplementing: #veryDeepCopyWith:) do: 
			[:aClass | 
			meth := aClass compiledMethodAt: #veryDeepCopyWith:.
			meth size > 20 & (meth literals includes: #veryDeepCopyWith:) not 
				ifTrue: 
					[(meth writesField: aClass instSize) 
						ifFalse: [self warnIverNotCopiedIn: aClass sel: #veryDeepCopyWith:]]]! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'tk 3/11/2003 13:56'!
fixDependents
	"They are not used much, but need to be right"

	| newDep newModel |
	DependentsFields associationsDo: [:pair |
		pair value do: [:dep | 
			newDep := references at: dep ifAbsent: [nil].
			newDep ifNotNil: [
				newModel := references at: pair key ifAbsent: [pair key].
				newModel addDependent: newDep]]].
! !

!CDeepCopier methodsFor: 'accessing' stamp: 'tk 6/2/2005 10:08'!
forbiddenIsland
	^ forbiddenIsland! !

!CDeepCopier methodsFor: 'accessing' stamp: 'tk 6/2/2005 10:09'!
forbiddenIsland: anIsland

	forbiddenIsland := anIsland! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'hg 11/23/1999 13:36'!
initialize

	self initialize: 4096.
! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'tk 2/7/2005 17:19'!
initialize: size

	references := IdentityDictionary new: size.
	newGuys := IdentitySet new: size.
	tweakScripts := IdentityDictionary new: size//8.
	uniClasses := IdentityDictionary new.	"UniClass -> new UniClass"
	"self isItTimeToCheckVariables ifTrue: [self checkVariables]."
		"no more checking at runtime"
	newUniClasses := true.! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'tk 11/24/1999 17:53'!
intervalForChecks
	"set delay interval for checking for new instance variables to 10 minutes. hg 11/23/1999"

	^600
! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'tk 11/25/1999 14:37'!
isItTimeToCheckVariables

	| now isIt |
	NextVariableCheckTime ifNil: [
		NextVariableCheckTime := Time totalSeconds.
		^ true].
	now := Time totalSeconds.
	isIt := NextVariableCheckTime < now.
	isIt ifTrue: ["update time for next check"
		NextVariableCheckTime := now + self intervalForChecks].
	^isIt
! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'tk 2/16/2005 12:02'!
mapIfCan: anObject
	"see if we have duplicated him.  map this reference to an old object."

	^ references at: anObject ifAbsent: [anObject]! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'ar 6/21/2010 11:19'!
mapUniClasses
	"For new Uniclasses, map their class vars to the new objects.  And their additional class instance vars.  (scripts slotInfo) and cross references like (player321)."
	"Players also refer to each other using associations in the References dictionary.  Search the methods of our Players for those.  Make new entries in References and point to them."
| pp oldPlayer newKey newAssoc oldSelList newSelList |

	newUniClasses ifFalse: [^ self].	"All will be siblings.  uniClasses is empty"
"Uniclasses use class vars to hold onto siblings who are referred to in code"
pp := (Smalltalk at: #Player ifAbsent:[^self]) class superclass instSize.
uniClasses do: [:playersClass | "values = new ones"
	playersClass classPool associationsDo: [:assoc |
		assoc value: (assoc value veryDeepCopyWith: self)].
	playersClass scripts: (playersClass privateScripts veryDeepCopyWith: self).	"pp+1"
	"(pp+2) slotInfo was deepCopied in copyUniClass and that's all it needs"
	pp+3 to: playersClass class instSize do: [:ii | 
		playersClass instVarAt: ii put: 
			((playersClass instVarAt: ii) veryDeepCopyWith: self)].
	].

"Make new entries in References and point to them."
References keys "copy" do: [:playerName |
	oldPlayer := References at: playerName.
	(references includesKey: oldPlayer) ifTrue: [
		newKey := (references at: oldPlayer) "new player" uniqueNameForReference.
		"now installed in References"
		(references at: oldPlayer) renameTo: newKey]].
uniClasses "values" do: [:newClass |
	oldSelList := OrderedCollection new.   newSelList := OrderedCollection new.
	newClass selectorsDo: [:sel | 
		(newClass compiledMethodAt: sel)	 literals do: [:assoc |
			assoc isVariableBinding ifTrue: [
				(References associationAt: assoc key ifAbsent: [nil]) == assoc ifTrue: [
					newKey := (references at: assoc value ifAbsent: [assoc value]) 
									externalName asSymbol.
					(assoc key ~= newKey) & (References includesKey: newKey) ifTrue: [
						newAssoc := References associationAt: newKey.
						newClass methodDictionary at: sel put: 
							(newClass compiledMethodAt: sel) clone.	"were sharing it"
						(newClass compiledMethodAt: sel)
							literalAt: ((newClass compiledMethodAt: sel) literals indexOf: assoc)
							put: newAssoc.
						(oldSelList includes: assoc key) ifFalse: [
							oldSelList add: assoc key.  newSelList add: newKey]]]]]].
	oldSelList with: newSelList do: [:old :new |
			newClass replaceSilently: old to: new]].	"This is text replacement and can be wrong"! !

!CDeepCopier methodsFor: 'accessing' stamp: 'tk 1/24/2005 13:21'!
newGuys
	"objects that are values in the references dictionary.  The newly created objects."
	^ newGuys! !

!CDeepCopier methodsFor: 'accessing' stamp: 'tk 3/11/2003 14:13'!
newUniClasses
	"If false, all new Players are merely siblings of the old players"

	^ newUniClasses! !

!CDeepCopier methodsFor: 'accessing' stamp: 'tk 3/4/2003 19:44'!
newUniClasses: newVal
	"If false, all new players are merely siblings of the old players"

	newUniClasses := newVal! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'tk 3/7/2001 15:29'!
objInMemory: ClassSymbol
	| cls |
	"Test if this global is in memory and return it if so."

	cls := Smalltalk at: ClassSymbol ifAbsent: [^ nil].
	^ cls isInMemory ifTrue: [cls] ifFalse: [nil].! !

!CDeepCopier methodsFor: 'accessing' stamp: 'tk 8/20/1998 22:13'!
references
	^ references! !

!CDeepCopier methodsFor: 'accessing' stamp: 'tk 2/7/2005 17:18'!
tweakScripts
	^ tweakScripts! !

!CDeepCopier methodsFor: 'accessing' stamp: 'tk 8/19/1998 15:48'!
uniClasses
	^uniClasses! !

!CDeepCopier methodsFor: 'like fullCopy' stamp: 'tk 3/7/2001 15:46'!
warnIverNotCopiedIn: aClass sel: sel
	"Warn the user to update veryDeepCopyWith: or veryDeepInner:"

	self inform: ('An instance variable was added to to class ', aClass name, ',\and it is not copied in the method ', sel, '.\Please rewrite it to handle all instance variables.\See DeepCopier class comment.') withCRs.
	Browser openMessageBrowserForClass: aClass selector: sel editString: nil.
! !

Object subclass: #CEllipseShape
	instanceVariableNames: 'myPrimCostume'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Costumes'!

!CEllipseShape methodsFor: 'drawing' stamp: 'tk 2/3/2005 17:28'!
computeBaseBounds
	"always use existing rectangle bounds"

	^ myPrimCostume bounds! !

!CEllipseShape methodsFor: 'drawing' stamp: 'tk 2/2/2005 10:14'!
drawOn: aCanvas in: drawingBounds
	"Draw an oval in my player's bounds."

	"Will be clipped.  Might want to go faster by using drawingBounds"
	aCanvas fillOval: (myPrimCostume localBounds translateBy: (1@1)) 
			fillStyle: myPrimCostume fill.
	aCanvas isShadowDrawing ifTrue: [^ self].
	(myPrimCostume border isNoBorder not and: [myPrimCostume border width > 0]) ifTrue: [
		myPrimCostume border frameOval: myPrimCostume localBounds 
							on: aCanvas].
! !

!CEllipseShape methodsFor: 'drawing' stamp: 'tk 2/2/2005 09:53'!
initialize
	"do nothing"! !

!CEllipseShape methodsFor: 'accessing' stamp: 'tk 2/2/2005 09:37'!
myPrimCostume
	"Answer the value of myPrimCostume"

	^ myPrimCostume! !

!CEllipseShape methodsFor: 'accessing' stamp: 'tk 2/2/2005 09:37'!
myPrimCostume: anObject
	"Set the value of myPrimCostume"

	myPrimCostume := anObject! !

Object subclass: #CGenericEvent
	instanceVariableNames: 'hand timeStamp'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Events'!

!CGenericEvent commentStamp: '<historical>' prior: 0!
I am the superclass for tweak ui events.!

CGenericEvent subclass: #CDropEvent
	instanceVariableNames: 'position contents wasHandled'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Events'!

!CDropEvent commentStamp: '<historical>' prior: 0!
I represent an event that indicates a drop operation.!

!CDropEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
contents
	^contents! !

!CDropEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:54'!
copyHandlerState: anEvent
	"Copy the handler state from anEvent. Used for quickly transferring handler information between transformed events."
	wasHandled := anEvent wasHandled.! !

!CDropEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
cursorPoint
	"For compatibility with mouse events"
	^position! !

!CDropEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:54'!
isDropEvent
	^true! !

!CDropEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
position
	^position! !

!CDropEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:54'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self position printString; space.
	aStream nextPutAll: self type.
	aStream nextPut: $].! !

!CDropEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:54'!
resetHandlerFields
	"Reset anything that is used to cross-communicate between two eventual handlers during event dispatch"
	wasHandled := false.! !

!CDropEvent methodsFor: 'dispatching' stamp: 'ar 8/28/2005 20:23'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	self type == #dropEvent ifTrue:[^anObject handleDrop: self].! !

!CDropEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:06'!
setPosition: pos contents: aPlayer hand: aHand
	position := pos.
	contents := aPlayer.
	hand := aHand.
	wasHandled := false.! !

!CDropEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 20:26'!
transformBy: aTransform
	"Transform the receiver into a local coordinate system."
	position :=  position transformedBy: aTransform.! !

!CDropEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 20:06'!
translateBy: offset
	position := position + offset.! !

!CDropEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 20:06'!
translatedBy: delta
	"add delta to cursorPoint, and return the new event"
	^self shallowCopy translateBy: delta! !

!CDropEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
type
	^#dropEvent! !

!CDropEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
wasHandled
	^wasHandled! !

!CDropEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
wasHandled: aBool
	wasHandled := aBool.! !

CDropEvent subclass: #CDropFilesEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Events'!

!CDropFilesEvent commentStamp: '<historical>' prior: 0!
I represent an event that indicates the drop of a file from the operating system.!

!CDropFilesEvent methodsFor: 'dispatching' stamp: 'ar 8/28/2005 19:54'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	self type == #dropFilesEvent ifTrue:[^anObject handleDropFiles: self].! !

!CDropFilesEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
type
	^#dropFilesEvent! !

!CGenericEvent class methodsFor: 'islands' stamp: 'ar 8/28/2005 19:53'!
howToPassAsArgument
	"Really, this ought to be passByCopy: but some events use handler returns etc."
	^#passByProxy:! !

!CGenericEvent class methodsFor: 'instance creation' stamp: 'ar 8/28/2005 20:23'!
readFrom: aStream
	"Read an event from the given stream."
	| typeString c |
	typeString := String streamContents:
		[:s |   [(c := aStream next) isLetter] whileTrue: [s nextPut: c]].
	typeString = 'mouseMove' ifTrue:[^CMouseMoveEvent type: #mouseMove readFrom: aStream].
	typeString = 'mouseDown' ifTrue:[^CMouseButtonEvent type: #mouseDown readFrom: aStream].
	typeString = 'mouseUp' ifTrue:[^CMouseButtonEvent type: #mouseUp readFrom: aStream].

	typeString = 'keystroke' ifTrue:[^CKeyboardEvent type: #keystroke readFrom: aStream].
	typeString = 'keyDown' ifTrue:[^CKeyboardEvent type: #keyDown readFrom: aStream].
	typeString = 'keyUp' ifTrue:[^CKeyboardEvent type: #keyUp readFrom: aStream].

	typeString = 'mouseOver' ifTrue:[^CMouseEvent type: #mouseOver readFrom: aStream].
	typeString = 'mouseEnter' ifTrue:[^CMouseEvent type: #mouseEnter readFrom: aStream].
	typeString = 'mouseLeave' ifTrue:[^CMouseEvent type: #mouseLeave readFrom: aStream].

	typeString = 'unknown' ifTrue:[^CUnknownEvent type: #unknown readFrom: aStream].

	^nil
! !

!CGenericEvent class methodsFor: 'instance creation' stamp: 'ar 8/28/2005 19:53'!
type: eventType readFrom: aStream
	^self new type: eventType readFrom: aStream! !

!CGenericEvent methodsFor: 'comparing' stamp: 'ar 8/28/2005 20:15'!
= anEvent
	^self type = anEvent type! !

!CGenericEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:53'!
copyHandlerState: anEvent
	"Copy the handler state from anEvent. Used for quickly transferring handler information between transformed events."
! !

!CGenericEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:53'!
cursorPoint
	"Backward compatibility. Use #position instead"
	^ self position! !

!CGenericEvent methodsFor: 'nil' stamp: 'jdm 11/20/2007 22:40'!
dump
	"Dump the event to the Transcript window for debugging"
	self printOn: Transcript.
	Transcript show: ' ts=', self timeStamp;
				show: ' pos=', self cursorPoint;
				cr.
! !

!CGenericEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 20:12'!
hand
	"Return the hand that generated the event"
	^hand! !

!CGenericEvent methodsFor: 'comparing' stamp: 'ar 8/28/2005 19:53'!
hash
	^self type hash! !

!CGenericEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:53'!
isDropEvent
	^false! !

!CGenericEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:53'!
isKeyboard
	^false! !

!CGenericEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:53'!
isKeystroke
	^false! !

!CGenericEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:53'!
isMouse
	^false! !

!CGenericEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:53'!
isMouseOver
	^self type == #mouseOver! !

!CGenericEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:53'!
resetHandlerFields
	"Reset anything that is used to cross-communicate between two eventual handlers during event dispatch"! !

!CGenericEvent methodsFor: 'dispatching' stamp: 'ar 8/28/2005 19:53'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	^anObject handleUnknownEvent: self! !

!CGenericEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:12'!
setHand: aHand
	hand := aHand! !

!CGenericEvent methodsFor: 'private' stamp: 'ar 8/28/2005 19:53'!
setTimeStamp: stamp
	timeStamp := stamp.! !

!CGenericEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:53'!
timeStamp
	"Return the millisecond clock value at which the event was generated"
	^timeStamp ifNil:[timeStamp := Time millisecondClockValue]! !

!CGenericEvent methodsFor: 'transforming' stamp: 'ar 8/29/2005 01:14'!
transformBy: aTransform! !

!CGenericEvent methodsFor: 'transforming' stamp: 'ar 8/29/2005 01:13'!
transformedBy: aTransform
	"Return the receiver transformed by the given transform."
	^self shallowCopy transformBy: aTransform! !

!CGenericEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:53'!
type
	"Return a symbol indicating the type this event."
	^self subclassResponsibility! !

!CGenericEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 20:22'!
type: eventType readFrom: aStream
	"Read an event from the given stream."
! !

!CGenericEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:53'!
wasHandled
	"Return true if this event was handled. May be ignored for some types of events."
	^false! !

!CGenericEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:53'!
wasHandled: aBool
	"Determine if this event was handled. May be ignored for some types of events."! !

CGenericEvent subclass: #CUnknownEvent
	instanceVariableNames: 'type argument'
	classVariableNames: ''
	poolDictionaries: 'EventSensorConstants'
	category: 'Tweak-Basic-Events'!

!CUnknownEvent commentStamp: '<historical>' prior: 0!
An unknown event.!

!CUnknownEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
argument
	^argument! !

!CUnknownEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
argument: arg
	argument := arg! !

!CUnknownEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
position
	^0@0! !

!CUnknownEvent methodsFor: 'private' stamp: 'ar 8/28/2005 19:54'!
setType: evtType argument: arg
	type := evtType.
	argument := arg.! !

!CUnknownEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:12'!
setType: evtType argument: arg hand: evtHand stamp: stamp
	type := evtType.
	argument := arg.
	hand := evtHand.
	timeStamp := stamp.! !

!CUnknownEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:54'!
storeOn: aStream
	aStream nextPutAll: 'unknown'.
	aStream space.
	self timeStamp storeOn: aStream.
	aStream space.
	{type. argument} storeOn: aStream.! !

!CUnknownEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:54'!
type
	^type! !

!CUnknownEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:54'!
type: eventType readFrom: aStream
	| typeAndArg |
	timeStamp := Integer readFrom: aStream.
	aStream skip: 1.
	typeAndArg := Object readFrom: aStream.
	type := typeAndArg first.
	argument := typeAndArg last.! !

CGenericEvent subclass: #CUserInputEvent
	instanceVariableNames: 'type buttons position handler wasHandled'
	classVariableNames: ''
	poolDictionaries: 'EventSensorConstants'
	category: 'Tweak-Basic-Events'!

!CUserInputEvent commentStamp: '<historical>' prior: 0!
The superclass for all mouse/keyboard events in Tweak.!

CUserInputEvent subclass: #CKeyboardEvent
	instanceVariableNames: 'keyValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Events'!

!CKeyboardEvent commentStamp: '<historical>' prior: 0!
I represent a keyboard event.!

!CKeyboardEvent methodsFor: 'comparing' stamp: 'ar 8/28/2005 20:25'!
= anEvent
	super = anEvent ifFalse:[^false].
	buttons = anEvent buttons ifFalse: [^ false].
	keyValue = anEvent keyValue ifFalse: [^ false].
	^ true
! !

!CKeyboardEvent methodsFor: 'comparing' stamp: 'ar 8/28/2005 19:54'!
hash
	^buttons hash + keyValue hash
! !

!CKeyboardEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:54'!
isKeyDown
	^self type == #keyDown! !

!CKeyboardEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:54'!
isKeyUp
	^self type == #keyUp! !

!CKeyboardEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:54'!
isKeyboard
	^true! !

!CKeyboardEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:54'!
isKeystroke
	^self type == #keystroke! !

!CKeyboardEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:54'!
isMouseMove
	^false! !

!CKeyboardEvent methodsFor: 'keyboard' stamp: 'ar 8/28/2005 19:54'!
keyString
	"Answer the string value for this keystroke. This is defined only for keystroke events."

	^ String streamContents: [ :s | self printKeyStringOn: s ]! !

!CKeyboardEvent methodsFor: 'keyboard' stamp: 'ar 8/28/2005 19:54'!
keyValue
	"Answer the ascii value for this keystroke. This is defined only for keystroke events."

	^ keyValue! !

!CKeyboardEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:54'!
printKeyStringOn: aStream
	"Print a readable string representing the receiver on a given stream"

	| kc inBrackets firstBracket keyString |
	kc := self keyCharacter.
	inBrackets := false.
	firstBracket := [ inBrackets ifFalse: [ aStream nextPut: $<. inBrackets := true ]].
	self controlKeyPressed ifTrue: [ 	firstBracket value. aStream nextPutAll: 'Ctrl-' ].
	self commandKeyPressed ifTrue: [ firstBracket value. aStream nextPutAll: 'Cmd-' ].
	(buttons anyMask: 32) ifTrue: [ firstBracket value. aStream nextPutAll: 'Opt-' ].
	(self shiftPressed and: [ keyValue between: 1 and: 31 ])
		ifTrue: [ firstBracket value. aStream nextPutAll: 'Shift-' ].

	(self controlKeyPressed and: [ keyValue <= 26 ])
			ifTrue:
				[aStream nextPut: (keyValue + $a asciiValue - 1) asCharacter]
			ifFalse: 
				[keyString := (kc caseOf: {
					[ Character space ] -> [ ' ' ].
					[ Character tab ] -> [ 'tab' ].
					[ Character cr ] -> [ 'cr' ].
					[ Character lf ] -> [ 'lf' ].
					[ Character enter ] -> [ 'enter' ].

					[ Character backspace ] -> [ 'backspace' ].
					[ Character delete ] -> [ 'delete' ].

					[ Character escape ] -> [ 'escape' ].

					[ Character arrowDown ] -> [ 'down' ].
					[ Character arrowUp ] -> [ 'up' ].
					[ Character arrowLeft ] -> [ 'left' ].
					[ Character arrowRight ] -> [ 'right' ].

					[ Character end ] -> [ 'end' ].
					[ Character home ] -> [ 'home' ].
					[ Character pageDown ] -> [ 'pageDown' ].
					[ Character pageUp ] -> [ 'pageUp' ].

					[ Character euro ] -> [ 'euro' ].
					[ Character insert ] -> [ 'insert' ].

				} otherwise: [ String with: kc ]).
				keyString size > 1 ifTrue: [ firstBracket value ].
				aStream nextPutAll: keyString].

	inBrackets ifTrue: [aStream nextPut: $> ]! !

!CKeyboardEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 20:15'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: type.
	self controlKeyPressed ifTrue: [
		aStream nextPutAll: ' ''^'.
		aStream nextPut: (keyValue + $a asciiValue - 1) asCharacter.
	] ifFalse: [
		aStream nextPutAll: ' '''.
		aStream nextPut: self keyCharacter.
	].
	aStream nextPut: $'.
	aStream space; print: keyValue.
	aStream nextPut: $].! !

!CKeyboardEvent methodsFor: 'dispatching' stamp: 'ar 8/28/2005 19:54'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #keystroke ifTrue:[^anObject handleKeystroke: self].
	type == #keyDown ifTrue:[^anObject handleKeyDown: self].
	type == #keyUp ifTrue:[^anObject handleKeyUp: self].
	^super sentTo: anObject.! !

!CKeyboardEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:13'!
setType: aSymbol buttons: anInteger position: pos keyValue: aValue hand: aHand stamp: stamp
	type := aSymbol.
	buttons := anInteger.
	position := pos.
	keyValue := aValue.
	hand := aHand.
	wasHandled := false.
	timeStamp := stamp.! !

!CKeyboardEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:54'!
storeOn: aStream

	aStream nextPutAll: type.
	aStream space.
	self timeStamp storeOn: aStream.
	aStream space.
	buttons storeOn: aStream.
	aStream space.
	keyValue storeOn: aStream.
! !

!CKeyboardEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:54'!
type: eventType readFrom: aStream
	type := eventType.
	timeStamp := Integer readFrom: aStream.
	aStream skip: 1.
	buttons := Integer readFrom: aStream.
	aStream skip: 1.
	keyValue := Integer readFrom: aStream.! !

CUserInputEvent subclass: #CMouseEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Events'!

!CMouseEvent commentStamp: '<historical>' prior: 0!
The common superclass for all mouse-related events.!

CMouseEvent subclass: #CMouseButtonEvent
	instanceVariableNames: 'whichButton nrClicks'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Events'!

!CMouseButtonEvent commentStamp: '<historical>' prior: 0!
An event indicating that some mouse button has changed.!

!CMouseButtonEvent methodsFor: 'printing' stamp: 'jdm 4/1/2008 11:27'!
nrClicks
	^nrClicks! !

!CMouseButtonEvent methodsFor: 'printing' stamp: 'jdm 4/1/2008 11:27'!
nrClicks: nClicks
	nrClicks := nClicks.! !

!CMouseButtonEvent methodsFor: 'dispatching' stamp: 'ar 8/28/2005 19:55'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseDown ifTrue:[^anObject handleMouseDown: self].
	type == #mouseUp ifTrue:[^anObject handleMouseUp: self].
	^super sentTo: anObject! !

!CMouseButtonEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:13'!
setType: evtType position: evtPos which: button buttons: evtButtons hand: evtHand stamp: stamp
	type := evtType.
	position := evtPos.
	buttons := evtButtons.
	hand := evtHand.
	wasHandled := false.
	whichButton := button.
	timeStamp := stamp.! !

!CMouseButtonEvent methodsFor: 'private' stamp: 'jdm 4/1/2008 11:20'!
setType: evtType position: evtPos which: button buttons: evtButtons nrClicks: nClicks hand: evtHand stamp: stamp
	type := evtType.
	position := evtPos.
	buttons := evtButtons.
	nrClicks := nClicks.
	hand := evtHand.
	wasHandled := false.
	whichButton := button.
	timeStamp := stamp.! !

!CMouseButtonEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:55'!
storeOn: aStream
	super storeOn: aStream.
	aStream space.
	whichButton storeOn: aStream.! !

!CMouseButtonEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:55'!
type: eventType readFrom: aStream
	super type: eventType readFrom: aStream.
	aStream skip: 1.
	whichButton := Integer readFrom: aStream.! !

!CMouseButtonEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 20:15'!
whichButton
	^whichButton! !

!CMouseEvent methodsFor: 'comparing' stamp: 'ar 8/28/2005 20:25'!
= anEvent
	super = anEvent ifFalse:[^false].
	position = anEvent position ifFalse: [^ false].
	buttons = anEvent buttons ifFalse: [^ false].
	^ true
! !

!CMouseEvent methodsFor: 'comparing' stamp: 'ar 8/28/2005 19:55'!
hash
	^ position hash + buttons hash! !

!CMouseEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:55'!
isMouse
	^true! !

!CMouseEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:55'!
isMouseDown
	^self type == #mouseDown! !

!CMouseEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:55'!
isMouseMove
	^self type == #mouseMove! !

!CMouseEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:55'!
isMouseUp
	^self type == #mouseUp! !

!CMouseEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:55'!
isMove
	^false! !

!CMouseEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:55'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self cursorPoint printString; space.
	aStream nextPutAll: type; space.
	aStream nextPutAll: self modifierString.
	aStream nextPutAll: self buttonString.
	aStream nextPutAll: timeStamp printString.
	aStream nextPut: $].! !

!CMouseEvent methodsFor: 'dispatching' stamp: 'ar 8/28/2005 19:55'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseOver ifTrue:[^anObject handleMouseOver: self].
	type == #mouseEnter ifTrue:[^anObject handleMouseEnter: self].
	type == #mouseLeave ifTrue:[^anObject handleMouseLeave: self].
	^super sentTo: anObject.! !

!CMouseEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:55'!
storeOn: aStream

	aStream nextPutAll: type.
	aStream space.
	self timeStamp storeOn: aStream.
	aStream space.
	position x storeOn: aStream.
	aStream space.
	position y storeOn: aStream.
	aStream space.
	buttons storeOn: aStream.! !

!CMouseEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:55'!
type: eventType readFrom: aStream
	| x y |
	type := eventType.
	timeStamp := Integer readFrom: aStream.
	aStream skip: 1.
	x := Integer readFrom: aStream.
	aStream skip: 1.
	y := Integer readFrom: aStream.
	aStream skip: 1.
	buttons := Integer readFrom: aStream.
	position := x@y.
! !

CMouseEvent subclass: #CMouseMoveEvent
	instanceVariableNames: 'startPoint trail'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Events'!

!CMouseMoveEvent commentStamp: '<historical>' prior: 0!
An event indicating that the mouse has moved.!

!CMouseMoveEvent methodsFor: 'comparing' stamp: 'ar 8/28/2005 20:27'!
= anEvent
	super = anEvent ifFalse:[^false].
	position = anEvent position ifFalse: [^ false].
	startPoint = anEvent startPoint ifFalse: [^ false].
	buttons = anEvent buttons ifFalse: [^ false].
	^ true
! !

!CMouseMoveEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:55'!
endPoint
	"Return the point where the movement ended."
	^position! !

!CMouseMoveEvent methodsFor: 'comparing' stamp: 'ar 8/28/2005 19:55'!
hash
	^ position hash + startPoint hash + buttons hash! !

!CMouseMoveEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 19:55'!
isMove
	^true! !

!CMouseMoveEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:55'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self startPoint printString; space.
	aStream nextPutAll: self endPoint printString; space.
	aStream nextPutAll: self type; space.
	aStream nextPutAll: self modifierString.
	aStream nextPutAll: self buttonString.
	aStream nextPutAll: timeStamp printString.
	aStream nextPut: $].! !

!CMouseMoveEvent methodsFor: 'dispatching' stamp: 'ar 8/28/2005 19:55'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseMove ifTrue:[^anObject handleMouseMove: self].
	^super sentTo: anObject.
! !

!CMouseMoveEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:16'!
setTrail: evtTrail
	trail := evtTrail.
! !

!CMouseMoveEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:14'!
setType: evtType startPoint: evtStart endPoint: evtEnd trail: evtTrail buttons: evtButtons hand: evtHand stamp: stamp
	type := evtType.
	startPoint := evtStart.
	position := evtEnd.
	trail := evtTrail.
	buttons := evtButtons.
	hand := evtHand.
	wasHandled := false.
	timeStamp := stamp.! !

!CMouseMoveEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:55'!
startPoint
	"Return the point where the movement started."
	^startPoint! !

!CMouseMoveEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:55'!
storeOn: aStream
	super storeOn: aStream.
	aStream space.
	self startPoint x storeOn: aStream.
	aStream space.
	self startPoint y storeOn: aStream.
	aStream space.
	"trail storeOn: aStream."! !

!CMouseMoveEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:55'!
trail
	"Return any immediate points that have been assembled along the move"
	^trail ifNil:[#()]! !

!CMouseMoveEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 20:27'!
transformBy: aTransform
	"Transform the receiver into a local coordinate system."
	position :=  position transformedBy: aTransform.
	startPoint :=  startPoint transformedBy: aTransform
! !

!CMouseMoveEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 19:55'!
translateBy: delta
	"add delta to cursorPoint, and return the new event"
	position := position + delta.
	startPoint := startPoint + delta.! !

!CMouseMoveEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:55'!
type: eventType readFrom: aStream
	| x y |
	super type: eventType readFrom: aStream.
	aStream skip: 1.
	x := Integer readFrom: aStream.
	aStream skip: 1.
	y := Integer readFrom: aStream.
	startPoint := x@y.! !

!CUserInputEvent methodsFor: 'comparing' stamp: 'ar 8/28/2005 20:20'!
= anEvent
	super = anEvent ifFalse:[^false].
	position = anEvent position ifFalse: [^ false].
	buttons = anEvent buttons ifFalse: [^ false].
	self keyValue = anEvent keyValue ifFalse: [^ false].
	^ true! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/28/2005 20:16'!
anyButtonPressed
	"Answer true if any mouse button is being pressed."
	^ buttons anyMask: 7! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/28/2005 19:56'!
anyModifierKeyPressed
	"ignore, however, the shift keys 'cause that's not REALLY a command key "

	^ self buttons anyMask: 16r70	"cmd | opt | ctrl"! !

!CUserInputEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 20:17'!
asMouseEnter
	^self clone setType: #mouseEnter! !

!CUserInputEvent methodsFor: 'converting' stamp: 'ar 8/28/2005 20:17'!
asMouseLeave
	^self clone setType: #mouseLeave! !

!CUserInputEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 20:17'!
asMouseMove
	"Convert the receiver into a mouse move"
	^CMouseMoveEvent new setType: #mouseMove startPoint: position endPoint: position trail: {position. position} buttons: buttons hand: hand stamp: Time millisecondClockValue.! !

!CUserInputEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 20:17'!
asMouseOver
	"Convert the receiver into a mouse over event"
	^CMouseButtonEvent new setType: #mouseOver position: position buttons: buttons hand: hand! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/29/2005 01:08'!
blueButtonChanged
	"Answer true if the blue mouse button has changed. This is the third mouse button or cmd+click on the Mac."
	^ self whichButton anyMask: 1! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/28/2005 20:18'!
blueButtonPressed
	"Answer true if the blue mouse button is being pressed. This is the third mouse button or cmd+click on the Mac."
	^ buttons anyMask: 1! !

!CUserInputEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:56'!
buttonString
	"Return a string identifying the currently pressed buttons"
	| string |
	string := ''.
	self redButtonPressed ifTrue:[string := string,'red '].
	self yellowButtonPressed ifTrue:[string := string,'yellow '].
	self blueButtonPressed ifTrue:[string := string,'blue '].
	^string! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:56'!
buttons
	"Return the a word encoding the mouse and modifier buttons for this event."

	^ buttons! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/28/2005 19:56'!
commandKeyPressed
	"Answer true if the command key on the keyboard was being held down when this event occurred."

	^ buttons anyMask: 64! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/28/2005 19:56'!
controlKeyPressed
	"Answer true if the control key on the keyboard was being held down when this event occurred."

	^ buttons anyMask: 16! !

!CUserInputEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:56'!
copyHandlerState: anEvent
	"Copy the handler state from anEvent. Used for quickly transferring handler information between transformed events."
	handler := anEvent handler.
	wasHandled := anEvent wasHandled.! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 20:18'!
cursorPoint
	"Answer the location of the cursor's hotspot when this event occured."
	^ position! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:56'!
handler
	^handler! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:56'!
handler: anObject
	handler := anObject! !

!CUserInputEvent methodsFor: 'comparing' stamp: 'ar 8/28/2005 20:18'!
hash
	^ position hash + buttons hash! !

!CUserInputEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 20:18'!
isMouseEnter
	^self type == #mouseEnter! !

!CUserInputEvent methodsFor: 'testing' stamp: 'ar 8/28/2005 20:18'!
isMouseLeave
	^self type == #mouseLeave! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 20:18'!
keyCharacter
	^self keyValue asCharacter! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 20:18'!
keyValue
	^0! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/28/2005 20:24'!
macOptionKeyPressed
	"Answer whether the option key on the Macintosh keyboard was being held down when this event occurred. Macintosh specific."

	Preferences macOptionKeyAllowed ifFalse: [self notifyWithLabel: 'Portability note:
Tests for macOptionKeyPressed are not portable.
Please use #yellowButtonPressed instead!!'].
	^ buttons anyMask: 32! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/28/2005 20:18'!
metaState
	"Answer the meta state for this event"
	^ (self buttons bitAnd: 16r78)	"cmd | opt | ctrl | shift"
		bitShift: -3! !

!CUserInputEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 19:56'!
modifierString
	"Return a string identifying the currently pressed modifiers"
	| string |
	string := ''.
	self commandKeyPressed ifTrue:[string := string,'CMD '].
	self shiftPressed ifTrue:[string := string,'SHIFT '].
	self controlKeyPressed ifTrue:[string := string,'CTRL '].
	^string! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 20:19'!
mouseButtons
	"answer the mouse buttons (but not the modifiers)"
	^buttons bitAnd: 7! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 20:19'!
mouseButtons: bitMask
	"set the mouse buttons (but not the modifiers)"
	buttons := (buttons bitClear: 7) bitOr: (bitMask bitAnd: 7).! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:56'!
position
	^position! !

!CUserInputEvent methodsFor: 'printing' stamp: 'ar 8/28/2005 20:19'!
printOn: aStream
	aStream nextPut: $[.
	aStream nextPutAll: self cursorPoint printString; space.
	aStream nextPutAll: type; space.
	aStream nextPutAll: self modifierString.
	aStream nextPutAll: self buttonString.
	aStream nextPutAll: timeStamp printString.
	aStream nextPut: $].! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/29/2005 01:10'!
redButtonChanged
	"Answer true if the red mouse button has changed. This is the first mouse button."
	^self whichButton anyMask: 4! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/28/2005 20:19'!
redButtonPressed
	"Answer true if the red mouse button is being pressed. This is the first mouse button."
	^ buttons anyMask: 4! !

!CUserInputEvent methodsFor: 'initialize' stamp: 'ar 8/28/2005 19:56'!
resetHandlerFields
	"Reset anything that is used to cross-communicate between two eventual handlers during event dispatch"
	handler := nil.
	wasHandled := false.! !

!CUserInputEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:19'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseOver ifTrue:[^anObject handleMouseOver: self].
	type == #mouseEnter ifTrue:[^anObject handleMouseEnter: self].
	type == #mouseLeave ifTrue:[^anObject handleMouseLeave: self].
	^super sentTo: anObject.! !

!CUserInputEvent methodsFor: 'private' stamp: 'ar 8/28/2005 19:56'!
setPosition: aPoint
	position := aPoint! !

!CUserInputEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:19'!
setType: aSymbol
	"For quick conversion between event types"
	type := aSymbol.! !

!CUserInputEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:19'!
setType: evtType position: evtPos buttons: evtButtons hand: evtHand
	type := evtType.
	position := evtPos.
	buttons := evtButtons.
	hand := evtHand.
	wasHandled := false.! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/28/2005 19:56'!
shiftPressed
	"Answer true if the shift key on the keyboard was being held down when this event occurred."

	^ buttons anyMask: 8
! !

!CUserInputEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:19'!
storeOn: aStream
	aStream nextPutAll: type.
	aStream space.
	self timeStamp storeOn: aStream.
	aStream space.
	position x storeOn: aStream.
	aStream space.
	position y storeOn: aStream.
	aStream space.
	buttons storeOn: aStream.! !

!CUserInputEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 20:27'!
transformBy: aTransform
	"Transform the receiver into a local coordinate system."
	position :=  position transformedBy: aTransform.! !

!CUserInputEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 19:56'!
translateBy: delta
	"add delta to cursorPoint, and return the new event"
	position := position + delta.! !

!CUserInputEvent methodsFor: 'transforming' stamp: 'ar 8/28/2005 19:56'!
translatedBy: delta
	"add delta to cursorPoint, and return the new event"
	^self shallowCopy translateBy: delta! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:56'!
type
	"Return a symbol indicating the type this event."

	^ type! !

!CUserInputEvent methodsFor: 'private' stamp: 'ar 8/28/2005 20:20'!
type: eventType readFrom: aStream
	| x y |
	type := eventType.
	timeStamp := Integer readFrom: aStream.
	aStream skip: 1.
	x := Integer readFrom: aStream.
	aStream skip: 1.
	y := Integer readFrom: aStream.
	aStream skip: 1.
	buttons := Integer readFrom: aStream.
	position := x@y.! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:56'!
wasHandled
	^wasHandled! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/28/2005 19:56'!
wasHandled: aBool
	wasHandled := aBool.! !

!CUserInputEvent methodsFor: 'accessing' stamp: 'ar 8/29/2005 01:09'!
whichButton
	^0! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/29/2005 01:12'!
yellowButtonChanged
	"Answer true if the yellow mouse button has changed. This is the second mouse button or option+click on the Mac."
	^self whichButton anyMask: 2! !

!CUserInputEvent methodsFor: 'modifier state' stamp: 'ar 8/28/2005 20:20'!
yellowButtonPressed
	"Answer true if the yellow mouse button is being pressed. This is the second mouse button or option+click on the Mac."
	^ buttons anyMask: 2! !

Object subclass: #CIconLibrary
	instanceVariableNames: 'icons'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Tweak-Basic-Support'!

!CIconLibrary class methodsFor: 'accessing' stamp: ''!
at: iconName
	^self default at: iconName.! !

!CIconLibrary class methodsFor: 'accessing' stamp: ''!
default
	^Default ifNil:[Default := self new].! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:04'!
formFromEncodedImage: base64String
	"Utility to recover the PNG's from base64-encoded strings of their image-saves,
	 converting white into transparent on the way.
	 (The depth-force is needed for the white-conversion....    
	  The old PNG-reader preserved index PNG depths; the new primitive one does not.) "

	^ColorForm mappingWhiteToTransparentFrom:(
			(Form fromBinaryStream: (ReadStream on: base64String base64Decoded asByteArray)) asFormOfDepth: 8)
 ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:20'!
initialize
	"  CIconLibrary initialize  "
	Default := self new.
	self initializeCenterOfRotation.
	self initializeCopyPlusIcon.
	self initializeCheckmarkIcon.
	self initializeSubMenuIcon.
	self initializeMiniUpArrowIcon.
	self initializeMiniLeftArrowIcon.
	self initializeMiniRightArrowIcon.
	self initializeMiniDownArrowIcon.
	self initializeTreePlusIcon.
	self initializeTreeMinusIcon.
	self initializeTreeOpenIcon.
	self initializeTreeOpenSelectedIcon.
	self initializeTreeClosedIcon.
	self initializeTreeClosedSelectedIcon.

	self initializeDownArrowIcon.
	self initializeBothArrowIcon.

	self initializeDropDownIcon.
	self initializeScrollDownIcon.
	self initializeScrollUpIcon.
	self initializeScrollLeftIcon.
	self initializeScrollRightIcon.

	self initializeFlagIcon.
	self initializeWarningIcon.

	self initializeNoEventIcon.
	self initializeDefaultEventIcon.
	self initializeChangeEventIcon.
	self initializeKeyboardEventIcon.
	self initializeMouseEventIcon.
	self initializeStructureEventIcon.
	self initializeListEventIcon.
	self initializeButtonEventIcon.
	self initializeTextEventIcon.
	self initializeTickEventIcon.
	self initializeUpArrowIcon.
	self initializeViewEventIcon.
	self initializeWindowEventIcon.! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:52'!
initializeBothArrowIcon
	^self default at: #BothArrowIcon put: (self formFromEncodedImage:
'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAILQCvCRQIYKBBgtUKHrwGoJpD
hQMbOnxoEIDFixAXLsR4seJEih5BHmyYceTBgAA7'
) ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:52'!
initializeButtonEventIcon
	^self default at: #ButtonEventIcon put: (self formFromEncodedImage:
'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIOQCvCRxIkCCAgwcLAqjGkCGA
gQulSZRI6OG1iBOlVRSIceLGi9UyarTYkaLFi4RSpjzJESHLggUDAgA7'
) ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:03'!
initializeCenterOfRotation
	"[	| file data |
		file := FileStream readOnlyFileNamed: 'C:\Tweak3.8\build\CenterOfRotation.gif'.
		[data := file binary contents asString base64Encoded] ensure:[file close].
		self class compile: 'initializeCenterOfRotation
	^self default at: #CenterOfRotation put: (ColorForm mappingWhiteToTransparentFrom:(
			Form fromBinaryStream: (ReadStream on: ((', data storeString,'
	) base64Decoded asByteArray))))' classified: 'class initialization'.
	]."
	^self default at: #CenterOfRotation put: (
			self formFromEncodedImage: 
'R0lGODlhCwALAPcAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A
/wD//////wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAMwAAZgAAmQAAzAAA/wAzAAAzMwAzZgAz
mQAzzAAz/wBmAABmMwBmZgBmmQBmzABm/wCZAACZMwCZZgCZmQCZzACZ/wDMAADMMwDMZgDM
mQDMzADM/wD/AAD/MwD/ZgD/mQD/zAD//zMAADMAMzMAZjMAmTMAzDMA/zMzADMzMzMzZjMz
mTMzzDMz/zNmADNmMzNmZjNmmTNmzDNm/zOZADOZMzOZZjOZmTOZzDOZ/zPMADPMMzPMZjPM
mTPMzDPM/zP/ADP/MzP/ZjP/mTP/zDP//2YAAGYAM2YAZmYAmWYAzGYA/2YzAGYzM2YzZmYz
mWYzzGYz/2ZmAGZmM2ZmZmZmmWZmzGZm/2aZAGaZM2aZZmaZmWaZzGaZ/2bMAGbMM2bMZmbM
mWbMzGbM/2b/AGb/M2b/Zmb/mWb/zGb//5kAAJkAM5kAZpkAmZkAzJkA/5kzAJkzM5kzZpkz
mZkzzJkz/5lmAJlmM5lmZplmmZlmzJlm/5mZAJmZM5mZZpmZmZmZzJmZ/5nMAJnMM5nMZpnM
mZnMzJnM/5n/AJn/M5n/Zpn/mZn/zJn//8wAAMwAM8wAZswAmcwAzMwA/8wzAMwzM8wzZswz
mcwzzMwz/8xmAMxmM8xmZsxmmcxmzMxm/8yZAMyZM8yZZsyZmcyZzMyZ/8zMAMzMM8zMZszM
mczMzMzM/8z/AMz/M8z/Zsz/mcz/zMz///8AAP8AM/8AZv8Amf8AzP8A//8zAP8zM/8zZv8z
mf8zzP8z//9mAP9mM/9mZv9mmf9mzP9m//+ZAP+ZM/+ZZv+Zmf+ZzP+Z///MAP/MM//MZv/M
mf/MzP/M////AP//M///Zv//mf//zP///ywAAAAACwALAAAILAAfCBxIsCCAgwUFAujHEIBB
hhAdKowYcWK/gwsvWrxYcWPDhx8TYkxIUmBAADs='
	)! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:53'!
initializeChangeEventIcon
	^self default at: #ChangeEventIcon put: (self formFromEncodedImage:
 'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIMwCvCRwIYKBBgQAKHmR1LaHD
gQwRKixo5aDCa1YqHhyYkeDBihMvRmzosOBIghc3phQYEAA7'
) ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:53'!
initializeCheckmarkIcon
	^self default at: #CheckMark put: (self formFromEncodedImage:
'R0lGODlhCwALAIAAAAAAAP///ywAAAAACwALAAACFIyPqWvAAdyRMCIbact6WtY9IlIAADs=')! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:55'!
initializeCopyPlusIcon
	^self default at: #CopyPlusIcon put: (self formFromEncodedImage:
'iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAABHNCSVQICAgIfAhkiAAAAAlw
SFlzAAALEgAACxIB0t1+/AAAACV0RVh0U29mdHdhcmUATWFjcm9tZWRpYSBGaXJld29ya3Mg
TVggMjAwNId2rM8AAAAWdEVYdENyZWF0aW9uIFRpbWUAMDUvMjAvMDVGT28oAAACY0lEQVR4
nLWVMW8TQRCFJwgp1+Uko2AKJJckCJnOluhSQkFpUEiTwmmAHgURoKSmwRJISEiYCgml4R8c
BcJFqOMgQS6iuYAUbscz+yh8Pu/57mxHgq12bvfetzOzM7sAgP7nODttcfDryIPyVQjXIHze
ivGhTBCOIObICveh3PPrN+MyjYUiD/Qk8qDctGIaEPYgTFAmK4YgTJBkPoTFEPPJCgfnrm3m
QDmAnkQ+lFtWuIqRoDJZYcraOVgI4XfLa/ejUoD+OfYhZssKe1AmDMxYfNIuhsVWzIsL17dT
yBlH3INwKxV3Tz7LltT2oNz6/v6BlwNAuAHlavqzTIgJ0/6rO+2D15vtGbAqhBsZgPz+mSQ1
c5qMOJSJQESEPDwPax682fJSAJTrhaGRrE0EIiAHL4B5UK6nACtcmyfuBBCG+0s9df6tuR5U
p5xmbFMSoimeOrAqUVLJEPYnN+y/XG8TQDSSHc0BOtx9mlurNNY7EzB/7EHBacrEkw+5tSLP
xx4oR1bYdzdcvP28M1lch7tP2gTQ8tq9TknbcMMauUkO500yzZ/k0E1yv6y4XJtOl+S+m4Me
hOOi4nJtIhCAedpGDOVeCli6csNAOSgqrgwsCdEclRysPPxi0iQncQ2gfAky7EdFApXmRmdG
3AnKoRUOcs2u0twwEO5COT5lB3VhsRXuXn781Yx0cw9O+PHZEpRvJV0xewXFTHsbQih3V3f2
jl29wifzx4dHixBuJs+mV3jfx7AYAxNAOVjd2TOTWoWA0fj29u4ixAwffeWqFeMnsMiKCZPr
3VvZ/pwTngvwL8ZfN/jXcSz/O+MAAAAASUVORK5CYII=')! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:55'!
initializeDefaultEventIcon
	^self default at: #DefaultEventIcon put: (self formFromEncodedImage:
 'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIMACvCbwGoODAgwCoKQRwkKDC
hwwHJnxILaLAiRAbYrSI0GDDi4o4Igz5EaRIhA0DAgA7'
)! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:55'!
initializeDownArrowIcon
	^self default at: #DownArrow put: (self formFromEncodedImage:
 'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIKACvCRxIsKDBgQASKgRAEEC1
hw8ZNoQocWLEgw4rFsx4UKDGjiANBgQAOw=='
) ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 2/13/2008 03:25'!
initializeDropDownIcon
	^self default at: #DropDownIcon put: (
			Form fromBinaryStream: (ReadStream on: (( 
'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAHlJREFUeNpi/P//PwMlgImBQjBqABUMYEEX4OTkPAukNIH4
HwsLC/efP39gUue+f/9uTIwLooEYZMhHKP4ExDeg4kR5AaR4OhAzQ/kgugsqTnQYbAbi81D2
cSBehTMQQEkZG+bg4FDl4eF5BqRVcKkBYcbRvMAAEGAAhgJCblI4VGEAAAAASUVORK5CYII='
) base64Decoded asByteArray)))! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:56'!
initializeFlagIcon
	^self default at: #FlagIcon put: (self formFromEncodedImage:
 'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAINACvCRwoEACAgQcNGhxx8BoA
hiMiRjToEKLEiRUfXmRY0CJDihkVEsw4kuDDhiVPlkRYMiAAOw=='
) ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:56'!
initializeKeyboardEventIcon
	^self default at: #KeyboardEventIcon put: (self formFromEncodedImage:
 'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIQACvWRlIkKDAa31OIFSYEICV
PgkhnpDo8ITFixcdLtzY8GHEhAsrYmR4QuO1kyA7Qlw5seNIjACuAZhJk+a1gAAAOw=='
)! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:56'!
initializeListEventIcon
	^self default at: #ListEventIcon put: (self formFromEncodedImage:
 'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAISgCvCRwIoODAgwBaKQRAsGCr
PhBPMARwrCIAiH0kUtSlq+DFjAWPdaxGEoDEaxQBVJPGkhBDgQVXtnwJU6Y0lwdREtpJE6HB
gQEBADs=') ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:56'!
initializeMiniDownArrowIcon
	^self default at: #MiniDownArrow put: (self formFromEncodedImage:
			'R0lGODlhCwALAIAAAAAAAP///ywAAAAACwALAAACEIyPqcttcCACtMpoV23X+VMAOw==') ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:57'!
initializeMiniLeftArrowIcon
	^self default at: #MiniLeftArrow put: (self formFromEncodedImage:'R0lGODlhCwALAIAAAAAAAP///ywAAAAACwALAAACD4yPqQprABdsbDp743W8FwA7')! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:57'!
initializeMiniRightArrowIcon
	^self default at: #MiniRightArrow put: (self formFromEncodedImage:
'R0lGODlhCwALAIAAAAAAAP///ywAAAAACwALAAACEYyPqQprAF2AbdGaqI0Oy58UADs=') ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:57'!
initializeMiniUpArrowIcon
	^self default at: #MiniUpArrow put:(self formFromEncodedImage: 'R0lGODlhCwALAIAAAAAAAP///ywAAAAACwALAAACEIyPqQprABecadp6mtu8uwIAOw==') ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:58'!
initializeMouseEventIcon
	^self default at: #MouseEventIcon put: (self formFromEncodedImage:
 'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIOQCvCRQIYKDBgQUPEgTAMCHC
ZxAhOgTwzImTYBaDJaSYMePGihc9EgSJ8eLHYChTnkyJcmLDhgIDAgA7'
) ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:58'!
initializeNoEventIcon
	^self default at: #NoEventIcon put: (self formFromEncodedImage:
 'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIFQCvCRxIsKDBgwgTKlzIsKHD
h9cCAgA7')! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 2/13/2008 03:13'!
initializeScrollDownIcon
	^self default at: #ScrollDownIcon put: (
			Form fromBinaryStream: (ReadStream on: (( 
'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAHNJREFUeNpi/P//PwMlgImBQjAMDGBBF+Dk5LwNpNhBhrOw
sPD8+fMHJvX6+/fvqsS4oByIf+IQJ+wFoC3rgNRqNOHFUHGiwwBkwHco+ysQzyMpEIG2nUdy
xUIoHzsApURcmIeH5wM+eRBmHE3KlBsAEGAA/fRSRPzRpAoAAAAASUVORK5CYII='
) base64Decoded asByteArray)))! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 2/13/2008 03:13'!
initializeScrollLeftIcon
	^self default at: #ScrollLeftIcon put: ((
			Form fromBinaryStream: (ReadStream on: (( 
'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAHNJREFUeNpi/P//PwMlgImBQjAMDGBBF+Dk5LwNpNhBhrOw
sPD8+fMHJvX6+/fvqsS4oByIf+IQJ+wFoC3rgNRqNOHFUHGiwwBkwHco+ysQzyMpEIG2nUdy
xUIoHzsApURcmIeH5wM+eRBmHE3KlBsAEGAA/fRSRPzRpAoAAAAASUVORK5CYII='
) base64Decoded asByteArray))) rotateBy: #right centerAt: 8@8).! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 2/13/2008 03:12'!
initializeScrollRightIcon
	^self default at: #ScrollRightIcon put: ((
			Form fromBinaryStream: (ReadStream on: (( 
'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAG9JREFUeNpi/P//PwMlgImBQjDcDeDl5f1AtgGcnJzNyDRJ
BgA16QOpRCg3HMonyQUgzTww84A4hGgDgLb5A6kENOFwqDgGYERPiUCFt4CUGIjNwsLC8OfP
n89A5j8g/vn9+3c1ggaMpkTSAUCAAQAwXxkJD8q86AAAAABJRU5ErkJggg=='
) base64Decoded asByteArray))) rotateBy: #right centerAt: 7@7).! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 2/13/2008 03:12'!
initializeScrollUpIcon
	^self default at: #ScrollUpIcon put: (
			Form fromBinaryStream: (ReadStream on: (( 
'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAG9JREFUeNpi/P//PwMlgImBQjDcDeDl5f1AtgGcnJzNyDRJ
BgA16QOpRCg3HMonyQUgzTww84A4hGgDgLb5A6kENOFwqDgGYERPiUCFt4CUGIjNwsLC8OfP
n89A5j8g/vn9+3c1ggaMpkTSAUCAAQAwXxkJD8q86AAAAABJRU5ErkJggg=='
) base64Decoded asByteArray)))! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 18:59'!
initializeStructureEventIcon
	^self default at: #StructureEventIcon put: (self formFromEncodedImage:
 'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIQQABCBwI4JrBawDgKCxQoODB
hNHgMHRoMOFCghUZNpTAUUJBABMlyPLQkSAACR5IcqR4suPKhy5fwix58OFAgwEBADs='
)! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:00'!
initializeSubMenuIcon
	^self default at: #SubMenu put: (self formFromEncodedImage:'R0lGODlhCwALAIAAAAAAAP///ywAAAAACwALAAACEoyPCanLBuILUbJqEX515juFBQA7') ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:00'!
initializeTextEventIcon
	^self default at: #TextEventIcon put: (self formFromEncodedImage:
'R0lGODlhDAAMAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAADAAMAAAIPAADCRwo8JrBawUTBjqI0GDB
hgcXIgRAsSLFiBczAsBo8OK1jQ5BegSJUKJHhwQDnWR4cCXLjyRZWrwYEAA7'
)! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:00'!
initializeTickEventIcon
	^self default at: #TickEventIcon put: (self formFromEncodedImage:
'R0lGODlhDAAMAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAADAAMAAAIOQCvCQRAkKDAgwcLIkwIYJFD
gwkdLmr4MOJDiQAGSpy4MSPFixivfQQ5UWPHkiY3ohz4EeJChQgDAgA7'
) ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 17:11'!
initializeTreeClosedIcon
	^self default at: #TreeClosedIcon put: (
			Form fromBinaryStream: (ReadStream on: (( 
'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAGBJREFUeNpi/P//PwMlgImBQkCxASzoApOnTQL5qRGIG3Bp
ys3KI+iCeiD+j88QYr1A0CBiwwCnQaQGIswgsg0ABS4j3ljAo7GBqGgkViMhAxqJiUIQYBz6
eYFiAwACDAB+VRYe/AOlAQAAAABJRU5ErkJggg=='
) base64Decoded asByteArray)))

"The magic string is produced via: 
 Clipboard default clipboardText: 
	 ((FileDirectory default directoryNamed: 'icons') readOnlyFileNamed: 'TreeClosed.png') 
		  binary contentsOfEntireFile asString base64Encoded printString
"! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 17:05'!
initializeTreeClosedSelectedIcon
	^self default at: #TreeClosedSelectedIcon put: (
			Form fromBinaryStream: (ReadStream on: (( 
'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAGBJREFUeNpi/P//PwMlgImBQkCxASzoAhcunQf5qRGIG3Bp
MtAzJOiCeiD+j88QYr1A0CBiwwCnQaQGIswgsg0ABS4j3ljAo7GBqGgkViMhAxqJiUIQYBz6
eYFiAwACDACX7RYe2zk8swAAAABJRU5ErkJggg=='
) base64Decoded asByteArray)))

"The magic string is produced via: 
 Clipboard default clipboardText: 
	 ((FileDirectory default directoryNamed: 'icons') readOnlyFileNamed: 'TreeClosedSelected.png') 
		  binary contentsOfEntireFile asString base64Encoded printString
"! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:11'!
initializeTreeMinusIcon
	^self default at: #TreeMinusIcon put: (self mappingRedToTransparentFrom:(
			(Form fromBinaryStream: (ReadStream on: (('R0lGODlhCwALAPcAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A
/wD//////wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAMwAAZgAAmQAAzAAA/wAzAAAzMwAzZgAz
mQAzzAAz/wBmAABmMwBmZgBmmQBmzABm/wCZAACZMwCZZgCZmQCZzACZ/wDMAADMMwDMZgDM
mQDMzADM/wD/AAD/MwD/ZgD/mQD/zAD//zMAADMAMzMAZjMAmTMAzDMA/zMzADMzMzMzZjMz
mTMzzDMz/zNmADNmMzNmZjNmmTNmzDNm/zOZADOZMzOZZjOZmTOZzDOZ/zPMADPMMzPMZjPM
mTPMzDPM/zP/ADP/MzP/ZjP/mTP/zDP//2YAAGYAM2YAZmYAmWYAzGYA/2YzAGYzM2YzZmYz
mWYzzGYz/2ZmAGZmM2ZmZmZmmWZmzGZm/2aZAGaZM2aZZmaZmWaZzGaZ/2bMAGbMM2bMZmbM
mWbMzGbM/2b/AGb/M2b/Zmb/mWb/zGb//5kAAJkAM5kAZpkAmZkAzJkA/5kzAJkzM5kzZpkz
mZkzzJkz/5lmAJlmM5lmZplmmZlmzJlm/5mZAJmZM5mZZpmZmZmZzJmZ/5nMAJnMM5nMZpnM
mZnMzJnM/5n/AJn/M5n/Zpn/mZn/zJn//8wAAMwAM8wAZswAmcwAzMwA/8wzAMwzM8wzZswz
mcwzzMwz/8xmAMxmM8xmZsxmmcxmzMxm/8yZAMyZM8yZZsyZmcyZzMyZ/8zMAMzMM8zMZszM
mczMzMzM/8z/AMz/M8z/Zsz/mcz/zMz///8AAP8AM/8AZv8Amf8AzP8A//8zAP8zM/8zZv8z
mf8zzP8z//9mAP9mM/9mZv9mmf9mzP9m//+ZAP+ZM/+ZZv+Zmf+ZzP+Z///MAP/MM//MZv/M
mf/MzP/M////AP//M///Zv//mf//zP///yH5BAEAAAkALAAAAAALAAsABwgzABMkAECwIACB
Ax8oXPjgIACGEAkqfDhx4kODBRtSxGhxYcaOHhlKhOgxIcmDAjESFBgQADs=') base64Decoded asByteArray))) asFormOfDepth:8))! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 17:14'!
initializeTreeOpenIcon
	^self default at: #TreeOpenIcon put: (
			Form fromBinaryStream: (ReadStream on: (( 
'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAGdJREFUeNpi/P//PwMlgImBQkCxASwwRv/kPgcWZpb9hDTk
ZuUxYnVBYW7RASDVSEB/IyEvNBAwoIGYMGgk1nZcBjQQazu+WGgk0lU4DWggNmyYiAjxRnIT
UgMxMcM49PMCxQYABBgAyjMUM/cJ+n8AAAAASUVORK5CYII='
) base64Decoded asByteArray)))


"The magic string is produced via: 
   Clipboard default clipboardText: 
	 ((FileDirectory default directoryNamed: 'icons') readOnlyFileNamed: 'TreeOpen.png') 
		  binary contentsOfEntireFile asString base64Encoded printString
"! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 17:03'!
initializeTreeOpenSelectedIcon
	^self default at: #TreeOpenSelectedIcon put: (
			Form fromBinaryStream: (ReadStream on: (( 
'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAGhJREFUeNpi/P//PwMlgImBQkCxASzInAuXzhP0j4GeISM+
FzQS0I8hz4gciEAXgCh8rmCEugJvGDTisx1ZMy4DGnAY0EBKLDQSGzZMRNrWQE46aCQmZvAZ
0EDIdoxoHJp5gWIDAAIMAHWtGK6Omo00AAAAAElFTkSuQmCC'
) base64Decoded asByteArray)))


"The magic string is produced via: 
   Clipboard default clipboardText: 
	 ((FileDirectory default directoryNamed: 'icons') readOnlyFileNamed: 'TreeOpenSelected.png') 
		  binary contentsOfEntireFile asString base64Encoded printString
"! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:01'!
initializeTreePlusIcon
	^self default at: #TreePlusIcon put: (self formFromEncodedImage:
'R0lGODlhCwALAPcAAAAAAIAAAACAAICAAAAAgIAAgACAgICAgMDAwP8AAAD/AP//AAAA//8A
/wD//////wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAMwAAZgAAmQAAzAAA/wAzAAAzMwAzZgAz
mQAzzAAz/wBmAABmMwBmZgBmmQBmzABm/wCZAACZMwCZZgCZmQCZzACZ/wDMAADMMwDMZgDM
mQDMzADM/wD/AAD/MwD/ZgD/mQD/zAD//zMAADMAMzMAZjMAmTMAzDMA/zMzADMzMzMzZjMz
mTMzzDMz/zNmADNmMzNmZjNmmTNmzDNm/zOZADOZMzOZZjOZmTOZzDOZ/zPMADPMMzPMZjPM
mTPMzDPM/zP/ADP/MzP/ZjP/mTP/zDP//2YAAGYAM2YAZmYAmWYAzGYA/2YzAGYzM2YzZmYz
mWYzzGYz/2ZmAGZmM2ZmZmZmmWZmzGZm/2aZAGaZM2aZZmaZmWaZzGaZ/2bMAGbMM2bMZmbM
mWbMzGbM/2b/AGb/M2b/Zmb/mWb/zGb//5kAAJkAM5kAZpkAmZkAzJkA/5kzAJkzM5kzZpkz
mZkzzJkz/5lmAJlmM5lmZplmmZlmzJlm/5mZAJmZM5mZZpmZmZmZzJmZ/5nMAJnMM5nMZpnM
mZnMzJnM/5n/AJn/M5n/Zpn/mZn/zJn//8wAAMwAM8wAZswAmcwAzMwA/8wzAMwzM8wzZswz
mcwzzMwz/8xmAMxmM8xmZsxmmcxmzMxm/8yZAMyZM8yZZsyZmcyZzMyZ/8zMAMzMM8zMZszM
mczMzMzM/8z/AMz/M8z/Zsz/mcz/zMz///8AAP8AM/8AZv8Amf8AzP8A//8zAP8zM/8zZv8z
mf8zzP8z//9mAP9mM/9mZv9mmf9mzP9m//+ZAP+ZM/+ZZv+Zmf+ZzP+Z///MAP/MM//MZv/M
mf/MzP/M////AP//M///Zv//mf//zP///yH5BAEAAAkALAAAAAALAAsABwg0ABMkAECwIACB
Ax8oXPjgIICFDxkSVFhQYsSGEyleNKgRYkaMGi+CpMiwYcKSJgUaLCgwIAA7') ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:01'!
initializeUpArrowIcon
	^self default at: #UpArrow put: (self formFromEncodedImage:'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIKgCvCRxIsKBBgwAOCgRQLeFB
hg4LMqzWECHFigQnXox4DYDHjxwVihwYEAA7'
) ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:01'!
initializeViewEventIcon
	^self default at: #ViewEventIcon put: (self formFromEncodedImage:
 'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIPwCvCQRAkKDAgwBYKVQIYCDB
hQwLWpkokeI1AFYARINzAmPDixnhcPQ4MOOJkyRBZiRoMSFEiAVjyjxIk2ZAAAA7'
) ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:02'!
initializeWarningIcon
	^self default at: #WarningIcon put: (self formFromEncodedImage:
'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAINACvCbwGoODAgwCkKQRwUGDC
hQ0fKpTGcODDixYnSqy4USPBiSAXSqQIsmDIkgQLqlR5LSAAOw=='
)! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'bgf 12/24/2008 19:02'!
initializeWindowEventIcon
	^self default at: #WindowEventIcon put: (self formFromEncodedImage:
'R0lGODlhCwALAPcAAAAAAAAAMwAAZgAAmQAAzAAA/zMAADMAMzMAZjMAmTMAzDMA/2YAAGYA
M2YAZmYAmWYAzGYA/5kAAJkAM5kAZpkAmZkAzJkA/8wAAMwAM8wAZswAmcwAzMwA//8AAP8A
M/8AZv8Amf8AzP8A/wAzAAAzMwAzZgAzmQAzzAAz/zMzADMzMzMzZjMzmTMzzDMz/2YzAGYz
M2YzZmYzmWYzzGYz/5kzAJkzM5kzZpkzmZkzzJkz/8wzAMwzM8wzZswzmcwzzMwz//8zAP8z
M/8zZv8zmf8zzP8z/wBmAABmMwBmZgBmmQBmzABm/zNmADNmMzNmZjNmmTNmzDNm/2ZmAGZm
M2ZmZmZmmWZmzGZm/5lmAJlmM5lmZplmmZlmzJlm/8xmAMxmM8xmZsxmmcxmzMxm//9mAP9m
M/9mZv9mmf9mzP9m/wCZAACZMwCZZgCZmQCZzACZ/zOZADOZMzOZZjOZmTOZzDOZ/2aZAGaZ
M2aZZmaZmWaZzGaZ/5mZAJmZM5mZZpmZmZmZzJmZ/8yZAMyZM8yZZsyZmcyZzMyZ//+ZAP+Z
M/+ZZv+Zmf+ZzP+Z/wDMAADMMwDMZgDMmQDMzADM/zPMADPMMzPMZjPMmTPMzDPM/2bMAGbM
M2bMZmbMmWbMzGbM/5nMAJnMM5nMZpnMmZnMzJnM/8zMAMzMM8zMZszMmczMzMzM///MAP/M
M//MZv/Mmf/MzP/M/wD/AAD/MwD/ZgD/mQD/zAD//zP/ADP/MzP/ZjP/mTP/zDP//2b/AGb/
M2b/Zmb/mWb/zGb//5n/AJn/M5n/Zpn/mZn/zJn//8z/AMz/M8z/Zsz/mcz/zMz/////AP//
M///Zv//mf//zP///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACwAAAAACwALAAAIMgCvCRxIUCCAgwgRGqTEsCEl
AAYTJjQIraJFaBCvAbhoMeNGjhgpgvQIMmREiQVTCgwIADs='
) ! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'ar 8/19/2005 00:52'!
mappingRedToTransparentFrom: aFormOrCursor
	"Return a ColorForm copied from the given Form or Cursor with white mapped to transparent."

	| f map |
	aFormOrCursor depth <= 8 ifFalse: [
		^ self error: 'argument depth must be 8-bits per pixel or less'].
	(aFormOrCursor isColorForm) ifTrue: [
		f := aFormOrCursor deepCopy.
		map := aFormOrCursor colors.
	] ifFalse: [
		f := ColorForm extent: aFormOrCursor extent depth: aFormOrCursor depth.
		f copyBits: aFormOrCursor boundingBox
			from: aFormOrCursor
			at: 0@0
			clippingBox: aFormOrCursor boundingBox
			rule: Form over
			fillColor: nil.
		map := Color indexedColors copyFrom: 1 to: (1 bitShift: aFormOrCursor depth)].
	map := map collect: [:c |
		c = Color red ifTrue: [Color transparent] ifFalse: [c]].
	f colors: map.
	^ f! !

!CIconLibrary class methodsFor: 'class initialization' stamp: 'ar 8/19/2005 00:51'!
mappingWhiteToTransparentFrom: aFormOrCursor
	"Return a ColorForm copied from the given Form or Cursor with white mapped to transparent."

	| f map |
	aFormOrCursor depth <= 8 ifFalse: [
		^ self error: 'argument depth must be 8-bits per pixel or less'].
	(aFormOrCursor isColorForm) ifTrue: [
		f := aFormOrCursor deepCopy.
		map := aFormOrCursor colors.
	] ifFalse: [
		f := ColorForm extent: aFormOrCursor extent depth: aFormOrCursor depth.
		f copyBits: aFormOrCursor boundingBox
			from: aFormOrCursor
			at: 0@0
			clippingBox: aFormOrCursor boundingBox
			rule: Form over
			fillColor: nil.
		map := Color indexedColors copyFrom: 1 to: (1 bitShift: aFormOrCursor depth)].
	map := map collect: [:c |
		c = Color white ifTrue: [Color transparent] ifFalse: [c]].
	f colors: map.
	^ f
! !

!CIconLibrary class methodsFor: 'instance creation' stamp: 'ar 1/30/2005 11:50'!
new
	^self basicNew initialize! !

!CIconLibrary methodsFor: 'accessing' stamp: 'ar 12/8/2002 02:49'!
at: iconName
	^icons at: iconName ifAbsent:[Form extent: 10@10 depth: 1]! !

!CIconLibrary methodsFor: 'accessing' stamp: 'ar 6/25/2005 15:58'!
at: iconName put: aForm
	^icons at: iconName put: aForm! !

!CIconLibrary methodsFor: 'accessing' stamp: ''!
icons
	^icons! !

!CIconLibrary methodsFor: 'initialize' stamp: ''!
initialize
	icons := Dictionary new.! !

Object subclass: #CKeyboardDispatchTable
	instanceVariableNames: 'metaTable'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Support'!

!CKeyboardDispatchTable commentStamp: 'ar 3/27/2004 18:01' prior: 0!
CKeyboardDispatchTable simplifies mapping from particular keyboard events to specific actions.!

!CKeyboardDispatchTable class methodsFor: 'instance creation' stamp: 'ar 1/30/2005 11:50'!
new
	^self basicNew initialize! !

!CKeyboardDispatchTable methodsFor: 'accessing' stamp: 'ar 3/24/2004 01:14'!
add: keySpec action: eventName
	| state |
	state := self analyze: keySpec.
	self bind: state last to: eventName metaState: state first.! !

!CKeyboardDispatchTable methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:27'!
addAll: specList
	specList do:[:each| self add: each first action: each last].! !

!CKeyboardDispatchTable methodsFor: 'private' stamp: 'bf 5/27/2005 16:42'!
analyze: keySpec
	| char meta modifier bit |
	keySpec class == Character 
		ifTrue:[^Array with: 0 with: keySpec].
	keySpec last isInteger ifTrue:[
		char := Character value: keySpec last.
	] ifFalse:[
		(keySpec size < 1 or:[keySpec last size ~= 1]) 
			ifTrue:[^self error: 'Character expected'].
		char := keySpec last first.
	].
	meta := 0.
	1 to: keySpec size - 1 do:[:i|
		modifier := keySpec at: i.
		bit := #(shift ctrl opt cmd) indexOf: modifier asLowercase.
		bit = 0 ifTrue:[self error:'Unknown modifier: ', modifier].
		meta := meta bitOr: (1 bitShift: bit-1).
	].
	^Array with: meta with: char! !

!CKeyboardDispatchTable methodsFor: 'private' stamp: ''!
bind: aCharacter to: anObject metaState: metaState
	"Install a character binding"
	| dispatchTable metaIndex |
	metaIndex := metaState + 1.
	dispatchTable := metaTable at: metaIndex.
	dispatchTable ifNil:[
		dispatchTable := IdentityDictionary new.
		metaTable at: metaIndex put: dispatchTable].
	anObject isNil ifTrue:[
		dispatchTable removeKey: aCharacter ifAbsent:[].
		dispatchTable isEmpty ifTrue:[metaTable at: metaIndex put: nil].
	] ifFalse:[
		dispatchTable at: aCharacter put: anObject.
	].! !

!CKeyboardDispatchTable methodsFor: 'initialize' stamp: ''!
initialize
	metaTable := Array new: 16.! !

!CKeyboardDispatchTable methodsFor: 'accessing' stamp: 'ar 3/24/2004 01:17'!
isEmpty
	^metaTable allSatisfy:[:each| each isEmptyOrNil].! !

!CKeyboardDispatchTable methodsFor: 'private' stamp: 'ar 3/24/2004 01:26'!
lookup: aCharacter meta: metaState
	| metaIndex dispatchTable |
	metaIndex := metaState + 1.
	"First look up modified entry"
	dispatchTable := metaTable at: metaIndex.
	^dispatchTable ifNotNil:[dispatchTable at: aCharacter ifAbsent:[nil]].! !

!CKeyboardDispatchTable methodsFor: 'accessing' stamp: ''!
lookupEvent: anEvent
	"Look up any associated short cut for the given event"
	| metaState character |
	metaState := anEvent metaState.
	character := anEvent keyCharacter.
	^self lookup: character meta: metaState! !

!CKeyboardDispatchTable methodsFor: 'accessing' stamp: 'ar 3/24/2004 01:16'!
remove: keySpec
	| state |
	state := self analyze: keySpec.
	self remove: state last meta: state first.! !

!CKeyboardDispatchTable methodsFor: 'private' stamp: 'ar 3/24/2004 01:15'!
remove: aCharacter meta: metaState
	| metaIndex dispatchTable |
	metaIndex := metaState + 1.
	"First look up modified entry"
	dispatchTable := metaTable at: metaIndex.
	^dispatchTable ifNotNil:[dispatchTable removeKey: aCharacter ifAbsent:[nil]].! !

Object subclass: #CMD5Hasher
	instanceVariableNames: 'state'
	classVariableNames: 'ABCDTable IndexTable ShiftTable SinTable'
	poolDictionaries: ''
	category: 'Tweak-Basic-Support'!

!CMD5Hasher commentStamp: 'DSM 1/20/2000 14:08' prior: 0!
This class implements the MD5 128-bit one-way hash function.  It relies on
the ThirtyTwoBitRegister class supplied as part of the "Digital Signatures"
functionality included in Squeak 2.7.  As of this date (1/20/2000), the
U.S. Government has lifted many of the previous restrictions on the export
of encryption software, but you should check before exporting anything
including this code.  MD5 is commonly used for some secure Internet
protocols, including authentication in HTTP, which is why I wrote it.

Submitted by Duane Maxwell

!

!CMD5Hasher class methodsFor: 'utilities' stamp: 'DSM 1/20/2000 18:03'!
hashMessage: aStringOrByteArray
	^ self new hashMessage: aStringOrByteArray! !

!CMD5Hasher class methodsFor: 'utilities' stamp: 'DSM 1/20/2000 18:03'!
hashStream: aPositionableStream
	^ self new hashStream: aPositionableStream! !

!CMD5Hasher class methodsFor: 'class initialization' stamp: 'ar 3/14/2004 01:35'!
initialize
	"MD5 initialize"
	"Obscure fact: those magic hex numbers that are hard to type in correctly are
	actually the result of a simple trigonometric function and are therefore
	easier to compute than proofread.  Laziness is sometimes a virtue."
	| c |
	c := 2 raisedTo: 32.
	SinTable := Array new: 64.
	1 to: 64 do: [ :i |
		SinTable
			at: i
			put: (ThirtyTwoBitRegister new load: (c * i sin abs) truncated)].
	ShiftTable := {
		#(7 12 17 22).
		#(5 9 14 20).
		#(4 11 16 23).
		#(6 10 15 21)
		}.
	IndexTable := {
		#(1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16).
		#(2 7 12 1 6 11 16 5 10 15 4 9 14 3 8 13).
		#(6 9 12 15 2 5 8 11 14 1 4 7 10 13 16 3).
		#(1 8 15 6 13 4 11 2 9 16 7 14 5 12 3 10)
		}.
	ABCDTable := {
		#(1 2 3 4).
		#(4 1 2 3).
		#(3 4 1 2).
		#(2 3 4 1)
		}! !

!CMD5Hasher class methodsFor: 'testing' stamp: 'ar 3/25/2004 19:27'!
test
	"CMD5Hasher test"

	(CMD5Hasher hashMessage: 'a') = 16r0CC175B9C0F1B6A831C399E269772661
		ifFalse: [ self error: 'failed'].
	(CMD5Hasher hashMessage: 'abc') = 16r900150983CD24FB0D6963F7D28E17F72
		ifFalse: [ self error: 'failed'].
	(CMD5Hasher hashMessage: 'message digest') = 16rF96B697D7CB7938D525A2F31AAF161D0
		ifFalse: [ self error: 'failed'].
	(CMD5Hasher hashMessage:
		'abcdefghijklmnopqrstuvwxyz') = 16rC3FCD3D76192E4007DFB496CCA67E13B
		ifFalse: [ self error: 'failed'].
	(CMD5Hasher hashMessage:
	'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789') =
		16rD174AB98D277D9F5A5611C2C9F419D9F
		ifFalse: [ self error: 'failed'].
	(CMD5Hasher hashMessage:
	'12345678901234567890123456789012345678901234567890123456789012345678901234567890') = 		16r57EDF4A22BE3C955AC49DA2E2107B67A
		ifFalse: [ self error: 'failed'].
! !

!CMD5Hasher methodsFor: 'private-functions' stamp: 'DSM 1/20/2000 01:47'!
fX: x Y: y Z: z
	" compute 'xy or (not x)z'"
	^ x copy bitAnd: y; bitOr: (x copy bitInvert; bitAnd: z)

	! !

!CMD5Hasher methodsFor: 'private-functions' stamp: 'DSM 1/20/2000 13:38'!
ffA: a B: b C: c D: d M: m S: s T: t
	"compute a = b + ((a + f(b,c,d) + m + t) <<< s)"
	^ a += (self fX: b Y: c Z: d); += m; += t; leftRotateBy: s; += b.
! !

!CMD5Hasher methodsFor: 'private-buffers' stamp: 'DSM 1/20/2000 17:56'!
finalValue
	"Concatenate the state values to produce the 128-bite result"
	^ (( state at: 1) asReverseInteger bitShift: 96) +
	  (( state at: 2) asReverseInteger bitShift: 64) +
	  (( state at: 3) asReverseInteger bitShift: 32) +
	  (( state at: 4) asReverseInteger)! !

!CMD5Hasher methodsFor: 'private-functions' stamp: 'DSM 1/20/2000 01:48'!
gX: x Y: y Z: z
	" compute 'xz or y(not z)'"
	^ x copy bitAnd: z; bitOr: (z copy bitInvert; bitAnd: y)

	! !

!CMD5Hasher methodsFor: 'private-functions' stamp: 'DSM 1/20/2000 13:38'!
ggA: a B: b C: c D: d M: m S: s T: t
	"compute a = b + ((a + g(b,c,d) + m + t) <<< s)"
	^ a += (self gX: b Y: c Z: d); += m; += t; leftRotateBy: s; += b.
! !

!CMD5Hasher methodsFor: 'private-functions' stamp: 'DSM 1/20/2000 01:48'!
hX: x Y: y Z: z
	" compute 'x xor y xor z'"
	^ x copy bitXor: y; bitXor: z

	! !

!CMD5Hasher methodsFor: 'public' stamp: 'DSM 1/20/2000 15:29'!
hashMessage: aStringOrByteArray
	"MD5 new hashMessage: 'foo'"
	^ self hashStream: (ReadStream on: aStringOrByteArray asByteArray)
! !

!CMD5Hasher methodsFor: 'public' stamp: 'ar 3/14/2004 01:34'!
hashStream: aPositionableStream
	"MD5 new hashStream: (ReadStream on: 'foo')"
	| startPosition buf bitLength |
	self initialize.

	"aPositionableStream atEnd ifTrue: [self error: 'empty stream']."

	startPosition := aPositionableStream position.
	[aPositionableStream atEnd] whileFalse: [
		buf := aPositionableStream next: 64.
		(aPositionableStream atEnd not and: [buf size = 64])
			ifTrue: [self processBuffer: buf]
			ifFalse: [
				bitLength := (aPositionableStream position - startPosition) * 8.
				self processFinalBuffer: buf bitLength: bitLength]].

	^ self finalValue
! !

!CMD5Hasher methodsFor: 'private-functions' stamp: 'DSM 1/20/2000 13:38'!
hhA: a B: b C: c D: d M: m S: s T: t
	"compute a = b + ((a + h(b,c,d) + m + t) <<< s)"
	^ a += (self hX: b Y: c Z: d); += m; += t; leftRotateBy: s; += b.
! !

!CMD5Hasher methodsFor: 'private-functions' stamp: 'DSM 1/20/2000 01:48'!
iX: x Y: y Z: z
	" compute 'y xor (x or (not z))'"
	^ y copy bitXor: (z copy bitInvert; bitOr: x)
! !

!CMD5Hasher methodsFor: 'private-functions' stamp: 'DSM 1/20/2000 13:39'!
iiA: a B: b C: c D: d M: m S: s T: t
	"compute a = b + ((a + i(b,c,d) + m + t) <<< s)"
	^ a += (self iX: b Y: c Z: d); += m; += t; leftRotateBy: s; += b.
! !

!CMD5Hasher methodsFor: 'initialization' stamp: 'ar 3/14/2004 01:57'!
initialize
	"Some magic numbers to get the process started"
	state := {
		(ThirtyTwoBitRegister new load: 16r67452301).
		(ThirtyTwoBitRegister new load: 16rEFCDAB89).
		(ThirtyTwoBitRegister new load: 16r98BADCFE).
		(ThirtyTwoBitRegister new load: 16r10325476)
	}.

! !

!CMD5Hasher methodsFor: 'private-buffers' stamp: 'ar 3/14/2004 01:34'!
processBuffer: aByteArray
	"Process a 64-byte buffer"

	| saveState data |
	saveState := state collect: [ :item | item copy ].
	data := Array new: 16.
	1 to: 16 do: [ :index |
		data at: index put:
			(ThirtyTwoBitRegister new reverseLoadFrom: aByteArray at: (index * 4) - 3)].
	self rounds: data.
	1 to: 4 do: [ :index | (state at: index) += (saveState at: index) ].! !

!CMD5Hasher methodsFor: 'private-buffers' stamp: 'DSM 1/20/2000 17:55'!
processFinalBuffer: aByteArray bitLength: bitLength
	"Pad the buffer until we have an even 64 bytes, then transform"

	| out |
	out := ByteArray new: 64.
	out replaceFrom: 1 to: aByteArray size with: aByteArray startingAt: 1.
	aByteArray size < 56 ifTrue: [
		out at: aByteArray size + 1 put: 128. "trailing bit"
		self storeLength: bitLength in: out.
		self processBuffer: out.
		^ self].

	"not enough room for the length, so just pad this one, then..."
	aByteArray size < 64 ifTrue: [ out at: aByteArray size + 1 put: 128 ].
	self processBuffer: out.

	"process one additional block of padding ending with the length"
	out := ByteArray new: 64.  "filled with zeros"
	aByteArray size = 64 ifTrue: [ out at: 1 put: 128].
	self storeLength: bitLength in: out.
	self processBuffer: out.
! !

!CMD5Hasher methodsFor: 'private-rounds' stamp: 'ar 3/14/2004 02:02'!
round: data selector: selector round: round
	"Do one round with the given function"

	| shiftIndex template abcd |
	1 to: 16 do: [ :i |
		shiftIndex := (i - 1) \\ 4 + 1.
		abcd := ABCDTable at: shiftIndex.
		template := Array new: 7.
		template at: 1 put: (abcd at: 1).
		template at: 2 put: (abcd at: 2).
		template at: 3 put: (abcd at: 3).
		template at: 4 put: (abcd at: 4).
		template at: 5 put: ((IndexTable at: round) at: i).
		template at: 6 put: ((ShiftTable at: round) at: shiftIndex).
		template at: 7 put: (SinTable at: round - 1 * 16 + i).
		self step: data template: template selector: selector ].
! !

!CMD5Hasher methodsFor: 'private-rounds' stamp: 'ar 3/14/2004 02:05'!
rounds: data
	"Perform the four rounds with different functions"
	self round: data selector: #ffA:B:C:D:M:S:T: round: 1.
	self round: data selector: #ggA:B:C:D:M:S:T: round: 2.
	self round: data selector: #hhA:B:C:D:M:S:T: round: 3.
	self round: data selector: #iiA:B:C:D:M:S:T: round: 4.
! !

!CMD5Hasher methodsFor: 'private-functions' stamp: 'ar 3/14/2004 01:56'!
step: data template: item selector: selector
	"Perform one step in the round"

	| args |
	args :=  Array new: 7.
	args at: 1 put: (state at: (item at: 1)).
	args at: 2 put: (state at: (item at: 2)).
	args at: 3 put: (state at: (item at: 3)).
	args at: 4 put: (state at: (item at: 4)).
	args at: 5 put: (data at: (item at: 5)).
	args at: 6 put: (item at: 6).
	args at: 7 put: (item at: 7).
	(self perform: selector withArguments: args).
! !

!CMD5Hasher methodsFor: 'private-buffers' stamp: 'DSM 1/20/2000 17:19'!
storeLength: bitLength in: aByteArray
	"Fill in the final 8 bytes of the given ByteArray with a 64-bit
	little-endian representation of the original message length in bits."

	| n i |
	n := bitLength.
	i := aByteArray size - 8 + 1.
	[n > 0] whileTrue: [
		aByteArray at: i put: (n bitAnd: 16rFF).
		n := n bitShift: -8.
		i := i + 1].
! !

Object subclass: #CPolygonShape
	instanceVariableNames: 'myPrimCostume vertexFirstSelector vertexLastSelector verticesSelector vertices closed filledForm arrows arrowForms smoothCurve curveState borderDashSpec handles borderForm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Costumes'!

!CPolygonShape commentStamp: 'tk 1/28/2005 12:13' prior: 0!
PolygonShape helps a primitive costume draw polygons and curves for a Player.  A polygon has straight segments and a curve is smooth.  Either may be open (a long wiggly line) or closed (makes a region).  The shape can be filled with a texture, can have a border, can have a dashed border, and can have arrows at either or both ends.

verticies are in same coordinate system as myPrimCostume player's position.  Must change them when the player moves.!

!CPolygonShape class methodsFor: 'instance creation' stamp: 'sw 10/3/2002 02:19'!
arrowPrototype
	"Answer an instance of the receiver that will serve as a prototypical arrow"

	| aa |
	aa := self new. 
	aa vertices: (Array with: 0@0 with: 40@40) 
		color: Color black 
		borderWidth: 2 
		borderColor: Color black.
	aa setProperty: #noNewVertices toValue: true.
	aa makeForwardArrow.		"is already open"
	aa computeBounds.
	^ aa! !

!CPolygonShape class methodsFor: 'parts bin' stamp: 'tk 11/16/2001 12:15'!
descriptionForPartsBin
	^ self partName:	'Polygon'
		categories:		#('Graphics' ' Basic 1 ')
		documentation:	'A series of connected line segments, which may be a closed solid, or a zig-zag line.  Shift-click to get handles and move the points.'! !

!CPolygonShape class methodsFor: 'instance creation' stamp: 'di 10/18/2001 03:56'!
fromHand: hand
	"Let the user draw a polygon, clicking at each vertex, and ending
		by clicking within 5 of the first point..."
	| p1 poly oldVerts pN opposite |
	Cursor crossHair showWhile:
		[[Sensor anyButtonPressed] whileFalse:
			[self currentWorld displayWorldSafely; runStepMethods].
		p1 := Sensor cursorPoint].
	opposite := (Display colorAt: p1) negated.
	opposite = Color transparent ifTrue: [opposite := Color red].
	(poly := LineMorph from: p1 to: p1 color: opposite width: 2) openInWorld.
	oldVerts := {p1}.
	self currentWorld displayWorldSafely; runStepMethods.
	[true] whileTrue:
		[[Sensor anyButtonPressed] whileTrue:
			[pN := Sensor cursorPoint.
			poly setVertices: (oldVerts copyWith: pN).
			self currentWorld displayWorldSafely; runStepMethods].
		(oldVerts size > 1 and: [(pN dist: p1) < 5]) ifTrue:
			[hand position: Sensor cursorPoint.  "Done -- update hand pos"
			^ (poly setVertices: (poly vertices copyWith: p1)) delete].
		oldVerts := poly vertices.
		[Sensor anyButtonPressed] whileFalse:
			[pN := Sensor cursorPoint.
			poly setVertices: (oldVerts copyWith: pN).
			self currentWorld displayWorldSafely; runStepMethods]].
! !

!CPolygonShape class methodsFor: 'instance creation' stamp: 'di 10/18/2001 04:42'!
fromHandFreehand: hand
	"Let the user draw a polygon, holding the mouse down, and ending
		by clicking within 5 of the first point..."
	| p1 poly pN opposite |
	Cursor crossHair showWhile:
		[[Sensor anyButtonPressed] whileFalse:
			[self currentWorld displayWorldSafely; runStepMethods].
		p1 := Sensor cursorPoint].
	opposite := (Display colorAt: p1) negated.
	opposite = Color transparent ifTrue: [opposite := Color red].
	(poly := LineMorph from: p1 to: p1 color: opposite width: 2) openInWorld.
	self currentWorld displayWorldSafely; runStepMethods.
	[Sensor anyButtonPressed] whileTrue:
			[pN := Sensor cursorPoint.
			(pN dist: poly vertices last) > 3 ifTrue:
				[poly setVertices: (poly vertices copyWith: pN).
				self currentWorld displayWorldSafely; runStepMethods]].
	hand position: Sensor cursorPoint.  "Done -- update hand pos"
	^ (poly setVertices: (poly vertices copyWith: p1)) delete! !

!CPolygonShape class methodsFor: 'instance creation' stamp: 'di 9/9/2000 11:41'!
shapeFromPen: penBlock color: c borderWidth: bw borderColor: bc
	"World addMorph: (CPolygonShape
		shapeFromPen: [:p | p hilbert: 4 side: 5. p go: 5.
						p hilbert: 4 side: 5. p go: 5]
		color: Color red borderWidth: 1 borderColor: Color black)"

	| pen |
	penBlock value: (pen := PenPointRecorder new).
	^ (self vertices: pen points asArray color: c borderWidth: bw borderColor: bc)
		quickFill: false! !

!CPolygonShape class methodsFor: 'parts bin' stamp: 'tk 11/16/2001 12:15'!
supplementaryPartsDescriptions
	^ {DescriptionForPartsBin
		formalName: 'Arrow'
		categoryList: #(' Basic 1 ' 'Graphics')
		documentation: 'A line with an arrowhead.  Shift-click to get handles and move the ends.'
		globalReceiverSymbol: #CPolygonShape
		nativitySelector: #arrowPrototype}
! !

!CPolygonShape class methodsFor: 'instance creation' stamp: 'di 9/7/2000 17:05'!
vertices: verts color: c borderWidth: bw borderColor: bc
	^ self basicNew beStraightSegments vertices: verts color: c borderWidth: bw borderColor: bc! !

!CPolygonShape methodsFor: 'menus' stamp: 'di 10/3/2000 07:17'!
addCustomMenuItems: aMenu hand: aHandMorph
	| lineName |
	super addCustomMenuItems: aMenu hand: aHandMorph.
	aMenu addUpdating: #handlesShowingPhrase target: self action: #showOrHideHandles.
	vertices size > 2 ifTrue:
		[aMenu addUpdating: #openOrClosePhrase target: self action: #makeOpenOrClosed.
		lineName := closed ifTrue: ['outline'] ifFalse: ['line'].
		self isCurve
			ifTrue: [aMenu add: 'make segmented ', lineName action: #toggleSmoothing]
			ifFalse: [aMenu add: 'make smooth ', lineName action: #toggleSmoothing]]. 
	aMenu add: 'specify dashed line' action:  #specifyDashedLine.

	self isOpen ifTrue:
		[aMenu addLine.
		aMenu addWithLabel: '---' enablement: [self isOpen and: [arrows ~~ #none]] action:  #makeNoArrows.
		aMenu addWithLabel: '-->' enablement: [self isOpen and: [arrows ~~ #forward]] action:  #makeForwardArrow.
		aMenu addWithLabel: '<--' enablement: [self isOpen and: [arrows ~~ #back]] action:  #makeBackArrow.
		aMenu addWithLabel: '<->' enablement: [self isOpen and: [arrows ~~ #both]] action:  #makeBothArrows.
		aMenu add: 'customize arrows' action: #customizeArrows:.
		(self hasProperty: #arrowSpec)
			ifTrue: [aMenu add: 'standard arrows' action: #standardArrows]].! !

!CPolygonShape methodsFor: 'editing' stamp: 'ar 3/17/2001 14:32'!
addHandles
	| handle newVert tri |
	self removeHandles.
	handles := OrderedCollection new.
	tri := Array with: 0@-4 with: 4@3 with: -3@3.
	vertices withIndexDo:
		[:vertPt :vertIndex |
		handle := EllipseMorph newBounds: (Rectangle center: vertPt extent: 8@8)
				color: Color yellow.
		handle on: #mouseMove send: #dragVertex:event:fromHandle:
				to: self withValue: vertIndex.
		handle on: #mouseUp send: #dropVertex:event:fromHandle:
				to: self withValue: vertIndex.
		self addMorph: handle.
		handles addLast: handle.
		(closed or: [vertIndex < vertices size]) ifTrue:
			[newVert := CPolygonShape
					vertices: (tri collect: [:p | p + (vertPt + (vertices atWrap: vertIndex+1) // 2)])
					color: Color green borderWidth: 1 borderColor: Color black.
			newVert on: #mouseDown send: #newVertex:event:fromHandle:
					to: self withValue: vertIndex.
			self addMorph: newVert.
			handles addLast: newVert]].
	smoothCurve ifTrue: [self updateHandles; layoutChanged].
	self changed! !

!CPolygonShape methodsFor: 'drawing' stamp: 'di 6/24/1998 14:36'!
areasRemainingToFill: aRectangle
	"Could be improved by quick check of inner rectangle"

	^ Array with: aRectangle! !

!CPolygonShape methodsFor: 'private' stamp: 'tk 1/28/2005 11:06'!
arrowBoundsAt: endPoint from: priorPoint 
	"Answer a triangle oriented along the line from priorPoint to endPoint."
	| d v angle wingBase arrowSpec length width |
	v := endPoint - priorPoint.
	angle := v degrees.
	d := myPrimCostume border width max: 1.
	arrowSpec := self valueOfProperty: #arrowSpec ifAbsent: [5@4].
	length := arrowSpec x abs.  width := arrowSpec y abs.
	wingBase := endPoint + (Point r: d * length degrees: angle + 180.0).
	arrowSpec x >= 0
		ifTrue: [^ {	endPoint.
					wingBase + (Point r: d * width degrees: angle + 125.0).
					wingBase + (Point r: d * width degrees: angle - 125.0) }]
		ifFalse: ["Negative length means concave base."
				^ {	endPoint.
					wingBase + (Point r: d * width degrees: angle + 125.0).
					wingBase.
					wingBase + (Point r: d * width degrees: angle - 125.0) }]! !

!CPolygonShape methodsFor: 'private' stamp: 'tk 1/28/2005 11:10'!
arrowForms
	"ArrowForms are computed only upon demand"
	arrowForms ifNotNil: [^ arrowForms].

	arrowForms := Array new.
	(closed or: [arrows == #none or: [vertices size < 2]]) ifTrue:
		[^ arrowForms].
	(arrows == #forward or: [arrows == #both]) ifTrue:
		[arrowForms := arrowForms copyWith:
			(self computeArrowFormAt: vertices last from: self nextToLastPoint)].
	(arrows == #back or: [arrows == #both]) ifTrue:
		[arrowForms := arrowForms copyWith:
			(self computeArrowFormAt: vertices first from: self nextToFirstPoint)].
	^ arrowForms! !

!CPolygonShape methodsFor: 'private' stamp: 'tk 1/27/2005 17:59'!
arrowForms: aValue
	"Modify the receiver's arrowForms"
	arrowForms := aValue! !

!CPolygonShape methodsFor: 'menu' stamp: 'sw 9/24/2002 19:23'!
arrowLength: aLength
	"Assumes that I have exactly two vertices"

	| theta horizontalOffset verticalOffset newTip delta |
	delta := vertices second - vertices first.
	theta := delta theta.
	horizontalOffset := aLength * (theta cos).
	verticalOffset := aLength * (theta sin).
	newTip := vertices first + (horizontalOffset @ verticalOffset).
	self verticesAt: 2 put: newTip! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 10/3/2000 09:09'!
arrowSpec: specPt
	"Specify a custom arrow for this line.
	specPt x abs gives the length of the arrow (point to base) in terms of borderWidth.
	If specPt x is negative, then the base of the arrow will be concave.
	specPt y abs gives the width of the arrow.
	The standard arrow is equivalent to arrowSpec: 5@4.
	See arrowBoundsAt:From: for details."

	self setProperty: #arrowSpec toValue: specPt.
	self computeBounds! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 17:59'!
arrows
	"Answer the arrows of the receiver"
	^arrows! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 17:59'!
arrows: aValue
	"Modify the receiver's arrows"
	arrows := aValue! !

!CPolygonShape methodsFor: 'initialization' stamp: 'tk 1/28/2005 11:02'!
beShapeFor: aCostume
	"Set myself up as a Shape in this costume"

	myPrimCostume := aCostume.
	self initialize.
	myPrimCostume shape: self.
	! !

!CPolygonShape methodsFor: 'initialization' stamp: 'di 9/8/2000 09:44'!
beSmoothCurve

	smoothCurve == true ifFalse:
		[smoothCurve := true.
		self computeBounds]! !

!CPolygonShape methodsFor: 'initialization' stamp: 'di 9/8/2000 09:45'!
beStraightSegments

	smoothCurve == false ifFalse:
		[smoothCurve := false.
		self computeBounds]! !

!CPolygonShape methodsFor: 'dashes' stamp: 'dgd 2/22/2003 18:55'!
borderDashOffset
	borderDashSpec size < 4 ifTrue: [^0.0].
	^(borderDashSpec fourth) asFloat! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 18:00'!
borderDashSpec
	"Answer the borderDashSpec of the receiver"
	^borderDashSpec! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 18:00'!
borderDashSpec: aValue
	"Modify the receiver's borderDashSpec"
	borderDashSpec := aValue! !

!CPolygonShape methodsFor: 'private' stamp: 'tk 1/28/2005 11:33'!
borderForm
	"A form must be created for drawing the border whenever the borderColor is translucent."

	| borderCanvas |
	borderForm ifNotNil: [^ borderForm].
	borderCanvas := (Display defaultCanvasClass extent: myPrimCostume bounds extent depth: 1)
		shadowColor: Color black.
	borderCanvas translateBy: myPrimCostume bounds topLeft negated
		during:[:tempCanvas| self drawBorderOn: tempCanvas].
	borderForm := borderCanvas form.
	self arrowForms do:
		[:f |  "Eliminate overlap between line and arrowheads if transparent."
		borderForm copy: f boundingBox from: f to: f offset - self position rule: Form erase].
	^ borderForm! !

!CPolygonShape methodsFor: 'private' stamp: 'tk 1/27/2005 18:00'!
borderForm: aValue
	"Modify the receiver's borderForm"
	borderForm := aValue! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/31/2005 13:58'!
borderStyle
	"Work around the borderWidth/borderColor pair"

	| style bordColor bordWidth |
	(bordColor := myPrimCostume border color) ifNil: [^BorderStyle default].
	(bordWidth := myPrimCostume border width) isZero ifTrue: [^BorderStyle default].
	style := myPrimCostume propertyValueAt: #pBorderStyle ifAbsent: [BorderStyle default].
	(bordWidth = style width and: 
			["Hah!! Try understanding this..."

			bordColor == style style or: 
					["#raised/#inset etc"

					#simple == style style and: [bordColor = style color]]]) 
		ifFalse: 
			[style := bordColor isColor 
				ifTrue: [BorderStyle width: bordWidth color: bordColor]
				ifFalse: [(BorderStyle perform: bordColor) width: bordWidth	"argh."].
			myPrimCostume propertyValueAt: #pBorderStyle put: style].
	^style trackColorFrom: self! !

!CPolygonShape methodsFor: 'access' stamp: 'tk 2/1/2005 23:36'!
cacheVertices
	"Bring the points into my local 'vertcies' variable.  May be fetched from my player, or may not!!"

	"all"
	verticesSelector ifNotNil: [^ vertices := myPrimCostume player perform: verticesSelector].

	"only first or last or both"
	vertexFirstSelector ifNotNil: [
		vertices at: 1 put: (myPrimCostume player perform: vertexFirstSelector)].
	vertexLastSelector ifNotNil: [
		vertices at: vertices size put: (myPrimCostume player perform: vertexLastSelector)].
	^ vertices! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 17:59'!
closed
	"Answer the closed of the receiver"
	^closed! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 2/4/2005 10:24'!
closed: aValue
	"closed tells if the polygon is a solid object, or just a jointed line.  true makes it a solid filled object.  false makes it a line or a curve.  When it is a line or a curve, color still controls its color, even though that is really the borderColor."

	| old |
	closed = aValue ifTrue: [^ self].
	old := closed.
	closed := aValue.
	old ifNil: [^ self].
	closed 
		ifTrue: ["became closed"
			"myPrimCostume color == myPrimCostume borderColor has been enforced"
			(myPrimCostume hasProperty: #oldBorderColor) ifTrue: [
				myPrimCostume border color: 
					(myPrimCostume propertyValueAt: #oldBorderColor)]]
		ifFalse: ["became open"
			myPrimCostume propertyValueAt: #oldBorderColor put: myPrimCostume border color.
			"myPrimCostume borderColor := myPrimCostume color will be enforced"].
! !

!CPolygonShape methodsFor: 'geometry' stamp: 'dgd 2/22/2003 14:14'!
closestPointTo: aPoint 
	| curvePoint closestPoint dist minDist |
	closestPoint := minDist := nil.
	self lineSegmentsDo: 
			[:p1 :p2 | 
			curvePoint := aPoint nearestPointOnLineFrom: p1 to: p2.
			dist := curvePoint dist: aPoint.
			(closestPoint isNil or: [dist < minDist]) 
				ifTrue: 
					[closestPoint := curvePoint.
					minDist := dist]].
	^closestPoint! !

!CPolygonShape methodsFor: 'smoothing' stamp: 'dgd 2/22/2003 14:14'!
coefficients
	"Compute an array for the coefficients.  This is copied from Flegal's old
	code in the Spline class."

	| length extras verts coefficients |
	curveState ifNotNil: [^curveState first].
	verts := closed 
				ifTrue: [vertices copyWith: vertices first]
				ifFalse: [vertices].
	length := verts size.
	extras := 0.
	coefficients := Array new: 8.
	1 to: 8 do: [:i | coefficients at: i put: (Array new: length + extras)].
	1 to: 5
		by: 4
		do: 
			[:k | 
			1 to: length
				do: 
					[:i | 
					(coefficients at: k) at: i
						put: (k = 1 
								ifTrue: [(verts at: i) x asFloat]
								ifFalse: [(verts at: i) y asFloat])].
			1 to: extras
				do: 
					[:i | 
					(coefficients at: k) at: length + i put: ((coefficients at: k) at: i + 1)].
			self 
				derivs: (coefficients at: k)
				first: (coefficients at: k + 1)
				second: (coefficients at: k + 2)
				third: (coefficients at: k + 3)].
	extras > 0 
		ifTrue: 
			[1 to: 8
				do: 
					[:i | 
					coefficients at: i put: ((coefficients at: i) copyFrom: 2 to: length + 1)]].
	curveState := { 
				coefficients.
				nil.
				nil}.
	self computeNextToEndPoints.
	^coefficients! !

!CPolygonShape methodsFor: 'private' stamp: 'dgd 2/22/2003 18:56'!
computeArrowFormAt: endPoint from: priorPoint 
	"Compute a triangle oriented along the line from priorPoint to  
	endPoint. Then draw those lines in a form and return that  
	form, with appropriate offset"

	| p1 pts box arrowForm bb origin |
	pts := self arrowBoundsAt: endPoint from: priorPoint.
	box := ((pts first rect: pts last) encompass: (pts second)) expandBy: 1.
	arrowForm := Form extent: box extent asIntegerPoint.
	bb := (BitBlt current toForm: arrowForm)
				sourceForm: nil;
				fillColor: Color black;
				combinationRule: Form over;
				width: 1;
				height: 1.
	origin := box topLeft.
	p1 := pts last - origin.
	pts do: 
			[:p | 
			bb drawFrom: p1 to: p - origin.
			p1 := p - origin].
	arrowForm convexShapeFill: Color black.
	^arrowForm offset: box topLeft! !

!CPolygonShape methodsFor: 'menu' stamp: 'tk 2/3/2005 11:58'!
computeBaseBounds
	"what box would enclose verticies and border and arrowheads?  Coordinate system of the verticies, which is local to this object."

	| box correction base |
	vertices ifNil: [^ 0@0 extent: 20@20].
	box := self curveBounds truncated.	"Both caller and curveBounds adding bounds?"
	self arrowForms do:
		[:f | box := (box merge: (f offset extent: f extent))].
	handles ifNotNil: [self updateHandles].	"set their position"
	base := myPrimCostume layoutInset + (myPrimCostume border width asPoint).
		"this will be added back by CPrimCost layoutBounds:"
	correction := box origin min: base.
	correction = base ifFalse: [
		vertices := vertices collect: [:vv | vv - correction "positive"].
		box := box translateBy: (base - correction)].
	^ box! !

!CPolygonShape methodsFor: 'menu' stamp: 'tk 2/3/2005 11:38'!
computeBounds
	"Called when a vertix moves.  (See older versions)"

	myPrimCostume layoutBounds: self computeBaseBounds.! !

!CPolygonShape methodsFor: 'smoothing' stamp: 'dgd 2/22/2003 14:15'!
computeNextToEndPoints
	| pointAfterFirst pointBeforeLast |
	pointAfterFirst := nil.
	self lineSegmentsDo: 
			[:p1 :p2 | 
			pointAfterFirst isNil ifTrue: [pointAfterFirst := p2 asIntegerPoint].
			pointBeforeLast := p1 asIntegerPoint].
	curveState at: 2 put: pointAfterFirst.
	curveState at: 3 put: pointBeforeLast! !

!CPolygonShape methodsFor: 'geometry testing' stamp: 'tk 1/28/2005 11:06'!
containsPoint: aPoint
	(super containsPoint: aPoint) ifFalse: [^ false].

	closed & myPrimCostume color isTransparent not ifTrue:
		[^ (self filledForm pixelValueAt: aPoint - myPrimCostume bounds topLeft + 1) > 0].

	self lineSegmentsDo:
		[:p1 :p2 |
		(aPoint onLineFrom: p1 to: p2 within: (3 max: myPrimCostume border width+1//2) asFloat)
				ifTrue: [^ true]].

	self arrowForms do:
		[:f | (f pixelValueAt: aPoint - f offset) > 0 ifTrue: [^ true]].

	^ false! !

!CPolygonShape methodsFor: 'private' stamp: 'tk 2/3/2005 11:35'!
curveBounds
	| curveBounds pointAfterFirst pointBeforeLast |
	smoothCurve ifFalse: 
		[^(Rectangle encompassing: vertices) expandBy: (myPrimCostume border width + 1) // 2].
		"allow more for acute point with a border?"

	"Compute the bounds from actual curve traversal, with leeway for border width.
	Also note the next-to-first and next-to-last points for arrow directions."
	curveState := nil.	"Force recomputation"
	curveBounds := vertices first corner: vertices last.
	pointAfterFirst := nil.
	self lineSegmentsDo: 
			[:p1 :p2 | 
			pointAfterFirst isNil 
				ifTrue: 
					[pointAfterFirst := p2 asIntegerPoint.
					curveBounds := curveBounds encompass: p1 asIntegerPoint].
			curveBounds := curveBounds encompass: p2 asIntegerPoint.
			pointBeforeLast := p1 asIntegerPoint].
	curveState at: 2 put: pointAfterFirst.
	curveState at: 3 put: pointBeforeLast.
	^curveBounds expandBy: (myPrimCostume border width + 1) // 2! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 17:59'!
curveState
	"Answer the curveState of the receiver"
	^curveState! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 17:59'!
curveState: aValue
	"Modify the receiver's curveState"
	curveState := aValue! !

!CPolygonShape methodsFor: 'menu' stamp: 'ar 10/3/2000 17:06'!
customizeArrows: evt
	| handle origin aHand |
	aHand := evt ifNil: [self primaryHand] ifNotNil: [evt hand].
	origin := aHand position.
	handle := HandleMorph new
		forEachPointDo:
			[:newPoint | handle removeAllMorphs.
			handle addMorph:
				(LineMorph from: origin to: newPoint color: Color black width: 1).
			self arrowSpec: (newPoint - origin) / 5.0]
		lastPointDo:
			[:newPoint | handle deleteBalloon.
			self halo doIfNotNil: [:halo | halo addHandles].].
	aHand attachMorph: handle.
	handle setProperty: #helpAtCenter toValue: true.
	handle showBalloon:
'Move cursor left and right
to change arrow length and style.
Move it up and down to change width.
Click when done.' hand: evt hand.
	handle startStepping! !

!CPolygonShape methodsFor: 'dashes' stamp: 'di 9/9/2000 09:20'!
dashedBorder: dashSpec
	"A dash spec is a 3- or 5-element array with
		{ length of normal border color.
		length of alternate border color.
		alternate border color.
		starting offset.
		amount to add to offset at each step }
	Starting offset is usually = 0, but changing it moves the dashes along the curve."

	borderDashSpec := dashSpec.
	self changed! !

!CPolygonShape methodsFor: 'smoothing' stamp: 'dgd 2/22/2003 14:16'!
derivs: a first: point1 second: point2 third: point3 
	"Compute the first, second and third derivitives (in coeffs) from
	the Points in this Path (coeffs at: 1 and coeffs at: 5)."

	| len v anArray |
	len := a size.
	len < 2 ifTrue: [^self].
	len > 2 
		ifTrue: 
			[v := Array new: len.
			v at: 1 put: 4.0.
			anArray := Array new: len.
			anArray at: 1 put: 6.0 * (a first - (a second * 2.0) + (a third)).
			2 to: len - 2
				do: 
					[:i | 
					v at: i put: 4.0 - (1.0 / (v at: i - 1)).
					anArray at: i
						put: 6.0 * ((a at: i) - ((a at: i + 1) * 2.0) + (a at: i + 2)) 
								- ((anArray at: i - 1) / (v at: i - 1))].
			point2 at: len - 1 put: (anArray at: len - 2) / (v at: len - 2).
			len - 2 to: 2
				by: 0 - 1
				do: 
					[:i | 
					point2 at: i
						put: ((anArray at: i - 1) - (point2 at: i + 1)) / (v at: i - 1)]].
	point2 at: 1 put: (point2 at: len put: 0.0).
	1 to: len - 1
		do: 
			[:i | 
			point1 at: i
				put: (a at: i + 1) - (a at: i) 
						- (((point2 at: i) * 2.0 + (point2 at: i + 1)) / 6.0).
			point3 at: i put: (point2 at: i + 1) - (point2 at: i)]! !

!CPolygonShape methodsFor: 'editing' stamp: 'ar 3/17/2001 14:30'!
dragVertex: ix event: evt fromHandle: handle
	| p |
	p := self isCurve
		ifTrue: [evt cursorPoint]
		ifFalse: [self griddedPoint: evt cursorPoint].
	handle position: p - (handle extent//2).
	self verticesAt: ix put: p.
! !

!CPolygonShape methodsFor: 'editing' stamp: 'ar 3/18/2001 17:28'!
dragVertex: arg1 fromHandle: arg2 vertIndex: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self dragVertex: arg1 event: arg2 fromHandle: arg3! !

!CPolygonShape methodsFor: 'drawing' stamp: 'tk 2/2/2005 00:08'!
drawArrowOn: aCanvas at: endPoint from: priorPoint 
	"Draw a triangle oriented along the line from priorPoint to  
	endPoint. Answer the wingBase."

	| pts spec wingBase |
	pts := self arrowBoundsAt: endPoint from: priorPoint.
	wingBase := pts size = 4 
				ifTrue: [pts third]
				ifFalse: [(pts copyFrom: 2 to: 3) average].
	spec := self valueOfProperty: #arrowSpec ifAbsent: [5 @ 4].
	spec x sign = spec y sign 
		ifTrue: [aCanvas drawPolygon: pts fillStyle: myPrimCostume fill]
		ifFalse: 
			[aCanvas 
				drawPolygon: pts
				fillStyle: Color transparent
				borderWidth: (myPrimCostume border width + 1) // 2
				borderColor: myPrimCostume border color].
	^wingBase! !

!CPolygonShape methodsFor: 'drawing' stamp: 'tk 1/28/2005 10:50'!
drawArrowsOn: aCanvas 
	"Answer (possibly modified) endpoints for border drawing"
	"ArrowForms are computed only upon demand"
	| array |
	(closed or: [arrows == #none or: [vertices size < 2]]) ifTrue: [^ self].
	"Nothing to do"
	myPrimCostume borderColor isColor ifFalse: [^ self].
	array := Array new: 2.
	"Prevent crashes for #raised or #inset borders"
	array at: 2 put: ((arrows == #forward
			or: [arrows == #both])
		ifTrue: [ self
				drawArrowOn: aCanvas
				at: vertices last
				from: self nextToLastPoint]
		ifFalse: [ vertices last ]).
	array at: 1 put: ((arrows == #back
			or: [arrows == #both])
		ifTrue: [self
				drawArrowOn: aCanvas
				at: vertices first
				from: self nextToFirstPoint]
		ifFalse: [ vertices first ]).
	^array! !

!CPolygonShape methodsFor: 'drawing' stamp: 'ar 11/26/2001 23:15'!
drawBorderOn: aCanvas 
	self
		drawClippedBorderOn: aCanvas
		usingEnds: (Array with: vertices first with: vertices last)! !

!CPolygonShape methodsFor: 'drawing' stamp: 'tk 1/28/2005 12:01'!
drawBorderOn: aCanvas usingEnds: anArray 
	"Display my border on the canvas."
	"NOTE: Much of this code is also copied in drawDashedBorderOn:  
	(should be factored)"
	| bigClipRect p1i p2i style |
	borderDashSpec
		ifNotNil: [^ self drawDashedBorderOn: aCanvas usingEnds: anArray].
	style := self borderStyle.
	bigClipRect := aCanvas clipRect expandBy: myPrimCostume border width + 1 // 2.
	self lineSegmentsDo: [:p1 :p2 | 
		p1i := p1 asIntegerPoint.
		p2i := p2 asIntegerPoint.
		(arrows ~= #none and: [closed not])
			ifTrue: ["Shorten line ends so as not to interfere with tip of arrow."
					((arrows == #back
								or: [arrows == #both])
							and: [p1 = vertices first])
						ifTrue: [p1i := anArray first asIntegerPoint].
					((arrows == #forward
								or: [arrows == #both])
							and: [p2 = vertices last])
						ifTrue: [p2i := anArray last asIntegerPoint]].
		(closed or: ["bigClipRect intersects: (p1i rect: p2i) optimized:"
			((p1i min: p2i) max: bigClipRect origin)
				<= ((p1i max: p2i) min: bigClipRect corner)])
				ifTrue: [style drawLineFrom: p1i to: p2i on: aCanvas]]! !

!CPolygonShape methodsFor: 'drawing' stamp: 'tk 1/28/2005 11:44'!
drawClippedBorderOn: aCanvas usingEnds: anArray 
	aCanvas clipBy: myPrimCostume bounds during:[:cc| self drawBorderOn: cc usingEnds: anArray].! !

!CPolygonShape methodsFor: 'drawing' stamp: 'nk 10/4/2000 12:23'!
drawDashedBorderOn: aCanvas 
	self
		drawDashedBorderOn: aCanvas
		usingEnds: (Array with: vertices first with: vertices last)! !

!CPolygonShape methodsFor: 'drawing' stamp: 'tk 1/28/2005 11:07'!
drawDashedBorderOn: aCanvas usingEnds: anArray 
	"Display my border on the canvas. NOTE: mostly copied from  
	drawBorderOn:"

	| lineColor bevel topLeftColor bottomRightColor bigClipRect p1i p2i segmentOffset |
	(myPrimCostume borderColor isNil 
		or: [myPrimCostume borderColor isColor and: 
				[myPrimCostume borderColor isTransparent]]) ifTrue: [^self].
	lineColor := myPrimCostume borderColor.
	bevel := false.
	"Border colors for bevelled effects depend on CW ordering of  
	vertices"
	myPrimCostume borderColor == #raised 
		ifTrue: 
			[topLeftColor := myPrimCostume fillColor lighter.
			bottomRightColor := myPrimCostume fillColor darker.
			bevel := true].
	myPrimCostume borderColor == #inset 
		ifTrue: 
			[topLeftColor := "owner colorForInsets" myPrimCostume fillColor darker.
			bottomRightColor := myPrimCostume fillColor lighter.
			bevel := true].
	bigClipRect := aCanvas clipRect expandBy: (myPrimCostume border width + 1) // 2.
	segmentOffset := self borderDashOffset.
	self lineSegmentsDo: 
			[:p1 :p2 | 
			p1i := p1 asIntegerPoint.
			p2i := p2 asIntegerPoint.
			(arrows ~= #none and: [closed not]) 
				ifTrue: 
					["Shorten line ends so as not to interfere with tip  
					of arrow."

					((arrows == #back or: [arrows == #both]) and: [p1 = vertices first]) 
						ifTrue: [p1i := anArray first asIntegerPoint].
					((arrows == #forward or: [arrows == #both]) and: [p2 = vertices last]) 
						ifTrue: [p2i := anArray last asIntegerPoint]].
			(closed or: 
					["bigClipRect intersects: (p1i rect: p2i) optimized:"

					((p1i min: p2i) max: bigClipRect origin) 
						<= ((p1i max: p2i) min: bigClipRect corner)]) 
				ifTrue: 
					[bevel 
						ifTrue: 
							[lineColor := (p1i quadrantOf: p2i) > 2 
										ifTrue: [topLeftColor]
										ifFalse: [bottomRightColor]].
					segmentOffset := aCanvas 
								line: p1i
								to: p2i
								width: myPrimCostume border width
								color: lineColor
								dashLength: borderDashSpec first
								secondColor: borderDashSpec third
								secondDashLength: borderDashSpec second
								startingOffset: segmentOffset]]! !

!CPolygonShape methodsFor: 'drawing' stamp: 'ar 11/26/2001 23:15'!
drawOn: aCanvas 
	"Display the receiver, a spline curve, approximated by straight 
	line segments."
	| array |
	vertices size < 1
		ifTrue: [self error: 'a polygon must have at least one point'].
	closed ifTrue:
		[aCanvas drawPolygon: self getVertices fillStyle: self fillStyle.
		aCanvas isShadowDrawing ifTrue: [^ self]].
	array := self drawArrowsOn: aCanvas.
	self drawClippedBorderOn: aCanvas usingEnds: array.
! !

!CPolygonShape methodsFor: 'drawing' stamp: 'tk 2/4/2005 10:54'!
drawOn: aCanvas in: drawingBounds
	"Either a closed filled polygon, or a jointed line segment (just the border).  In addition, the boundary is either smooth and curved (a spline), or it is a collection of straight line segments."
	| array |

	"Will be clipped.  Might want to go faster by using drawingBounds"
	self cacheVertices.	"fetch from player if necessary"
	vertices size < 1
		ifTrue: [self error: 'a polygon must have at least one point'].
	closed ifTrue:
		[aCanvas drawPolygon: self getVertices fillStyle: myPrimCostume fill.
		aCanvas isShadowDrawing ifTrue: [^ self]].
	array := self drawArrowsOn: aCanvas.
	(myPrimCostume border isNoBorder not and: [myPrimCostume border width > 0]) ifTrue: [
		closed
			ifTrue: [myPrimCostume border framePolygon: self getVertices on: aCanvas]
			ifFalse: ["when open, use the color as the borderColor"
				myPrimCostume border color: myPrimCostume fill color.
					"continually force it"
				myPrimCostume border framePolyline: self getVertices on: aCanvas].
		"add arrow heads here"
		array
		].! !

!CPolygonShape methodsFor: 'drawing' stamp: 'tk 1/28/2005 10:38'!
drawOnFormCanvas: aCanvas 
	"Display the receiver, a spline curve, approximated by straight line segments."

	vertices size < 1 ifTrue: [self error: 'a polygon must have at least one point'].
	closed & myPrimCostume fillColor isTransparent not
		ifTrue: [aCanvas stencil: self filledForm at: myPrimCostume bounds topLeft - 1 
					color: myPrimCostume fillColor].
	(myPrimCostume borderColor isColor and: [myPrimCostume borderColor isTranslucentColor])
		ifTrue: [aCanvas stencil: self borderForm at: myPrimCostume bounds topLeft
						color: myPrimCostume borderColor]
		ifFalse: [self drawBorderOn: aCanvas].
	self arrowForms do:
		[:f | aCanvas stencil: f at: f offset
			color: (myPrimCostume borderColor isColor 
					ifTrue: [myPrimCostume borderColor] 
					ifFalse: [myPrimCostume fillColor])]! !

!CPolygonShape methodsFor: 'drawing' stamp: 'tk 1/28/2005 11:07'!
drawPostscriptOn: aCanvas 
	"Display the receiver, a spline curve, approximated by straight 
	line segments."
	| array |
	vertices size < 1
		ifTrue: [self error: 'a polygon must have at least one point'].
	array := self drawArrowsOn: aCanvas.
	closed
		ifTrue: [aCanvas
				drawPolygon: self getVertices
				color: self color
				borderWidth: self border width
				borderColor: self borderColor]
		ifFalse: [self drawClippedBorderOn: aCanvas usingEnds: array].
! !

!CPolygonShape methodsFor: 'editing' stamp: 'ar 3/17/2001 14:31'!
dropVertex: ix event: evt fromHandle: handle
	| p |
	p := vertices at: ix.
	(((vertices atWrap: ix-1) dist: p) < 3 or:
		[((vertices atWrap: ix+1) dist: p) < 3])
		ifTrue: ["Drag a vertex onto its neighbor means delete"
				self setVertices: (vertices copyReplaceFrom: ix to: ix with: Array new)].
	evt shiftPressed
		ifTrue: [self removeHandles]
		ifFalse: [self addHandles "remove then add to recreate"]! !

!CPolygonShape methodsFor: 'editing' stamp: 'ar 3/18/2001 17:28'!
dropVertex: arg1 fromHandle: arg2 vertIndex: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self dropVertex: arg1 event: arg2 fromHandle: arg3! !

!CPolygonShape methodsFor: 'visual properties' stamp: 'tk 1/28/2005 11:38'!
fillStyle

	self isOpen
		ifTrue: [^ myPrimCostume borderColor  "easy access to line color from halo"]
		ifFalse: [^ myPrimCostume fill style]! !

!CPolygonShape methodsFor: 'menu' stamp: 'tk 1/28/2005 11:12'!
filledForm
	"Note: The filled form is actually 2 pixels bigger than bounds, and the point corresponding to this morphs' position is at 1@1 in the form.  This is due to the details of the fillig routines, at least one of which requires an extra 1-pixel margin around the outside.  Computation of the filled form is done only on demand."
	| bb origin |
	closed ifFalse: [^ filledForm := nil].
	filledForm ifNotNil: [^ filledForm].
	filledForm := Form extent: myPrimCostume bounds extent+2.

	"Draw the border..."
	bb := (BitBlt current toForm: filledForm) sourceForm: nil; fillColor: Color black;
			combinationRule: Form over; width: 1; height: 1.
	origin := myPrimCostume bounds topLeft asIntegerPoint-1.
	self lineSegmentsDo: [:p1 :p2 | bb drawFrom: p1 asIntegerPoint-origin
										to: p2 asIntegerPoint-origin].

	"Fill it in..."
	filledForm convexShapeFill: Color black.

	(myPrimCostume borderColor isColor and: [myPrimCostume borderColor isTranslucentColor]) ifTrue:
		["If border is stored as a form, then erase any overlap now."
		filledForm copy: self borderForm boundingBox from: self borderForm
			to: 1@1 rule: Form erase].

	^ filledForm! !

!CPolygonShape methodsFor: 'private' stamp: 'tk 1/27/2005 17:59'!
filledForm: aValue
	"Modify the receiver's filledForm"
	filledForm := aValue! !

!CPolygonShape methodsFor: 'private' stamp: 'di 9/7/2000 16:17'!
getVertices

	smoothCurve ifFalse: [^ vertices].

	"For curves, enumerate the full set of interpolated points"
	^ Array streamContents:
		[:s | self lineSegmentsDo: [:pt1 :pt2 | s nextPut: pt1]]! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 18:00'!
handles
	"Answer the handles of the receiver"
	^handles! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 18:00'!
handles: aValue
	"Modify the receiver's handles"
	handles := aValue! !

!CPolygonShape methodsFor: 'menu' stamp: 'sw 8/19/2000 15:17'!
handlesShowingPhrase
	^ self showingHandles
		ifTrue:
			['hide handles']
		ifFalse:
			['show handles']! !

!CPolygonShape methodsFor: 'initialization' stamp: 'tk 2/1/2005 22:05'!
initialize
"initialize the state of the receiver"

	| offset |
	vertices ifNil: [
		offset := 0@0.  "myPrimCostume ifNil: [0@0] ifNotNil: [myPrimCostume player position].	"	
		vertices := Array
				with: (5 @ 0) + offset
				with: (30 @ 15) + offset
				with: (0 @ 30) + offset].
	closed ifNil: [closed := true].
	smoothCurve ifNil: [smoothCurve := false].
	arrows ifNil: [arrows := #none].
	self computeBounds.! !

!CPolygonShape methodsFor: 'access' stamp: 'sw 8/25/2000 22:37'!
isClosed
	^ closed! !

!CPolygonShape methodsFor: 'access' stamp: 'di 9/7/2000 16:18'!
isCurve
	^ smoothCurve! !

!CPolygonShape methodsFor: 'access' stamp: 'jm 11/19/97 18:55'!
isOpen
	^ closed not! !

!CPolygonShape methodsFor: 'private' stamp: 'di 11/21/97 21:29'!
lineSegments
	| lineSegments |
	lineSegments := OrderedCollection new.
	self lineSegmentsDo: [:p1 :p2 | lineSegments addLast: (Array with: p1 with: p2)].
	^ lineSegments! !

!CPolygonShape methodsFor: 'smoothing' stamp: 'dgd 2/22/2003 14:21'!
lineSegmentsDo: endPointsBlock 
	"Emit a sequence of segment endpoints into endPointsBlock."

	| n t x y x1 x2 x3 y1 y2 y3 beginPoint endPoint cs |
	smoothCurve 
		ifFalse: 
			[beginPoint := nil.
			vertices do: 
					[:vert | 
					beginPoint ifNotNil: [endPointsBlock value: beginPoint value: vert].
					beginPoint := vert].
			(closed or: [vertices size = 1]) 
				ifTrue: [endPointsBlock value: beginPoint value: vertices first].
			^self].

	"For curves we include all the interpolated sub segments."
	vertices size < 1 ifTrue: [^self].
	cs := self coefficients.
	beginPoint := (x := cs first first) @ (y := cs fifth first).
	1 to: cs first size - 1
		do: 
			[:i | 
			"taylor series coefficients"

			x1 := cs second at: i.
			y1 := cs sixth at: i.
			x2 := (cs third at: i) / 2.0.
			y2 := (cs seventh at: i) / 2.0.
			x3 := (cs fourth at: i) / 6.0.
			y3 := ((cs eighth) at: i) / 6.0.
			"guess n"
			n := 5 
						max: (((x2 abs + y2 abs) * 2.0 + (cs third at: i + 1) abs 
								+ (cs seventh at: i + 1) abs) / 100.0) 
								rounded.
			1 to: n - 1
				do: 
					[:j | 
					t := j asFloat / n.
					endPoint := (((x3 * t + x2) * t + x1) * t + x) 
								@ (((y3 * t + y2) * t + y1) * t + y).
					endPointsBlock value: beginPoint value: endPoint.
					beginPoint := endPoint].
			endPoint := (x := cs first at: i + 1) @ (y := cs fifth at: i + 1).
			endPointsBlock value: beginPoint value: endPoint.
			beginPoint := endPoint]! !

!CPolygonShape methodsFor: 'caching' stamp: 'di 11/13/97 15:16'!
loadCachedState
	"Prepare for fast response -- next page of a book?"
	self filledForm.
	self arrowForms! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 6/9/97 21:32'!
makeBackArrow
	arrows := #back.
	self computeBounds! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 6/9/97 21:32'!
makeBothArrows
	arrows := #both.
	self computeBounds! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 8/20/2000 14:27'!
makeClosed
	closed := true.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 6/9/97 21:32'!
makeForwardArrow
	arrows := #forward.
	self computeBounds! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 6/9/97 21:32'!
makeNoArrows
	arrows := #none.
	self computeBounds! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 8/20/2000 14:27'!
makeOpen
	closed := false.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!CPolygonShape methodsFor: 'access' stamp: 'sw 8/23/2000 16:16'!
makeOpenOrClosed
	"toggle the open/closed status of the receiver"
	closed ifTrue: [self makeOpen] ifFalse: [self makeClosed]! !

!CPolygonShape methodsFor: 'geometry' stamp: 'dgd 2/22/2003 18:56'!
merge: aPolygon 
	"Expand myself to enclose the other polygon.  (Later merge overlapping or disjoint in a smart way.)  For now, the two polygons must share at least two vertices.  Shared vertices must come one after the other in each polygon.  Polygons must not overlap."

	| shared mv vv hv xx |
	shared := vertices select: [:mine | aPolygon vertices includes: mine].
	shared size < 2 ifTrue: [^nil].	"not sharing a segment"
	mv := vertices asOrderedCollection.
	[shared includes: mv first] whileFalse: 
			["rotate them"

			vv := mv removeFirst.
			mv addLast: vv].
	hv := aPolygon vertices asOrderedCollection.
	[mv first = hv first] whileFalse: 
			["rotate him until same shared vertex is first"

			vv := hv removeFirst.
			hv addLast: vv].
	[shared size > 2] whileTrue: 
			[shared := shared asOrderedCollection.
			(self 
				mergeDropThird: mv
				in: hv
				from: shared) ifNil: [^nil]].
	"works by side effect on the lists"
	(mv second) = hv last 
		ifTrue: 
			[mv
				removeFirst;
				removeFirst.
			^self setVertices: (hv , mv) asArray].
	(hv second) = mv last 
		ifTrue: 
			[hv
				removeFirst;
				removeFirst.
			^self setVertices: (mv , hv) asArray].
	(mv second) = (hv second) 
		ifTrue: 
			[hv removeFirst.
			mv remove: (mv second).
			xx := mv removeFirst.
			^self setVertices: (hv , (Array with: xx) , mv reversed) asArray].
	mv last = hv last 
		ifTrue: 
			[mv removeLast.
			hv removeFirst.
			^self setVertices: (mv , hv reversed) asArray].
	^nil! !

!CPolygonShape methodsFor: 'geometry' stamp: 'dgd 2/22/2003 18:57'!
mergeDropThird: mv in: hv from: shared 
	"We are merging two polygons.  In this case, they have at least three identical shared vertices.  Make sure they are sequential in each, and drop the middle one from vertex lists mv, hv, and shared.  First vertices on lists are identical already."

	"know (mv first = hv first)"

	| mdrop vv |
	(shared includes: (mv at: mv size - 2)) 
		ifTrue: [(shared includes: mv last) ifTrue: [mdrop := mv last]]
		ifFalse: 
			[(shared includes: mv last) 
				ifTrue: [(shared includes: mv second) ifTrue: [mdrop := mv first]]].
	(shared includes: (mv third)) 
		ifTrue: [(shared includes: mv second) ifTrue: [mdrop := mv second]].
	mdrop ifNil: [^nil].
	mv remove: mdrop.
	hv remove: mdrop.
	shared remove: mdrop.
	[shared includes: mv first] whileFalse: 
			["rotate them"

			vv := mv removeFirst.
			mv addLast: vv].
	[mv first = hv first] whileFalse: 
			["rotate him until same shared vertex is first"

			vv := hv removeFirst.
			hv addLast: vv]! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 17:58'!
myPrimCostume
	"Answer the myPrimCostume of the receiver"
	^myPrimCostume! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 17:58'!
myPrimCostume: aValue
	"Modify the receiver's myPrimCostume"
	myPrimCostume := aValue! !

!CPolygonShape methodsFor: 'editing' stamp: 'sw 9/25/2002 01:16'!
newVertex: ix event: evt fromHandle: handle
	"Insert a new vertex and fix everything up!! Install the drag-handle of the new vertex as recipient of further mouse events."

	| pt |
	(self hasProperty: #noNewVertices) ifFalse:
		[pt := evt cursorPoint.
		self setVertices: (vertices copyReplaceFrom: ix + 1 to: ix with: (Array with: pt)).
		evt hand newMouseFocus: (handles at: ((ix + 1) * 2) - 1)]
! !

!CPolygonShape methodsFor: 'editing' stamp: 'ar 3/18/2001 17:28'!
newVertex: arg1 fromHandle: arg2 afterVert: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self newVertex: arg1 event: arg2 fromHandle: arg3! !

!CPolygonShape methodsFor: 'smoothing' stamp: 'dgd 2/22/2003 18:57'!
nextToFirstPoint
	"For arrow direction"

	smoothCurve 
		ifTrue: 
			[curveState ifNil: [self coefficients].
			^curveState second]
		ifFalse: [^vertices second]! !

!CPolygonShape methodsFor: 'smoothing' stamp: 'dgd 2/22/2003 18:58'!
nextToLastPoint
	"For arrow direction"

	smoothCurve 
		ifTrue: 
			[curveState ifNil: [self coefficients].
			^curveState third]
		ifFalse: [^vertices at: vertices size - 1]! !

!CPolygonShape methodsFor: 'access' stamp: 'di 9/7/2000 16:53'!
openOrClosePhrase
	| curveName |
	curveName := self isCurve ifTrue: ['curve'] ifFalse: ['polygon'].
	^ closed
		ifTrue: ['make open ' , curveName]
		ifFalse: ['make closed ' , curveName]! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 9/7/2000 13:10'!
quickFill: ignored! !

!CPolygonShape methodsFor: 'caching' stamp: 'di 9/4/2000 13:36'!
releaseCachedState
	super releaseCachedState.
	filledForm := nil.
	arrowForms := nil.
	borderForm := nil.
	curveState := nil.
	(self hasProperty: #flex) ifTrue:
		[self removeProperty: #unflexedVertices;
			removeProperty: #flex].
! !

!CPolygonShape methodsFor: 'menu' stamp: 'tk 9/2/97 16:04'!
removeHandles
	"tk 9/2/97 allow it to be called twice (when nil already)"

	handles ifNotNil: [
		handles do: [:h | h delete].
		handles := nil].! !

!CPolygonShape methodsFor: 'private' stamp: 'di 9/8/2000 10:36'!
setVertices: newVertices
	vertices := newVertices.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!CPolygonShape methodsFor: 'menu' stamp: 'sw 8/19/2000 15:16'!
showOrHideHandles
	self showingHandles
		ifTrue:	[self removeHandles]
		ifFalse:	[self addHandles]! !

!CPolygonShape methodsFor: 'menu' stamp: 'sw 8/19/2000 15:17'!
showingHandles
	^ handles notNil! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 17:59'!
smoothCurve
	"Answer the smoothCurve of the receiver"
	^smoothCurve! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 1/27/2005 17:59'!
smoothCurve: aValue
	"Modify the receiver's smoothCurve"
	smoothCurve := aValue! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 9/9/2000 11:38'!
specifyDashedLine

	| executableSpec newSpec |
	executableSpec := FillInTheBlank
		request:
'Enter a dash specification as
{ major dash length. minor dash length. minor dash color }
The major dash will have the normal border color.
A blank response will remove the dash specification.
[Note: You may give 5 items as, eg, {10. 5. Color white. 0. 3}
where the 4th ityem is zero, and the 5th is the number of pixels
by which the dashes will move in each step of animation]'
		initialAnswer: '{ 10. 5. Color red }'.
	executableSpec isEmpty ifTrue:
		[^ self stopStepping; dashedBorder: nil].
	newSpec := [Compiler evaluate: executableSpec] ifError:
		[^ self stopStepping; dashedBorder: nil].
	newSpec first isNumber & newSpec second isNumber & newSpec third isColor ifFalse:
		[^ self stopStepping; dashedBorder: nil].
	newSpec size = 3 ifTrue:
		[^ self stopStepping; dashedBorder: newSpec].
	(newSpec size = 5 and: [newSpec fourth isNumber & newSpec fifth isNumber]) ifTrue:
		[^ self dashedBorder: newSpec; startStepping].
! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 10/3/2000 07:12'!
standardArrows

	self removeProperty: #arrowSpec.
	self computeBounds! !

!CPolygonShape methodsFor: 'stepping and presenter' stamp: 'dgd 2/22/2003 18:58'!
step
	borderDashSpec ifNil: [^super step].
	borderDashSpec size < 5 ifTrue: [^super step].

	"Only for dashed lines with creep"
	borderDashSpec at: 4 put: (borderDashSpec fourth) + borderDashSpec fifth.
	self changed.
	^super step! !

!CPolygonShape methodsFor: 'testing' stamp: 'di 9/9/2000 09:24'!
stepTime

	^ 100! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 8/20/2000 14:31'!
toggleHandles

	handles ifNil: [self addHandles] ifNotNil: [self removeHandles].

! !

!CPolygonShape methodsFor: 'menu' stamp: 'di 9/7/2000 15:43'!
toggleSmoothing

	smoothCurve := smoothCurve not.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!CPolygonShape methodsFor: 'geometry' stamp: 'ar 10/6/2000 15:40'!
transformedBy: aTransform
	self setVertices: (self vertices collect:[:v| aTransform localPointToGlobal: v])! !

!CPolygonShape methodsFor: 'menu' stamp: 'sw 9/24/2002 18:18'!
unrotatedWidth: aWidth
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"

	self borderWidth: aWidth! !

!CPolygonShape methodsFor: 'editing' stamp: 'nk 10/4/2000 12:56'!
updateHandles
	| newVert oldVert midPts nextVertIx tweens |
	smoothCurve
		ifTrue: [
			handles first center: vertices first.
			handles last center: vertices last.
			midPts := OrderedCollection new.
			nextVertIx := 2.
			tweens := OrderedCollection new.
			self
				lineSegmentsDo: [:p1 :p2 | 
					tweens addLast: p2 asIntegerPoint.
					p2
							= (vertices atWrap: nextVertIx)
						ifTrue: ["Found endPoint."
							midPts addLast: (tweens at: tweens size // 2)
									+ (tweens at: tweens size + 1 // 2) // 2.
							tweens := OrderedCollection new.
							nextVertIx := nextVertIx + 1]].
			midPts
				withIndexDo: [:midPt :vertIndex | (closed
							or: [vertIndex < vertices size])
						ifTrue: [newVert := handles at: vertIndex * 2.
							newVert position: midPt - (newVert extent // 2)]]]
		ifFalse: [vertices
				withIndexDo: [:vertPt :vertIndex | 
					oldVert := handles at: vertIndex * 2 - 1.
					oldVert position: vertPt - (oldVert extent // 2).
					(closed
							or: [vertIndex < vertices size])
						ifTrue: [newVert := handles at: vertIndex * 2.
							newVert position: vertPt
									+ (vertices atWrap: vertIndex + 1) - newVert extent // 2 + (1 @ -1)]]]! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 2/1/2005 23:38'!
vertexFirstSelector: aValue
	"the selector to send to the Player to get the first vertex.  nil means our local vertices."
	vertexFirstSelector := aValue! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 2/1/2005 23:39'!
vertexLastSelector: aValue
	"the selector to send to the Player to get the last vertex.  nil means our local vertices."
	vertexLastSelector := aValue! !

!CPolygonShape methodsFor: 'access' stamp: 'tk 1/27/2005 17:58'!
vertices
	"Answer the vertices of the receiver"
	^vertices! !

!CPolygonShape methodsFor: 'access' stamp: 'tk 1/27/2005 17:58'!
vertices: aValue
	"Modify the receiver's vertices"
	vertices := aValue! !

!CPolygonShape methodsFor: 'initialization' stamp: 'tk 1/28/2005 10:43'!
vertices: verts color: aColor borderWidth: borderWidthInteger borderColor: anotherColor 

	vertices := verts.
	myPrimCostume fillColor: aColor.
	myPrimCostume borderWidth: borderWidthInteger.
	myPrimCostume borderColor: anotherColor.
	closed := vertices size > 2.
	arrows := #none.
	self computeBounds! !

!CPolygonShape methodsFor: 'editing' stamp: 'di 9/8/2000 10:39'!
verticesAt: ix put: newPoint
	vertices at: ix put: newPoint.
	self computeBounds! !

!CPolygonShape methodsFor: 'accessing' stamp: 'tk 2/1/2005 23:38'!
verticesSelector: aValue
	"the selector to send to the Player to get the vertices.  nil means we always use our local copy."
	verticesSelector := aValue! !

!CPolygonShape methodsFor: 'testing' stamp: 'dgd 2/22/2003 18:58'!
wantsSteps
	super wantsSteps ifTrue: [^true].

	"For crawling ants effect of dashed line."
	borderDashSpec ifNil: [^false].
	^borderDashSpec size = 5 and: [(borderDashSpec fifth) > 0]! !

Object subclass: #CPrimitiveCostume
	instanceVariableNames: 'container contents bounds fill border flags properties'
	classVariableNames: 'DefaultBorder DefaultFill FlagCached FlagClipEnabled FlagCustomDraw FlagExcessBounds FlagHasFullBounds FlagHasGraphic FlagHasGrid FlagHasHighlights FlagHasMatrix FlagHasOverlay FlagHasPivot FlagHasScroller FlagHasShape FlagHasText FlagHidden FlagIntrinsicFullBounds FlagInvalidated FlagInvalidationSuppressed FlagLayoutNeeded FlagLocked FlagRegistry FlagRoundedCorners'
	poolDictionaries: ''
	category: 'Tweak-Basic-Costumes'!

!CPrimitiveCostume commentStamp: 'ar 9/21/2003 20:11' prior: 0!
TODO:
	* add player as inst var
	* add text as inst var!

CPrimitiveCostume subclass: #CHandCostume
	instanceVariableNames: 'mouseCursor'
	classVariableNames: ''
	poolDictionaries: 'EventSensorConstants'
	category: 'Tweak-Basic-Worlds'!

!CHandCostume commentStamp: '<historical>' prior: 0!
This class is obsolete. It will be removed in the future.!

!CHandCostume methodsFor: 'mouse cursor' stamp: 'ar 6/28/2005 18:57'!
mouseCursor
	^super mouseCursor ifNil:[Cursor normal].! !

!CHandCostume methodsFor: 'mouse cursor' stamp: 'ar 6/28/2005 18:57'!
showHardwareCursor
	^false! !

!CPrimitiveCostume class methodsFor: 'instance creation' stamp: 'ar 5/13/2005 13:24'!
asCostumeFor: aPlayer
	"Create a new costume for the given player"
	^(super new) initialize; player: aPlayer; yourself! !

!CPrimitiveCostume class methodsFor: 'accessing' stamp: 'ar 3/1/2005 09:32'!
eventsGenerated
	^#(
		mouseDown mouseEnter mouseMove mouseStillDown mouseLeave mouseUp 
		mouseDrag dragEnter dragMove dragLeave dragDrop
		redButtonDown redButtonMove redButtonUp 
		yellowButtonDown yellowButtonMove yellowButtonUp 
		blueButtonDown blueButtonMove blueButtonUp 
		mouseClick mouseDoubleClick
		layoutChanged
		opened closed
	)! !

!CPrimitiveCostume class methodsFor: 'class initialization' stamp: ''!
flagRegistry
	^FlagRegistry ifNil:[FlagRegistry := Array new: SmallInteger maxVal highBit].! !

!CPrimitiveCostume class methodsFor: 'class initialization' stamp: 'ar 6/25/2005 17:03'!
initialize
	"CPrimitiveCostume initialize"
	DefaultFill := CCostumeFill new.
	DefaultBorder := CCostumeBorder new.
	FlagHidden := self requestFlag: #hidden.
	FlagCustomDraw := self requestFlag: #customDraw.
	FlagInvalidationSuppressed := self requestFlag: #invalidationSuppressed.
	FlagExcessBounds := self requestFlag: #excessBounds.
	FlagLocked := self requestFlag: #locked.
	FlagLayoutNeeded := self requestFlag: #layoutNeeded.
	FlagCached := self requestFlag: #cacheEnabled.
	FlagHasScroller := self requestFlag: #hasScroller.
	FlagClipEnabled := self requestFlag: #clipEnabled.
		FlagHasMatrix := self requestFlag: #hasMatrix. "OBSOLETE"
	FlagHasHighlights := self requestFlag: #hasHighlights.
	FlagRoundedCorners := self requestFlag: #hasRoundedCorners.
	FlagHasGraphic := self requestFlag: #hasGraphic.
	FlagHasText := self requestFlag: #hasText.
	FlagHasShape := self requestFlag: #hasShape.
	FlagHasOverlay := self requestFlag: #hasOverlay.
	FlagHasGrid := self requestFlag: #hasGrid.
	FlagInvalidated := self requestFlag: #invalidated.
	FlagHasPivot := self requestFlag: #hasPivot.

	"FlagHasFullBounds: OBSOLETE.
		If present, it indicates that we cache the receiver's fullBounds in the #fullBounds
		property. If absent, no fullBounds are cached."
		FlagHasFullBounds := self requestFlag: #hasFullBounds. "OBSOLETE"

	"Define a mask when 'bounds' define accurate full bounds"
	FlagIntrinsicFullBounds := 0. "OBSOLETE-OBSOLETE-OBSOLETE"
	"If there's a scroller it doesn't (due to scroll offset)"
	FlagIntrinsicFullBounds := FlagIntrinsicFullBounds bitOr: FlagHasScroller.
	"If children exceed the receiver it doesn't"
	FlagIntrinsicFullBounds := FlagIntrinsicFullBounds bitOr: FlagExcessBounds.
	"If a layout is needed, it doesn't"
	FlagIntrinsicFullBounds := FlagIntrinsicFullBounds bitOr: FlagLayoutNeeded.
	"If it's got a transform it doesn't"
	FlagIntrinsicFullBounds := FlagIntrinsicFullBounds bitOr: FlagHasMatrix.
	"If it's got highlights it doesn't (they may exceed beyound bounds)"
	FlagIntrinsicFullBounds := FlagIntrinsicFullBounds bitOr: FlagHasHighlights.
! !

!CPrimitiveCostume class methodsFor: 'instance creation' stamp: 'ar 1/30/2005 11:50'!
new
	^self basicNew initialize! !

!CPrimitiveCostume class methodsFor: 'class initialization' stamp: ''!
releaseFlag: flagName
	"Release an entry in Costume's flags."
	"@@@@: TODO. Clean up existing flag values in costumes. :@@@@"
	| registry flag |
	registry := self flagRegistry.
	1 to: registry size do:[:i|
		flag := registry at: i.
		flag == flagName ifTrue:[registry at: i put: nil].
	].! !

!CPrimitiveCostume class methodsFor: 'class initialization' stamp: ''!
requestFlag: flagName				 
	"Request an entry in Costume's flags. Return the bit assigned 
	for this flag or nil if no space is available"
	| registry flag flagIndex |
	registry := self flagRegistry.
	flagIndex := nil.
	1 to: registry size do:[:i|
		flag := registry at: i.
		flag == flagName ifTrue:[^1 bitShift: (i-1)].
		flag ifNil:[flagIndex ifNil:[flagIndex := i]].
	].
	flagIndex ifNotNil:[registry at: flagIndex put: flagName].
	^1 bitShift: (flagIndex-1)! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: 'ar 3/24/2004 03:11'!
abruptly: x
	x > 1.0 ifTrue:[^1.0] ifFalse:[^x]! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'ar 4/2/2004 14:27'!
add: aCostume
	self addLast: aCostume.! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'tk 1/4/2005 13:06'!
add: newObject after: oldObject				 
	"Add the argument, newObject, as an element of the receiver.  Put it in the sequence just after oldObject."
	self add: newObject at: (self indexOf: oldObject)+1! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'ar 3/16/2004 12:36'!
add: aCostume at: index	
	^self privateAdd: aCostume beforeIndex: index! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'ar 12/18/2002 14:15'!
add: aCostume atIndex: index				 	 
	^self privateAdd: aCostume atIndex: index! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'tk 1/4/2005 13:01'!
add: newObject before: oldObject				 	 
	"Add the argument, newObject, as an element of the receiver. Put it in the sequence just preceding oldObject.  Answer newObject."
	self add: newObject at: (self indexOf: oldObject ifAbsent:[1])! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'ar 12/7/2002 22:26'!
add: aCostume frame: relFrame
	| frame |
	frame := CLayoutFrame new.
	frame 
		leftFraction: relFrame left; 
		rightFraction: relFrame right; 
		topFraction: relFrame top; 
		bottomFraction: relFrame bottom.
	self add: aCostume fullFrame: frame.! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'ar 12/7/2002 22:26'!
add: aCostume frame: relFrame offsets: offsets
	| frame |
	frame := CLayoutFrame new.
	frame 
		leftFraction: relFrame left offset: offsets left;
		rightFraction: relFrame right offset: offsets right;
		topFraction: relFrame top offset: offsets top;
		bottomFraction: relFrame bottom offset: offsets bottom.
	self add: aCostume fullFrame: frame.! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'ar 3/21/2004 13:39'!
add: aCostume fullFrame: aLayoutFrame
	aCostume layoutFrame: aLayoutFrame.
	aCostume hResizing: #spaceFill; vResizing: #spaceFill.
	self add: aCostume.! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'tk 1/4/2005 13:03'!
addAll: aCollection					 	 
	"Add all the costumes in aCollection after the current elements in this object."
	self privateReplaceFrom: contents size+1 to: contents size 
		with: (aCollection collect:[:each| each asPrimCostume]).! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'tk 1/4/2005 13:04'!
addAll: aCollection after: prevCostume					 	 	 
	"Add all the costumes in aCollection after prevCostume, which is a element in me."
	^self addAll: aCollection 
		at: (contents indexOf: prevCostume ifAbsent:[contents size])+1.
! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'tk 1/4/2005 13:07'!
addAll: aCollection at: index					 	 
	"Make all the costumes in aCollection be elements of me.  Put them after the current element whose sequence number is index."
	self privateReplaceFrom: index to: index-1 
		with: (aCollection collect:[:each| each asPrimCostume]).! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'tk 1/4/2005 13:08'!
addFirst: aCostume
	"Add newObject, and make it be the first element of the receiver."
	self add: aCostume at: 1! !

!CPrimitiveCostume methodsFor: 'highlights' stamp: 'ar 3/24/2004 02:00'!
addHighlight: aHighlight
	"Add the given highlight to the receiver's list of highlights.
	propertyName describes the particular name of this highlight such as
		* mouseDown, mouseOver, drop, grab (UI highlights)
		* selection (widget highlights)
		* dropShadow (static highlights)
	"
	| list priority index highlight aCostume |
	(highlight := aHighlight) ifNil:[^self].
	aCostume := highlight costume.
	aCostume ifNotNil:[
		aCostume == self ifFalse:[highlight := highlight copy].
	].
	self removeHighlight: highlight name.
	highlight costume: self.
	priority := highlight priority.
	index := 1. "default insertion index"
	list := self highlights.
	"this should use binary search but I'm too lazy to do it now"
	(list size = 0 or:[list first priority >= priority]) ifFalse:[
		list last priority <= priority ifTrue:[index := list size + 1].
		index = 1 ifTrue:[ "e.g., above didn't work"
			1 to: list size do:[:i|
				(list at: i) priority < priority ifTrue:[index := i+1]]]].
	list := list copyWith: highlight before: index.
	self highlights: list.
	"Activate highlight if it is active"
	highlight changed.
	highlight isActive ifTrue:[highlight highlight].! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'tk 1/4/2005 13:08'!
addLast: aCostume
	"Add newObject to the end of the elements in me."
	self add: aCostume at: contents size+1! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/23/2004 23:06'!
adjustLayoutBounds					 
	"Adjust the receivers bounds depending on the resizing strategy imposed"
	| hFit vFit clip box myExtent extent partsBox |
	self ignoreShrinkWrapConstraints ifTrue:[^self].
	partsBox := self contentsBounds.
	clip := self clipping.
	hFit := self hResizing.
	vFit := self vResizing.
	(self localBounds containsRect: partsBox)
		ifFalse:[flags := flags bitOr: FlagExcessBounds].
	((hFit == #rigid or:[hFit == #spaceFill and:[clip]]) and:[
		(vFit == #rigid or:[vFit == #spaceFill and:[clip]])])
			ifTrue:[^self].
	box := self layoutBounds.
	myExtent := box extent.
	"Something is definitely screwed up here... but what???"
	extent := partsBox extent.
	"extent := partsBox corner - box origin."
	hFit == #shrinkWrap ifTrue:[myExtent := extent x @ myExtent y].
	vFit == #shrinkWrap ifTrue:[myExtent := myExtent x @ extent y].
	clip ifFalse:[
		hFit == #spaceFill ifTrue:[myExtent := (extent x max: myExtent x) @ myExtent y].
		vFit == #spaceFill ifTrue:[myExtent := myExtent x @ (extent y max: myExtent y)].
	].
	"Make sure we don't get smaller than minWidth/minHeight"
	myExtent x < self minWidth ifTrue:[
		myExtent := (myExtent x max: 
			(self minWidth - self localBounds width + self layoutBounds width)) @ myExtent y].
	myExtent y < self minHeight ifTrue:[
		myExtent := myExtent x @ (myExtent y max:
			(self minHeight - self localBounds height + self layoutBounds height))].
	self layoutBounds: (box origin extent: myExtent).
	(self localBounds containsRect: partsBox)
		ifTrue:[flags := flags bitClear: FlagExcessBounds]
		ifFalse:[flags := flags bitOr: FlagExcessBounds].! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 3/24/2004 03:17'!
align: aPoint1 with: aPoint2
	"Translate by aPoint2 - aPoint1."
	^self moveBy: (aPoint2 - aPoint1)! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 12/8/2003 21:08'!
allContainersDo: aBlock
	"Evaluate aBlock with all containers of the receiver"
	| parent |
	parent := container.
	[parent == nil] whileFalse:[
		aBlock value: parent.
		parent := parent container].! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/24/2004 03:14'!
allocateBorder
	border isNoBorder ifTrue:[border := CCostumeBorder forCostume: self].
	^border! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/24/2004 03:15'!
allocateFill
	fill isNoFill ifTrue:[fill := CCostumeFill forCostume: self].
	^fill! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/27/2005 14:45'!
angle
	"Answer the heading of the receiver, e.g., 
	its direction measured counter-clockwise from east."
	| deg |
	deg := self direction degrees negated \\ 360.0.
	^deg! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/27/2005 14:44'!
angle: aNumber
	"Indicate the angle (in degrees) the receiver is rotated by"
	self rotateBy: (aNumber - self angle) negated. "angle is ccw"
	^aNumber! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 6/24/2005 21:27'!
areasRemainingToFill: aRectangle
	"Answer the damage regions needed to be restored after the receiver has been drawn in aRectangle."
	| outer box rects cRect |
	self userDraw 
		ifTrue: [^ Array with: aRectangle].
	"translucent fills need to restore background first"
	self hasTranslucentFill
		ifTrue: [^ Array with: aRectangle].
	"transformed objects need to restore background first"
	self transform isPureTranslation
		ifFalse:[^ Array with: aRectangle].
	outer := box := self bounds.
	"inset by border if translucent border is present"
	"bw := self borderWidth.
	(bw > 0 and:[self borderColor isTranslucent]) 
		ifTrue:[box := box insetBy: bw]."
	border isTranslucent ifTrue:[box := border insetBounds: box].
	"if no rounded corners take what's left"
	self wantsRoundedCorners 
		ifFalse: [^aRectangle areasOutside: box].
	outer == box ifFalse:[
		"don't spend more work on translucent borders"
		box := box intersect: (CCornerRounder rectWithinCornersOf: outer).
		^aRectangle areasOutside: box].
	"Eliminate most of the damage for the common case
	when just the receiver is being redrawn."
	aRectangle = box
		ifTrue:[^box cornerRectsOfSize: CCornerRounder cornerSize].
	"Otherwise take the area outside of the receiver"
	rects := WriteStream on: (Array new: 8).
	"aRectangle areasOutside: box into: rects."
	rects nextPutAll: (aRectangle areasOutside: box).
	"And the corner rectangles (if any)"
	(box cornerRectsOfSize: CCornerRounder cornerSize) do:[:r|
		cRect := r intersect: aRectangle.
		cRect area > 0 ifTrue:[rects nextPut: cRect].
	].
	^rects contents! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/24/2004 03:12'!
asClipMask
	"Answer the receiver as a clip mask"
	| mask maskCanvas |
	mask := Form extent: self localBounds extent depth: 1.
	maskCanvas := CTransformCanvas on: mask.
	maskCanvas shadowColor: Color black.
	self drawOn: maskCanvas.
	^mask! !

!CPrimitiveCostume methodsFor: 'converting' stamp: 'tk 7/14/2004 11:31'!
asCostume

	"*** Later remove all implementor of asCostume ***"
	self inform: 'Please call asPrimCostume instead of asCostume'.
	^self! !

!CPrimitiveCostume methodsFor: 'initialize' stamp: 'ar 5/13/2005 13:28'!
asCostumeFor: aPlayer
	self player: aPlayer.
	^self! !

!CPrimitiveCostume methodsFor: 'converting' stamp: 'ar 12/16/2002 17:22'!
asForm
	"Convert me into a form"
	^self asFormOfDepth: Display depth! !

!CPrimitiveCostume methodsFor: 'converting' stamp: 'ar 12/16/2002 17:34'!
asFormExtent: maxExtent
	"Convert me into a form of at most maxExtent size.
	Scale the receiver to fit that extent if needed."
	^self asFormOfDepth: Display depth extent: maxExtent! !

!CPrimitiveCostume methodsFor: 'converting' stamp: 'ar 8/5/2005 20:30'!
asFormOfDepth: depth
	"Convert me into a form"
	| box form canvas |
	box := self outerBounds.
	form := Form extent: box extent depth: depth.
	canvas := CTransformCanvas on: form.
	canvas translateBy: box origin negated during:[:cc|
		self localFullDrawOn: cc.
	].
	^form! !

!CPrimitiveCostume methodsFor: 'converting' stamp: 'ar 12/16/2002 17:33'!
asFormOfDepth: depth extent: maxExtent
	"Convert me into a form of at most maxExtent size.
	Scale the receiver to fit that extent if needed."
	| form size thumb |
	form := self asFormOfDepth: depth.
	form extent <= maxExtent ifTrue:[^form].
	size := form extent.
	size y > maxExtent y 
		ifTrue:[size := (size x * maxExtent y // size y) @ maxExtent y].
	size x > maxExtent x 
		ifTrue:[size := maxExtent x @ (size y * maxExtent x // size x)].
	thumb := Form extent: size depth: depth.
	form displayScaledOn: thumb.
	^thumb! !

!CPrimitiveCostume methodsFor: 'converting' stamp: 'ar 3/24/2004 03:16'!
asPlayer
	^(self player ifNil:[CPlayer new costume: self]) asPlayer.! !

!CPrimitiveCostume methodsFor: 'converting' stamp: 'tk 7/14/2004 11:20'!
asPrimCostume

	^self! !

!CPrimitiveCostume methodsFor: 'graphic' stamp: 'ar 9/4/2003 23:50'!
aspectRatio
	^self propertyValueAt: #aspectRatio ifAbsent:[nil].! !

!CPrimitiveCostume methodsFor: 'graphic' stamp: 'ar 6/25/2005 20:21'!
aspectRatio: newRatio
	newRatio == nil
		ifTrue:[self removeProperty: #aspectRatio]
		ifFalse:[self propertyValueAt: #aspectRatio put: newRatio].
	self extent: self extent. "so we keep the constraints accordingly"
! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 12/7/2002 22:27'!
assureLayoutProperties
	| props |
	props := self layoutProperties.
	props == self ifTrue:[props := nil].
	props ifNil:[
		props := CLayoutProperties new initializeFrom: self.
		self layoutProperties: props].
	^props! !

!CPrimitiveCostume methodsFor: 'pen my motion' stamp: 'tk 4/27/2005 13:33'!
assurePen
	"Create and return the object that holds the Pen parameters, if we do not already own one"
	| newPen |
	^ self propertyValueAt: #pen ifAbsent: [
		self pen: (newPen := CCostumePen new).
		newPen costume: self player costume.
		newPen].! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 3/24/2004 03:18'!
at: index
	^contents at: index! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:13'!
balloonHelp
	^self propertyValueAt: #balloonHelp! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:13'!
balloonHelp: aText
	self propertyValueAt: #balloonHelp put: aText.
	self signal: #balloonHelpChanged.
	^aText! !

!CPrimitiveCostume methodsFor: 'initialize' stamp: 'bgf 3/17/2009 03:26'!
beOnScreen
	"Make sure the receiver shows on the screen (and not outside)"
	| trial box world |
	world := self world ifNil:[^self].
	box := world localBounds.
	trial := self fullBoundsForPlacement.
	self moveBy: (trial amountToTranslateWithin: box).! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:14'!
border
	^border! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:14'!
border: aBorder
	"Give the receiver a new border"
	aBorder == border ifTrue:[^aBorder].
	aBorder ifNil:[
		border isNoBorder ifTrue:[^nil].
		border costume: nil.
		border := CNoBorder forCostume: self.
		^nil].
	border isNoBorder ifTrue:[
		border := CCostumeBorder new.
		border costume: self.
	].
	border loadFrom: aBorder.
	^aBorder! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/24/2004 03:14'!
borderWidthForRounding
	^border width! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:25'!
bottom
	^ self bounds bottom! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:26'!
bottom: aNumber
	self moveBy: 0@(aNumber - self bottom).
	^aNumber
! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'bgf 2/23/2008 12:40'!
bottomCenter
	^self bounds bottomCenter! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'bgf 2/23/2008 12:40'!
bottomCenter: aPoint
	self moveBy: (aPoint - self bottomCenter).
! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:25'!
bottomLeft
	^ self bounds bottomLeft! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:26'!
bottomLeft: aPoint
	self bottom: aPoint y.
	self left: aPoint x.! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:26'!
bottomRight
	^ self bounds bottomRight! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:26'!
bottomRight: aPoint
	self bottom: aPoint y.
	self right: aPoint x! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:26'!
bounds
	^self transform transformRect: self localBounds! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:26'!
bounds: newBounds
	self topLeft: newBounds origin.
	self resizeTo: newBounds extent.! !

!CPrimitiveCostume methodsFor: 'collision detection' stamp: 'ar 3/23/2003 22:56'!
boxCollisionPairs
	^Array streamContents:[:s|
		self boxCollisionPairsDo:[:a1 :a2|
			s nextPut: (Array with: a1 value player with: a2 value player).
		].
	].
! !

!CPrimitiveCostume methodsFor: 'collision detection' stamp: 'ar 3/23/2004 23:49'!
boxCollisionPairsDo: aBlock
	| count sorted box objectA boxA j continue objectB boxB |
	sorted := Array new: contents size.
	count := 0.
	contents do:[:aCostume|
		aCostume collisionsEnabled ifTrue:[
			box := aCostume fullBounds.
			sorted at: (count := count + 1) put: (Association key: box value: aCostume).
		].
	].
	count = 0 ifTrue:[^#()].
	sorted mergeSortFrom: 1 
		to: count 
		src: (sorted copyFrom: 1 to: count)
		dst: sorted
		by:[:a1 :a2| a1 key left <= a2 key left].
	1 to: count-1 do:[:i|
		objectA := sorted at: i.
		boxA := objectA key.
		j := i.
		continue := true.
		[continue and:[(j := j+1) <= count]] whileTrue:[
			objectB := sorted at: j.
			boxB := objectB key.
			continue := boxB left < boxA right.
			continue ifTrue:[
				(boxA bottom <= boxB top or:[boxB bottom <= boxA top]) ifFalse:[
					aBlock value: objectA value: objectB.
				].
			].
		].
	].
! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:12'!
cacheEnabled
	^(flags bitAnd: FlagCached) == FlagCached! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:12'!
cacheEnabled: aBool
	aBool = self localCache notNil ifFalse:[
		^self localCache: (aBool ifTrue:[self newFormCache] ifFalse:[nil]).
	].
	flags := aBool ifTrue:[flags bitOr: FlagCached] ifFalse:[flags bitClear: FlagCached].
! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/23/2004 23:16'!
canBltMove: aCostume from: oldRect to: newRect
	"Private. Answer whether it is possible to relocate the given costume
	from oldRect to newRect using a direct screen blt. If possible, prepare
	everything that's needed for the blt (which may include transferring
	some damage etc.)"
	"Note: This version is currently :=very:= conservative in its assumptions."
	| cache dirty dirtyArea index frontCostume debug |
true ifTrue:[^false].
	debug := false.
	debug ifTrue:[Transcript cr; show: '<bltMove: ',aCostume,'>'].
	"Make sure this *is* a blt"
	oldRect extent = newRect extent ifFalse:[
		debug ifTrue:[Transcript cr; show: '<no blit>'].
		^false].
	(cache := self localCache) ifNotNil:[
		"See if there is any damage in the old rectangle. 
		If so, don't blt since the costume itself may not be up to date.
		Note: For further optimization we could take the dirty region
		into account and translate it into the new region (for non-front
		costumes this would probably have to introduce addl. damage)."
		(cache damageRecorder hasDamageInRect: oldRect) ifTrue:[
			debug ifTrue:[Transcript cr; show: '<oldRect damage>'].
			^false].

		"See if there is any damage in the new rectangle.
		If so, don't blt since we'd be drawing this area more than once.
		Note: For further optimization we should take the size of the 
		dirty area into account since it can be small."
		(cache damageRecorder hasDamageInRect: newRect) ifTrue:[
			debug ifTrue:[Transcript cr; show: '<newRect damage>'].
			^false].
	].
	"See how much we'd have to redraw after the blt.
	Note: for non-front costumes this must take the overlap into account
	since we :=do:= have to restore everything before aCostume."
	dirty := aCostume areasRemainingToFill: newRect.
	dirtyArea := dirty inject: 0 into:[:sum :rect| sum + rect area].
	"If we have to restore more than half of aCostume after the blt it's pointless"
	newRect area > (dirtyArea * 2) ifFalse:[
		debug ifTrue:[Transcript cr; show: '<tiny area>'].
		^false].
	"Only the front-most costume is allowed to blt.
	Note: For further optimization this could be relaxed but requires
	more effort in some other parts."
	index := 1.
	[index <= contents size and:[(contents at: index) ~~ aCostume]] whileTrue:[
		frontCostume := contents at: index.
		frontCostume visible ifTrue:[
			"see if front costume overlaps either old or new rect"
			frontCostume layoutNeeded ifTrue:[
				"Uh, ah ... this guy needs a layout which may change its size.
				I can't say for sure what this means at this point; probably some
				other than the 'fullBounds' measure is needed."
				debug ifTrue:[Transcript cr; show: '<layout needed>'].
				^false].
			"Check if its fullBounds overlap either the the new area"
			(frontCostume fullBounds intersects: newRect) ifTrue:[
				debug ifTrue:[Transcript cr; show: '<costume overlap>'].
				^false]].
		index := index + 1].
	"Okay, so this looks good. Let's go for it"
	debug ifTrue:[Transcript cr; show: '<OKAY>'].
	^true! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/24/2004 03:32'!
cellPositioning
	^(self layout ifNil:[^#center]) cellPositioning! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:26'!
center
	^ self bounds center! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:26'!
center: aPoint
	self moveBy: aPoint - self center.! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/25/2005 20:27'!
childToParent
	"Answer a transformation mapping from my coordinate frame 
	into my parent's coordinate frame."
	^self transform! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/25/2005 20:27'!
childToParent: anObject
	^anObject transformedBy: self childToParent! !

!CPrimitiveCostume methodsFor: 'pen my motion' stamp: 'ar 6/25/2005 19:22'!
clearPenTrails
	"Look up the container hierarchy to find a Player who has penTrails.  Erase the Form that has the marks."

	| primPen |
	(primPen := self primitivePen) 
		ifNil: [container ifNotNil: [container clearPenTrails]]	
						"else stop looking"
		ifNotNil: ["Yes, wipe out the Form.  Small in case never used again."
			primPen setDestForm: (Form extent: 16@16 depth: Display depth).
			self invalidate]. ! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/23/2004 23:08'!
clipping
	"Drawing specific. If this property is set, clip the receiver's children to the receiver's clipping bounds."
	^(flags bitAnd: FlagClipEnabled) == FlagClipEnabled! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/23/2004 23:09'!
clipping: aBool
	"Drawing specific. If this property is set, clip the receiver's children to the receiver's clipping bounds."
	aBool = self clipping ifTrue:[^self].
	self invalidate.
	flags := aBool ifTrue:[flags bitOr: FlagClipEnabled] ifFalse:[flags bitClear: FlagClipEnabled].
	self invalidate.
	self signal: #flagsChanged.! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 6/25/2005 14:58'!
clippingBounds
	"Return the bounds to which any parts should be clipped if the property is set"
	^self clipping ifTrue:[self innerBounds] ifFalse:[self drawingBounds]! !

!CPrimitiveCostume methodsFor: 'initialize' stamp: 'ar 1/19/2003 02:41'!
close
	"Close this costume"
	container ifNotNil:[container remove: self].! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'tk 2/1/2005 21:35'!
closed: aBoolean
	"closed tells if the object's polygon shape will be a filled area, or just a segmented line (or a curve).  You must give this object a shape for this command to have any effect (self shape: CPolygonShape new)."

	^(self shape ifNil: [^ self]) closed: aBoolean.
! !

!CPrimitiveCostume methodsFor: 'collision detection' stamp: 'ar 3/23/2003 23:24'!
collisionPairs
	^Array streamContents:[:s|
		self collisionPairsDo:[:a1 :a2|
			s nextPut: (Array with: a1 value player with: a2 value player).
		].
	].
! !

!CPrimitiveCostume methodsFor: 'collision detection' stamp: 'ar 3/23/2003 23:57'!
collisionPairsDo: aBlock
	"this method makes a fine-grained distinction between the colliding objects actually checking the outlines"
	| collide formA formB rect costumeA preciseA costumeB preciseB pixels |
	formA := formB := nil.
	self boxCollisionPairsDo:[:a1 :a2|
		rect := a1 key intersect: a2 key.
		costumeA := a1 value.
		preciseA := costumeA preciseCollisions.

		costumeB := a2 value.
		preciseB := costumeB preciseCollisions.

		preciseA ifTrue:[
			(formA notNil and:[formA extent >= rect extent]) 
				ifFalse:[formA := Form extent: rect extent depth: 1].
			(CTransformCanvas on: formA) translateBy: rect origin negated during:[:cc|
				cc clipBy: rect during:[:clipped|
					clipped shadowColor: Color black.
					a1 value fullDrawOn: clipped.
				].
			].
		].

		preciseB ifTrue:[
			(formB notNil and:[formB extent >= rect extent]) 
				ifFalse:[formB := Form extent: rect extent depth: 1].
			(CTransformCanvas on: formB) translateBy: rect origin negated during:[:cc|
				cc clipBy: rect during:[:clipped|
					clipped shadowColor: Color black.
					a2 value fullDrawOn: clipped.
				].
			].
		].

		preciseA ifTrue:[
			preciseB ifTrue:[
				(BitBlt toForm: formA)
					sourceForm: formB;
					width: rect width; height: rect height;
					combinationRule: Form and;
					copyBits.
			].
			pixels := formA tallyPixelValuesInRect: (0@0 extent: rect extent) 
							into: (Bitmap new: 2).
			collide := (pixels at: 2) > 0.
		] ifFalse:[
			preciseB ifTrue:[
				pixels := formB tallyPixelValuesInRect: (0@0 extent: rect extent) 
								into: (Bitmap new: 2).
				collide := (pixels at: 2) > 0.
			] ifFalse:[collide := true].
		].
		collide ifTrue:[aBlock value: a1 value player value: a2 value player].
	].! !

!CPrimitiveCostume methodsFor: 'collision detection' stamp: 'ar 3/23/2004 23:49'!
collisionsEnabled
	^self propertyValueAt: #collisionsEnabled ifAbsent:[false]! !

!CPrimitiveCostume methodsFor: 'collision detection' stamp: 'ar 3/23/2004 23:49'!
collisionsEnabled: aBool
	^self propertyValueAt: #collisionsEnabled put: aBool! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 4/2/2004 14:17'!
colorAt: aPoint behind: aPlayer
	"Answer the color of the receiver at the given point"
	| aForm aCanvas rect pv index |
	aPlayer
		ifNil:[index := contents size] "don't bother searching"
		ifNotNil:[index := (self indexOf: aPlayer costume) - 1].
	rect := aPoint-1 extent: 3@3.
	aForm := Form extent: 3@3 depth: 32. "always pick in 32bpp"
	aCanvas := CTransformCanvas on: aForm.
	aCanvas translateBy: rect origin negated.
	aCanvas clipBy: rect.
	self topDownFullDrawOn: aCanvas 
		damageList: (Array with: rect) 
		dirtyRect: aCanvas clipRect 
		startingAt: index.
	aCanvas finish.
	pv := aForm bits at: 5.
	pv = 0 ifTrue:[^Color transparent].
	^Color colorFromPixelValue: (pv bitOr: 16rFF000000) depth: 32! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 4/2/2004 15:00'!
comeToFront
	container ifNotNil:[container add: self].! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/15/2004 20:32'!
componentBinding
	"Answer the receiver's part binding"
	^self propertyValueAt: #componentBinding! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/15/2004 20:32'!
componentBinding: aBinding
	"Change the component binding associated with the receiver."
	| myBinding |
	"Carefully release the previous binding first"
	(myBinding := self componentBinding) ifNotNil:[myBinding value: nil].
	^self propertyValueAt: #componentBinding put: aBinding.! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 1/8/2003 16:57'!
container
	"Answer the container of this costume"
	^ container! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 6/25/2005 14:58'!
containsEvent: localPt
	"Return true if aPoint is considered to be inside the receiver for event handling. The default implementation treats locked contents as integral part of their owners."
	self obsolete.
	self infinite ifTrue:[^true].
	(self drawingBounds containsPoint: localPt)
		ifFalse:[^false]. "try quick elimination"
	(self containsPoint: localPt) 
		ifTrue:[^true]. "try quick acceptance"
	contents do:[:m|
		(m locked and:[m fullContainsPoint: localPt])
			ifTrue:[^true]].
	^false! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'bf 4/29/2004 02:15'!
containsPoint: localPoint
	"Answer whether aPoint is in the receiver. Query player for custom drawing components"
	| box |
	box := self localBounds.
	(box containsPoint: localPoint) ifFalse:[^false].
	(flags bitAnd: FlagCustomDraw) == 0 ifTrue:[
		fill isTransparent ifFalse:[^true].
		self preciseCollisions ifFalse:[^true].
		(flags bitAnd: FlagHasShape) = 0 ifTrue:[
			"Check to see if it's on the border"
			border isTransparent ifFalse:[
				((border insetBounds: box) containsPoint: localPoint) ifFalse:[^true].
			].
		] ifFalse:[
			(self shape containsPoint: localPoint) ifTrue:[^true].
		].
		(flags bitAnd: FlagHasGraphic) = 0 ifFalse:[
			(self graphic containsPoint: localPoint) ifTrue:[^true].
		].
		(flags bitAnd: FlagHasText) = 0 ifFalse:[
			(self textHolder containsPoint: localPoint) ifTrue:[^true].
		].
		^false
	].
	"Disable custom draw while querying player"
	flags := flags bitClear: FlagCustomDraw.
	^[self asPlayer containsPoint: localPoint]
		ensure:[flags := flags bitOr: FlagCustomDraw]! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 2/13/2003 11:37'!
contents
	^contents collect:[:each| each]! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'tk 7/14/2004 11:26'!
contents: aCollection
	self removeAll.
	self addAll: (aCollection collect:[:each| each asPrimCostume]).! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 6/25/2005 14:58'!
contentsAt: aPoint behind: aCostume unlocked: aBool
	"Return all costumes at aPoint that are behind aCostume; if aBool is true return only unlocked, visible costumes."
	| isBack found all |
	(aCostume == nil or:[container == nil]) ifTrue:["Traverse down"
		(self drawingBounds containsPoint: aPoint) ifFalse:[^#()].
		(aBool and:[self locked or:[self visible not]]) ifTrue:[^#()].
		all := nil.
	] ifFalse:[ "Traverse up"
		all := container contentsAt: (self childToParent: aPoint) behind: self unlocked: aBool.
		all := WriteStream with: all.
	].
	isBack := aCostume == nil.
	contents reverseDo:[:m|
		isBack ifTrue:[
			found := m contentsAt: (m parentToChild: aPoint) behind: nil unlocked: aBool.
			found size > 0 ifTrue:[
				all ifNil:[all := WriteStream on: #()].
				all nextPutAll: found]].
		m == aCostume ifTrue:[isBack := true]].
	(isBack and:[self containsPoint: aPoint]) ifTrue:[
		all ifNil:[^Array with: self].
		all nextPut: self].
	^all ifNil:[#()] ifNotNil:[all contents].! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/23/2004 23:16'!
contentsBounds
	"Answer the bounds of my contents"
	| box subBox |
	box := nil.
	(flags bitAnd: FlagHasText) = 0 ifFalse:[
		box := self textBounds origin extent: self textHolder bounds extent.
	].
	(flags bitAnd: FlagHasGraphic) = 0 ifFalse:[
		| tex |
		tex := self graphic.
		tex fit = #rigid ifTrue:[
			box ifNil:[box := tex bounds]
				ifNotNil:[box := box quickMerge: tex bounds].
		].
	].
	contents do:[:p |
		p visible ifTrue:[
			subBox := p fullBounds.
			box 
				ifNil:[box := subBox copy]
				ifNotNil:[box := box quickMerge: subBox].
		].
	].
	box ifNil:[^0@0 corner: 0@0]. "e.g., having children but not visible"
	^ box origin asIntegerPoint corner: box corner asIntegerPoint! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:27'!
corner
	^self bounds corner! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:27'!
corner: aPoint
	^self moveBy: aPoint - self corner! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 12/8/2003 19:36'!
costume
	^self! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 6/25/2005 20:10'!
costumeMoved: aCostume
	self invalidateLayout.
	self isTextCostume ifTrue:[self adjustTextAnchor: self].! !

!CPrimitiveCostume methodsFor: 'pen trails in me' stamp: 'ar 6/25/2005 19:22'!
createOrResizeTrailsForm
	"If necessary, create a new turtleTrailsForm or resize the existing one to fill my bounds.
	On return, turtleTrailsForm exists and is the correct size.
	Use the Display depth so that color comparisons (#color:sees: and #touchesColor:) will work right."

	| newForm primPen turtleTrailsForm |
	"Know we have something in the primitivePen"
	primPen := self primitivePen.
	turtleTrailsForm := primPen destForm.
	(turtleTrailsForm isNil or: [ turtleTrailsForm extent ~= self extent ]) ifTrue:
		["resize TrailsForm if my size has changed"
		newForm := Form extent: self extent depth: Display depth.
		turtleTrailsForm ifNotNil:[turtleTrailsForm displayOn: newForm at: 0@0 rule: Form paint].
		primPen setDestForm: (turtleTrailsForm := newForm)].
! !

!CPrimitiveCostume methodsFor: 'styling' stamp: 'ar 7/7/2005 12:12'!
cssTag
	^self propertyValueAt: #cssTag! !

!CPrimitiveCostume methodsFor: 'styling' stamp: 'ar 7/7/2005 12:13'!
cssTag: aSymbol
	^self propertyValueAt: #cssTag put: aSymbol! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 3/24/2004 03:19'!
cursorPoint
	"Answer the active's hand position in local coordinates"
	^self globalToLocal: self hand position! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 7/25/2003 11:52'!
debugDraw
	^self fullDrawOn: (CTransformCanvas on: Display).! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 3/24/2004 03:19'!
deselect
	"Deselect the receiver"
! !

!CPrimitiveCostume methodsFor: 'initialize' stamp: 'ar 2/24/2004 00:43'!
destroy
	"Dump the receiver"
	| cc |
	self close.
	(cc := self player) ifNotNil:[
		self player: nil.
		cc destroy].! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'bf 7/22/2005 15:54'!
direction
	"Answer the direction of the receiver"
	| tfm |
	tfm := self transform.
	^tfm isPureTranslation
		ifTrue: [self nose]
		ifFalse: [(tfm transformDirection: self nose) normalized]! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:27'!
direction: aPoint
	"Indicate the new direction for the receiver"
	| oldDirection newDirection sin cos angle |
	oldDirection := self direction.
	"Compute the delta angle between current and new heading"
	newDirection := [aPoint normalized] on: ZeroDivide do:[:ex| ex return: nil].
	newDirection ifNil:[^oldDirection].

	sin := oldDirection crossProduct: newDirection.
	cos := oldDirection dotProduct: newDirection.
	angle := cos arcCos radiansToDegrees.
	sin < 0.0 ifTrue:[angle := angle negated].
	self rotateBy: angle.
	^aPoint! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'bgf 7/8/2009 04:04'!
disable
	| highlight |
	highlight := CFillingHighlight new.
	highlight name: #disableHighlight.
	highlight fill: ((Color gray: 0.8) alpha: 0.45).
	self addHighlight: highlight.
	self locked: true.! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 6/25/2005 14:58'!
dispatchEvent: anEvent
	"Dispatch the given event. The event will be passed to the front-most visible part that contains the event."
	| localEvent index inside child shield |
	(self rejectsEvent: anEvent) ifTrue:[^#rejected].
	localEvent := anEvent transformedBy: self parentToChild. "into local frame"
	self infinite ifFalse:[
		(self drawingBounds containsPoint: localEvent position) 
			ifFalse:[^#rejected]. "outside"
	].

	"Check for event shield"
	(shield := self eventShield) ifNotNil:[
		(self containsEvent: localEvent position) 
			ifTrue:[^shield handleEvent: localEvent]
			ifFalse:[^#rejected].
	].

	"Traverse children"
	index := contents size + 1. inside := false.
	[(index := index-1) > 0] whileTrue:[
		child := contents at: index.
		(child dispatchEvent: localEvent) == #rejected ifFalse:[
			"Inside child of receiver"
			inside := true.
			localEvent wasHandled ifTrue:[
				"Copy handler state for return"
				anEvent copyHandlerState: localEvent
			].
			index := 0. "e.g., break"
		].
	].
	"Check for inside of the receiver itself if necessary"
	inside := inside or:[self infinite or:[self containsPoint: localEvent position]].
	inside ifTrue:[
		self handleEvent: localEvent.
		localEvent wasHandled ifTrue:[
			anEvent copyHandlerState: localEvent
		].
		^self].
	^#rejected! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 4/14/2004 20:54'!
displayOn: gDev
	"Draw the receiver in the presence of a full transform"
	| highlights index hl transformedCanvas |
	transformedCanvas := gDev asCanvas.
	highlights := self highlights.
	"Step 1: Draw highlights below the receiver"
	index := 1.
	[index <= highlights size and:[(hl := highlights at: index) priority < 0]]
		whileTrue:[	hl drawHighlightOn: transformedCanvas.
					index := index + 1].
	"Skip zero-priority highlights"
	[index <= highlights size and:[(highlights at: index) priority = 0]]
		whileTrue:[index := index + 1].
	transformedCanvas roundCornersOf: self in: self localBounds during:[
		"Step 2: Draw receiver itself"
		self drawOn: transformedCanvas.
	].

	"Step 3: Draw pen trails"
	self drawPenTrailsOn: transformedCanvas.
	"Step 4: Draw all the parts"
	"transformedCanvas clipBy: self clippingBounds during:[:clip|
		self drawContentsOn: clip.
	]."
	(flags bitAnd: FlagHasOverlay) = 0 
		ifFalse:[self drawOverlayOn: transformedCanvas].
	"Step 5: Draw highlights on top of the receiver"
	index to: highlights size do:[:i|
		(highlights at: i) drawHighlightOn: transformedCanvas.
	].
! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'bgf 7/14/2009 12:58'!
do: aBlock
	"In-loop size-check is for safety  if async-ness leads to list change during the evaluations."
	1 to: contents size do:[:i| contents size >= i ifTrue: [ aBlock value: (contents at: i) ] ].
! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/24/2004 02:50'!
doLayoutIn: layoutBounds
	"Compute a new layout based on the given (local) layout bounds."
	| layout box priorBounds |
	self layoutNeeded: true. "don't bother broadcasting changes from here"
	"A costume with #spaceFill constraints needs to update its own bounds
	before the layout can be computed since it will grow anyways to match
	the layout bounds."
	(self hResizing == #spaceFill or:[self vResizing == #spaceFill]) ifTrue:[
		box := self layoutBounds.
		self hResizing == #spaceFill 
			ifTrue:[box := box withWidth: layoutBounds width].
		self vResizing == #spaceFill 
			ifTrue:[box := box withHeight: layoutBounds height].
		self layoutBounds: box.
	].
	"Note: Testing for #bounds or #layoutBounds would be sufficient to
	figure out if we need an invalidation afterwards but #outerBounds
	is what we need for all leaf nodes so we use that."
	flags := flags bitClear: FlagExcessBounds.
	contents size = 0 ifTrue:[
		self adjustLayoutBounds.
	] ifFalse:[
		"Send #containerChanged to our children"
		contents do:[:m| m signal: #containerLayoutChanged with: m with: self].
		priorBounds := self outerBounds.
		layout := self layout.
		layout ifNotNil:[layout layout: self in: layoutBounds].
		self adjustLayoutBounds.
		box := self outerBounds.
		box = priorBounds ifFalse:[self invalidate: (priorBounds quickMerge: box)].
	].
	self layoutNeeded: false.
	self signal: #layoutChanged.! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/24/2004 03:12'!
drawClipMaskOn: aCanvas in: clippingBounds
	"Draw a clipping mask (form of depth 1) on aCanvas"
	| mask maskCanvas |
	mask := Form extent: clippingBounds extent depth: 1.
	maskCanvas := CTransformCanvas on: mask.
	maskCanvas transformBy: self localToGlobal.
	maskCanvas shadowColor: Color white.
	self localFullDrawOn: maskCanvas.
	^mask! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 8/25/2010 14:31'!
drawContentsOn: aCanvas
	"Display children back to front"
	contents size = 0 ifTrue:[^self].
	contents do:[:each| each fullDrawOn: aCanvas].
! !

!CPrimitiveCostume methodsFor: 'graphic' stamp: 'ar 1/10/2003 20:55'!
drawGraphic: aGraphic on: aCanvas in: drawingBounds
	"Draw the receiver's texture"
	aGraphic ifNil:[^self].
	aCanvas clipBy: drawingBounds during:[:cc|
		aGraphic drawOn: aCanvas in: drawingBounds.
	].! !

!CPrimitiveCostume methodsFor: 'graphic' stamp: 'ar 3/18/2003 22:45'!
drawGraphicOn: aCanvas in: drawingBounds
	"Draw the receiver's texture"
	| aGraphic |
	aGraphic := self graphic ifNil:[^self].
	aCanvas clipBy: drawingBounds during:[:cc|
		aGraphic drawOn: aCanvas in: drawingBounds.
	].! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/24/2004 03:10'!
drawGrid: aGrid on: aCanvas in: drawingBounds
	aGrid ifNil:[^self].
	aGrid drawOn: aCanvas in: drawingBounds.! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 6/24/2005 21:28'!
drawOn: aCanvas
	| box |
	box := self localBounds.
	(aCanvas isVisible: box) ifFalse:[^self].
	fill isOrientedFill ifTrue:[fill updateFrame: box].
	border trackColorFrom: self.
	aCanvas pushState.
	(flags bitAnd: FlagCustomDraw) == 0 ifTrue:[
		self drawOn: aCanvas in: box.
	] ifFalse:[
		flags := flags bitClear: FlagCustomDraw.
		[aCanvas clipBy: box.
		self asPlayer drawOn: aCanvas in: box]
			ensure:[flags := flags bitOr: FlagCustomDraw].
	].
	aCanvas popState.! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'tk 1/31/2005 15:12'!
drawOn: aCanvas in: drawingBounds
	"Draw only the receiver on the canvas using the given drawing bounds"

	"TODO: Optimize fill for the case of opaque textures (no or only partial background fill needs to be drawn in this case)"

	"draw fill (color/gradient) and border if present"
	(fill isTransparent not or:[border width > 0]) ifTrue:[
		(flags bitAnd: FlagHasShape) = 0 ifTrue: [
			aCanvas fillRectangle: drawingBounds fill: fill border: self border.
	]].
	"draw vector shape if present"
	(flags bitAnd: FlagHasShape) = 0 ifFalse:[
		self drawShape: self shape on: aCanvas in: drawingBounds.	"does its own border"
	].
	"draw texture if present"
	(flags bitAnd: FlagHasGraphic) = 0 ifFalse:[
		self drawGraphic: self graphic on: aCanvas in: (drawingBounds insetBy: border width).
	].
	"draw text if present"
	(flags bitAnd: FlagHasText) = 0 ifFalse:[
		self drawText: self textHolder on: aCanvas in: (drawingBounds insetBy: border width).
	].
	"draw grid if present"
	(flags bitAnd: FlagHasGrid) = 0 ifFalse:[
		self drawGrid: self grid on: aCanvas in: (drawingBounds insetBy: border width).
	].! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/24/2004 03:11'!
drawOverlay: aPlayer on: aCanvas in: drawingBounds
	"Draw the receiver's texture"
	aPlayer ifNil:[^self].
	aCanvas clipBy: drawingBounds during:[:cc|
		aPlayer drawOverlayOn: aCanvas in: drawingBounds.
	].! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 6/24/2005 21:28'!
drawOverlayOn: aCanvas
	| box |
	box := self localBounds.
	(aCanvas isVisible: box) ifFalse:[^self].
	aCanvas pushState.
	aCanvas clipBy: box.
	self drawOverlay: self overlay on: aCanvas in: box.
	aCanvas popState.! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 6/25/2005 14:58'!
drawPatchOn: canvas without: stoppee andNothingAbove: stopThere
	| box highlights index hl found |
	stoppee == self ifTrue:[^true]. "you found me"
	(stoppee isIn: self) ifFalse:[
		self fullDrawOn: canvas.
		^false].
	"Need to draw individual elements"
	self visible ifFalse:[^true].
	(canvas isVisible: self invalidationBounds) ifFalse:[^true].
	box := self drawingBounds.
	canvas transformBy: self transform clippingTo: box during:[:cc|
		"A variant on #localFullDrawOn:"
		self flag: #localFullDrawOn:.
		highlights := self highlights. index := 1.
		[index <= highlights size and:[(hl := highlights at: index) priority < 0]]
			whileTrue:[	hl drawHighlightOn: cc. index := index + 1].
		[index <= highlights size and:[(highlights at: index) priority = 0]]
			whileTrue:[index := index + 1].
		cc roundCornersOf: self in: self localBounds during:[self drawOn: cc].
		self drawPenTrailsOn: cc.
		cc clipBy: self clippingBounds during:[:clip|
			"---- here is the difference ---"
			contents size = 0 ifTrue:[^self].
			1 to: contents size do:[:i| 
				found := (contents at: i) drawPatchOn: clip without: stoppee andNothingAbove: stopThere.
				found ifTrue:[stopThere ifTrue:[^true]].
			].
		].
		(flags bitAnd: FlagHasOverlay) = 0 ifFalse:[self drawOverlayOn: cc].
		index to: highlights size do:[:i| (highlights at: i) drawHighlightOn: cc].
	].
	^true! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'tk 4/26/2005 11:54'!
drawPenTrailsOn: aCanvas
	"I need to draw my pen trails on the given canvas"

	self indicatePenTrails ifFalse:[^self].

	"Now blt on the turtleTrailsForm"
	aCanvas paintImage: self turtleTrailsForm at: 0@0.! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/24/2004 03:15'!
drawShape: aShape on: aCanvas in: drawingBounds
	"Draw the receiver's texture"
	aShape ifNil:[^self].
	aCanvas clipBy: drawingBounds during:[:cc|
		aShape drawOn: aCanvas in: drawingBounds.
	].! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/24/2004 03:17'!
drawText: textContainer on: aCanvas in: drawingBounds
	textContainer ifNil:[^self].
	aCanvas clipBy: drawingBounds during:[:cc|
		textContainer drawOn: aCanvas in: drawingBounds.
	].
! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 5/3/2004 14:57'!
drawTextOn: aCanvas in: drawingBounds
	^self drawText: self textHolder on: aCanvas in: drawingBounds! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 6/25/2005 14:57'!
drawingBounds
	"Return the (local) bounds for drawing the receiver."
	| box  |
	box := self outerBounds.
	contents size = 0 ifTrue: [^box].
	(flags bitAnd: FlagExcessBounds) == 0 ifTrue:[^box].
	(flags bitAnd: FlagClipEnabled) == 0 ifFalse:[^box].
	box := box copy.
	contents do: [:m | m visible ifTrue: [
		box := box quickMerge: m invalidationBounds
	]].
	^box! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:14'!
dropShadow
	^self highlightNamed: #dropShadow! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:14'!
dropShadow: aDropShadow
	aDropShadow ifNil:[^self removeHighlight: #dropShadow].
	aDropShadow name: #dropShadow.
	self addHighlight: aDropShadow.! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/16/2004 13:42'!
elements
	"Answer the players represented by this costume"
	^self propertyValueAt: #elements ifAbsentPut:[CPlayerCollection costume: self].! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/16/2004 13:42'!
elements: aCollection
	"Give this costume a set of new elements"
	self elements contents: aCollection! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 3/24/2004 03:14'!
enable
	self removeHighlight: #disableHighlight.
	self locked: false.! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 7/19/2003 20:54'!
eventShield
	"Answer the event shield for the receiver (if any)"
	^self propertyValueAt: #eventShield.! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 7/19/2003 20:55'!
eventShield: anObject
	"Set the event shield for the receiver (if any)"
	^self propertyValueAt: #eventShield put: anObject! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/25/2005 20:24'!
extent
	^bounds extent! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/27/2005 23:11'!
extent: aPoint
	| newExtent aspect oldExtent |
	newExtent := aPoint asNonFractionalPoint.
	(aspect := self aspectRatio) ifNotNil:[
		(newExtent y * aspect) > newExtent x
			ifFalse:[newExtent := newExtent x @ (newExtent x / aspect)]
			ifTrue:[newExtent := (newExtent y * aspect) @ newExtent y]].
	newExtent := newExtent asIntegerPoint max: self minWidth @ self minHeight.
	oldExtent := bounds extent.
	oldExtent = newExtent ifTrue:[^aPoint].

	oldExtent <= newExtent ifFalse:[self invalidate: self outerBounds].
	bounds := bounds origin extent: newExtent.
	newExtent <= oldExtent ifFalse:[self invalidate: self outerBounds].
	self signal: #extentChanged.
	self signal: #geometryChanged.
	1 to: contents size do:[:i| (contents at: i) signal: #geometryChanged]. "for cartesian coordinates"
	self frameChanged.
	self invalidateLayout.
	(flags bitAnd: FlagHasText) = 0 ifFalse:[
		self textHolder updateParagraphExtent
	].
	^aPoint! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:15'!
fill
	^fill! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:15'!
fill: aFill
	aFill == fill ifTrue:[^aFill].
	aFill ifNil:[
		fill isNoFill ifTrue:[^nil].
		fill costume: nil.
		fill := CNoFill forCostume: self.
		^nil].
	fill isNoFill 
		ifTrue:[fill := CCostumeFill forCostume: self].
	fill loadFrom: aFill.
	^aFill! !

!CPrimitiveCostume methodsFor: 'drag and drop' stamp: 'ar 5/22/2005 18:05'!
findDraggablePlayerAt: aPoint
	"Answer a child of me which can be dragged when the mouse is clicked"
	| obj |
	(self containsPoint: aPoint) ifFalse:[^nil].
	contents size to: 1 by: -1 do:[:i|
		obj := contents at: i.
		(obj locked not
			and:[obj visible 
			and:[obj sticky not
			and:[obj fullContainsPoint: aPoint]]]) ifTrue:[^obj asPlayer].
	].
	^nil! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 2/7/2005 02:00'!
findPlayerAt: aPoint
	"Search the receiver to find the first player located at the given position."
	| aCostume |
	(self containsPoint: aPoint) ifFalse:[^nil].
	contents size to: 1 by: -1 do:[:i|
		aCostume := contents at: i.
		(aCostume fullContainsPoint: aPoint) ifTrue:[^aCostume player].
	].
	^nil! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 3/24/2004 03:20'!
first
	^contents at: 1! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 6/25/2005 17:32'!
fixGeometryAfterPivotChange
	"Fix my geometry after the pivot changes"
	"CPrimitiveCostume allSubInstancesDo:[:each| each fixGeometryAfterPivotChange]"
	((flags bitAnd: FlagHasFullBounds) ~= 0 
		or:[(flags bitAnd: FlagHasMatrix) = 0]) ifTrue:[
			"Flush out old bounds"
			self propertyValueAt: #transform put: (COffsetTransform withOffset: bounds origin).
			bounds := 0@0 extent: bounds extent.
			self removeProperty: #fullBounds.
			flags := flags bitClear: FlagHasFullBounds.
			flags := flags bitOr: FlagHasMatrix.
	].! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/25/2004 20:01'!
flash
	"Flash the receiver"
	| prevFill |
	prevFill := fill copy.
	fill color: Color black.
	self wait: 0.1.
	self fill: prevFill.! !

!CPrimitiveCostume methodsFor: 'drag and drop' stamp: 'ar 3/23/2004 23:11'!
floating
	"Answer whether the receiver is a 'floating' object, e.g., one which will never generate a drop event when dragged around."
	^self propertyValueAt: #floating ifAbsent:[false]! !

!CPrimitiveCostume methodsFor: 'drag and drop' stamp: 'ar 3/23/2004 23:11'!
floating: aBool
	"Answer whether the receiver is a 'floating' object, e.g., one which will never generate a drop event when dragged around."
	self propertyValueAt: #floating put: (aBool ifTrue:[true] ifFalse:[nil]).
	self signal: #flagsChanged! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: ''!
forwardBy: distance
	^self forwardBy: 5 duration: 0 style: #gently! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
forwardBy: distance duration: duration style: style
	^self move: #forward distance: distance duration: duration style: style! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 3/21/2004 05:22'!
frameChanged
	"Either my or one of my owner's frames have changed. Notify my children.
	TODO: At some point we may want to introduce a flag that tells us if
	someone is really interested in handling this. If not, there is no point in
	traversing all the children 'cause there can be very many."
	self signal: #frameChanged.
	contents do:[:p| p frameChanged].! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 2/24/2009 14:40'!
fullBounds
	self layoutNeeded ifTrue:[
		"Errors at this point can be critical so make sure we catch 'em all right"
		[self doLayoutIn: self layoutBounds] on: Error do:[:ex| ex pass].
	].
	^self transform transformRect: self outerBounds! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'bgf 3/17/2009 03:30'!
fullBoundsForPlacement
	"Like fullbounds, but ignores hilight expansion,
	 so drop shadows don't push menus out of alignment."
	self layoutNeeded ifTrue:[
		"Errors at this point can be critical so make sure we catch 'em all right"
		[self doLayoutIn: self layoutBounds] on: Error do:[:ex| ex pass].
	].
	^self transform transformRect: self localBounds! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 6/25/2005 14:59'!
fullContainsPoint: globalPt
	"NOTE: Contrary to #containsPoint: this method assumes the target point to be in the parents reference frame."
	| localPt |
	localPt := self parentToChild: globalPt.
	(self drawingBounds containsPoint: localPt) 
		ifFalse: [^ false].  "try quick elimination"
	(self containsPoint: localPt)
		ifTrue:[^true]. "try quick acceptance"
	contents reverseDo: [:m | (m fullContainsPoint: localPt) ifTrue: [^true]].
	^ false! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 6/26/2005 13:42'!
fullDrawOn: aCanvas
	"Draw the full structure of the receiver on the given Canvas.
	There are many properties that come into play here some of which
	are pretty heavily used and some of which aren't.
	The more heavily used properties are:
		* clipping
		* rounded corners
		* cache enabled
	The more rarely used properties are
		* a full transformation
		* highlights
	And therefore the following optimizes for the first three and falls
	through into a generic version in the presence of the latter two."
	| box tfm |
	self visible ifFalse: [^ self]. "(a)"
	(aCanvas isVisible: self invalidationBounds) ifFalse:[^self].
	"Bail out to the generic version if needed"
	tfm := self transform.
	(tfm isPureTranslation and:[(flags bitAnd: FlagHasHighlights) == 0]) ifFalse:[
		box := self drawingBounds.
		^aCanvas 
			transformBy: tfm clippingTo: box during:[:cc|
				cc cache: box using: self localCache during:[:cc2| self localFullDrawOn: cc2]].
	].
	"Otherwise go straight for it"
	flags := flags bitClear: FlagInvalidated.
	box := self localBounds.
	aCanvas pushState.
	aCanvas translateBy: tfm offset.
	aCanvas cache: self drawingBounds using: self localCache during:[:cache |
		cache roundCornersOf: self in: box during:[
			self drawOn: cache.
			self drawPenTrailsOn: cache.
		].
		contents size = 0 ifFalse:[
			self clipping 
				ifTrue:[cache clipBy: self clippingBounds during:[:cc| self drawContentsOn: cc]]
				ifFalse:[self drawContentsOn: cache].
		].
		(flags bitAnd: FlagHasOverlay) = 0 
			ifFalse:[self drawOverlayOn: aCanvas].
	].
	aCanvas popState.! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: 'ar 3/24/2004 03:11'!
gently: x lowerBound: lower upperBound: upper
	"This method converts a linear proportion done to a slow in - slow out proportion. If x is less than A then the animation is in the slow in part, while if x is greater than B then the animation is in the slow out part."
	| a3 b3 c3 m b2 |
	x > 1.0 ifTrue:[^1.0].
  	(x < lower) ifTrue: [ 
		^((upper - 1)/(lower *  ((upper * upper) - (lower * upper) + lower - 1))) * x * x
	].
	(x > upper) ifTrue: [
		a3 := 1 / ((upper * upper) - (lower * upper) + lower - 1).
		b3 := -2 * a3.
		c3 := 1 + a3.
  		^(a3 * x * x) + (b3 * x) + c3.
	].
	m := 2 * (upper - 1) / ((upper * upper) - (lower * upper) + lower - 1).
	b2 := (0 - m) * lower / 2.
	^m * x + b2.! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/23/2004 22:08'!
geometry
	^self! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 7/28/2004 22:01'!
getCostume
	^nil! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:28'!
globalBounds
	^self localToGlobal: self localBounds! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:28'!
globalBounds: rect
	container ifNil:[^self bounds: rect].
	self bounds: (container globalToLocal: rect)! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/26/2005 12:07'!
globalPosition
	^self localToGlobal: self pivot! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:41'!
globalPosition: aPoint
	container ifNil:[^self position: aPoint].
	self position: (container globalToLocal: aPoint)! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:39'!
globalToLocal
	"Answer a transformation from the global into the local frame"
	^self localToGlobal inverseTransformation! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:28'!
globalToLocal: anObject
	"Transform anObject from global (world) coordinates into local coordinates"
	^anObject transformedBy: self globalToLocal.! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 4/2/2004 15:01'!
goBehind
	container addFirst: self.! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 3/24/2004 03:16'!
grabMe
	self hand attach: self.! !

!CPrimitiveCostume methodsFor: 'graphic' stamp: 'ar 9/4/2003 23:43'!
graphic
	^self propertyValueAt: #graphic! !

!CPrimitiveCostume methodsFor: 'graphic' stamp: 'ar 7/31/2004 18:32'!
graphic: aGraphic
	| myGraphic |
	myGraphic := self graphic.
	myGraphic == aGraphic ifTrue:[^self].
	aGraphic ifNil:[
		"Remove existing graphic"
		self invalidate.
		self removeProperty: #graphic.
		flags := flags bitClear: FlagHasGraphic.
		^self signal: #graphicChanged].

	myGraphic ifNil:[
		"Create new graphic"
		myGraphic := CCostumeGraphic forCostume: self.
		flags := flags bitOr: FlagHasGraphic.
		self propertyValueAt: #graphic put: myGraphic
	].

	myGraphic loadFrom: aGraphic.
	^aGraphic! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:10'!
grid
	^self propertyValueAt: #grid! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/26/2004 21:08'!
grid: newGrid
	| oldGrid |
	oldGrid := self grid.
	oldGrid ifNotNil:[oldGrid costume: nil].
	newGrid ifNotNil:[newGrid costume: self].
	self setFlag: FlagHasGrid value: newGrid notNil.
	self signal: #gridChanged.
	^self propertyValueAt: #grid put: newGrid.
! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'tk 1/4/2005 13:18'!
hResizing
	"Layout specific. This property describes how the receiver should be resized with respect to its container and its elements. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit my elements
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#rigid] ifNotNil:[props hResizing].! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 2/17/2004 18:53'!
hResizing: aSymbol
	"Layout specific. This property describes how the receiver should be resized with respect to its owner and its children. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit children
	"
	self hResizing == aSymbol ifTrue:[^self].
	self assureLayoutProperties hResizing: aSymbol.
	self signal: #hResizingChanged.
	self invalidateLayout.! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/29/2003 17:16'!
hand
	^Processor activeProcess hand! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 2/17/2004 18:57'!
handleChildEnter: anEvent with: aChild
	"System level event handling."
	(anEvent hand hasDragContents) 
		ifTrue:[self signal: #dragEnterChild with: anEvent with: aChild with: self].
	self signal: #mouseEnterChild with: anEvent with: aChild with: self.! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 2/17/2004 18:57'!
handleChildLeave: anEvent with: aChild
	"System level event handling."
	(anEvent hand hasDragContents) 
		ifTrue:[self signal: #dragLeaveChild with: anEvent with: aChild with: self].
	self signal: #mouseLeaveChild with: anEvent with: aChild with: self.! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'bf 4/27/2005 15:52'!
handleDrop: anEvent
	"Handle a drop event."
	anEvent wasHandled ifTrue:[^self].
	(self handlesDrop: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	self signal: #dragDrop with: anEvent contents with: anEvent! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 2/17/2004 18:55'!
handleDropFiles: anEvent
	"Handle a drop event."
	anEvent wasHandled ifTrue:[^self].
	(self handlesDropFiles: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	self signal: #dropFiles with: anEvent.! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: ''!
handleEvent: anEvent
	"Handle the given event"
	^anEvent sentTo: self.! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 4/2/2004 14:19'!
handleFocusEvent: localEvent
	"Dispatch the given event. The event will be passed to the front-most visible part that contains the event."
	| index |

	(self rejectsEvent: localEvent) ifTrue:[^self].
	"Traverse children"
	index := contents size + 1.
	[(index := index-1) > 0] whileTrue:[
		(( contents at: index) player dispatchEvent: localEvent) == #rejected 
			ifFalse:[index := 0]. "e.g., break"
	].
	self handleEvent: localEvent.! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 8/3/2004 17:21'!
handleKeyDown: anEvent
	"System level event handling."
	| map event |
	self isWorld ifTrue:[self pressedKeys add: anEvent keyValue].
	anEvent wasHandled ifTrue:[^self].
	(map := self keyDownMap) ifNotNil:[
		event := map lookupEvent: anEvent.
		event ifNotNil:[
			anEvent wasHandled: true.
			^self signal: event with: anEvent with: self.
		].
	].
	(self handlesEvent: #keyDown) ifTrue:[
		anEvent wasHandled: true.
		self signal: #keyDown with: anEvent with: self
	].! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 8/3/2004 17:21'!
handleKeyUp: anEvent
	"System level event handling."
	| map event |
	self isWorld ifTrue:[self pressedKeys remove: anEvent keyValue ifAbsent:[]].
	anEvent wasHandled ifTrue:[^self].
	(map := self keyUpMap) ifNotNil:[
		event := map lookupEvent: anEvent.
		event ifNotNil:[
			anEvent wasHandled: true.
			^self signal: event with: anEvent with: self.
		].
	].
	(self handlesEvent: #keyUp) ifTrue:[
		anEvent wasHandled: true.
		self signal: #keyUp with: anEvent with: self
	].! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 3/25/2004 20:08'!
handleKeystroke: anEvent
	"System level event handling."
	| map event |
	anEvent wasHandled ifTrue:[^self].
	(map := self keyStrokeMap) ifNotNil:[
		event := map lookupEvent: anEvent.
		event ifNotNil:[
			anEvent wasHandled: true.
			^self signal: event with: anEvent with: self.
		].
	].
	(self handlesEvent: #keyStroke) ifTrue:[
		anEvent wasHandled: true.
		self signal: #keyStroke with: anEvent with: self
	].! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 3/1/2005 09:31'!
handleMouseDown: anEvent
	"System level event handling."
	self signal: #mouseDownListen with: anEvent.
	anEvent wasHandled ifTrue:[^self]. "not interested"

	(self handlesEvent: #rawMouseDown) ifTrue:[
		"if you handle raw events everything is up to you - we don't care"
		anEvent wasHandled: true.
		^self signal: #rawMouseDown with: anEvent with: self.
	].
	anEvent hand releaseBalloonTarget: self.
	(self handlesMouseDown: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	self signal: #anyButtonDown with: anEvent with: self.
	anEvent redButtonChanged ifTrue:[
		(self handlesAnyOf:#(mouseClick mouseDrag mouseDoubleClick)) 
			ifTrue:[self startScript: #waitForClicksOrDrag:event:
						withArguments: {anEvent hand. anEvent clone}].
		self signal: #mouseDown with: anEvent with: self.
		self signal: #redButtonDown with: anEvent with: self.
		(self handlesEvent: #mouseStillDown) ifTrue:[
			self startScript: #onMouseStillDown: withArguments: {anEvent}.
		].
	].
	anEvent yellowButtonChanged 
		ifTrue:[self signal: #yellowButtonDown with: anEvent with: self].
	anEvent blueButtonChanged 
		ifTrue:[self signal: #blueButtonDown with: anEvent with: self].
	"Make me modal during mouse transitions"
	anEvent hand mouseFocus: self.! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 6/18/2005 17:31'!
handleMouseEnter: anEvent
	"System level event handling."

	(self handlesEvent: #rawMouseEnter) ifTrue:[
		"if you handle raw events everything is up to you - we don't care"
		^self signal: #rawMouseEnter with: anEvent with: self.
	].

	(anEvent hand hasDragContents) 
		ifTrue:[^self signal: #dragEnter with: anEvent with: self].

	self signal: #mouseEnter with: anEvent with: self.
	(anEvent anyButtonPressed or:[self balloonHelp == nil])
		ifFalse:[anEvent hand balloonTarget: self].! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 6/18/2005 17:31'!
handleMouseLeave: anEvent
	"System level event handling."

	(self handlesEvent: #rawMouseLeave) ifTrue:[
		"if you handle raw events everything is up to you - we don't care"
		^self signal: #rawMouseLeave with: anEvent with: self.
	].

	anEvent hand releaseBalloonTarget: self.
	(anEvent hand hasDragContents) 
		ifTrue:[^self signal: #dragLeave with: anEvent with: self]
		ifFalse:[^self signal: #mouseLeave with: anEvent with: self]! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 2/17/2004 18:56'!
handleMouseMove: anEvent
	"System level event handling."
	anEvent wasHandled: true.
	(self handlesEvent: #rawMouseMove) ifTrue:[
		"if you handle raw events everything is up to you - we don't care"
		^self signal: #rawMouseMove with: anEvent with: self.
	].
	(anEvent hand hasDragContents) 
		ifTrue:[^self signal: #dragMove with: anEvent with: self].
	self signal: #mouseMove with: anEvent with: self.
	anEvent redButtonPressed 
		ifTrue:[self signal: #redButtonMove with: anEvent with: self].
	anEvent yellowButtonPressed 
		ifTrue:[self signal: #yellowButtonMove with: anEvent with: self].
	anEvent blueButtonPressed 
		ifTrue:[self signal: #blueButtonMove with: anEvent with: self].! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 6/28/2005 17:51'!
handleMouseOver: anEvent
	"@@@@: Check this - should we generate any #mouseOver events here?!! @@@@"
	"System level event handling."
	anEvent hand noticeMouseOver: self event: anEvent! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 11/19/2004 08:48'!
handleMouseUp: anEvent
	"System level event handling."
	anEvent wasHandled ifTrue:[^self]. "not interested"
	(self handlesEvent: #rawMouseUp) ifTrue:[
		"if you handle raw events everything is up to you - we don't care"
		^self signal: #rawMouseUp with: anEvent with: self.
	].
	anEvent hand mouseFocus == self ifFalse:[
		container ifNotNil:[^self]. "Not interested in other parties"
	].
	anEvent anyButtonPressed ifFalse:[anEvent hand releaseMouseFocus: self].
	anEvent wasHandled: true.
	self signal: #anyButtonUp with: anEvent with: self.
	anEvent redButtonChanged ifTrue:[
		self signal: #redButtonUp with: anEvent with: self.
		self signal: #mouseUp with: anEvent with: self].
	anEvent yellowButtonChanged 
		ifTrue:[self signal: #yellowButtonUp with: anEvent with: self].
	anEvent blueButtonChanged 
		ifTrue:[self signal: #blueButtonUp with: anEvent with: self].
! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 3/24/2004 03:09'!
handlesAnyOf: eventNames
	| myPlayer |
	(super handlesAnyOf: eventNames) ifTrue:[^true].
	myPlayer := self player ifNil:[^false].
	^myPlayer handlesAnyOf: eventNames! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 2/17/2004 18:52'!
handlesDrop: anEvent
	^self handlesEvent: #dragDrop! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 2/17/2004 18:52'!
handlesDropFiles: anEvent
	^self handlesEvent: #dropFiles! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 3/24/2004 03:09'!
handlesEvent: eventName
	| myPlayer |
	(super handlesEvent: eventName) ifTrue:[^true].
	myPlayer := self player ifNil:[^false].
	^myPlayer handlesEvent: eventName! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 2/17/2004 18:51'!
handlesKeyboard: evt
	^self isWorld or:[self handlesAnyOf: #(keyDown keyUp keyStroke)]! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 3/1/2005 09:31'!
handlesMouseDown: evt
	"This really means: Am I interested in any kinds of events starting with a mouseDown"
	(self handlesAnyOf: #(anyButtonDown anyButtonUp)) 
		ifTrue:[^true].
	evt ifNil:[
		^self handlesAnyOf: #(
			mouseDown mouseUp mouseStillDown
			redButtonDown redButtonUp 
			yellowButtonDown yellowButtonUp 
			blueButtonDown blueButtonUp 
			mouseClick mouseDoubleClick mouseDrag
		).
	].
	evt redButtonChanged ifTrue:[
		^self handlesAnyOf: #(
			mouseDown mouseUp mouseStillDown
			redButtonDown redButtonUp
			mouseClick mouseDoubleClick mouseDrag
		).
	].
	evt yellowButtonChanged ifTrue:[
		^self handlesAnyOf: #(
			yellowButtonDown yellowButtonUp
			yellowButtonClick yellowButtonDrag
		).
	].
	evt blueButtonChanged ifTrue:[
		^self handlesAnyOf: #(
			blueButtonDown blueButtonUp
			blueButtonClick blueButtonDrag
		).
	].! !

!CPrimitiveCostume methodsFor: 'highlights' stamp: 'ar 3/23/2004 21:16'!
hasHighlight: propertyName
	self highlights do:[:highlight|
		highlight name = propertyName ifTrue:[^true]].
	^false! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 6/1/2005 20:56'!
hasKeyboardFocus
	^self propertyValueAt: #hasKeyboardFocus ifAbsent:[false].! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 6/1/2005 20:56'!
hasKeyboardFocus: aBool
	^self propertyValueAt: #hasKeyboardFocus put: aBool! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 2/17/2004 19:03'!
hasMouseFocus
	^self hand mouseFocus == self! !

!CPrimitiveCostume methodsFor: 'pen my motion' stamp: ''!
hasPen
	^self hasProperty: #pen! !

!CPrimitiveCostume methodsFor: 'properties' stamp: ''!
hasProperty: propName
	"Answer whether the receiver has the given property.  Deemed to have it only if I have a property dictionary entry for it and that entry is neither nil nor false"
	^properties ifNil:[false] ifNotNil:[properties includesKey: propName]! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 2/24/2009 14:21'!
hasTranslucentFill
	"Answer whether the receiver has any parts where the background must be restored"
	^fill isTranslucent or:[fill isTransparent]! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 22:27'!
heading
	"Answer the heading of the receiver, e.g., its angle measured clockwise from its direction."
	| deg |
	deg := self direction degrees + 90 \\ 360.0.
	deg > 180.0 ifTrue:[deg := deg-360].
	^deg! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:37'!
heading: aHeading
	"Indicate the new heading for the receiver"
	self rotateBy: aHeading - self heading.
	^aHeading! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
headingFor: aDirection
	| fwd |
	fwd := (self transform transformDirection: 0@-1) normalized.
	aDirection == #forward ifTrue:[^0@-1].
	aDirection == #back ifTrue:[^fwd negated].
	fwd := fwd y @ fwd x negated.
	aDirection == #left ifTrue:[^fwd].
	aDirection == #right ifTrue:[^fwd negated].
	self error:'No such direction: ', aDirection.! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/25/2005 20:55'!
height
	^self extent y! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/25/2005 20:25'!
height: aNumber
	self extent: self width@aNumber.
	^aNumber! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: ''!
hidden
	^(flags bitAnd: FlagHidden) == FlagHidden! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 8/10/2004 11:33'!
hidden: aBool
	self hidden = aBool ifTrue:[^self].
	self invalidate.
	flags := aBool ifTrue:[flags bitOr: FlagHidden] ifFalse:[flags bitClear: FlagHidden].
	self invalidate.
	self signal: #flagsChanged.
	container ifNotNil:[container invalidateLayout].! !

!CPrimitiveCostume methodsFor: 'highlights' stamp: 'ar 9/4/2003 23:50'!
highlightBounds
	"Return the bounds used for highlighting the receiver"
	^self propertyValueAt: #highlightBounds ifAbsent:[self localBounds]! !

!CPrimitiveCostume methodsFor: 'highlights' stamp: 'ar 9/4/2003 23:37'!
highlightBounds: newBounds
	"Set the bounds used for highlighting the receiver"
	newBounds == nil
		ifTrue:[self removeProperty: #highlightBounds]
		ifFalse:[self propertyValueAt: #highlightBounds put: newBounds].! !

!CPrimitiveCostume methodsFor: 'highlights' stamp: 'ar 3/23/2004 21:17'!
highlightNamed: propertyName
	self highlights do:[:highlight|
		highlight name = propertyName ifTrue:[^highlight]].
	^nil! !

!CPrimitiveCostume methodsFor: 'highlights' stamp: 'ar 9/24/2003 16:46'!
highlights
	"Return all the highlights of the receiver"
	^self propertyValueAt: #highlights ifAbsent:[#()]! !

!CPrimitiveCostume methodsFor: 'highlights' stamp: 'ar 9/4/2003 23:37'!
highlights: anArray
	"Install all the highlights of the receiver"
	self invalidate.
	(anArray == nil or:[anArray isEmpty]) 
		ifTrue:[self removeProperty: #highlights.
				flags := flags bitClear: FlagHasHighlights]
		ifFalse:[self propertyValueAt: #highlights put: anArray.
				flags := flags bitOr: FlagHasHighlights].
	self invalidateLayout.
	self invalidate.! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 8/18/2007 20:51'!
ignoreShrinkWrapConstraints
	"Answer whether shrink wrap constraints should be ignored for the receiver. "
	^contents size = 0 "e.g., ignore shrinkWrap if there are no children"
		and:[(flags bitAnd: FlagHasText) = 0] "and if there is no text"! !

!CPrimitiveCostume methodsFor: 'adding' stamp: ''!
indexOf: anElement
	"Answer the index of the first occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer 0."
	^ self indexOf: anElement ifAbsent: [0]! !

!CPrimitiveCostume methodsFor: 'adding' stamp: ''!
indexOf: anElement ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the 
	result of evaluating the argument, exceptionBlock."
	^ self indexOf: anElement startingAt: 1 ifAbsent: exceptionBlock! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'ar 1/8/2003 17:12'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement after start
	within the receiver. If the receiver does not contain anElement, 
	answer the 	result of evaluating the argument, exceptionBlock."
	start to: contents size do:[:index |
		(contents at: index) = anElement ifTrue: [^ index]].
	^ exceptionBlock value! !

!CPrimitiveCostume methodsFor: 'pen trails in me' stamp: 'ar 6/25/2005 19:35'!
indicatePenTrails
	self flag: #fixMe. "should be ^self primitivPen notNil but this broke the first time around since #primitivePen was just added. Remove this after the next config map which ensures that #primitivePen exists."
	^(self propertyValueAt: #primitivePen) "aka: self primitivePen" notNil! !

!CPrimitiveCostume methodsFor: 'pen trails in me' stamp: 'ar 6/25/2005 19:23'!
indicatePenTrails: aBool
	"Presence of a Pen tells that I record trails.  Make the Form small in case it is never used.  Will be expanded when drawing begins."

	| primBlt |
	aBool & (self indicatePenTrails not) 
		ifTrue: [
			(primBlt := BitBlt new) setDestForm: (Form extent: 16@16 depth: Display depth);
					sourceOrigin: 0@0.
			self primitivePen: primBlt].
	aBool not & (self indicatePenTrails) 
		ifTrue: [self primitivePen: nil.
				self invalidate].

	^ aBool! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 6/25/2005 20:25'!
infinite
	"Answer whether the receiver should be treated as an infinite plane"
	^self propertyValueAt: #infinite ifAbsent:[false]! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 6/25/2005 20:25'!
infinite: aBool
	"Indicate whether the receiver should be treated as an infinite plane"
	self propertyValueAt: #infinite put: (aBool ifTrue:[true] ifFalse:[nil]).
	self signal: #flagsChanged! !

!CPrimitiveCostume methodsFor: 'initialize' stamp: 'ar 6/24/2005 21:31'!
initialize
	bounds ifNil:[bounds := 0@0 corner: 50@40].
	contents ifNil:[contents := #()].
	fill ifNil:[fill := CNoFill forCostume: self].
	border ifNil:[border := CNoBorder forCostume: self].
	flags ifNil:[flags := 0].
	self transform ifNil:[self transform: CIdentityTransform new].
! !

!CPrimitiveCostume methodsFor: 'text anchoring' stamp: 'ar 6/25/2005 19:07'!
inlineTextAnchorPosition: aPoint
	"Set the new inline text anchor position for the receiver. This method is currently a hack since not all of the transformation model has been figured out with respect to text handling. Currently, the point describes the left and the baseline position of the receiver with respect to a text origin of 0@0. This text origin is relative to the containers text bounds so we need to do a bit of conversion here."
	| distance pos |
true ifTrue:[^self].
	container isTextCostume ifFalse:[^self]. "should never happen"
	distance := (self position y - self top) asInteger.
	pos := container textBounds topLeft + (aPoint x @ (aPoint y - distance)).
	self position: pos! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/24/2004 03:20'!
innerBounds
	"Return the inner rectangle enclosed by the bounds of this costume excluding the space taken by its borders."
	"^self localBounds insetBy: self borderWidth"
	^border insetBounds: self localBounds! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 12/8/2003 19:42'!
insetColor
	"Return the color to be used for shading inset borders.  The default is my own color, but it might want to be, eg, my container's color.  Whoever's color ends up prevailing, the color itself gets the last chance to determine, so that when, for example, an InfiniteForm serves as the color, callers won't choke on some non-Color object being returned"
	(fill isTransparent and:[container notNil]) ifTrue:[^container insetColor].
	^fill asColor! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: 'ar 3/24/2004 03:11'!
interpolationForStyle: aStyle
	aStyle == #gently ifTrue:[
		^[:x| self gently: x lowerBound: 0.3 upperBound: 0.8].
	].
	^[:x| self abruptly: x]! !

!CPrimitiveCostume methodsFor: 'initialize' stamp: 'ar 10/18/2004 20:30'!
intoWorld: aWorld
	"This costume has just appeared in a new world. Note:
		* aWorld can be nil (due to optimizations in other places)
		* owner is already set
		* owner's children may not include receiver yet.
	Important: Keep this method fast - it is run whenever objects are added."
	aWorld ifNil:[^self].
	self asPlayer resumeTickingScripts.
	self signal: #opened with: aWorld.
	self signal: #open with: aWorld.
	contents do:[:m| m intoWorld: aWorld].! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 6/25/2005 15:00'!
invalidate
	"Invalidate the receiver so that it will be redrawn"
	(flags bitAnd: FlagHidden) == 0 ifFalse:[^self].
	self invalidate: self drawingBounds. "locally"
! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/23/2004 21:27'!
invalidate: aRectangle
	"Invalidate the rectangle so that it will be redrawn within the receiver"
	^self invalidate: aRectangle from: self! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 6/21/2005 18:35'!
invalidate: aRectangle from: aCostume
	| damageRect |
	(flags bitAnd: FlagInvalidationSuppressed) == 0 ifFalse:[^self].
	(flags bitAnd: FlagHidden) == 0 ifFalse:[^self].
	damageRect := aRectangle.
	(aCostume ~~ self and:[self clipping]) ifTrue:[
		"Clip to receiver's clipping bounds if the damage came from a part"
		damageRect := damageRect intersect: self clippingBounds.
	].
	"Notify our cache if any"
	(flags bitAnd: FlagCached) == 0 
		ifFalse:[self localCache invalidate: damageRect].
	"Notify our highlights if any"
	(flags bitAnd: FlagHasHighlights) = 0 
		ifFalse:[self highlights do:[:hl| hl noteInvalidationOf: aRectangle]].
	"Notify any clients"
	(flags bitAnd: FlagInvalidated) = 0 ifTrue:[
		self signal: #invalidate.
		flags := flags bitOr: FlagInvalidated.
	].
	"Pass invalidation area to parent"
	container ifNotNil: [
		damageRect := self transform transformRect: damageRect.
		container invalidate: damageRect from: self.
	].! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 6/25/2005 14:51'!
invalidateLayout
	| layout |
	self layoutNeeded ifTrue:[^self]. "layout will be recomputed so don't bother"
	self layoutNeeded: true.
	layout := self layout.
	layout ifNotNil:[layout flushLayoutCache].
	container ifNotNil: [container invalidateLayout].
! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 6/25/2005 14:56'!
invalidationBounds
	"Return the bounds for invalidating the receiver, e.g., the transformed drawing bounds."
	^self transform transformRect: self drawingBounds! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: ''!
invalidationSuppressed
	^(flags bitAnd: FlagInvalidationSuppressed) = FlagInvalidationSuppressed! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: ''!
invalidationSuppressed: aBool
	"Temporarily suppress invalidation requests"
	flags := aBool 
		ifTrue:[flags bitOr: FlagInvalidationSuppressed]
		ifFalse:[flags bitClear: FlagInvalidationSuppressed].! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 1/8/2003 17:15'!
isEmpty
	^contents size = 0! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 9/10/2003 18:15'!
isIn: aContainer
	"Return true if the receiver is contained in aContainer."
	| cc |
	aContainer ifNil:[^true].
	"Open-coded for speed."
	cc := container.
	[cc == nil] whileFalse:[
		aContainer == cc ifTrue:[^true].
		cc := cc container].
	^false! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 3/11/2005 21:11'!
isInKeyboardFocus
	self hasKeyboardFocus ifTrue:[^true].
	container ifNil:[^false].
	^container isInKeyboardFocus! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 7/15/2003 16:18'!
isInWorld
	"Remove this at some point"
	self isWorld ifFalse:[^self world notNil].
	container ifNil:[^false].
	^container world notNil! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 2/17/2004 19:05'!
isOrIn: aCostume
	"Return true if the receiver is contained in aContainer"
	aCostume == nil ifTrue:[^true].
	self == aCostume ifTrue:[^true].
	self allContainersDo:[:aContainer| aContainer == aCostume ifTrue:[^true]].
	^false! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'tk 6/22/2004 17:46'!
isPrimitiveCostume
	"well, of course"
	^true! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 1/25/2007 14:55'!
isTabStop
	^false! !

!CPrimitiveCostume methodsFor: 'testing' stamp: ''!
isTextCostume
	^false! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 9/4/2003 23:50'!
isWorld
	^self propertyValueAt: #isWorld ifAbsent:[false]! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 9/4/2003 23:38'!
isWorld: aBool
	^self propertyValueAt: #isWorld put: aBool! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 3/24/2004 01:23'!
keyDownMap
	^self propertyValueAt: #keyDownMap! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 3/24/2004 01:23'!
keyDownMap: aMap
	^self propertyValueAt: #keyDownMap put: aMap! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 3/24/2004 01:24'!
keyStrokeMap
	^self propertyValueAt: #keyStrokeMap! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 3/24/2004 01:24'!
keyStrokeMap: aMap
	^self propertyValueAt: #keyStrokeMap put: aMap! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 3/24/2004 01:23'!
keyUpMap
	^self propertyValueAt: #keyUpMap! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 3/24/2004 01:24'!
keyUpMap: aMap
	^self propertyValueAt: #keyUpMap put: aMap! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 6/1/2005 20:57'!
keyboardFocusEnter
	"The receiver has become the new keyboard focus"
	self hasKeyboardFocus: true.
	(flags bitAnd: FlagHasText) = 0 ifFalse:[
		self textHolder keyboardFocusEnter.
	].
	self signal: #keyboardFocusEnter.
! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 6/1/2005 20:57'!
keyboardFocusLeave
	"The receiver has become the new keyboard focus"
	self hasKeyboardFocus: false.
	(flags bitAnd: FlagHasText) = 0 ifFalse:[
		self textHolder keyboardFocusLeave.
	].
	self signal: #keyboardFocusLeave.! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'tk 6/17/2004 15:28'!
keysAndValuesDo: aBlock
	"index numbers are the keys"
	1 to: contents size do:[:ii | aBlock value: ii value: (contents at: ii)].
! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'tk 1/4/2005 13:19'!
layout
	"Return the layout policy describing how elements of the receiver should be arranged on the screen."
	^self propertyValueAt: #layout! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'tk 1/4/2005 12:51'!
layout: aLayout
	"layout is the policy describing how the elements of this object should appear.  An example is TableLayout."
	| myLayout |
	(myLayout := self layout) == aLayout ifTrue:[^aLayout].
	myLayout ifNotNil:[myLayout costume: nil].
	self propertyValueAt: #layout put: aLayout.
	aLayout ifNotNil:[aLayout costume: self].
	self signal: #layoutChanged.
	self invalidateLayout.
	^aLayout! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'tk 1/4/2005 13:19'!
layoutBounds
	"Return the bounds for laying out elements of the receiver.
	Note: The layout bounds always start relative to the scroll origin."
	| box |
	box := self innerBounds.
	box := box insetBy: self layoutInset.
	^box translateBy: self scrollOffset negated! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 4/19/2005 21:37'!
layoutBounds: newBounds
	"Set the bounds for laying out elements of the receiver"
	| box local keptExtent |
	box := newBounds.
	box := box expandBy: self layoutInset.
	box := border expandBounds: box.
	box = (local := self localBounds) ifTrue:[^self].
	keptExtent := box extent = local extent.

	self invalidate.
	self privateBounds:(bounds origin + box origin extent: box extent).
	self invalidate.
	self invalidateLayout.
	keptExtent ifFalse:[self signal: #extentChanged].
	self signal: #geometryChanged.
	self frameChanged.! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 9/4/2003 23:43'!
layoutFrame
	"Layout specific. Return the layout frame describing where the receiver should appear in a proportional layout"
	^self propertyValueAt: #layoutFrame! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'bf 4/27/2004 01:14'!
layoutFrame: aFrame
	"Layout specific. Return the layout frame describing where the receiver should appear in a proportional layout"
	| layoutFrame |
	layoutFrame := aFrame asLayoutFrame.
	self propertyValueAt: #layoutFrame put: layoutFrame.
	self invalidateLayout.
	^ layoutFrame! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 6/24/2005 21:04'!
layoutInBounds: cellBounds
	"Layout specific. Apply the given global bounds to the receiver after being layed out in its container."
	| box aSymbol |
	self layoutNeeded ifTrue:[
		"We need to get our own layout right first"
		box := cellBounds origin extent: "adjust for #rigid receiver"
			(self hResizing == #rigid ifTrue:[self bounds extent x] ifFalse:[cellBounds extent x]) @
			(self vResizing == #rigid ifTrue:[self bounds extent y] ifFalse:[cellBounds extent y]).
		self bounds: box.
		self doLayoutIn: self layoutBounds].
	cellBounds = self fullBounds ifTrue:[^self]. "already up to date"
	cellBounds extent = self fullBounds extent "nice fit"
		ifTrue:[^self topLeft: cellBounds origin].
	box := self bounds.
	"match #spaceFill constraints"
	self hResizing == #spaceFill 
		ifTrue:[box := box withWidth: cellBounds width].
	self vResizing == #spaceFill
		ifTrue:[box := box withHeight: cellBounds height].
	"align accordingly"
	aSymbol := (container ifNil:[self]) cellPositioning.
	box := box align: (box perform: aSymbol) with: (cellBounds perform: aSymbol).
	"install new bounds"
	box extent = self bounds extent 
		ifTrue:[^self topLeft: box origin].
	"Since the extent is subject to change. layout right away"
	self layoutNeeded: true. "don't bother broadcasting"
	self bounds: box.
	"and do the layout"
	self doLayoutIn: self layoutBounds.! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/24/2004 03:35'!
layoutInset
	| props |
	props := self layoutProperties.
	^props ifNil:[0@0] ifNotNil:[props layoutInset].! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/24/2004 03:36'!
layoutInset: newInset
	self layoutInset = newInset ifTrue:[^self].
	self assureLayoutProperties layoutInset: newInset.
	self invalidateLayout.! !

!CPrimitiveCostume methodsFor: 'layout' stamp: ''!
layoutNeeded				 
	^(flags bitAnd: FlagLayoutNeeded) == FlagLayoutNeeded! !

!CPrimitiveCostume methodsFor: 'layout' stamp: ''!
layoutNeeded: aBool				 
	flags := aBool ifTrue:[flags bitOr: FlagLayoutNeeded] ifFalse:[flags bitClear: FlagLayoutNeeded].! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 9/4/2003 23:43'!
layoutProperties
	"Return the current layout properties associated with the receiver"
	^self propertyValueAt: #layoutProperties! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 9/4/2003 23:39'!
layoutProperties: newProperties
	"Return the current layout properties associated with the receiver"
	self propertyValueAt: #layoutProperties put: newProperties.! !

!CPrimitiveCostume methodsFor: 'layout' stamp: ''!
layoutProportionallyIn: newBounds
	"Layout specific. Apply the given bounds to the receiver."
	| box frame |
	frame := self layoutFrame ifNil:[^self].
	"before applying the proportional values make sure the receiver's layout is computed"
	self fullBounds. "sigh..."
	"compute the cell size the receiver has given its layout frame"
	box := frame layout: self bounds in: newBounds.
	(box = self bounds) ifTrue:[^self]. "no change"
	^self layoutInBounds: box.! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/15/2003 17:08'!
layoutSimplyIn: newBounds
	"Layout specific. Apply the given bounds to the receiver."
	| box hFit vFit myBounds |
	hFit := self hResizing.
	vFit := self vResizing.
	(hFit == #spaceFill or:[vFit == #spaceFill]) ifFalse:[^self].
	myBounds := self bounds.
	box := myBounds.
	hFit == #spaceFill ifTrue:[box := box withWidth: newBounds width].
	vFit == #spaceFill ifTrue:[box := box withHeight: newBounds height].
	box = myBounds ifTrue:[^self].
	^self layoutInBounds: box.! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 3/23/2004 23:16'!
layoutTargets
	^contents select:[:cc| cc visible].! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:28'!
left
	^self bounds left! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:29'!
left: aNumber
	self moveBy: (aNumber - self left) @ 0.
	^aNumber
! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'bgf 2/23/2008 12:42'!
leftCenter
	^self bounds leftCenter! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'bgf 2/23/2008 12:42'!
leftCenter: aPoint
	self moveBy: (aPoint - self leftCenter).
! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/29/2004 14:54'!
localAreasRemainingToFill: aRectangle
	"Answer the damage regions needed to be restored after the receiver has been drawn in aRectangle."
	| outer box |
	self userDraw 
		ifTrue: [^ Array with: aRectangle].
	"translucent fills need to restore background first"
	self hasTranslucentFill
		ifTrue: [^ Array with: aRectangle].
	outer := box := self localBounds.
	"inset by border if translucent border is present"
	"bw := self borderWidth.
	(bw > 0 and:[self borderColor isTranslucent]) 
		ifTrue:[box := box insetBy: bw]."
	border isTranslucent ifTrue:[box := border insetBounds: box].
	"if no rounded corners take what's left"
	self wantsRoundedCorners 
		ifFalse: [^aRectangle areasOutside: box].
	"Eliminate most of the damage for the common case when just the receiver is being redrawn."
	aRectangle = box ifTrue:[
		^aRectangle cornerRectsOfSize: CCornerRounder cornerSize
	].
	"Otherwise take the interior"
	box := box intersect: (CCornerRounder rectWithinCornersOf: outer).
	^aRectangle areasOutside: box.! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/25/2005 20:34'!
localBounds
	"Answer the receiver's local bounds"
	(flags bitAnd: FlagHasScroller) == 0 ifTrue:[^bounds].
	^bounds translateBy: self scrollOffset! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'bf 7/22/2005 15:51'!
localBounds: aBox
	"Indicate the receiver's local bounds"
	| delta |
	delta := aBox origin - self localBounds origin.
	delta isZero ifFalse:[self moveBy: (self direction * delta x) + (self normal * delta y)].
	self extent: aBox extent.
	^aBox! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:12'!
localCache
	^self propertyValueAt: #localCache! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:12'!
localCache: aCache
	self propertyValueAt: #localCache put: aCache.
	self cacheEnabled: aCache notNil.! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 6/25/2005 14:50'!
localFullBounds
	"Return the bounding box of the receiver and all its parts. Recompute the layout if necessary."
	self layoutNeeded ifTrue:[
		"Errors at this point can be critical so make sure we catch 'em all right"
		[self doLayoutIn: self layoutBounds] on: Error do:[:ex| ex pass].
	].
	^self localBounds! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 6/21/2005 18:34'!
localFullDrawOn: transformedCanvas
	"Draw the receiver in the presence of a full transform"
	| highlights index hl |
	flags := flags bitClear: FlagInvalidated.
	highlights := self highlights.
	"Step 1: Draw highlights below the receiver"
	index := 1.
	[index <= highlights size and:[(hl := highlights at: index) priority < 0]]
		whileTrue:[	hl drawHighlightOn: transformedCanvas.
					index := index + 1].
	"Skip zero-priority highlights"
	[index <= highlights size and:[(highlights at: index) priority = 0]]
		whileTrue:[index := index + 1].
	transformedCanvas roundCornersOf: self in: self localBounds during:[
		"Step 2: Draw receiver itself"
		self drawOn: transformedCanvas.
	].
	"Step 3: Draw pen trails"
	self drawPenTrailsOn: transformedCanvas.
	"Step 4: Draw all the parts"
	transformedCanvas clipBy: self clippingBounds during:[:clip|
		self drawContentsOn: clip.
	].
	(flags bitAnd: FlagHasOverlay) = 0 
		ifFalse:[self drawOverlayOn: transformedCanvas].
	"Step 5: Draw highlights on top of the receiver"
	index to: highlights size do:[:i|
		(highlights at: i) drawHighlightOn: transformedCanvas.
	].
! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:39'!
localToGlobal
	"Answer a transformation from the local into the global frame"
	^container 
		ifNil:[self childToParent]
		ifNotNil:[self childToParent transformedBy: container localToGlobal]! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:39'!
localToGlobal: anObject
	"Transform anObject from local coordinates into global (world) coordinates"
	^anObject transformedBy: self localToGlobal.! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:39'!
localToWorld
	"Answer a transformation from the local into the global frame"
	self isWorld ifTrue:[^CIdentityTransform new].
	^container 
		ifNil:[self childToParent]
		ifNotNil:[self childToParent transformedBy: container localToWorld]! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:39'!
localToWorld: anObject
	"Transform anObject from global (world) coordinates into local coordinates"
	^anObject transformedBy: self localToWorld.! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: ''!
localTransformBy: aTransform
	self transform: (aTransform transformedBy: self transform)! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 3/23/2004 23:15'!
lockContents: aBool
	contents do:[:each| each locked: aBool].! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 3/23/2004 23:39'!
locked				 
	^(flags bitAnd: FlagLocked) == FlagLocked! !

!CPrimitiveCostume methodsFor: 'testing' stamp: 'ar 3/23/2004 23:39'!
locked: aBool
	flags := aBool ifTrue:[flags bitOr: FlagLocked] ifFalse:[flags bitClear: FlagLocked].
	self signal: #flagsChanged.! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 2/16/2003 14:56'!
makeThumbnail: extent
	"Create a thumbnail of the receiver with the given extent"
	^self makeThumbnail: extent depth: Display depth! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 2/16/2003 20:26'!
makeThumbnail: extent depth: depth
	"Create a thumbnail of the receiver with the given extent"
	| myPicture form scale thumb |
	myPicture := self asFormOfDepth: depth.
	form := Form extent: extent depth: depth.
	scale := 1.
	(myPicture width > 0 and:[myPicture height > 0])
		ifTrue:[scale := (extent x / myPicture width) min: (extent y / myPicture height)].
	scale < 1.0 ifTrue:[
		thumb := Form extent: (myPicture extent * scale) asIntegerPoint depth: myPicture depth.
		myPicture displayScaledOn: thumb.
		thumb displayOn: form at: (form extent - thumb extent // 2) rule: Form paint.
	] ifFalse:[
		myPicture displayOn: form at: (form extent - myPicture extent // 2) rule: Form paint.
	].
	^form! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/23/2004 23:07'!
minExtent
	"Layout specific. Return the minimum size the receiver can be represented in."
	| minExtent extra hFit vFit |
	hFit := self hResizing.
	vFit := self vResizing.
	"If both are rigid then just take my full bounds"
	(hFit == #rigid and:[vFit == #rigid]) 
		ifTrue:[^self fullBounds extent].
	"Compute the minimal extent of my children"
	minExtent := self minimalPartsExtent.
	"receiver with #shrinkWrap constraints but no children is being treated #rigid"
	self ignoreShrinkWrapConstraints ifTrue:[
		hFit == #shrinkWrap ifTrue:[hFit := #rigid].
		vFit == #shrinkWrap ifTrue:[vFit := #rigid]].
	"Ignore the minExtent for clipping #spaceFill costumes. 
	This allows #spaceFill costumes to shrink below their parts size,
	but ONLY if clipping."
	self clipping ifTrue:[
		hFit == #spaceFill ifTrue:[minExtent := 0 @ minExtent y].
		vFit == #spaceFill ifTrue:[minExtent := minExtent x @ 0].
	].
	"Fix up #rigid costumes which will not resize at all. 
	For all others outset the minExtent by the costume's layout inset."
	hFit == #rigid
		ifTrue:[	minExtent := self fullBounds extent x @ minExtent y]
		ifFalse:[extra := self bounds width - self layoutBounds width.
				minExtent := (minExtent x + extra) @ minExtent y].
	vFit == #rigid
		ifTrue:[minExtent := minExtent x @ self fullBounds extent y]
		ifFalse:[extra := self bounds height - self layoutBounds height.
				minExtent := minExtent x @ (minExtent y + extra)].
	"Constrain minExtent by user-supplied minimal width / height"
	minExtent := minExtent max: (self minWidth@self minHeight).
	^minExtent! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/23/2004 20:17'!
minHeight
	| minH |
	minH := self propertyValueAt: #minHeight ifAbsent:[1].
	(flags bitAnd: FlagHasText) = 0 ifFalse:[
		minH := minH max: self textHolder minHeight.
	].
	(flags bitAnd: FlagHasGraphic) = 0 ifFalse:[
		minH := minH max: self graphic minHeight.
	].
	^minH! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 9/4/2003 23:39'!
minHeight: aNumber
	aNumber == nil
		ifTrue:[self removeProperty: #minHeight]
		ifFalse:[self propertyValueAt: #minHeight put: aNumber].
	self invalidateLayout.! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 6/24/2005 21:33'!
minWidth
	| minW |
	minW := self propertyValueAt: #minWidth ifAbsent:[1].
	(flags bitAnd: FlagHasText) = 0 ifFalse:[
		minW := minW max: self textHolder minWidth.
	].
	(flags bitAnd: FlagHasGraphic) = 0 ifFalse:[
		minW := minW max: self graphic minWidth.
	].
	^minW! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 9/4/2003 23:39'!
minWidth: aNumber
	aNumber == nil
		ifTrue:[self removeProperty: #minWidth]
		ifFalse:[self propertyValueAt: #minWidth put: aNumber].
	self invalidateLayout.! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'tk 1/4/2005 13:20'!
minimalPartsExtent
	"Compute the minimal extent of my elements. This method has been factored out since Text costumes need to handle this slightly differently."
	| layout minExtent |
	layout := self layout.
	layout 
		ifNil:[minExtent := self contentsBounds corner - self innerBounds origin]
		ifNotNil:[minExtent := layout minExtentOf: self in: self layoutBounds].
	(flags bitAnd: FlagHasText) = 0 ifFalse:[
		"Include the text itself. Note: For both, wrapped and non-wrapped text
		we take the current editor extent. Obviously, this is correct for wrapped
		text but also for non-wrapped since the editor has layed out the text
		in its current extent which means that it is possible to represent the
		text in these minimal bounds."
		minExtent := minExtent max: self textHolder minimalTextExtent.
	].
	^minExtent! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:13'!
mouseCursor
	"Answer the associated mouse cursor for this costume"
	^self propertyValueAt: #mouseCursor! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 6/28/2005 17:56'!
mouseCursor: newCursor
	"Answer the associated mouse cursor for this costume"
	self signalChanged: #mouseCursorChanged from: self mouseCursor to: newCursor.
	self propertyValueAt: #mouseCursor put: newCursor.
	^ newCursor! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
move
	^self move: #forward distance: 10 duration: 1 style: #gently! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
move: direction
	^self move: direction distance: 10 duration: 1 style: #gently! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
move: direction distance: dist
	^self move: direction distance: dist duration: 1 style: #gently! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
move: direction distance: dist duration: dur
	^self move: direction distance: dist duration: dur style: #gently! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: 'auto 6/3/2005 16:45'!
move: 		direction
distance: 	distance
duration: 	duration
style: 		style
	"Move me according the given attributes."
	| msecDuration nowTime startTime styleFunc lastState startState endState deltaTime proportion newState |
	"Get the ultimate state we're interested in"
	endState := (self headingFor: direction) * distance.
	duration <= 0
		ifTrue:[^self localTransformBy: (COffsetTransform withOffset: endState)].
	"Compute msecs since that's what we'll be working on here"
	msecDuration := duration * 1000.
	"Get our starting time to compute incremental changes accordingly"
	nowTime := startTime := Time millisecondClockValue.
	"Get the interpolation function (evaluating from 0-1 and returning values from 0-1)"
	styleFunc := self interpolationForStyle: style.
	"Reset the interpolation state"
	lastState := startState := 0@0.
	"And go moving until we're out of time"
	Processor activeProject
		forEach: #tick 
		until:[nowTime >= (startTime + msecDuration)]
		do:[
			"Compute the time at the tick and the delta to when we started"
			nowTime := Time millisecondClockValue.
			deltaTime := (nowTime - startTime) min: msecDuration.
			"Compute the proportion of time that's over"
			proportion := deltaTime asFloat / msecDuration asFloat.
			"Map it through the interpolation style"
			proportion := styleFunc value: proportion.
			"Compute the state (e.g., offset) at our new proportion"
			newState := startState interpolateTo: endState at: proportion.
			"Transform incrementally - this is so that we can superimpose animations."
			self localTransformBy: (COffsetTransform withOffset: newState - lastState).
			"Remember last state for the next round"
			lastState := newState.
	].
	"Apply the remaining part of the transform (if any)"
	self localTransformBy: (COffsetTransform withOffset: endState - lastState).! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: ''!
move: direction duration: aDuration
	^self move: direction distance: 1 duration: aDuration style: #gently! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 3/7/2006 00:49'!
moveBy: delta
	| box oldRect newRect suppressed |
	delta isZero ifTrue:[^self].
	box := self drawingBounds.
	oldRect := self childToParent transformRect: box.
	suppressed := self invalidationSuppressed.
	self invalidationSuppressed: true.
	self transform: ((COffsetTransform withOffset: delta) composedWithLocal: self transform).
	self invalidationSuppressed: suppressed.
	container ifNil:[^self].
	newRect := self childToParent transformRect: box.
	suppressed ifFalse:[container privateBltMove: self from: oldRect to: newRect].
	(self hasPen and: [self pen isDown]) ifTrue:[
		container ifNotNil: [container notePen: self movedBy: delta].
	].
	container invalidateLayout.! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 6/25/2005 19:07'!
moveBy: delta duration: duration
	"Move the receiver by delta over the given duration"
	^self moveTo: self position + delta duration: duration! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: 'auto 6/3/2005 14:47'!
moveBy: 		distance
duration: 	duration
style: 		style
	"Move me according the given attributes."
	| msecDuration nowTime startTime styleFunc lastState startState endState deltaTime proportion newState |
	"Compute msecs since that's what we'll be working on here"
	msecDuration := duration * 1000.
	"Get our starting time to compute incremental changes accordingly"
	nowTime := startTime := Time millisecondClockValue.
	"Get the interpolation function (evaluating from 0-1 and returning values from 0-1)"
	styleFunc := self interpolationForStyle: style.
	"Reset the interpolation state"
	lastState := startState := 0@0.
	"Get the ultimate state we're interested in"
	endState := distance.
	"And go moving until we're out of time"
	Processor activeProject
		forEach: #tick 
		until:[nowTime >= (startTime + msecDuration)]
		do:[
			"Compute the time at the tick and the delta to when we started"
			nowTime := Time millisecondClockValue.
			deltaTime := (nowTime - startTime) min: msecDuration.
			"Compute the proportion of time that's over"
			proportion := deltaTime asFloat / msecDuration asFloat.
			"Map it through the interpolation style"
			proportion := styleFunc value: proportion.
			"Compute the state (e.g., offset) at our new proportion"
			newState := startState interpolateTo: endState at: proportion.
			"Transform incrementally - this is so that we can superimpose animations."
			self localTransformBy: (COffsetTransform withOffset: newState - lastState).
			"Remember last state for the next round"
			lastState := newState.
	].
	"Apply the remaining part of the transform (if any)"
	self localTransformBy: (COffsetTransform withOffset: endState - lastState).
	"Renormalize transform after completing the move"
	self normalizeWithRotationDegrees.! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 3/24/2004 03:13'!
moveDownLayer
	container add: self at: (container indexOf: self) + 1! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: ''!
moveTo: aPoint
	self moveBy: aPoint - self position.! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 6/25/2005 19:08'!
moveTo: target duration: duration
	"Move the receiver to match its reference position with the given reference position"
	| lastTime nowTime myPos itsPos delta startTime myStart itsStart |
	duration <= 0 ifTrue:[
		^self moveBy: target position - self position].
	startTime := lastTime := Time millisecondClockValue.
	myStart := self position.
	itsStart := target position.
	Processor activeProject forEach: #tick do:[
		nowTime := Time millisecondClockValue.
		(nowTime - startTime * 0.001) >= duration ifTrue:[
			^self moveTo: target position duration: 0.
		].
		itsPos := target position.
		(itsPos = itsStart) ifTrue:[
			"Simple case since 'it' has not moved"
			delta := nowTime - startTime * 0.001 / duration.
			self moveTo: myStart + (itsStart - myStart * delta) duration: 0.
		] ifFalse:[
			"Tricky case - 'it' has moved so adjust our strategy."
			myPos := self position.
			delta := nowTime - lastTime * 0.001 / (duration - (lastTime - startTime * 0.001)).
			self moveTo: myPos + (itsPos - myPos * delta) duration: 0.
		].
		lastTime := nowTime.
	].! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 6/25/2005 19:08'!
moveTo: aReferenceObject speed: speed
	"Move the receiver to match its reference position with the given reference position"
	| mySpeed lastTime nowTime distance myPos itsPos direction delta |
	lastTime := Time millisecondClockValue.
	Processor activeProject forEach: #tick do:[
		nowTime := Time millisecondClockValue.
		mySpeed := speed ifNil:[self speed].
		distance := (nowTime - lastTime) * mySpeed * 0.001. "dealing with msecs here"
		myPos := self position.
		itsPos := aReferenceObject position.
		(myPos dist: itsPos) <= distance ifTrue:[ "done"
			self position: itsPos.
			^self].
		direction := (itsPos - myPos) normalized.
		delta := direction * distance.
		self moveBy: delta.
		lastTime := nowTime.
	].! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 3/24/2004 03:13'!
moveUpLayer
	container add: self at: (container indexOf: self) - 1! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 9/4/2003 23:43'!
myEventMap
	^self propertyValueAt: #myEventMap! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 9/4/2003 23:39'!
myEventMap: eventMap
	self propertyValueAt: #myEventMap put: eventMap! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 9/4/2003 23:43'!
myScripts
	^(self propertyValueAt: #myScripts) ifNil:[#()]! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 9/4/2003 23:39'!
myScripts: scriptList
	scriptList isEmptyOrNil
		ifTrue:[self removeProperty: #myScripts]
		ifFalse:[self propertyValueAt: #myScripts put: scriptList].! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 2/5/2003 15:16'!
name
	^self player ifNil:[super name] ifNotNil:[self player name]! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/24/2004 03:12'!
newFormCache
	^CFormCache new! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:29'!
normal
	"Answer the direction rotated by 90 degrees"
	| dir |
	dir := self direction.
	^dir y @ dir x negated! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:29'!
normal: aPoint
	"Indicate the direction rotated by 90 degrees"
	self direction: aPoint y negated @ aPoint x.
	^aPoint! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 6/27/2005 15:10'!
normalizeWithRotationDegrees
	"Set the receiver's rotation degrees and renormalize the transformation."
	| degrees |
	degrees := self rotation.
	(degrees - degrees rounded abs < 0.001 and:[(degrees rounded \\ 90) = 0]) 		ifTrue:[^self normalizeWithRotationDegrees: degrees rounded].! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 6/25/2005 14:59'!
normalizeWithRotationDegrees: degrees
	"Set the receiver's rotation degrees and renormalize the transformation.
	WARNING: Only call this method with multiple of 90 degrees!!"
	| tfm tfmOffset tfmDegrees tfmScale s c a11 a12 a21 a22 quadrant |
	tfm := self transform.
	tfmOffset := tfm offset rounded.
	tfmDegrees := degrees truncated.
	tfmScale := tfm scale.
	(tfmScale x  < 0.0 and:[tfmScale x * tfmScale y > 0.0]) ifTrue:[
		"if both scales are < 0.0 then we have a rotation and reverse the scale"
		tfmScale := tfmScale negated].
	tfmScale := [tfmScale roundTo: (1.0 asPoint / self drawingBounds extent)]
		on: ZeroDivide do:[:ex| ex return: ex dividend].
	"Compute new transform but manually to have full 64 bit precision"
	quadrant := tfmDegrees \\ 360 // 90 + 1.
	s := #(0.0 1.0 0.0 -1.0) at: quadrant.
	c := #(1.0 0.0 -1.0 0.0) at: quadrant.
	a11 := c * tfmScale x.
	a12 := (0.0 - s) * tfmScale y. 
	a21 := s * tfmScale x.
	a22 := c * tfmScale y.
	"done here; install new transform and be happy"
	tfm := CMatrixTransform new.
	tfm 	a11: a11; a12: a12; a13: tfmOffset x;
			a21: a21; a22: a22; a23: tfmOffset y.
	self transform: tfm.! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/25/2005 22:29'!
nose
	"Answer the 'nose' of the receiver, a unit-vector determining 
	which direction (relative to the local coordinate frame) should 
	be considered the 'forward direction' for this object."
	^self propertyValueAt: #nose ifAbsent: [0@-1]! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/26/2005 14:31'!
nose: aPointOrNil
	"Answer the 'nose' of the receiver, a unit-vector determining 
	which direction (relative to the local coordinate frame) should 
	be considered the 'forward direction' for this object."
	| rs |
	(aPointOrNil isNil or: [aPointOrNil = (0@-1)])
		ifTrue:[self removeProperty: #nose]
		ifFalse:[| normalized |
			[normalized := aPointOrNil normalized]
				on: ZeroDivide do: [:ex | ^nil].
			self propertyValueAt: #nose put: normalized.
		].
	(rs := self rotationStyle) == #smooth ifFalse:[self updateRotationStyleTransform: rs].
	self signal: #geometryChanged.
	^aPointOrNil
! !

!CPrimitiveCostume methodsFor: 'pen my motion' stamp: 'ar 6/25/2005 19:23'!
notePen: aPrimCostume movedBy: delta
	"Look up the container hierarchy to find a Player who has penTrails.  Draw from startPoint to the aspect's costume's current position.  Make the real pen have all the properties of the moving Player's pen.  A Player can't draw on himself."

	| startPt endPt rect pen tfm pos |
	self indicatePenTrails ifFalse:["pass to parent"
		container ifNotNil: [container notePen: aPrimCostume movedBy: delta].
		^self].
	pen := aPrimCostume pen ifNil:[^self].
	tfm := aPrimCostume container localToGlobal.
	pos := aPrimCostume position.
	startPt := tfm transformPoint: pos.
	endPt := tfm transformPoint: pos - delta.
	self createOrResizeTrailsForm.
	rect := pen drawPenTrailFrom: startPt to: endPt using: self primitivePen.
	rect ifNotNil:[self invalidate: rect].! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 3/24/2004 03:33'!
onKeyDown: keySpec signal: eventName
	"Signal the given event when the key specified by keySpec is pressed"
	| map |
	(map := self keyDownMap) ifNil:[
		map := CKeyboardDispatchTable new.
		self keyDownMap: map.
	].
	eventName ifNil:[
		map remove: keySpec.
		map isEmpty ifTrue:[self keyDownMap: nil].
	] ifNotNil:[
		map add: keySpec action: eventName.
	].! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 3/24/2004 03:33'!
onKeyStroke: keySpec signal: eventName
	"Signal the given event when the key specified by keySpec is pressed"
	| map |
	(map := self keyStrokeMap) ifNil:[
		map := CKeyboardDispatchTable new.
		self keyStrokeMap: map.
	].
	eventName ifNil:[
		map remove: keySpec.
		map isEmpty ifTrue:[self keyStrokeMap: nil].
	] ifNotNil:[
		map add: keySpec action: eventName.
	].! !

!CPrimitiveCostume methodsFor: 'keyboard events' stamp: 'ar 3/24/2004 03:33'!
onKeyUp: keySpec signal: eventName
	"Signal the given event when the key specified by keySpec is pressed"
	| map |
	(map := self keyUpMap) ifNil:[
		map := CKeyboardDispatchTable new.
		self keyUpMap: map.
	].
	eventName ifNil:[
		map remove: keySpec.
		map isEmpty ifTrue:[self keyUpMap: nil].
	] ifNotNil:[
		map add: keySpec action: eventName.
	].! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 12/19/2004 17:59'!
onMouseStillDown: anEvent
	| hand evt |
	self signal: #mouseStillDown with: anEvent with: self.
	hand := anEvent hand.
	(Delay forMilliseconds: 500) wait.
	[hand redButtonPressed] whileTrue:[
		evt := hand lastEvent transformedBy: self globalToLocal.
		self signal: #mouseStillDown with: evt with: self.
		self waitTick.
	].! !

!CPrimitiveCostume methodsFor: 'drag and drop' stamp: 'ar 3/29/2003 17:18'!
onMove
	"Move the receiver"
	| lastPos nextPos cursor delta |
	container ifNil:[^self].
	cursor := self hand mouseCursor.
	self hand mouseCursor: cursor asForm.
	lastPos := container cursorPoint.
	self hand forEach: #mouseMove do:[
		nextPos := container cursorPoint.
		delta := nextPos - lastPos.
		lastPos := nextPos.
		self moveBy: delta.
	] until: #mouseUp.
	self hand mouseCursor: cursor.! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/24/2004 15:05'!
onPlayerChanged
! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:30'!
origin
	^self bounds origin! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:30'!
origin: aPoint
	^self moveBy: aPoint - self origin! !

!CPrimitiveCostume methodsFor: 'initialize' stamp: 'ar 10/18/2004 20:31'!
outOfWorld: aWorld
	"The receiver has just appeared in a new world. Notes:
		* aWorld can be nil (due to optimizations in other places)
		* owner is still valid
	Important: Keep this method fast - it is run whenever objects are removed."
	aWorld ifNil:[^self].
	contents do:[:m| m outOfWorld: aWorld].
	self signal: #closed with: aWorld.
	self signal: #close with: aWorld.
	self cacheEnabled ifTrue:[self localCache release].
	self asPlayer suspendTickingScripts.! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'tk 2/3/2005 17:24'!
outerBounds
	"Return the 'outer' bounds of the receiver, e.g., the bounds that need to be invalidated when the receiver changes."
	| box highlights |
	(flags bitAnd: FlagHasShape) == 0 
		ifFalse:[self privateBounds: 
			(bounds origin extent: self shape computeBaseBounds extent)].
	(flags bitAnd: FlagHasHighlights) = 0 ifTrue:[^self localBounds].
	box := self localBounds.
	highlights := self highlights.
	highlights do:[:highlight| box := highlight expandBounds: box].
	^box! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:11'!
overlay
	^self propertyValueAt: #overlay! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:11'!
overlay: aPlayer
	aPlayer ifNil:[
		self removeProperty: #overlay.
		flags := flags bitClear: FlagHasOverlay.
	] ifNotNil:[
		self propertyValueAt: #overlay put: aPlayer.
		flags := flags bitOr: FlagHasOverlay.
	].
	self invalidate.
	self signal: #overlayChanged.
	^aPlayer! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/25/2005 20:30'!
parentToChild
	^self childToParent inverseTransformation! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/25/2005 20:30'!
parentToChild: anObject
	^anObject transformedBy: self parentToChild! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 4/19/2005 17:09'!
patchAt: patchRect without: stoppee andNothingAbove: stopThere
	"Return a complete rendering of this patch of the display screen
	without stoppee, and possibly without anything above it."
	| form canvas |
	form := Form extent: patchRect extent depth: Display depth.
	canvas := CTransformCanvas on: form.
	canvas translateBy: patchRect topLeft negated.

	(self bounds containsRect: patchRect) ifFalse:[
		"Need to fill area outside bounds with black."
		canvas form fillColor: Color black].

	(self bounds intersects: patchRect) ifFalse:[
		"Nothing within bounds to show."
		^ canvas form].

	self drawPatchOn: canvas without: stoppee asPrimCostume andNothingAbove: stopThere.
	^canvas form
! !

!CPrimitiveCostume methodsFor: 'pen my motion' stamp: 'tk 4/27/2005 13:31'!
pen
	^self propertyValueAt: #pen ifAbsent: [nil]! !

!CPrimitiveCostume methodsFor: 'pen my motion' stamp: 'ar 9/4/2003 23:40'!
pen: aPen
	self propertyValueAt: #pen put: aPen.! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/27/2005 23:17'!
pivot
	"Answer the pivot point, e.g., the distance by which the position 
	is offset from the origin of the local coordinate system"
	(flags bitAnd: FlagHasPivot) == 0 ifTrue:[^bounds center].
	^(((self propertyValueAt: #relativePivot) ifNil:[^bounds center]) * self extent) rounded	! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/27/2005 23:44'!
pivot: aPointOrSymbol
	"Set the pivot point, e.g., the distance by which the position 
	is offset from the origin of the local coordinate system"
	| aPoint |
	aPointOrSymbol isSymbol 
		ifTrue:[aPoint := bounds perform: aPointOrSymbol]
		ifFalse:[aPoint := aPointOrSymbol asFloatPoint].
	aPoint = bounds center ifTrue:[
		flags := flags bitClear: FlagHasPivot.
		self removeProperty: #relativePivot.
	] ifFalse:[
		flags := flags bitOr: FlagHasPivot.
		[aPoint := aPoint / self extent asFloatPoint] on: ZeroDivide do:[:ex| ex resume: 0].
		self propertyValueAt: #relativePivot put: aPoint.
	].
	self signal: #geometryChanged.
	1 to: contents size do:[:i| (contents at: i) signal: #geometryChanged]. "for cartesian coordinates"
	^aPoint! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/26/2005 12:12'!
pivotPoint
	"Backstop while converting to #pivot"
	^bounds center! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:10'!
player
	^self propertyValueAt: #player! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'bf 3/30/2004 10:35'!
player: aPlayer
	^ self propertyValueAt: #player put: aPlayer! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/26/2005 12:07'!
position
	"Answer the receiver's position (in parent coordinates)"
	^self childToParent: self pivot
! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:30'!
position: aPoint
	"Change the receiver's position"
	self moveTo: aPoint.
	^aPoint! !

!CPrimitiveCostume methodsFor: 'serialization' stamp: 'tk 7/14/2004 11:27'!
postBuildWith: aMaker
	contents := contents collect:[:x| x asPrimCostume].
	contents do:[:cc|
		cc ifNil:[contents := contents reject:[:any| any == nil]]
		ifNotNil:[cc setContainer: self]].
	fill = DefaultFill ifTrue:[fill := CNoFill forCostume: self].
	border = DefaultBorder ifTrue:[border := CNoBorder forCostume: self].
	fill costume: self.
	border costume: self.
	self graphic ifNotNil:[
		self graphic costume: self.
		flags := flags bitOr: FlagHasGraphic.
	].

	self textHolder ifNotNil:[
		self textHolder costume: self.
		flags := flags bitOr: FlagHasText.
	].

	self highlights size = 0 ifFalse:[
		flags := flags bitOr: FlagHasHighlights.
	].! !

!CPrimitiveCostume methodsFor: 'collision detection' stamp: 'ar 3/23/2003 23:33'!
preciseCollisions
	^self propertyValueAt: #preciseCollisions ifAbsent:[false]! !

!CPrimitiveCostume methodsFor: 'collision detection' stamp: 'ar 3/23/2003 23:33'!
preciseCollisions: aBool
	^self propertyValueAt: #preciseCollisions put: aBool! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 8/18/2005 21:28'!
preferredExtent
	"Answer the preferred extent for the receiver"
	^self extent "by default use the current extent"! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 8/3/2004 17:22'!
pressedKeys
	self isWorld ifFalse:[^nil].
	^self propertyValueAt: #pressedKeys ifAbsentPut:[Set new].! !

!CPrimitiveCostume methodsFor: 'pen my motion' stamp: 'ar 6/25/2005 19:20'!
primitivePen
	^self propertyValueAt: #primitivePen! !

!CPrimitiveCostume methodsFor: 'pen my motion' stamp: 'ar 6/25/2005 19:20'!
primitivePen: aPen
	^self propertyValueAt: #primitivePen put: aPen! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/24/2004 03:10'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut:$[; print: self player; nextPut: $].! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 2/17/2004 18:55'!
privateAdd: aCostume atIndex: index
	"@@@@: Fix this. When adding a costume we should check if it obtrudes the receiver and update its obtrusion flag (****check wording ****) appropriately before any invalidations are triggered. Might need to update the 'i am fully invalidated' flag too. :@@@@"
	"@@@@: Check this. See how to avoid invalidations when we (for example) pick up costumes or drop them again. Should check the bounds before and after the add operation if aCostume is in the world."
	"@@@@: Check this. Should we add costumes based on their world position??? Makes sense in many cases but can be a pain in others... :@@@@"
	| oldIndex newIndex myWorld itsWorld |
	newIndex := index.
	newIndex < 1 ifTrue:[newIndex := 1].
	newIndex > (contents size + 1) ifTrue:[newIndex := contents size + 1].
	aCostume == self ifTrue:[^self]. "@@@@: Should this raise an error or not??? :@@@@"
	aCostume container == self ifTrue:[
		"costume's position changes within in the contents"
		newIndex > contents size ifTrue:[newIndex := contents size].
		oldIndex := self indexOf: aCostume.
		oldIndex = newIndex ifTrue:[^self].
		oldIndex < newIndex ifTrue:["moving aCostume to the back"
			oldIndex+1 to: newIndex do:[:i| contents at: i-1 put: (contents at: i)].
			contents at: newIndex put: aCostume.
		] ifFalse:["moving aCostume to front"
			oldIndex-1 to: newIndex by: -1 do:[:i| contents at: i+1 put: (contents at: i)].
			contents at: newIndex put: aCostume.
		]
	] ifFalse:["a new costume"
		myWorld := self world.
		aCostume container ifNotNil:[
			itsWorld := aCostume world.
			itsWorld == myWorld ifFalse: [aCostume outOfWorld: itsWorld].
			aCostume container privateRemove: aCostume.
		].
		aCostume setContainer: self.
		contents := contents copyReplaceFrom: newIndex to: newIndex-1 with: (Array with: aCostume).
		itsWorld == myWorld ifFalse:[aCostume intoWorld: myWorld].
	].
	aCostume invalidate.
	self invalidateLayout.
	self signal: #added with: aCostume.
	self structureChanged.! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 9/30/2009 16:56'!
privateAdd: aCostume beforeIndex: index
	| newIndex oldIndex |
	self ifTweak:[] ifNot:[
		"Guard to catch the manipulation of Tweak stuff from background processes."
		CUnsynchronizedModification 
			signal:'Trying to manipulate Tweak from a foreign process'.
	].

	aCostume == self ifTrue:[^self]. "@@@@: Should this raise an error or not??? :@@@@"
	(self isIn: aCostume) ifTrue:[
		"Need to change my container to aCostume's container"
		aCostume container 
			ifNil:[self container remove: self]
			ifNotNil:[aCostume container elements replace: aCostume with: self].
	].
	newIndex := index.
	newIndex < 1 ifTrue:[newIndex := 1].
	newIndex > (contents size + 1) ifTrue:[newIndex := contents size + 1].
	aCostume container == self 
		ifFalse:[^self privateReplaceFrom: newIndex 
					to: newIndex-1 
					with: (Array with: aCostume)].
	"move that guy to the appropriate position"
	"costume's position changes within in the contents"
	oldIndex := self indexOf: aCostume.
	oldIndex = newIndex ifTrue:[^self].
	oldIndex < newIndex ifTrue:["moving aCostume to the back"
		oldIndex+1 to: newIndex-1 do:[:i| contents at: i-1 put: (contents at: i)].
		contents at: newIndex-1 put: aCostume.
	] ifFalse:["moving aCostume to front"
		oldIndex-1 to: newIndex by: -1 do:[:i| contents at: i+1 put: (contents at: i)].
		contents at: newIndex put: aCostume.
	].
	aCostume invalidate.
	self invalidateLayout.
	self structureChanged.! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/7/2004 02:34'!
privateBltMove: aCostume from: oldRect to: newRect
	"Private. Try to relocate the given costume by delta using a direct screen blt.
	By default, just invalidate the portion for later updates."
	| aCanvas cache form tfm |
	(self canBltMove: aCostume from: oldRect to: newRect) ifFalse:[
		^self invalidate: (oldRect merge: newRect) from: aCostume.
	].
	cache := self localCache ifNil:[
		container ifNil:[^self].
		tfm := self transform.
		^container privateBltMove: self 
			from: (tfm transformRect: oldRect) 
			to: (tfm transformRect: newRect)].
	form := cache form.
	form deferUpdates: true. "must defer from here"
	aCanvas := CTransformCanvas on: form.
	aCanvas clipBy: (oldRect merge: newRect).
	(BitBlt toForm: form)
		sourceForm: form;
		combinationRule: 3;
		sourceRect: oldRect;
		destOrigin: newRect origin;
		copyBits.
	"Now that the blt is complete, record the remaining dirty areas.
	Note: For non-front costumes this needs to take overlapping
	areas into account."
	(aCostume areasRemainingToFill: newRect)
		do:[:rect| self invalidate: rect].
	"And of course, the old rectangle"
	(oldRect areasOutside: newRect) do:[:rect| self invalidate: rect].
	"Notify parent about the change in newRect"
	container ifNotNil:[container invalidate: newRect from: self].
! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/16/2003 00:06'!
privateBounds
	^bounds! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 3/16/2003 00:06'!
privateBounds: newBounds
	bounds := newBounds.! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 2/17/2004 18:55'!
privateRemove: aCostume
	"Private!! Should only be used by methods which:
		* maintain the ower/part invariant
		* notify costumes about leaving worlds
	"
	self assert:[aCostume container == self].
	aCostume invalidate.
	contents := contents copyWithout: aCostume.
	self invalidateLayout.
	self signal: #removed with: aCostume.
	self structureChanged.! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 2/17/2004 18:56'!
privateReplaceFrom: firstIndex to: lastIndex with: aCollection
	"The core of modifying costume contents"
	| myWorld itsWorld startIndex stopIndex |
	myWorld := self world.
	startIndex := firstIndex.
	stopIndex := lastIndex.
	aCollection do:[:aCostume|
		aCostume == self ifTrue:[^self error: 'Cannot add myself to my children'].
		itsWorld := nil.
		aCostume container ifNil:[
			"the simple case - merely transfer it"
			aCostume setContainer: self.
			aCostume intoWorld: myWorld.
			aCostume invalidate.
		] ifNotNil:[
			"more complex - must distinguish between containers"
			aCostume container == self ifTrue:[
				"this guy is being moved around - fix up my start/stop if needed"
				| oldIndex |
				oldIndex := self indexOf: aCostume.
				oldIndex < startIndex ifTrue:[startIndex := startIndex-1].
				oldIndex <= stopIndex ifTrue:[stopIndex := stopIndex-1].
				itsWorld := myWorld.
				"simply remove it from my contents without any further invalidation"
				contents := contents copyReplaceFrom: oldIndex to: oldIndex with: #().
			] ifFalse:[
				"the guy comes from somewhere else"
				itsWorld := aCostume world.
				itsWorld == myWorld ifFalse:[aCostume outOfWorld: itsWorld].
				aCostume container privateRemove: aCostume.
				"transfer aCostume to me"
				aCostume setContainer: self.
				itsWorld == myWorld ifFalse:[aCostume intoWorld: myWorld].
				aCostume invalidate.
			].
		].
	].
	"Now bump everything from start to stop"
	contents from: startIndex to: stopIndex do:[:aCostume|
		aCostume outOfWorld: myWorld.
		aCostume invalidate.
		self signal: #removed with: aCostume.
		aCostume setContainer: nil.
	].
	contents := contents copyReplaceFrom: startIndex to: stopIndex with: aCollection.
	self invalidateLayout.
	self structureChanged.! !

!CPrimitiveCostume methodsFor: 'scrolling' stamp: 'ar 6/24/2005 21:35'!
privateScrollBy: delta
	"Scroll by the specified amount. Note: At some point this should be optimized to use a direct blt for the non-overlapping regions if the receiver is cached."
	self invalidate.
	self transform: (self transform composedWithLocal: (COffsetTransform withOffset: delta negated)).
	contents do:[:m| m scrollingDisabled ifTrue:[m moveBy: delta]].
	(flags bitAnd: FlagHasText) = 0 ifFalse:[
		self textHolder offset: self textHolder offset - delta.
	].
	self invalidate.! !

!CPrimitiveCostume methodsFor: 'properties' stamp: 'ar 9/4/2003 23:50'!
propertyValueAt: propName
	^self propertyValueAt: propName ifAbsent:[nil]! !

!CPrimitiveCostume methodsFor: 'properties' stamp: 'ar 9/4/2003 23:51'!
propertyValueAt: propName ifAbsent: aBlock
	^properties ifNil:[aBlock value] ifNotNil:[properties at: propName ifAbsent: aBlock].! !

!CPrimitiveCostume methodsFor: 'properties' stamp: 'ar 9/9/2003 15:35'!
propertyValueAt: propName ifAbsentPut: aBlock
	properties ifNil:[^self propertyValueAt: propName put: aBlock value].
	^properties at: propName ifAbsentPut: aBlock! !

!CPrimitiveCostume methodsFor: 'properties' stamp: 'ar 9/4/2003 23:42'!
propertyValueAt: propName put: aValue
	aValue ifNil: [^self removeProperty: propName].
	properties ifNil:[properties := IdentityDictionary new].
	properties at: propName put: aValue.
	^aValue! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 12/8/2003 19:42'!
raisedColor
	"Return the color to be used for shading raised borders.  The default is my own color, but it might want to be, eg, my container's color.  Whoever's color ends up prevailing, the color itself gets the last chance to determine, so that when, for example, an InfiniteForm serves as the color, callers won't choke on some non-Color object being returned"
	(fill isTransparent and:[container notNil]) ifTrue:[^container raisedColor].
	^fill asColor! !

!CPrimitiveCostume methodsFor: 'serialization' stamp: 'ar 3/29/2004 14:54'!
readFieldsFrom: deserializer
	| v origin extent |
	self initialize.
	"basic"
	origin := deserializer readField: #origin ifAbsent:[bounds origin].
	extent := deserializer readField: #extent ifAbsent:[bounds extent].
	bounds := origin extent: extent.

	fill := deserializer readField: #fill ifAbsent:[fill].
	border := deserializer readField: #border ifAbsent:[border].

	contents := deserializer readField: #contents.
	self player: (deserializer readField: #player).

	"flags"
	v := deserializer readField: #hidden ifAbsent:[self hidden].
	v ifNotNil:[self hidden: v].
	v := deserializer readField: #userDraw ifAbsent:[self userDraw].
	v ifNotNil:[self userDraw: v].
	v := deserializer readField: #locked ifAbsent:[self locked].
	v ifNotNil:[self locked: v].
	v := deserializer readField: #cacheEnabled ifAbsent:[self cacheEnabled].
	v ifNotNil:[self cacheEnabled: v].
	v := deserializer readField: #clipEnabled ifAbsent:[self clipping].
	v ifNotNil:[self clipping: v].
	v := deserializer readField: #roundCorners ifAbsent:[self roundCorners].
	v ifNotNil:[self roundCorners: v].
	v := deserializer readField: #hasScroller ifAbsent:[false].
	self setFlag: FlagHasScroller value: v.

	"flags := 0.
	1 to: FlagRegistry size do:[:bit|
		flagBit := 1 bitShift: bit-1.
		flagName := FlagRegistry at: bit.
		flagName ifNotNil:[
			v := deserializer readField: flagName ifAbsent:[flags anyMask: flagBit].
			v ifNotNil:[
				v == true 
					ifTrue:[flags := flags bitOr: flagBit]
					ifFalse:[flags := flags bitClear: flagBit].
			].
		].
	]."

	self serializableFields do:[:fieldName|
		v := deserializer readField: fieldName ifAbsent:[nil].
		v ifNotNil:[self propertyValueAt: fieldName put: v].
	].! !

!CPrimitiveCostume methodsFor: 'tweaking' stamp: 'ar 2/27/2004 19:35'!
readTweakedFieldsFrom: deSerializer
	"FIXME!!!!!!"
	| oldValue newValue |
	deSerializer fieldsDo:[:fieldName|
		oldValue := self perform: fieldName.
		newValue := deSerializer tweakField: fieldName value: oldValue.
		(newValue equals: oldValue) ifFalse:[
			self perform: (fieldName,':') asSymbol with: newValue.
			self tweaked: fieldName from: oldValue to: newValue.
		].
	].! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 3/23/2004 23:13'!
rejectsEvent: anEvent
	"Return true to reject the given event. Rejecting an event means neither the receiver nor any of it's parts will be given any chance to handle it."
	^self locked or:[self visible not]! !

!CPrimitiveCostume methodsFor: 'text anchoring' stamp: 'ar 9/4/2003 23:48'!
relativeTextAnchorPosition
	^self propertyValueAt: #relativeTextAnchorPosition! !

!CPrimitiveCostume methodsFor: 'text anchoring' stamp: 'ar 9/4/2003 23:40'!
relativeTextAnchorPosition: aPoint
	^self propertyValueAt: #relativeTextAnchorPosition put: aPoint! !

!CPrimitiveCostume methodsFor: 'removing' stamp: 'ar 2/12/2003 22:24'!
remove: aCostume
	"Remove oldObject from the receiver's elements."
	^self remove: aCostume ifAbsent:[]! !

!CPrimitiveCostume methodsFor: 'removing' stamp: 'ar 2/12/2003 22:24'!
remove: aCostume ifAbsent: aBlock
	"Remove oldObject from the receiver's elements."
	aCostume container == self ifFalse:[^aBlock value].
	aCostume outOfWorld: self world.
	self privateRemove: aCostume.
	aCostume setContainer: nil.! !

!CPrimitiveCostume methodsFor: 'removing' stamp: 'ar 2/25/2004 20:06'!
removeAll
	"Remove all the contents of the receiver"
	| world |
	self invalidate.
	world := self world.
	contents do:[:aCostume|
		aCostume outOfWorld: world.
		aCostume setContainer: nil.
	].
	contents := #().
	self invalidateLayout.
	self structureChanged.
! !

!CPrimitiveCostume methodsFor: 'removing' stamp: 'ar 2/12/2003 22:24'!
removeAll: aCollection
	"Remove all the elements found in aCollection"
	aCollection do:[:each| self remove: each].! !

!CPrimitiveCostume methodsFor: 'removing' stamp: 'ar 6/28/2003 21:52'!
removeAt: index
	"Remove oldObject from the receiver's elements."
	^self remove: (self at: index) ifAbsent:[]! !

!CPrimitiveCostume methodsFor: 'removing' stamp: 'ar 3/13/2005 11:17'!
removeFrom: firstIndex to: lastIndex
	"Remove oldObject from the receiver's elements."
	| oldData |
	oldData := contents copyFrom: firstIndex to: lastIndex.
	firstIndex to: lastIndex do:[:i| self removeAt: firstIndex].
	^oldData! !

!CPrimitiveCostume methodsFor: 'highlights' stamp: 'ar 3/24/2004 02:03'!
removeHighlight: name
	"Remove the given highlight from the list of highlights.
	Either aHighlight or propertyName may be nil."
	| list drop anyDropped |
	anyDropped := false.
	list := self highlights reject:[:highlight|
		drop := false.
		drop := highlight name = name.
		(drop and:[highlight isActive]) ifTrue:[
			highlight unhighlight.
			highlight costume: nil.
		].
		drop ifTrue:[anyDropped := true].
		drop].
	anyDropped ifTrue:[self highlights: list].! !

!CPrimitiveCostume methodsFor: 'removing' stamp: 'ar 4/13/2005 00:57'!
removeLast
	^self remove: contents last! !

!CPrimitiveCostume methodsFor: 'properties' stamp: 'ar 3/10/2003 07:47'!
removeProperty: propName
	| old |
	properties ifNotNil:[
		old := properties removeKey: propName ifAbsent:[].
		properties isEmpty ifTrue:[properties := nil].
	].
	^old! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'ar 6/21/2005 22:56'!
replace: cc1 with: cc2
	| index costume1 costume2 |
	costume1 := cc1 asPrimCostume.
	costume2 := cc2 asPrimCostume.
	costume1 == costume2 ifTrue:[^self].
	costume2 ifNil:[^self remove: costume1].
	costume1 ifNil:[^self addLast: costume2].
	index := self indexOf: costume1 ifAbsent:[^nil].
	self replaceFrom: index to: index with: (Array with: costume2).! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'ar 2/16/2003 19:20'!
replaceFrom: firstIndex to: lastIndex with: aCollection
	^self privateReplaceFrom: firstIndex to: lastIndex with: aCollection! !

!CPrimitiveCostume methodsFor: 'adding' stamp: 'ar 3/16/2004 13:40'!
replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: replStart
	replStart = 1 
		ifTrue:[^self privateReplaceFrom: firstIndex to: lastIndex with: aCollection]
		ifFalse:[^self privateReplaceFrom: firstIndex to: lastIndex with: (aCollection copyFrom: replStart to: aCollection size)].! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 6/25/2005 20:12'!
resizable
	"Return true if the receiver is resizable. Returning false will result in scaling."
	^self propertyValueAt: #resizable ifAbsent:[true]! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 6/25/2005 20:12'!
resizable: aBool
	"Return true if the receiver is resizable. Returning false will result in scaling."
	aBool
		ifTrue:[self removeProperty: #resizable]
		ifFalse:[self propertyValueAt: #resizable put: false].! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 6/25/2005 20:21'!
resizeTo: extentPoint
	"Resize the receiver to match the given (global) extent"
	| tfm oldBox origin corner relOrigin relCorner |
	self resizable ifFalse:[^self scaleTo: extentPoint].
	tfm := self childToParent.
	tfm isPureTranslation 
		ifTrue:[^self extent: extentPoint].
	"TODO: This doesn't work"
	"If our transform is not a pure translation we must inscribe a local rectangle into the global one defined by our (global) origin and the given extent point."
	oldBox := self bounds.
	origin := tfm transformPoint: self localBounds origin.
	corner := tfm transformPoint: self localBounds corner.
	"origin and corner are on some border of oldBox; compute its relative position"
	[
		relOrigin := (origin - oldBox origin) / oldBox extent.
		relCorner := (corner - oldBox origin) / oldBox extent.
	] on: ZeroDivide do:[:ex| ex resume: ex dividend].
	"map origin and corner into the new extent"
	origin := oldBox origin + (relOrigin * extentPoint).
	corner := oldBox origin + (relCorner * extentPoint).
	"transform them back into local frame"
	origin := tfm invertPoint: origin.
	corner := tfm invertPoint: corner.
	"set our new local bounds"
	"self localBounds: (origin corner: corner)."
	self extent: corner - origin.! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 3/29/2005 23:52'!
reverseDo: aBlock
	^contents reverseDo: aBlock! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:30'!
right
	^self bounds right! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:30'!
right: aNumber
	self moveBy: (aNumber - self right) @ 0.
	^aNumber
! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'bgf 2/23/2008 12:42'!
rightCenter
	^self bounds rightCenter! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'bgf 2/23/2008 12:42'!
rightCenter: aPoint
	self moveBy: (aPoint - self rightCenter).
! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 6/1/2005 20:58'!
root
	container 
		ifNil:[^self asPlayer]
		ifNotNil:[^container root]! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 6/26/2005 12:07'!
rotateBy: degrees
	^self rotateBy: degrees around: self pivot! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 6/26/2005 14:27'!
rotateBy: degrees around: center
	| tfm dir |
	"For non-standard rotation style just change the nose"
	(self rotationStyle) == #smooth ifFalse:[
		"Note: This uses direction since nose might be swapped"
		dir := (Point r: 1.0 degrees: self direction degrees + degrees).
		^self nose: (self transform invertDirection: dir).
	].
	tfm := COffsetTransform withOffset: center.
	tfm := tfm composedWithLocal: (CMatrixTransform withAngle: degrees).
	tfm := tfm composedWithLocal: (COffsetTransform withOffset: center negated).
	self localTransformBy: tfm! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'bf 7/22/2005 15:32'!
rotation
	"Answer the angle (in degrees) the receiver is rotated by"
	| deg tfm |
	tfm := self transform.
	tfm isPureTranslation ifTrue:[^0].
	"Actual rotation is helpful when shearing is involved"
	deg := (tfm transformDirection: 1@0) degrees.
	deg < 180.0 ifFalse: [deg := deg - 360.0].
	^deg! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/27/2005 15:10'!
rotation: aNumber
	"Indicate the angle (in degrees) the receiver is rotated by"
	self rotateBy: aNumber - self rotation.
	^aNumber! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 6/26/2005 13:19'!
rotationStyle
	"Answer the style by which the receiver should be rotated"
	^(self propertyValueAt: #rotationStyle) ifNil:[#smooth].! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 6/27/2005 15:10'!
rotationStyle: aSymbol
	"Indicate the style by which the receiver should be rotated"
	| rotation |
	aSymbol == self rotationStyle ifTrue:[^aSymbol].
	self signalChanged: #rotationStyleChanged from: self rotationStyle to: aSymbol.
	aSymbol == #smooth ifTrue:[
		self removeProperty: #rotationStyle.
		^aSymbol.
	].
 	"if not smooth, absorb current rotation"
	rotation := self rotation.
	rotation isZero ifFalse:[
		self rotateBy: rotation negated.
		self nose: (Point r: 1.0 degrees: self nose degrees + rotation).
	].
	self updateRotationStyleTransform: aSymbol.
	^self propertyValueAt: #rotationStyle put: aSymbol.
! !

!CPrimitiveCostume methodsFor: 'corners' stamp: 'ar 4/20/2004 21:47'!
roundCornerFlags
	self roundCorners ifFalse:[^0].
	^self propertyValueAt: #roundCornerFlags ifAbsent:[15]! !

!CPrimitiveCostume methodsFor: 'corners' stamp: 'ar 4/20/2004 21:48'!
roundCornerFlags: cornerFlags
	cornerFlags = 0 ifTrue:[
		self removeProperty: #roundCornerFlags.
		^self roundCorners: false].
	cornerFlags = 15 ifTrue:[
		self removeProperty: #roundCornerFlags.
		^self roundCorners: true].
	self propertyValueAt: #roundCornerFlags put: cornerFlags.
	self roundCorners: true.
	self invalidate.
	self signal: #roundCornersChanged.! !

!CPrimitiveCostume methodsFor: 'corners' stamp: 'ar 4/20/2004 21:54'!
roundCornerSpec
	^(1 to: 4) select:[:i| self roundCornerFlags anyMask: (1 bitShift: i-1)]! !

!CPrimitiveCostume methodsFor: 'corners' stamp: 'ar 5/19/2010 13:59'!
roundCornerSpec: anArray
	^self roundCornerFlags: (anArray inject: 0 into:[:sum :next| 
		| bit |
		bit := next isSymbol 
			ifTrue:[#(topLeft bottomLeft bottomRight topRight) indexOf: next]
			ifFalse:[next].
		sum bitOr: (1 bitShift: bit-1)])! !

!CPrimitiveCostume methodsFor: 'corners' stamp: 'ar 3/24/2004 14:14'!
roundCorners
	"Return true if the receiver wants its corners rounded"
	^(flags bitAnd: FlagRoundedCorners) = FlagRoundedCorners! !

!CPrimitiveCostume methodsFor: 'corners' stamp: 'ar 3/24/2004 14:26'!
roundCorners: aBool
	"Indicate whether the receiver wants its corners rounded"
	self roundCorners = aBool ifTrue:[^self].
	flags := aBool 
		ifTrue:[flags bitOr: FlagRoundedCorners] 
		ifFalse:[flags bitClear: FlagRoundedCorners].
	self invalidate.
	self signal: #roundCornersChanged.! !

!CPrimitiveCostume methodsFor: 'corners' stamp: 'ar 3/24/2004 14:26'!
roundedCorners
	"Compatibility"
	^self roundCornerSpec! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/25/2005 20:31'!
scale
	"Warning: scale is numerically unstable so round it to .01"
	| scale |
	self transform isPureTranslation ifTrue:[^1].
	scale := self transform scale roundTo: 0.01.
	scale x = scale y ifTrue:[^scale x].
	^scale! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/25/2005 20:31'!
scale: aNumber
	self scaleBy: aNumber / self scale.
	^aNumber! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 6/26/2005 12:08'!
scaleBy: scale
	"Scale the receiver"
	^self scaleBy: scale centerAt: self pivot! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 6/26/2005 12:07'!
scaleBy: scale centerAt: centerPt
	"Scale the receiver"
	| tfm refPos scaleX scaleY |
	scaleX := scale asPoint x.
	scaleY := scale asPoint y.
	(scaleX = 1 and:[scaleY = 1]) ifTrue:[^self].
	(scaleX > -1.0e-3 and:[scaleX < 1.0e-3]) 
		ifTrue:[scaleX := 1].
	(scaleY > -1.0e-3 and:[scaleY < 1.0e-3]) 
		ifTrue:[scaleY := 1].
	refPos := centerPt.
	tfm := COffsetTransform withOffset: refPos.
	tfm := tfm composedWithLocal: (CMatrixTransform withScale: scaleX@scaleY).
	tfm := tfm composedWithLocal: (COffsetTransform withOffset: refPos negated).
	self localTransformBy: tfm! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 12/7/2002 22:14'!
scaleTo: aPoint
	"Scale the receiver to match the given extent"
	| x y |
	x := self width.
	(x > -1 and:[x < 1]) ifTrue:[x := 1].
	y := self height.
	(y > -1 and:[x < 1]) ifTrue:[y := 1].
	self localTransformBy: (CMatrixTransform withScale: aPoint / (x@y))! !

!CPrimitiveCostume methodsFor: 'scrolling' stamp: 'ar 3/26/2004 01:09'!
scrollArea
	"Answer the area that should be scrolled in the receiver"
	^self contentsBounds encompass: 0@0! !

!CPrimitiveCostume methodsFor: 'scrolling' stamp: 'ar 3/20/2003 02:51'!
scrollBox
	^self scrollOffset extent: bounds extent.! !

!CPrimitiveCostume methodsFor: 'scrolling' stamp: 'ar 3/20/2003 02:53'!
scrollOffset
	"Return the current scroll offset if any"
	(flags bitAnd: FlagHasScroller) == 0 ifTrue:[^0@0].
	^self propertyValueAt: #scrollOffset ifAbsent:[0@0]! !

!CPrimitiveCostume methodsFor: 'scrolling' stamp: 'ar 9/8/2004 16:24'!
scrollOffset: aPoint
	"Change the current scroll offset"
	| newOffset delta oldOffset |
	oldOffset := self scrollOffset.
	newOffset := aPoint asPoint truncated.
	delta := newOffset - oldOffset.
	delta isZero ifTrue:[^self].
	self privateScrollBy: delta.
	self signalChanged: #scrollOffsetChanged from: oldOffset to: newOffset.
	newOffset isZero ifTrue:[
		flags := flags bitClear: FlagHasScroller.
		self removeProperty: #scrollOffset.
	] ifFalse:[
		flags := flags bitOr: FlagHasScroller.
		self propertyValueAt: #scrollOffset put: newOffset.
	].
	self frameChanged.! !

!CPrimitiveCostume methodsFor: 'scrolling' stamp: 'ar 3/15/2003 01:32'!
scrollableExtent
	"Return the area that should be scrolled"
	^(self contentsBounds encompass: 0@0) extent! !

!CPrimitiveCostume methodsFor: 'scrolling' stamp: 'ar 9/4/2003 23:51'!
scrollingDisabled
	^self propertyValueAt: #scrollingDisabled ifAbsent:[false]! !

!CPrimitiveCostume methodsFor: 'scrolling' stamp: 'ar 9/8/2004 16:25'!
scrollingDisabled: aBool
	self signalChanged: #scrollingDisabledChanged from: self scrollingDisabled to: aBool.
	self propertyValueAt: #scrollingDisabled put: aBool.! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 3/24/2004 03:23'!
select
	"Select the receiver"
! !

!CPrimitiveCostume methodsFor: 'serialization' stamp: 'ar 3/24/2004 14:21'!
serializableFields
	^#(
		transform 
		shape 

		roundedCorners 
		aspectRatio 
		graphic
		highlights
		overlay 
		mouseCursor

		textHolder
		textAnchorType

		scrollOffset
		scrollingDisabled

		layoutProperties
		layout
		layoutFrame
		minWidth
		minHeight

		floating
		infinite
		collisionsEnabled
		sticky
		isWorld
		resizable

		preciseCollisions
		balloonHelp
	).
! !

!CPrimitiveCostume methodsFor: 'serialization' stamp: 'ar 2/25/2004 16:18'!
serializeOn: aSerializer
	^aSerializer writeCostume: self! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 9/8/2004 16:25'!
setContainer: aCostume
	"Private!! Should only be used by methods that maintain the ower/part invariant."
	| oldOwner |
	oldOwner := container.
	container := aCostume.
	self signalChanged: #containerChanged from: oldOwner to: container.! !

!CPrimitiveCostume methodsFor: 'serialization' stamp: 'ar 3/14/2004 15:55'!
setFlag: flagBit value: aBool
	aBool ifNil:[^self].
	flags := aBool ifTrue:[flags bitOr: flagBit] ifFalse:[flags bitClear: flagBit].! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 12/8/2002 00:13'!
shadowForm
	"Return a form representing the 'shadow' of the receiver - e.g., all pixels that are occupied by the receiver are one, all others are zero."
	| canvas |
	canvas := (CTransformCanvas extent: self extent depth: 1)
				asShadowDrawingCanvas: Color black. "Color black represents one for 1bpp"
	canvas translateBy: self topLeft negated
		during:[:tempCanvas| tempCanvas fullDraw: self].
	^ canvas form offset: self topLeft! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'tk 2/1/2005 21:33'!
shape
	^self propertyValueAt: #shape ifAbsent:[nil]! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'tk 2/3/2005 17:14'!
shape: aShape

	aShape 
		ifNil: [
			self removeProperty: #shape.
			flags := flags bitClear: FlagHasShape]
		ifNotNil: [
			self propertyValueAt: #shape put: aShape.
			flags := flags bitOr: FlagHasShape.
			aShape myPrimCostume: self.	"backpointer"
			aShape initialize.	"only changes parts that are nil"
			self invalidate].! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'tk 2/1/2005 21:38'!
showHandles: aBoolean
	"showHandles tells if the object's polygon shape will have knobs on each corner.  You can drag a knob to move the corner.  Drag a knob in the middle of a line to get a new corner (also called a vertex).  
	You must give this object a shape for this command to have any effect (self shape: CPolygonShape new)."

	^(self shape ifNil: [^ self]) showHandles: aBoolean.
! !

!CPrimitiveCostume methodsFor: 'initialize' stamp: 'ar 5/27/2005 20:31'!
shutDownProject: quitting
	"This project is shutting down"
	contents do:[:each| each shutDownProject: quitting].
	self cacheEnabled ifTrue:[self localCache release].
	fill releaseCachedState.
	border releaseCachedState.
	self asPlayer suspendTickingScripts.! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 6/21/2005 20:13'!
signalEvent: anEvent
	| myPlayer |
	super signalEvent: anEvent.
	myPlayer := self player ifNil:[^self].
	myPlayer getCostume == self ifFalse:[^nil].
	anEvent setSender: myPlayer.
	myPlayer signalEvent: anEvent.! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 3/24/2004 03:23'!
size
	^contents size! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'tk 2/1/2005 21:35'!
smoothCurve: aBoolean
	"smoothCurve tells if the object's polygon shape will a smooth curve as its edge, or straight line segments.  You must give this object a shape for this command to have any effect (self shape: CPolygonShape new)."

	^(self shape ifNil: [^ self]) smoothCurve: aBoolean.
! !

!CPrimitiveCostume methodsFor: 'initialize' stamp: 'ar 6/25/2005 17:20'!
startUpProject: resuming
	"This project is just coming up"
	self fixGeometryAfterPivotChange.
	self asPlayer resumeTickingScripts.
	contents do:[:each| each startUpProject: resuming].! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/23/2004 23:37'!
sticky
	^self propertyValueAt: #sticky ifAbsent:[false]! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/23/2004 23:37'!
sticky: aBool
	self propertyValueAt: #sticky put: aBool.
	self signal: #flagsChanged.! !

!CPrimitiveCostume methodsFor: 'private' stamp: 'ar 2/17/2004 18:56'!
structureChanged
	"My structure has changed (by adding or removing some entitiy)."
	self signal: #structureChanged with: self.! !

!CPrimitiveCostume methodsFor: 'styling' stamp: 'ar 7/7/2005 09:26'!
styleSheet
	"Answer the receiver's style sheet"
	^self propertyValueAt: #styleSheet! !

!CPrimitiveCostume methodsFor: 'styling' stamp: 'ar 7/7/2005 09:26'!
styleSheet: aStyleSheet
	"Answer the receiver's style sheet"
	^self propertyValueAt: #styleSheet put: aStyleSheet! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 3/9/2003 22:11'!
swap: costumeA with: costumeB
	"Exchange costumeA and costumeB"
	| posA posB containerA containerB indexA indexB |
	posA := costumeA position.
	posB := costumeB position.
	containerA := costumeA container.
	containerB := costumeB container.
	containerA ifNotNil:[indexA := containerA indexOf: costumeA].
	containerB ifNotNil:[indexB := containerB indexOf: costumeB].
	containerA ifNotNil:[
		containerA remove: costumeA.
		containerA add: costumeB atIndex: indexA.
		costumeB position: posA.
	].

	containerB ifNotNil:[
		containerB remove: costumeB.
		containerB add: costumeA atIndex: indexB.
		costumeA position: posB.
	].
! !

!CPrimitiveCostume methodsFor: 'text anchoring' stamp: 'ar 9/24/2003 13:54'!
textAnchorType
	^(self propertyValueAt: #textAnchorType) ifNil:[#document]! !

!CPrimitiveCostume methodsFor: 'text anchoring' stamp: 'ar 9/4/2003 23:41'!
textAnchorType: aSymbol
	aSymbol == #document
		ifTrue:[^self removeProperty: #textAnchorType]
		ifFalse:[^self propertyValueAt: #textAnchorType put: aSymbol].! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/26/2004 01:12'!
textBounds
	"Answer the bounds for drawing text"
	^border insetBounds: (0@0 extent: bounds extent)! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/24/2004 03:18'!
textHolder
	^self propertyValueAt: #textHolder! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/25/2004 17:25'!
textHolder: aTextHolder
	| myText |
	myText := self textHolder.
	myText == aTextHolder ifTrue:[^self].
	aTextHolder ifNil:[
		"Remove existing text"
		self removeProperty: #textHolder.
		flags := flags bitClear: FlagHasText.
		^self signal: #textChanged].

	myText ifNil:[
		"Create new text"
		myText := CCostumeTextHolder forCostume: self.
		flags := flags bitOr: FlagHasText.
		self propertyValueAt: #textHolder put: myText
	].

	myText loadFrom: aTextHolder.
	^aTextHolder! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:31'!
top
	^ self bounds top! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:31'!
top: aNumber
	self moveBy: 0 @ (aNumber - self top).
	^aNumber
! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'bgf 2/23/2008 12:41'!
topCenter
	^self bounds topCenter! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'bgf 2/23/2008 12:41'!
topCenter: aPoint
	self moveBy: (aPoint - self topCenter).
! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 4/2/2004 14:17'!
topDownFullDrawOn: aCanvas damageList: damageList
	"Draw the receiver and all its parts on the given canvas. This method uses a top-down strategy for redrawing which will eliminate much of the actual work if many parts overlap and are (at least partly) transparent. This strategy is useful, for instance, in worlds (using overlapping windows) but others may use it as well. Note that for optimal performance, aCanvas' clip rect should be set to encapsulate the minimum bounding area of all the rectangles in damage list."
	^self topDownFullDrawOn: aCanvas damageList: damageList dirtyRect: aCanvas clipRect startingAt: contents size! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 4/2/2004 14:18'!
topDownFullDrawOn: aCanvas damageList: damageList dirtyRect: damageBounds startingAt: idx
	"Draw the receiver and all its contents on the given canvas. This method uses a top-down strategy for redrawing which will eliminate much of the actual work if many costumes overlap and are (at least partly) transparent. This strategy is useful, for instance, in worlds (using overlapping windows) but others may use it as well. Note that for optimal performance, aCanvas' clip rect should be set to encapsulate the minimum bounding area of all the rectangles in damage list."
	| child stillDirty index |
	damageList size = 0 ifTrue:[^self]. "no damage no work"
	"Find the next visible child intersecting the damage area"
	index := idx. child := nil.
	[index > 0 and:[child == nil]] whileTrue:[
		child := contents at: index.
		(child visible and:[child fullBounds intersects: damageBounds])
			ifFalse:[child := nil].
		index := index-1].
	"If we have no more children draw the receiver"
	child ifNil:[^aCanvas draw: self damageList: damageList].
	"Collect left-over damage for this child"
	stillDirty := OrderedCollection new: damageList size.
	damageList do:[:rect|
		(child areasRemainingToFill: rect) do:[:dirty| 
			stillDirty add: (dirty origin truncated corner: dirty corner ceiling)].
	].
	"Now recursively update everything behind the child"
	self topDownFullDrawOn: aCanvas damageList: stillDirty dirtyRect: damageBounds startingAt: index.
	"Everything below the child has been restored, now restore the child itself"
	^aCanvas fullDraw: child damageList: damageList.! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:31'!
topLeft
	^ self bounds topLeft! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:31'!
topLeft: aPoint
	self moveBy: aPoint - self topLeft.! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:31'!
topRight
	^ self bounds topRight! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:31'!
topRight: aPoint
	self top: aPoint y.
	self right: aPoint x! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/25/2005 20:31'!
transform
	(flags bitAnd: FlagHasMatrix) == 0 ifTrue:[self fixGeometryAfterPivotChange].
	^self propertyValueAt: #transform! !

!CPrimitiveCostume methodsFor: 'geometry-transform' stamp: 'ar 6/28/2005 19:08'!
transform: newTransform
	"Change the current transform. This is the core of all transformations."
	(self invalidationSuppressed or:[container == nil]) ifFalse:[
		"Invalidate old position"
		container invalidate: (self childToParent: self drawingBounds) from: self.
	].
	self propertyValueAt: #transform put: newTransform.
	flags := flags bitOr: FlagHasMatrix. "OBSOLETE - just so we can identify old stuff"
	self signal: #geometryChanged.
	self frameChanged.
	(self invalidationSuppressed or:[container == nil]) ifFalse:[
		"Invalidate new position"
		container invalidate: (self childToParent: self drawingBounds) from: self.
		container costumeMoved: self.
	].
! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: ''!
transformBy: aTransform
	self transform: (self transform transformedBy: aTransform)! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: ''!
transformTo: aCostume
	"Compute a transform that allows relocating costumes from myself to aCostume"
	^self localToGlobal transformedBy: aCostume globalToLocal! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
turn
	^self turn: #left turns: 1 duration: 1 style: #gently.! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
turn: dir
	^self turn: dir turns: 1 duration: 1 style: #gently.! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
turn: dir turns: turns
	^self turn: dir turns: turns duration: 1 style: #gently.! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
turn: dir turns: turns duration: dur
	^self turn: dir turns: turns duration: dur style: #gently.! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: 'ar 6/27/2005 15:11'!
turn: direction turns: turns duration: duration style: style
	"Move me according the given attributes."
	| msecDuration nowTime startTime styleFunc lastState startState endState deltaTime proportion newState |
	"Get the ultimate state we're interested in"
	direction == #left 
		ifTrue:[endState := -360 * turns]
		ifFalse:[endState := 360 * turns].	
	duration <= 0 ifTrue:[^self rotation: endState].
	"Compute msecs since that's what we'll be working on here"
	msecDuration := duration * 1000.
	"Get our starting time to compute incremental changes accordingly"
	nowTime := startTime := Time millisecondClockValue.
	"Get the interpolation function (evaluating from 0-1 and returning values from 0-1)"
	styleFunc := self interpolationForStyle: style.
	"Reset the interpolation state"
	lastState := startState := 0.
	"And go moving until we're out of time"
	Processor activeProject 
		forEach: #tick 
		until:[nowTime >= (startTime + msecDuration)]
		do:[
			"Compute the time at the tick and the delta to when we started"
			nowTime := Time millisecondClockValue.
			"Truncate deltaTime to fit into the given proportion"
			deltaTime := (nowTime - startTime) min: msecDuration.
			"Compute the proportion of time that's over"
			proportion := deltaTime asFloat / msecDuration asFloat.
			"Map it through the interpolation style"
			proportion := styleFunc value: proportion.
			"Compute the state (e.g., offset) at our new proportion"
			newState := startState interpolateTo: endState at: proportion.
			"Apply transform incrementally - this is so that we can superimpose animations."
			self rotateBy: (newState - lastState).
			"Remember last state for the next round"
			lastState := newState.
	].
	"Apply the remaining part of the transform (if any)"
	self rotateBy: (endState - lastState).! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 3/28/2004 16:10'!
turnBy: angle
	self rotateBy: angle asNumber.! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: 'ar 6/27/2005 14:46'!
turnBy: angle duration: duration style: style
	"Move me according the given attributes."
	duration <= 0 ifTrue:[^self rotateBy: angle].
	^self turn: #right turns: (angle / 360.0) duration: duration style: style! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: ''!
turnBy: turns speed: speed! !

!CPrimitiveCostume methodsFor: 'primitive behaviors' stamp: ''!
turnTo: target duration: duration style: style
	"Move me according the given attributes."
self error:'not implemented'! !

!CPrimitiveCostume methodsFor: 'pen trails in me' stamp: 'ar 6/25/2005 19:24'!
turtleTrailsForm
	"return the Form with the pen marks"
	^(self primitivePen ifNil:[^nil]) destForm! !

!CPrimitiveCostume methodsFor: 'tweaking' stamp: 'ar 2/27/2004 20:02'!
tweaked: propName from: oldValue to: newValue
	"Remember the change of the given property from its previous to its new value.
	Answer true if the receiver is newly changed or reverted back to unchanged."
	(super tweaked: propName from: oldValue to: newValue) ifFalse:[^false].
	self player tweakedCostume: self.
	^true! !

!CPrimitiveCostume methodsFor: 'tweaking' stamp: 'ar 2/27/2004 18:35'!
tweaks
	"Answer the changes associated with the receiver"
	^self propertyValueAt: #myTweaks! !

!CPrimitiveCostume methodsFor: 'tweaking' stamp: 'ar 2/27/2004 18:36'!
tweaks: aDictionary
	"Set the changes associated with the receiver"
	^self propertyValueAt: #myTweaks put: aDictionary! !

!CPrimitiveCostume methodsFor: 'move/resize/scale' stamp: 'ar 6/26/2005 19:22'!
updateRotationStyleTransform: rs
	"Update the receiver's transform based on its rotation style after its nose has changed."
	| dir deg scale tfm offset nose |

	"Note: This uses direction to escape any problems with the nose being swapped"
	dir := (Point r: 1.0 degrees: self direction degrees).
	deg := dir degrees + 90 \\ 360.0.
	deg > 180.0 ifTrue:[deg := deg-360].

	scale := 1@1.
	(rs == #'left-right' and:[deg < 0.0]) ifTrue:[scale := -1@1].
	(rs == #'up-down' and:[deg abs > 90.0]) ifTrue:[scale := 1@-1].

	"Construct the appropriate transform"
	tfm := self transform.
	offset := tfm offset.
	scale := tfm scale abs * scale.
	scale = (1@1) ifTrue:[
		tfm := COffsetTransform withOffset: offset.
	] ifFalse:[
		tfm := CMatrixTransform identity.
		tfm setScale: scale.
		tfm setOffset: offset.
	].
	"Adjust position to stay the same"
	offset := self position - (tfm transformPoint: self pivot). "adjusts position to match"
	tfm := (COffsetTransform withOffset: offset) composedWithLocal: tfm.
	"Install new transform, keeping position and heading in place"
	self transform: tfm.
	[nose := (tfm invertDirection: dir) normalized]
		on: ZeroDivide do: [:ex | ^nil].
	self propertyValueAt: #nose put: nose.! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/29/2004 14:55'!
userDraw
	"Return true if the receiver should act as a custom drawing area"
	^(flags bitAnd: FlagCustomDraw) == FlagCustomDraw! !

!CPrimitiveCostume methodsFor: 'drawing' stamp: 'ar 3/29/2004 14:55'!
userDraw: aBool
	"Determine whether the receiver should act as a custom drawing area"
	flags := aBool ifTrue:[flags bitOr: FlagCustomDraw] ifFalse:[flags bitClear: FlagCustomDraw].
	self signal: #flagsChanged.
	self invalidate.! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'ar 3/21/2004 04:44'!
vResizing
	"Layout specific. This property describes how the receiver should be resized with respect to its owner and its children. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit children
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#rigid] ifNotNil:[props vResizing].! !

!CPrimitiveCostume methodsFor: 'layout' stamp: 'tk 1/4/2005 13:18'!
vResizing: aSymbol
	"Layout specific. This property describes how the receiver should be resized with respect to its container and its elements. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit my elements
	"
	self vResizing == aSymbol ifTrue:[^self].
	self assureLayoutProperties vResizing: aSymbol.
	self signal: #vResizingChanged.
	self invalidateLayout.! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/16/2004 19:17'!
value
	^self player value! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 3/16/2004 19:17'!
value: ignored
	^self player value! !

!CPrimitiveCostume methodsFor: 'copying' stamp: 'tk 2/11/2005 14:12'!
veryDeepFixupWith: deepCopier
	"If some fields were weakly copied, fix new copy here."

	"If my container is being duplicated too, then store his duplicate.  If I am contained outside the duplicated tree, then I am no longer owned!!"
	(deepCopier newGuys includes: container) ifFalse: [
		container := deepCopier references at: container ifAbsent: [nil]].

	super veryDeepFixupWith: deepCopier.
! !

!CPrimitiveCostume methodsFor: 'copying' stamp: 'tk 4/21/2005 16:23'!
veryDeepInner: deepCopier 
	"The inner loop, so it can be overridden when a field should not  
	be traced."

	super veryDeepInner: deepCopier.
	"container := container.		special. see veryDeepFixupWith:"
	contents ifNotNil: [
		contents := contents veryDeepCopyWith: deepCopier. 	"array. enter all in refs"
		1 to: contents size do: [:ii | 
			contents at: ii put: ((contents at: ii) veryDeepForSure: deepCopier)]].
			"force the duplication of CPlayers we own"
	bounds := bounds clone.		"If there is ever a CRectangle, change this"
			"Points are shared with original"
	fill := fill veryDeepCopyWith: deepCopier.
	border := border veryDeepCopyWith: deepCopier.
	flags := flags veryDeepCopyWith: deepCopier.
	self veryDeepProperties: deepCopier.	"properties"
	properties := properties.	"to show the checker we did it"! !

!CPrimitiveCostume methodsFor: 'copying' stamp: 'tk 1/27/2005 14:42'!
veryDeepProperties: deepCopier
	"Special case for when a class wants one of its fields to be shared and not duplicated."

	| old |
	(deepCopier newGuys includes: properties) ifTrue: [^ self].  	"already done"
	old := properties.
	properties := properties copy.	"has old values"  "does duplicate the associations"
	deepCopier references at: old put: properties.
	deepCopier newGuys add: properties.	"remember"

	"Skip some special fields"
	properties removeKey: #myEventMap ifAbsent: [].	"there may be others too"
	properties associationsDo: [:assoc |
			assoc value: (assoc value veryDeepForSure: deepCopier).
				"watch for ones that should be weakly copied"
			deepCopier references at: (old associationAt: assoc key) put: assoc.
			deepCopier newGuys add: assoc].
! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: ''!
visible
	^(flags bitAnd: FlagHidden) == 0! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: ''!
visible: aBool
	^self hidden: aBool not! !

!CPrimitiveCostume methodsFor: 'scrolling' stamp: 'ar 8/18/2003 00:07'!
visibleBounds
	"Answer the (scrolled) bounds encompassing the currently visible portion of this object"
	^self innerBounds! !

!CPrimitiveCostume methodsFor: 'scrolling' stamp: 'ar 3/24/2004 01:53'!
visibleScrollArea
	"Answer the visible portion of the scroll area"
	^self innerBounds! !

!CPrimitiveCostume methodsFor: 'event handling' stamp: 'ar 4/22/2005 10:40'!
waitForClicksOrDrag: aHand event: firstEvent
	"Synthesize #mouseClick, #mouseDoubleClick or #mouseDrag events"
	| pos dragThreshold timeOut startTime nowTime evt timer firstPos |
	firstPos := aHand position.
	dragThreshold := 10. "pixels before drag"
	timeOut := 350. "double click time out"
	startTime := Time millisecondClockValue.
	(self handlesEvent: #mouseDrag) ifTrue:[
		pos := aHand position.
		evt := aHand forEach: #mouseMove do:[:event|
			(aHand position dist: pos) > dragThreshold ifTrue:[
				aHand position: firstPos.
				^self signal: #mouseDrag 
					with: firstEvent with: self.
			].
		] until: #mouseUp.
	] ifFalse:[evt := aHand waitUntil: #mouseUp].

	(self containsPoint: self cursorPoint) ifFalse:[^self].

	(self handlesEvent: #mouseDoubleClick) ifFalse:[
		"If the player doesn't handle #mouseDoubleClick don't wait for #mouseClick to time out"
		aHand position: firstPos.
		^self signal: #mouseClick
				with: firstEvent with: self.
	].

	nowTime := Time millisecondClockValue.
	(nowTime - startTime) > timeOut ifTrue:[
		"timeout after mouseUp -> mouseClick"
		aHand position: firstPos.
		^self signal: #mouseClick
				with: firstEvent with: self.
	].
	"this is ugly... we should be able to wait on a delay but we aren't"
	timer := self startScript:[
		(Delay forMilliseconds: timeOut - (nowTime - startTime)) wait.
		self signal: #doubleClickTimeout 
			with: firstEvent with: self.
	].
	evt := self waitUntilAnyOf: {
		self. #mouseDown.
		self. #doubleClickTimeout.
	}.
	self stopScript: timer.
	aHand position: firstPos.
	evt selector == #mouseDown ifTrue:[
		^self signal: #mouseDoubleClick
			with: firstEvent with: self.
	] ifFalse:[
		^self signal: #mouseClick
			with: firstEvent with: self.
	].! !

!CPrimitiveCostume methodsFor: 'corners' stamp: 'ar 3/24/2004 14:26'!
wantsRoundedCorners
	"Compatibility"
	^self roundCorners! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/25/2005 20:32'!
width
	^self extent x! !

!CPrimitiveCostume methodsFor: 'geometry-local' stamp: 'ar 6/25/2005 20:32'!
width: aNumber
	self extent: aNumber@self height.
	^aNumber! !

!CPrimitiveCostume methodsFor: 'other' stamp: 'ar 3/24/2004 03:24'!
windowColor: aColor
	(border width > 0 and:[border color isTransparent not])
		ifTrue:[border color: aColor].
! !

!CPrimitiveCostume methodsFor: 'structure' stamp: 'ar 12/8/2003 19:41'!
withAllContainersDo: aBlock
	"Evaluate aBlock with the receiver and all of its containers"
	aBlock value: self.
	container ifNotNil:[^container withAllContainersDo: aBlock].! !

!CPrimitiveCostume methodsFor: 'accessing' stamp: 'ar 7/15/2003 16:35'!
world
	self isWorld ifTrue:[^self player].
	container ifNotNil:[^container world].
	^nil! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:32'!
worldBounds
	^self localToWorld: self localBounds
! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:40'!
worldBounds: rect
	container ifNil:[^self bounds: rect].
	self bounds: (container worldToLocal: rect)! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/26/2005 12:08'!
worldPosition
	^self localToWorld: self pivot! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:41'!
worldPosition: aPoint
	container ifNil:[^self position: aPoint].
	self position: (container worldToLocal: aPoint)! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:32'!
worldToLocal
	"Answer a transformation from the local into the global frame"
	^self localToWorld inverseTransformation! !

!CPrimitiveCostume methodsFor: 'geometry-global' stamp: 'ar 6/25/2005 20:32'!
worldToLocal: anObject
	"Transform anObject from global (world) coordinates into local coordinates"
	^anObject transformedBy: self worldToLocal.! !

!CPrimitiveCostume methodsFor: 'serialization' stamp: 'ar 3/22/2004 02:34'!
writeContentsOn: aSerializer
	| aPlayer |
	^aSerializer writeField: #contents value: (contents collect:[:each| 
		aPlayer := each player.
		[aPlayer player notNil and:[aPlayer player getCostume == aPlayer]]
			whileTrue:[aPlayer := aPlayer player].
		aPlayer]).! !

!CPrimitiveCostume methodsFor: 'serialization' stamp: 'ar 3/29/2004 14:54'!
writeFieldsOn: aSerializer

	"basic"
	aSerializer writeField: #origin value: bounds origin.
	aSerializer writeField: #extent value: bounds extent.
	aSerializer writeField: #fill value: fill.
	aSerializer writeField: #border value: border.

	aSerializer writeField: #player value: self player.
	self writeContentsOn: aSerializer.
	"aSerializer writeField: #contents value: contents."

	"flags"
	aSerializer writeField: #hidden value: (flags anyMask: FlagHidden).
	aSerializer writeField: #userDraw value: (flags anyMask: FlagCustomDraw).
	aSerializer writeField: #locked value: (flags anyMask: FlagLocked).
	aSerializer writeField: #cacheEnabled value: (flags anyMask: FlagCached).
	aSerializer writeField: #clipEnabled value: (flags anyMask: FlagClipEnabled).
	aSerializer writeField: #roundCorners value: (flags anyMask: FlagRoundedCorners).
	aSerializer writeField: #hasScroller value: (flags anyMask: FlagHasScroller).

	"1 to: FlagRegistry size do:[:bit|
		flagBit := 1 bitShift: bit-1.
		flagName := FlagRegistry at: bit.
		flagName ifNotNil:[
			(flags bitAnd: flagBit) = 0 ifFalse:[
				aSerializer writeField: flagName value: true.
			].
		].
	]."

	self serializableFields do:[:fieldName|
		aSerializer writeField: fieldName value: (self propertyValueAt: fieldName).
	].
! !

!CPrimitiveCostume methodsFor: 'tweaking' stamp: 'ar 2/27/2004 18:40'!
writeTweakedFieldsOn: aSerializer
	"FIXME!!!!!!"
	| changeList newValue |
	changeList := self tweaks ifNil:[#()].
	changeList keysAndValuesDo:[:fieldName :chgRec|
		fieldName == #contents ifTrue:[
			self writeContentsOn: aSerializer.
		] ifFalse:[
			newValue := self perform: fieldName.
			aSerializer writeField: fieldName value: newValue.
		].
	].! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:33'!
x
	^self position x! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:33'!
x: aNumber
	self position: aNumber @ self y.
	^aNumber! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:33'!
y
	^self position y! !

!CPrimitiveCostume methodsFor: 'geometry-parent' stamp: 'ar 6/25/2005 20:33'!
y: aNumber
	self position: self x @ aNumber.
	^aNumber! !

CPrimitiveCostume subclass: #CWorldCostume
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Worlds'!

!CWorldCostume methodsFor: 'accessing' stamp: 'ar 7/15/2003 16:35'!
isWorld
	^true! !

!CWorldCostume methodsFor: 'accessing' stamp: 'ar 7/15/2003 16:16'!
root
	^self player! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 1/22/2005 18:39'!
asPlayer
	^CValuePlayer on: self! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 1/22/2005 18:40'!
deselect
	"Graphically deselect the receiver"! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 5/13/2005 11:21'!
displayText
	^self asText! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 1/22/2005 19:06'!
fieldNamed: fieldName
	^nil! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 1/22/2005 18:39'!
isPrimitiveCostume
	^false! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 2/27/2004 18:32'!
isTweaked
	"Answer whether I have any tweaks to write out"
	^self tweaks notNil! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 2/27/2004 18:33'!
isTweakedFrom: oldValue
	"Answer true if I have been tweaked from oldValue"
	self == oldValue ifTrue:[^self isTweaked].
	self class == oldValue class ifFalse:[^true]. "definitely changed"
	^(self = oldValue) not! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 2/27/2004 18:33'!
readTweakedFieldsFrom: deSerializer
	"I don't record changes so read all my fields"
	^self readFieldsFrom: deSerializer! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 1/22/2005 18:40'!
select
	"Graphically select the receiver"! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 2/27/2004 18:30'!
tweak: propertyName to: newValue
	| oldValue |
	oldValue := self perform: propertyName.
	self perform: (propertyName,':') asSymbol with: newValue.
	self tweaked: propertyName from: oldValue to: newValue.! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 2/27/2004 18:30'!
tweaked: propName after: aBlock
	| oldValue newValue |
	oldValue := self perform: propName.
	aBlock value.
	newValue := self perform: propName.
	self tweaked: propName from: oldValue to: newValue.! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 3/18/2005 16:22'!
tweaked: propName from: oldValue to: newValue
	"Remember the change of the given property from its previous to its new value.
	Answer true if the receiver is newly changed or reverted back to unchanged."
	| tweaks chgRec update |
	propName ifNil:[self error: 'What are you trying to do???'. ^false].
	propName == #contents ifTrue:[
		"Temporarily disable tweaking #contents since this too easily leads to projects hanging on to tons of old and completely useless stuff, when, for example, closing a window through halo interaction. We need to rethink this."
		^false].
	update := false.
	(tweaks := self tweaks) ifNil:[
		update := update not.
		self tweaks: (tweaks := Dictionary new)].
	chgRec := tweaks at: propName ifAbsentPut:[CPropertyChange new value: oldValue].
	(newValue isTweakedFrom: chgRec value) ifFalse:[
		tweaks removeKey: propName.
		tweaks isEmpty ifTrue:[
			update := update not.
			self tweaks: nil].
	].
	^update! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 2/27/2004 18:32'!
tweaks
	"Answer the tweaks associated with the receiver. By default we answer nil here as it is a responsibility of our subclasses to provide the place where to put tweaks."
	^nil! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 2/27/2004 18:32'!
tweaks: aDictionary
	"Set the tweaks associated with the receiver. By default we ignore this message as it is a responsibility of our subclasses to provide the place where to put tweaks."! !

!Object methodsFor: '*Tweak-Basic-Copying-override' stamp: 'ar 3/18/2005 14:06'!
veryDeepCheck: copier
	"Give CObject a chance to look for unmapped elements"
! !

!Object methodsFor: '*Tweak-Basic-Copying-override' stamp: 'ar 3/18/2005 14:06'!
veryDeepCopy
	"Do a complete tree copy using a dictionary.  An object in the tree twice is only copied once.  All references to the object in the copy of the tree will point to the new copy."

	| copier new cnt cycles |
	copier := CDeepCopier new initialize: self initialDeepCopierSize.
	"Smalltalk at: #CC put: copier.	debug"
	new := self veryDeepForSure: copier. 	"always duplicate the top of the tree"
	copier mapUniClasses.
	cnt := copier references size.  cycles := 0.
	[copier references array copy do: [:assoc | "safe for adding during loop"
			assoc ifNotNil: [
				"assoc key == assoc value ifTrue: ['debug'.
					assoc key isCObject ifTrue: [self error: 'mapping to self']]. "
				assoc value veryDeepFixupWith: copier.
				"(cb := assoc value componentBinding) ifNotNil: [
					cb value == assoc value ifFalse: [self error: 'pointing across']] "
				]].
	 cycles := cycles + 1.
	 cnt = copier references size "did we discover any new objects?"] 
		whileFalse: [cnt := copier references size].
	"Transcript show: 'Did fixup pass this many times: ', cycles printString; cr."
	copier fixDependents.

	copier references associationsDo: [:assoc | 
			assoc ifNotNil: [assoc value veryDeepCheck: copier]].
	^ new! !

!Object methodsFor: '*Tweak-Basic-Copying-override' stamp: 'ar 3/18/2005 14:07'!
veryDeepCopySibling
	"Do a complete tree copy using a dictionary.  Substitute a clone of oldPlayer for the root.  Normally, a Player or non systemDefined object would have a new class.  We do not want one this time.  An object in the tree twice, is only copied once.  All references to the object in the copy of the tree will point to the new copy."

	| copier new |
	copier := CDeepCopier new initialize: self initialDeepCopierSize.
	copier newUniClasses: false.
	new := self veryDeepCopyWith: copier.
	copier mapUniClasses.
	copier references associationsDo: [:assoc | 
		assoc value veryDeepFixupWith: copier].
	copier fixDependents.
	^ new! !

!Object methodsFor: '*Tweak-Basic-Copying-override' stamp: 'ar 3/18/2005 14:06'!
veryDeepCopyUsing: copier
	"Do a complete tree copy using a dictionary.  An object in the tree twice is only copied once.  All references to the object in the copy of the tree will point to the new copy.
	Same as veryDeepCopy except copier (with dictionary) is supplied.
	** do not delete this method, even if it has no callers **"

	| new refs newDep newModel cnt |
	new := self veryDeepForSure: copier. 	"Always duplicate top of tree"
	copier mapUniClasses.
	cnt := copier references size.
	[copier references array copy do: [:assoc | "safe for adding during loop"
			assoc ifNotNil: [
				assoc value veryDeepFixupWith: copier]].
	 cnt = copier references size "did we discover any new objects?"] 
		whileFalse: [cnt := copier references size].
	"Fix dependents"
	refs := copier references.
	DependentsFields associationsDo: [:pair |
		pair value do: [:dep | 
			(newDep := refs at: dep ifAbsent: [nil]) ifNotNil: [
				newModel := refs at: pair key ifAbsent: [pair key].
				newModel addDependent: newDep]]].

	copier references associationsDo: [:assoc | "safe for adding during loop"
			assoc ifNotNil: [assoc value veryDeepCheck: copier]].
	^ new
! !

!Object methodsFor: '*Tweak-Basic-Copying-override' stamp: 'ar 3/18/2005 14:06'!
veryDeepCopyWith: deepCopier
	"Copy me and the entire tree of objects I point to.  An object in the tree twice is copied once, and both references point to him.  deepCopier holds a dictionary of objects we have seen.  Some classes refuse to be copied.  Some classes are picky about which fields get deep copied."
	| class index sub assoc new uc sup has mine |
	deepCopier references at: self ifPresent: [:newer | ^ newer]. 	"already did him"
	(deepCopier newGuys includes: self) ifTrue: [^ self].  	"I am the new guy"
	class := self class.
	class isMeta ifTrue: [^ self].		"a class"
	new := self clone.
	(class isSystemDefined not and: [deepCopier newUniClasses "allowed"]) ifTrue: [
		uc := deepCopier uniClasses at: class ifAbsent: [nil].
		uc ifNil: [
			deepCopier uniClasses at: class put: (uc := self copyUniClassWith: deepCopier).
			deepCopier references at: class put: uc].	"remember"
		new := uc new.
		new copyFrom: self].	"copy inst vars in case any are weak"
	deepCopier references at: self put: new.	"remember"
	deepCopier newGuys add: new.	"remember"
	(class isVariable and: [class isPointers]) ifTrue: 
		[index := self basicSize.
		[index > 0] whileTrue: 
			[sub := self basicAt: index.
			(assoc := deepCopier references associationAt: sub ifAbsent: [nil])
				ifNil: [new basicAt: index put: (sub veryDeepCopyWith: deepCopier)]
				ifNotNil: [new basicAt: index put: assoc value].
			index := index - 1]].
	"Ask each superclass if it wants to share (weak copy) any inst vars"
	new veryDeepInner: deepCopier.		"does super a lot"

	"other superclasses want all inst vars deep copied"
	sup := class.  index := class instSize.
	[has := sup compiledMethodAt: #veryDeepInner: ifAbsent: [nil].
	has := has ifNil: [class isSystemDefined not "is a uniClass"] ifNotNil: [true].
	mine := sup instVarNames.
	has ifTrue: [index := index - mine size]	"skip inst vars"
		ifFalse: [1 to: mine size do: [:xx |
				sub := self instVarAt: index.
				(assoc := deepCopier references associationAt: sub ifAbsent: [nil])
						"use association, not value, so nil is an exceptional value"
					ifNil: [new instVarAt: index put: 
								(sub veryDeepCopyWith: deepCopier)]
					ifNotNil: [new instVarAt: index put: assoc value].
				index := index - 1]].
	(sup := sup superclass) == nil] whileFalse.
	new rehash.	"force Sets and Dictionaries to rehash"
	^ new
! !

!Object methodsFor: '*Tweak-Basic-Copying-override' stamp: 'ar 3/18/2005 14:06'!
veryDeepFixupWith: deepCopier
	"The contents of every field is tested against the references dictionary after all duplicates have been created.  Map any object that has been duplicated, no matter which field it appears in (within the duplicated objects)."

	"See if any of my fields have been duplicated since I was created as a duplicate."
	| class migheBeOld new |
	class := self class.
 	1 to: class instSize do: [:ii |
		migheBeOld := self instVarAt: ii.
		new := deepCopier references at: migheBeOld ifAbsent: [migheBeOld].
		migheBeOld == new ifFalse: [self instVarAt: ii put: new]].	

	"The same for variable pointers"
	(class isVariable and: [class isPointers]) ifTrue: [
		1 to: self basicSize do: [:ii |
			migheBeOld := self basicAt: ii.
			new := deepCopier references at: migheBeOld ifAbsent: [migheBeOld].
			migheBeOld == new ifFalse: [self basicAt: ii put: new]]].
! !

!Object methodsFor: '*Tweak-Basic-Copying-override' stamp: 'ar 3/18/2005 14:06'!
veryDeepForSure: deepCopier
	"a backstop for when we try to force CPlayers to duplicate"

	^ self veryDeepCopyWith: deepCopier
! !

!Object methodsFor: '*Tweak-Basic-Copying-override' stamp: 'ar 3/18/2005 14:07'!
veryDeepInner: deepCopier
	"No special treatment for inst vars of my superclasses.  Override when some need to be weakly copied.  Object>>veryDeepCopyWith: will veryDeepCopy any inst var whose class does not actually define veryDeepInner:"
! !

!Object methodsFor: '*Tweak-Basic' stamp: 'ar 2/27/2004 18:33'!
writeTweakedFieldsOn: aSerializer
	"I don't record tweaks so write all my fields"
	^self writeFieldsOn: aSerializer
! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 12/8/2002 00:37'!
actualScreenSize
	"Polymorphic implementation for DisplayScreen"
	^self extent! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 12/8/2002 01:55'!
asForm
	^self! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 3/23/2004 19:35'!
asPlayer
	^(CPlayer new)
		graphicForm: self;
		extent: self extent;
		value: self; 
		yourself! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 12/8/2002 00:49'!
deferUpdates: aBool
	"Polymorphic implementation for DisplayScreen"
	^false! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 1/8/2003 19:31'!
drawOn: aCanvas in: drawingBounds
	drawingBounds extent = self extent ifTrue:[
		self depth < 32
			ifTrue:[aCanvas paintImage: self at: drawingBounds origin]
			ifFalse:[aCanvas translucentImage: self at: drawingBounds origin].
		^self].
	^aCanvas 
		warpImage: self
		transform: (CMatrixTransform localRect: self boundingBox toGlobal: drawingBounds).

! !

!Form methodsFor: '*Tweak-Basic' stamp: 'bf 4/3/2004 00:31'!
drawOn: aCanvas in: drawingBounds sourceRect: aRectangle
	drawingBounds extent = aRectangle extent ifTrue:[
		self depth < 32
			ifTrue:[aCanvas paintImage: self at: drawingBounds origin sourceRect: aRectangle]
			ifFalse:[aCanvas translucentImage: self at: drawingBounds origin sourceRect: aRectangle].
		^self].
	^aCanvas 
		warpImage: self
		transform: (CMatrixTransform localRect: aRectangle toGlobal: drawingBounds)
		sourceRect: aRectangle.
! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 12/8/2002 00:49'!
forceDamageToScreen: rectList
	"Polymorphic implementation for DisplayScreen"! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 12/8/2002 00:50'!
forceDisplayUpdate
	"Polymorphic implementation for DisplayScreen"! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 12/8/2002 00:50'!
forceToScreen
	"Polymorphic implementation for DisplayScreen"! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 12/8/2002 00:50'!
forceToScreen: aRectangle
	"Polymorphic implementation for DisplayScreen"! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 12/8/2002 01:31'!
isCursor
	^false! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 6/24/2003 14:53'!
isOpaque
	"Answer whether this form contains any transparent or translucent colors"
	self depth < 32 ifTrue:[
		"Depth < 32 means only zero pixels are mapped to transparent. Count those."
		^self tallyPixelValues first = 0
	].
	"Depth = 32 means we need to check for alpha.
	ARRRGHHH!!!!!! I knew it!! Years back I knew that aliasing BitBlt's tallyMap with its colorMap can't be a good idea. Now it's clear why - you can't count pixels *after* they've been transferred. Ho hum."
	^false! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 12/8/2002 18:41'!
supportedDisplayDepths
	"Polymorphic implementation for DisplayScreen"
	^#(1 2 4 8 16 32)! !

!Form methodsFor: '*Tweak-Basic' stamp: 'ar 12/8/2002 18:41'!
supportsDisplayDepth: pixelDepth
	"Polymorphic implementation for DisplayScreen"
	^true! !

Exception subclass: #CUnsynchronizedModification
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tweak-Basic-Support'!

!CUnsynchronizedModification commentStamp: '<historical>' prior: 0!
Resumable error to capture unsynchronized modifications inside Tweak.!

!CUnsynchronizedModification methodsFor: 'actions' stamp: 'ar 9/30/2009 16:57'!
defaultAction
	self resignalAs: (Error new messageText: self messageText)! !

!CUnsynchronizedModification methodsFor: 'testing' stamp: 'ar 9/30/2009 16:38'!
isResumable
	^true! !
CDirectoryEntry initialize!
CFieldReference initialize!
CSpriteCostume initialize!
CHandPlayer initialize!
CCursorLibrary initialize!
CIconLibrary initialize!
CMD5Hasher initialize!
CPrimitiveCostume initialize!
